/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "actionactionitem.h"
#include "actionexecutioninfo.h"


sh::actions::ActionActionItem::ActionActionItem(QString text, bool enabled, QString icon, int defaultActionPrecedence, bool checkable, bool ischecked) :
    sh::actions::AbstractActionItem(text, enabled, icon, defaultActionPrecedence, checkable, ischecked)
{
}

sh::tools::AtomicCounter sh::actions::ActionActionItem::_runningActions;

int sh::actions::ActionActionItem::runningActions()
{
    return _runningActions.value();
}

sh::actions::ActionActionItem::ActionActionItem() :
    sh::actions::ActionActionItem(QString())
{
}

bool sh::actions::ActionActionItem::allowExitWhileRunning()
{
    return false;
}

void sh::actions::ActionActionItem::execute(sh::actions::ActionExecutionInfo* info)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    FORBID_ASYNC_CALLS
    auto foo = allowExitWhileRunning() ? 0 : _runningActions.increment(); //automatically decrements when it goes out of scope
    if (!isInitialized())
        initializeSync();
    if (enabled() && visible())
    {
        info->startExecution(shared_from_this());
        try
        {
            action(info);
            info->respectCancel(); // last chance before committing :)
        }
        catch(...)
        {
            info->finishExecution();
            throw;
        }
        info->finishExecution();
    }
}

void sh::actions::ActionActionItem::execute()
{
    std::shared_ptr<sh::actions::AbstractActionItem> _sharedptr = shared_from_this();
    std::weak_ptr<sh::actions::AbstractActionItem> ai = shared_from_this();
    while (true)
    {
        std::shared_ptr<sh::actions::AbstractActionItem> sai = ai.lock();
        if (sai == 0)
            return;
        if (sai == sh::actions::AbstractActionItem::NO_PARENT)
            break;
        if (!sai->enabled() || !sai->visible())
            return;
        ai = sai->parentAction();
    }
    sh::tools::ThreadPool::enqueueForce([=]() {
        sh::actions::ActionExecutionInfo* info;
        sh::tools::MainThread::dispatcher()->invokeSync([&](){
            info = new sh::actions::ActionExecutionInfo();
        });
        info->withExecuteGuards_infodlg([&](){
            bool execsuccess = false;
            sh::exceptions::Exception::executeGuarded([&](){
                sh::exceptions::Exception::executeGuarded([&](){
                    std::shared_ptr<sh::actions::AbstractActionItem> __sharedptr = _sharedptr;
                    execute(info);
                    execsuccess = true;
                }, sh::exceptions::ExecuteGuardFlag::CANCELABLE_UP_TO_HERE);
            }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
            info->shutdown(execsuccess && !info->isCancelled());
        }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        sh::tools::MainThread::dispatcher()->invokeSync([&](){
            delete info;
        });
    });
}

QKeySequence sh::actions::ActionActionItem::shortcuthint()
{
    return _shortcuthint;
}

bool sh::actions::ActionActionItem::shortcuthintTriggersOnCurrentDirectory()
{
    return _shortcuthintTriggersOnCurrentDirectory;
}

void sh::actions::ActionActionItem::setShortcuthint(QKeySequence shortcuthint, bool triggersOnCurrentDirectory)
{
    _shortcuthint = shortcuthint;
    _shortcuthintTriggersOnCurrentDirectory = triggersOnCurrentDirectory;
}
