/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ACTIONACTIONITEM_H
#define ACTIONACTIONITEM_H

#include <QShortcut>
#include "tools/atomiccounter.h"
#include "abstractactionitem.h"


namespace sh{
    namespace actions{

        class ActionExecutionInfo;

        /** \brief
         * Abstract base class for an executable action.
         *
         * It can be made visible in menus or the toolbar or executed directly.
         *
         * Subclasses provide the actual implementation by implementing `action`.
         * Execution can be triggered from outside by calling `execute`.
         */
        class ActionActionItem : public sh::actions::AbstractActionItem
        {
            Q_OBJECT
        public:
            /** \brief Is (for subclasses) intended to be directly constructed from everywhere or by
             * registering a factory somewhere.
             */
            ActionActionItem(QString text, bool enabled = true, QString icon = QString(), int defaultActionPrecedence = 0, bool checkable = false, bool ischecked = false);
            /** \brief
             * Executes this action.
             */
            void execute();
            /** \brief
             * Executes this action.
             */
            void execute(sh::actions::ActionExecutionInfo *info);

            /** \brief
             * Returns the keyboard shortcut for triggering this action.
             */
            QKeySequence shortcuthint();
            /** \brief
             * Checks if using the keyboard shortcut shall trigger the
             * action on the current directory (or on the entry selection).
             */
            bool shortcuthintTriggersOnCurrentDirectory();
            /** \brief
             * Sets the keyboard shortcut for triggering this action.
             */
            void setShortcuthint(QKeySequence shortcut, bool triggersOnCurrentDirectory=false);

            /** \brief
             * Returns the number of currently running actions.
             */
            static int runningActions();

        signals:

        public slots:

        protected:
            ActionActionItem();

            /** \brief
             * Checks if such an action would prevent shallot from exiting when it is executing.
             */
            virtual bool allowExitWhileRunning();
            /** \brief
             * The action implementation, i.e. what the actions should actually do.
             */
            virtual void action(sh::actions::ActionExecutionInfo* info) = 0;
            QKeySequence _shortcuthint = 0;
            bool _shortcuthintTriggersOnCurrentDirectory = false;
            static sh::tools::AtomicCounter _runningActions;

        private:

        };

    }
}

#endif // ACTIONACTIONITEM_H
