/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ACTIONFACTORY_H
#define ACTIONFACTORY_H

#include <memory>
#include <QKeySequence>
#include <QList>
#include "actions/actioncategory.h"


namespace sh{
    namespace base{class FilesystemNode;}

    namespace actions{

        class AbstractActionFactory;
        class AbstractActionItem;

        /** \brief
         * Holds various data around action creation.
         *
         * Action creation is a process, which begins with querying all actions available for certain nodes.
         * The action factories evaluate availabilities and store some construction information here.
         * Many parties are involved in working with it.
         * The created action, is stored in this object as well.
         */
        class ActionInstantiation
        {
        public:
            /** \brief
             * Constructor for a typical fresh instance, containing infos about selected items and some more.
             */
            ActionInstantiation(QList<std::shared_ptr<sh::base::FilesystemNode>> selectedNodes={}, std::shared_ptr<sh::base::FilesystemNode> currentDirectory=0);
            /** \brief
             * Clones an existing instance.
             */
            ActionInstantiation(const ActionInstantiation& s) = default;
            /** \brief
             * Clones an existing instance and sets createdAction. This is a convenience constructor used for some special cases.
             */
            ActionInstantiation(const ActionInstantiation& s, std::shared_ptr<sh::actions::AbstractActionItem> createdAction);
            /** \brief
             * The selected nodes (initially: what's selected in the active file list).
             */
            QList<std::shared_ptr<sh::base::FilesystemNode>> selectedNodes;
            /** \brief
             * The current directory (which is visible in the active file list).
             */
            std::shared_ptr<sh::base::FilesystemNode> currentDirectory;
            /** \brief
             * If links shall be resolved for evaluation and action creation.
             */
            bool resolveLinks = true;
            /** \brief
             * If this instantiation refers to a lookup for actions on 'current directory level' (instead of 'selection level').
             */
            bool isLookupOnCurrentDirectoryLevel = false;
            /** \brief
             * The action faction.
             */
            AbstractActionFactory* actionfactory = 0;
            /** \brief
             * The action category.
             */
            std::shared_ptr<ActionCategory> category = 0;
            /** \brief
             * The created action.
             */
            std::shared_ptr<AbstractActionItem> createdAction = 0;
            /** \brief
             * The requested keyboard shortcut for the action.
             */
            QKeySequence shortcut;
            /** \brief
             * The position information index.
             */
            int positionIndex = 0;
        };

        /** \brief
         * Controls when and how an action is created in the action factory.
         */
        class Predicate
        {
        public:
            Predicate();
            /** \brief
             * Prepares the evaluation.
             */
            virtual void prepare(ActionInstantiation* instantiation);
            /** \brief
             * Evaluates if the predicate is solved.
             *
             * If it returns `true`, the evaluation proceeds and eventually succeeds.
             * If `false`, it stops at this place.
             */
            virtual bool evaluate(ActionInstantiation* instantiation);
            virtual ~Predicate();
        };

        /** \brief
         * Shows actions only on directories.
         */
        class OnDirectoriesPredicate : public Predicate
        {
        public:
            OnDirectoriesPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only on files.
         */
        class OnFilesPredicate : public Predicate
        {
        public:
            OnFilesPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only on links.
         */
        class OnLinksPredicate : public Predicate
        {
        public:
            OnLinksPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only on single-entry selections.
         */
        class OnSingleEntrySelectionPredicate : public Predicate
        {
        public:
            OnSingleEntrySelectionPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only when the selection paths matches a regular expression.
         */
        class ByRegExpPredicate : public Predicate
        {
        public:
            ByRegExpPredicate(QString regexp);
            bool evaluate(ActionInstantiation* instantiation);
        private:
            QRegExp filterRegExp;
        };

        /** \brief
         * Disables links resolving.
         */
        class DontResolveLinksPredicate : public Predicate
        {
        public:
            DontResolveLinksPredicate();
            void prepare(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only when not searching for 'current directory level' actions.
         */
        class HideOnCurrentDirectoryLevelPredicate : public Predicate
        {
        public:
            HideOnCurrentDirectoryLevelPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Shows actions only when not searching for 'selection level' actions.
         */
        class HideOnSelectionLevelPredicate : public Predicate
        {
        public:
            HideOnSelectionLevelPredicate();
            bool evaluate(ActionInstantiation* instantiation);
        };

        /** \brief
         * Sets a keyboard shortcut.
         */
        class KeyShortcutPredicate : public Predicate
        {
        public:
            KeyShortcutPredicate(QKeySequence shortcut, bool triggersOnCurrentDirectoryLevel);
            void prepare(ActionInstantiation* instantiation);
        private:
            QKeySequence shortcut;
            bool triggersOnCurrentDirectoryLevel;
        };

        /** \brief
         * Sets a positioning information index.
         */
        class PositionIndexPredicate : public Predicate
        {
        public:
            PositionIndexPredicate(int positionIndex);
            void prepare(ActionInstantiation* instantiation);
        private:
            int positionIndex;
        };

        /** \brief
         * Abstract base class for an action factory.
         *
         * It creates an instance of a certain subclass of AbstractActionItem. It can also check if it
         * is defined to be created in a given situation (e.g. not all actions are visible in toolbar).
         */
        class AbstractActionFactory
        {
        public:
            /** \brief Is (for subclasses) intended to be directly constructed and registered once.
              */
            AbstractActionFactory(std::shared_ptr<sh::actions::ActionCategory> category, QList<std::shared_ptr<Predicate>> predicates);
            virtual ~AbstractActionFactory();
            std::shared_ptr<ActionInstantiation> actionAvailable(ActionInstantiation *instantiation);
            virtual std::shared_ptr<sh::actions::AbstractActionItem> construct(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes) =0;
        private:
            std::shared_ptr<sh::actions::ActionCategory> category;
            QList<std::shared_ptr<Predicate>> predicates;
        };

        /** \brief A typical implementation for an action factory.
         *
         * See base class for more.
         */
        template<class T>
        class ActionFactory : public sh::actions::AbstractActionFactory
        {
        public:
            ActionFactory(std::shared_ptr<sh::actions::ActionCategory> category, QList<std::shared_ptr<Predicate>> predicates)
                : sh::actions::AbstractActionFactory(category, predicates) {}
            std::shared_ptr<sh::actions::AbstractActionItem> construct(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
            {
                return std::shared_ptr<sh::actions::AbstractActionItem>(new T(nodes));
            }
        };

    }
}

#endif // ACTIONFACTORY_H
