/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QBoxLayout>
#include <QDesktopWidget>
#include <QLabel>
#include <QShortcut>
#include <QStack>
#include <QWidgetAction>
#include "actions/abstractactionitem.h"
#include "actions/actionactionitem.h"
#include "actions/headeractionitem.h"
#include "actions/actioncategory.h"
#include "actions/common/actiongroups.h"
#include "actions/common/actionnavigation.h"
#include "actions/common/actionbookmarks.h"
#include "actions/submenuactionitem.h"
#include "actions/viewmenu/actionview.h"
#include "actions/mainmenu/actionmain.h"
#include "base/filesystemhandler.h"
#include "base/eurl.h"
#include "base/filesystemnode.h"
#include "configuration/configurationmanager.h"
#include "settings/setting.h"
#include "settings/settingsmanager.h"
#include "tools/atomiccounter.h"
#include "tools/iconmanager.h"
#include "tools/threadpool.h"
#include "tools/singletoninitializer.h"
#include "tools/benchmarking.h"
#include "tools/mainthread.h"
#include "ui/toolbarbutton.h"
#include "ui/mainwindow.h"
#include "ui/filesystempanel.h"
#include "ui/fileview/fileview.h"
#include "ui/fileview/fileviewcontrol.h"
#include "abstractactionitem.h"
#include "actionactionitem.h"
#include "actionsmanager.h"
#include "separatoractionitem.h"
#include "submenuactionitem.h"


namespace sh{namespace actions{INITIALIZE_SINGLETON_IN_MAINTHREAD(50060, ActionsManager)}}

sh::actions::ActionsManager::ActionsManager(QObject *parent) :
    QObject(parent)
{
}

void sh::actions::ActionsManager::_initialize()
{
}

void sh::actions::ActionsManager::_shutdown()
{
}

sh::actions::ActionsManager::~ActionsManager()
{
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        if (toolbarManagedAction)
            delete toolbarManagedAction;
        foreach (QAction* a, toolbarPermanentActions)
        {
            auto itr = qaction2managedaction.find(a);
            if (itr != qaction2managedaction.end())
                unregisterManagedAction(*itr);
        }
        _isdeleted = true;
    });
}

bool sh::actions::ActionsManager::_isdeleted = false;

sh::actions::ActionsManager* sh::actions::ActionsManager::instance()
{
    if (_isdeleted)
        return 0;
    static sh::actions::ActionsManager* _instance = new sh::actions::ActionsManager();
    return _instance;
}

//#### GENERAL STUFF ####

void sh::actions::ActionsManager::slot_actionactiontriggered()
{
    CHECK_CORRECT_CALLER_THREAD
    QAction *actionaction = static_cast<QAction*>(sender());
    std::shared_ptr<sh::actions::AbstractActionItem> itmAction = qaction2managedaction[actionaction]->actionitem();
    sh::actions::ActionActionItem* action = static_cast<sh::actions::ActionActionItem*>(itmAction.get());
    action->execute();
}

void sh::actions::ActionsManager::registerActionFactory(std::shared_ptr<sh::actions::AbstractActionFactory> actionFactory)
{
    QMutexLocker _l(&_actionfactoriesMutex);
    _actionfactories.append(actionFactory);
}

void sh::actions::ActionsManager::_apply_baseproperties_to_button(ManagedAction *managedaction)
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::actions::AbstractActionItem> _itm = managedaction->actionitem();
    sh::ui::ToolbarButton* btn = ((ManagedAction_ToolbarItem*) managedaction)->button();
    btn->setText(_itm->text());
    btn->setIcon(sh::tools::IconManager::instance()->getIconByFullname(_itm->icon()));
    bool _itmenabled = _itm->enabled();
    if (btn->hasFocus() && !_itmenabled) {  // we must do that, otherwise the last file list gets focus by the following
        auto fspanel = sh::ui::MainWindow::mainWindow()->filesystemPanel();
        fspanel->viewwidget(fspanel->currentViewIndex())->focus();
    }
    btn->setEnabled(_itmenabled);
    managedaction->widgetaction()->setVisible(_itm->visible());
}

void sh::actions::ActionsManager::_apply_baseproperties_to_qaction(sh::actions::AbstractActionItem* itm, QAction* _widgetaction)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::actions::SeparatorActionItem* itmSeparator = qobject_cast<sh::actions::SeparatorActionItem*>(itm);
    sh::actions::HeaderActionItem* itmHeader = qobject_cast<sh::actions::HeaderActionItem*>(itm);
    _widgetaction->setVisible(itm->visible());
    if (!itmSeparator) {
        _widgetaction->setText(itm->text());
        _widgetaction->setIcon(sh::tools::IconManager::instance()->getIconByFullname(itm->icon()));
        if (!itmHeader) {
            _widgetaction->setEnabled(itm->enabled());
            _widgetaction->setCheckable(itm->isCheckable());
            _widgetaction->setChecked(itm->isChecked());
        }
    }
}

void sh::actions::ActionsManager::sortAndSeparateActions(QList<std::shared_ptr<sh::actions::ActionInstantiation> >* actions)
{
    qSort(actions->begin(), actions->end(), [](const std::shared_ptr<sh::actions::ActionInstantiation>& a, const std::shared_ptr<sh::actions::ActionInstantiation>& b){
        return a->positionIndex < b->positionIndex;
    });
    int lastgroup=-1;
    for(int i=actions->length()-1;i>=0;i--)
    {
        std::shared_ptr<sh::actions::ActionInstantiation> acsi = actions->at(i);
        int group = acsi->positionIndex / 100000;
        if (group != lastgroup)
        {
            auto sepsit = std::make_shared<sh::actions::ActionInstantiation>(QList<std::shared_ptr<sh::base::FilesystemNode>>(), (std::shared_ptr<sh::base::FilesystemNode>)0);
            sepsit->createdAction = std::make_shared<sh::actions::SeparatorActionItem>();
            actions->insert(i+1, sepsit);
            lastgroup = group;
        }
        else if (lastgroup == -1)
            lastgroup = group;
    }
}

void sh::actions::ActionsManager::_getDefaultAction_helper(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::shared_ptr<GDAHstruct> gdah)
{
    QList<std::shared_ptr<sh::actions::AbstractActionItem>> interestingList;
    foreach(auto x, actionList) {
        if (x->defaultActionPrecedence() > 0) {
            gdah->counter.doIncValue();
            interestingList.append(x);
        }
    }
    foreach(auto x, interestingList)
        sh::tools::ThreadPool::enqueueForce([=](){
            if (!x->isInitialized())
                x->initializeSync();
            if (x->enabled() && x->visible()) {
                sh::actions::SubmenuActionItem* itmSubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(x.get());
                sh::actions::ActionActionItem* itmAction = qobject_cast<sh::actions::ActionActionItem*>(x.get());
                if (itmSubmenu)
                    _getDefaultAction_helper(itmSubmenu->subitems(), gdah);
                else if (itmAction) {
                    if (x->defaultActionPrecedence() > gdah->currentValue) {
                        gdah->currentAction = x;
                        gdah->currentValue = x->defaultActionPrecedence();
                    }
                }
            }
            if (gdah->counter.doDecValue() == 0)
                gdah->callback(std::dynamic_pointer_cast<sh::actions::ActionActionItem>(gdah->currentAction));
        });
}

void sh::actions::ActionsManager::getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback)
{
    std::shared_ptr<GDAHstruct> gdah = std::make_shared<GDAHstruct>();
    gdah->callback = callback;
    gdah->initialActionList = actionList;
    _getDefaultAction_helper(actionList, gdah);
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::actions::ActionsManager::getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem> > actionList)
{
    CHECK_CORRECT_CALLER_THREAD
    int lastPrecedence = 0;
    std::shared_ptr<sh::actions::AbstractActionItem> result = std::shared_ptr<sh::actions::AbstractActionItem>(0);
    foreach( std::shared_ptr<sh::actions::AbstractActionItem> item, actionList)
    {
        if (item->visible() && item->enabled() && item->defaultActionPrecedence() > lastPrecedence)
        {
            lastPrecedence = item->defaultActionPrecedence();
            sh::actions::SubmenuActionItem* itmSubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(item.get());
            if (itmSubmenu)
                result = getDefaultAction(itmSubmenu->subitems());
            else
                result = item;
        }
    }
    return result;
}

QString sh::actions::ActionsManager::selectionLabel(QList<std::shared_ptr<sh::base::FilesystemNode>> selection)
{
    int sellen = selection.length();
    if (sellen == 1) {
        QString sellbl = selection.first()->eurl()->basename();
        if (sellbl.length() > 15)
            sellbl = sellbl.left(6) + " ⋯ " + sellbl.right(6);
        return tr("Selection '%1'").arg(sellbl);
    }
    else
        return tr("%1 selected items").arg(sellen);
}

QString sh::actions::ActionsManager::directoryLabel(std::shared_ptr<sh::base::FilesystemNode> directory)
{
    QString dirlbl = directory->eurl()->basename();
    if (dirlbl.length() > 15)
        dirlbl = dirlbl.left(6) + " ⋯ " + dirlbl.right(6);
    return tr("Directory '%1'").arg(dirlbl);
}

//#### GET ACTIONS ####

QList<std::shared_ptr<sh::actions::ActionInstantiation>> sh::actions::ActionsManager::_getActions_raw(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<sh::actions::ActionInstantiation>> result;
    // get actions from handler (if we have a common one)
    bool sameHandler = true;
    sh::base::FilesystemHandler* handler = 0;
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, instantiation->selectedNodes) {
        if (handler == 0 && sameHandler)
            handler = node->handler();
        else if (handler != node->handler())
        {
            sameHandler = false;
            break;
        }
    }
    if (sameHandler && handler)
    {
        QList<std::shared_ptr<const sh::base::Eurl>> eurls;
        foreach(std::shared_ptr<sh::base::FilesystemNode> node, instantiation->selectedNodes)
            eurls.append(node->eurl());
        {
            FORBID_ASYNC_CALLS
            QList< std::shared_ptr<sh::actions::AbstractActionItem> > actions;
            sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                actions = handler->getActions(eurls);
            });
            foreach(std::shared_ptr<sh::actions::AbstractActionItem> actionitem, actions) {
                auto hai = std::make_shared<ActionInstantiation>(*instantiation, actionitem);
                hai->category = sh::actions::ActionCategory::getByName("manage");
                auto aactionitem = std::dynamic_pointer_cast<ActionActionItem>(actionitem);
                if (aactionitem && instantiation->isLookupOnCurrentDirectoryLevel == aactionitem->shortcuthintTriggersOnCurrentDirectory())
                    hai->shortcut = aactionitem->shortcuthint();
                result.append(hai);
            }
        }
    }
    // get actions from the registered action factories
    QList<std::shared_ptr<sh::actions::ActionInstantiation>> flist;
    {
        QMutexLocker _l(&_actionfactoriesMutex);
        foreach(std::shared_ptr<sh::actions::AbstractActionFactory> actionfactory, _actionfactories) {
            auto rsituation = actionfactory->actionAvailable(instantiation.get());
            if (rsituation)
                flist.append(rsituation);
        }
    }
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        foreach(std::shared_ptr<sh::actions::ActionInstantiation> rsituation, flist)
        {
            sh::exceptions::Exception::executeGuarded([&](){
                std::shared_ptr<sh::actions::AbstractActionItem> act;
                {
                    FORBID_ASYNC_CALLS
                    rsituation->createdAction = rsituation->actionfactory->construct(rsituation->selectedNodes);
                    std::shared_ptr<sh::actions::ActionActionItem> itmAction = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(rsituation->createdAction);
                    if (rsituation->shortcut.isEmpty() && itmAction && itmAction->shortcuthintTriggersOnCurrentDirectory() == rsituation->isLookupOnCurrentDirectoryLevel)
                        rsituation->shortcut = itmAction->shortcuthint();
                }
                if (rsituation->createdAction)
                    result.append(rsituation);
            }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
        }
    });
    return result;
}

void sh::actions::ActionsManager::getActions(std::shared_ptr<sh::actions::ActionInstantiation> instantiation, std::function<void(QList<std::shared_ptr<sh::actions::ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>)> callback, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> initializedcallback)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::tools::ThreadPool::enqueueForce([=](){
        QList<std::shared_ptr<sh::actions::ActionInstantiation> > selacts = _getActions_raw(instantiation);
        auto dsituation = std::make_shared<sh::actions::ActionInstantiation>(*instantiation.get());
        dsituation->selectedNodes = {instantiation->currentDirectory};
        dsituation->isLookupOnCurrentDirectoryLevel = true;
        QList<std::shared_ptr<sh::actions::ActionInstantiation> > diracts = _getActions_raw(dsituation);
        for(auto actions: {&selacts, &diracts})
            for(std::shared_ptr<sh::actions::ActionInstantiation> action: *actions)
                action->createdAction->initializeAsync([=](){
                    sh::tools::MainThread::dispatcher()->invokeSync([=](){
                        initializedcallback(action);
                    });
                });
        callback(selacts, diracts);
    });
}

//#### SHORTCUTS ####

void sh::actions::ActionsManager::recursivelyInitializeAction(std::shared_ptr<sh::actions::ActionInstantiation> actsit, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> callback, QList<std::shared_ptr<sh::actions::ActionInstantiation>> sits)
{
    sits.append(actsit);
    sh::tools::ThreadPool::enqueueForce([=](){
        QList<std::shared_ptr<sh::actions::ActionInstantiation>> _sits = sits;  // just preserve the parent pointers until the bitter end
        std::shared_ptr<sh::actions::AbstractActionItem> action = actsit->createdAction;
        action->initializeSync();
        sh::actions::SubmenuActionItem* asubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(action.get());
        sh::actions::ActionActionItem* aaction = qobject_cast<sh::actions::ActionActionItem*>(action.get());
        if (asubmenu) {
            for(auto subitm : asubmenu->subitems()){
                auto subsit = std::make_shared<sh::actions::ActionInstantiation>(*actsit);
                subsit->createdAction = subitm;
                auto asubitm = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(subitm);
                if (asubitm && actsit->isLookupOnCurrentDirectoryLevel == asubitm->shortcuthintTriggersOnCurrentDirectory())
                    subsit->shortcut = asubitm->shortcuthint();
                recursivelyInitializeAction(subsit, callback, sits);
            }
        }
        if (aaction || asubmenu)
            callback(actsit);
    });
}

void sh::actions::ActionsManager::getActionForShortcut(QList<std::shared_ptr<sh::base::FilesystemNode>> selectedNodes, std::shared_ptr<sh::base::FilesystemNode> folderNode,
                                          int key, Qt::KeyboardModifiers modifiers,
                                          std::function<void(std::shared_ptr<sh::actions::AbstractActionItem>)> callback)
{
    CHECK_CORRECT_CALLER_THREAD
    BENCHMARK_SCOPE("ActionsManager::getActionForShortcut")
    int keyInt = key;
    if(modifiers & Qt::ShiftModifier)
        keyInt += Qt::SHIFT;
    if(modifiers & Qt::ControlModifier)
        keyInt += Qt::CTRL;
    if(modifiers & Qt::AltModifier)
        keyInt += Qt::ALT;
    if(modifiers & Qt::MetaModifier)
        keyInt += Qt::META;
    QKeySequence pressedSequence(keyInt);
    auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectedNodes, folderNode);
    getActions(instantiation, [](QList<std::shared_ptr<sh::actions::ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>){},
                                    [=](std::shared_ptr<sh::actions::ActionInstantiation> action){
        recursivelyInitializeAction(action, [=](std::shared_ptr<sh::actions::ActionInstantiation> x){
            if (qobject_cast<sh::actions::ActionActionItem*>(x->createdAction.get()) && x->shortcut.matches(pressedSequence) == QKeySequence::ExactMatch)
                callback(x->createdAction);
        });
    });
}

void sh::actions::ActionsManager::_setupGlobalShortcut(std::shared_ptr<sh::actions::AbstractActionItem> action)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::actions::ActionActionItem* _action = qobject_cast<sh::actions::ActionActionItem*>(action.get());
    QKeySequence seq = _action->shortcuthint();
    if (!seq.isEmpty())
    {
        QShortcut* shortcut = new QShortcut(seq, sh::ui::MainWindow::mainWindow());
        shortcut->setContext(Qt::WindowShortcut);
        connect(shortcut, &QShortcut::activated, this, [=](){
            _action->execute();
        });
    }
}

void sh::actions::ActionsManager::_setupGlobalShortcut_recursive(std::shared_ptr<sh::actions::AbstractActionItem> submenu)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::actions::SubmenuActionItem* _submenu = qobject_cast<sh::actions::SubmenuActionItem*>(submenu.get());
    if (_submenu) {
        foreach(std::shared_ptr<sh::actions::AbstractActionItem> aai, _submenu->subitems())
        {
            sh::actions::SubmenuActionItem* _aai_submenu = qobject_cast<sh::actions::SubmenuActionItem*>(aai.get());
            sh::actions::ActionActionItem* _aai_action = qobject_cast<sh::actions::ActionActionItem*>(aai.get());
            if (_aai_submenu)
                _setupGlobalShortcut_recursive(aai);
            else if (_aai_action)
                _setupGlobalShortcut(aai);
        }
    }
}

//#### TOOLBAR ####

void sh::actions::ActionsManager::setToolbar(QToolBar* toolbar)
{
    CHECK_CORRECT_CALLER_THREAD
    if (toolbarManagedAction)
        delete toolbarManagedAction;
    toolbarManagedAction = new ManagedAction_Toolbar(toolbar);
    this->toolbar = toolbar;
    QWidget* hspacer = new QWidget(toolbar);
    QBoxLayout* hlayout = new QHBoxLayout(hspacer);
    hlayout->addStretch(1);
    hlayout->setMargin(0);
    hspacer->setLayout(hlayout);
    QWidget* vspacer = new QWidget(toolbar);
    QBoxLayout* vlayout = new QVBoxLayout(vspacer);
    vlayout->addStretch(1);
    vlayout->setMargin(0);
    vspacer->setLayout(vlayout);
    addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionHistoryNavigateBackward>());
    addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionHistoryNavigateForward>());
    addPermanentActionToToolbar(std::make_shared<sh::actions::common::ActionBookmarks>());
    addPermanentActionToToolbar(std::make_shared<sh::actions::SeparatorActionItem>());
    for(auto category : sh::actions::ActionCategory::categories()) {
        auto agroup = std::make_shared<sh::actions::common::ActionGroups>(category);
        actionGroupActions[category->name()] = addPermanentActionToToolbar(agroup, false, true);
    }
    toolbarPermanentActions.insert(toolbar->addWidget(hspacer));
    toolbarPermanentActions.insert(toolbar->addWidget(vspacer));
    addPermanentActionToToolbar(sh::actions::viewmenu::ActionView::instance(), true);
    addPermanentActionToToolbar(sh::actions::mainmenu::ActionMain::instance(), true);
}

void sh::actions::ActionsManager::refreshSituation(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    toolbarManagedAction->set(instantiation);
}

sh::actions::ActionsManager::ManagedAction_ToolbarItem* sh::actions::ActionsManager::addPermanentActionToToolbar(std::shared_ptr<sh::actions::AbstractActionItem> item, bool rightSide, bool preventDefaultAction)
{
    item->initializeAsync([=](){
        sh::tools::MainThread::dispatcher()->invokeSync([=](){
            _setupGlobalShortcut_recursive(item);
        });
    });
    auto instantiation = std::make_shared<sh::actions::ActionInstantiation>();
    instantiation->createdAction = item;
    ManagedAction_ToolbarItem* helper = new ManagedAction_ToolbarItem(instantiation, toolbar, preventDefaultAction);
    QAction* vaction = helper->widgetaction();
    if (rightSide) {
        toolbar->removeAction(vaction);
        toolbar->addAction(vaction);
    }
    toolbarPermanentActions.insert(vaction);
    return helper;
}

//#### MENUS ####

std::shared_ptr<QMenu> sh::actions::ActionsManager::mountActionItemsInMenu(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    CHECK_CORRECT_CALLER_THREAD
    ManagedAction_Menu* helper = new ManagedAction_Menu();
    auto shelper = helper->takeStrongQMenuReference();
    sh::actions::ActionsManager::instance()->getActions(instantiation, [=](QList<std::shared_ptr<sh::actions::ActionInstantiation>> selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>> diracts){
        sh::tools::MainThread::dispatcher()->invokeAsync([=](){
            auto __shelper = shelper;  // just for not loosing the object
            QList<std::shared_ptr<sh::actions::ActionInstantiation> > _actions;
            if (!instantiation->selectedNodes.isEmpty()) {
                _actions.append(std::make_shared<sh::actions::ActionInstantiation>(*instantiation, std::make_shared<sh::actions::HeaderActionItem>(selectionLabel(instantiation->selectedNodes), "tool-pointer")));
                for(auto category : sh::actions::ActionCategory::categories()) {
                    QList<std::shared_ptr<sh::actions::ActionInstantiation> > __actions;
                    for(std::shared_ptr<sh::actions::ActionInstantiation> item : selacts)
                        if (item->category == category)
                            __actions.append(item);
                    sortAndSeparateActions(&__actions);
                    _actions.append(std::make_shared<sh::actions::ActionInstantiation>(*instantiation, std::make_shared<sh::actions::SeparatorActionItem>()));
                    _actions.append(__actions);
                }
            }
            _actions.append(std::make_shared<sh::actions::ActionInstantiation>(*instantiation, std::make_shared<sh::actions::HeaderActionItem>(directoryLabel(instantiation->currentDirectory), "folder")));
            for(auto category : sh::actions::ActionCategory::categories()) {
                QList<std::shared_ptr<sh::actions::ActionInstantiation> > __actions;
                for(std::shared_ptr<sh::actions::ActionInstantiation> item : diracts)
                    if (item->category == category)
                        __actions.append(item);
                sortAndSeparateActions(&__actions);
                _actions.append(std::make_shared<sh::actions::ActionInstantiation>(*instantiation, std::make_shared<sh::actions::SeparatorActionItem>()));
                _actions.append(__actions);
            }
            helper->setFixedActionItems(_actions);
        });
    });
    return shelper;
}

//#### HELPERS ####

//###### ManagedAction ##

void sh::actions::ActionsManager::registerManagedAction(ManagedAction *ma)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    qaction2managedaction[ma->widgetaction()] = ma;
    actionitem2managedaction[ma->actionitem()] = ma;
}

void sh::actions::ActionsManager::unregisterManagedAction(ManagedAction *ma)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    qaction2managedaction.remove(ma->widgetaction());
    actionitem2managedaction.remove(ma->actionitem());
    delete ma;
}

sh::actions::ActionsManager::ManagedAction::ManagedAction(std::shared_ptr<sh::actions::ActionInstantiation> acs) : _acs(acs)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    mgr = sh::actions::ActionsManager::instance();
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::actions::ActionsManager::ManagedAction::actionitem()
{
    return _acs ? _acs->createdAction : 0;
}

QAction* sh::actions::ActionsManager::ManagedAction::widgetaction()
{
    return _widgetaction;
}

sh::ui::ToolbarButton* sh::actions::ActionsManager::ManagedAction::button()
{
    return _button;
}

std::shared_ptr<sh::actions::ActionInstantiation> sh::actions::ActionsManager::ManagedAction::actintst()
{
    return _acs;
}

void sh::actions::ActionsManager::ManagedAction::populateUI()
{
}

sh::actions::ActionsManager::ManagedAction::~ManagedAction()
{
}

//###### ManagedAction_Toolbar ##

sh::actions::ActionsManager::ManagedAction_Toolbar::ManagedAction_Toolbar(QToolBar *toolbar)
            : sh::actions::ActionsManager::ManagedAction((std::shared_ptr<sh::actions::ActionInstantiation>)0), toolbar(toolbar)
{
}

void sh::actions::ActionsManager::ManagedAction_Toolbar::set(std::shared_ptr<sh::actions::ActionInstantiation> instantiation)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static int _requestcounter = 0;
    int _myrequest = ++_requestcounter;
    mgr->getActions(instantiation, [=](QList<std::shared_ptr<sh::actions::ActionInstantiation>> selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>> diracts){
        sh::tools::MainThread::dispatcher()->invokeSync([=]() {
            if (_myrequest != _requestcounter)
                return;
            for(auto category : sh::actions::ActionCategory::categories()) {
                auto actiongr = dynamic_cast<sh::actions::common::ActionGroups*>(mgr->actionGroupActions[category->name()]->actionitem().get());
                actiongr->setGroupActions(instantiation, selacts, diracts);
            }
        });
    });
}

sh::actions::ActionsManager::ManagedAction_Toolbar::~ManagedAction_Toolbar()
{
    foreach(QAction* a, toolbar->actions())
    {
        auto itr = mgr->qaction2managedaction.find(a);
        if (itr != mgr->qaction2managedaction.end())
            mgr->unregisterManagedAction(*itr);
    }
}

//###### ManagedAction_ToolbarItem ##

sh::actions::ActionsManager::ManagedAction_ToolbarItem::ManagedAction_ToolbarItem(std::shared_ptr<sh::actions::ActionInstantiation> acs, QToolBar *toolbar, bool preventDefaultAction)
                : sh::actions::ActionsManager::ManagedAction(acs), toolbar(toolbar), preventDefaultAction(preventDefaultAction)
{
    sh::actions::ActionActionItem* itmAction = qobject_cast<sh::actions::ActionActionItem*>(acs->createdAction.get());
    sh::actions::SeparatorActionItem* itmSeparator = qobject_cast<sh::actions::SeparatorActionItem*>(acs->createdAction.get());
    sh::actions::SubmenuActionItem* itmSubmenu = qobject_cast<sh::actions::SubmenuActionItem*>(acs->createdAction.get());
    if (itmAction || itmSubmenu)
    {
        _button = new sh::ui::ToolbarButton(toolbar);
        QString sp = sh::ui::MainWindow::mainWindow()->toolbarPosition();
        sh::ui::ToolbarButton::Position p = sh::ui::ToolbarButton::Position::TOP;
        if (sp == "N")
            p = sh::ui::ToolbarButton::Position::TOP;
        else if (sp == "S")
            p = sh::ui::ToolbarButton::Position::BOTTOM;
        else if (sp == "W")
            p = sh::ui::ToolbarButton::Position::LEFT;
        else if (sp == "E")
            p = sh::ui::ToolbarButton::Position::RIGHT;
        _button->setLocation(p);
        _widgetaction = toolbar->addWidget(_button);
    }
    else if (itmSeparator)
        _widgetaction = toolbar->addSeparator();
    set(acs);
    mgr->registerManagedAction(this);
}

void sh::actions::ActionsManager::ManagedAction_ToolbarItem::populateUI()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (menuhelper)
    {
        sh::actions::ActionActionItem* d = 0;
        if (!preventDefaultAction) {
            std::shared_ptr<sh::actions::AbstractActionItem> defaultAction = mgr->getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem> >(((sh::actions::SubmenuActionItem*)menuhelper->actionitem().get())->subitems()));
            if (defaultAction.get())
                d = qobject_cast<sh::actions::ActionActionItem*>( defaultAction.get() );
        }
        if (d)
            _button->setClickAction( [=](){ d->execute(); } );
        else
            _button->unsetClickAction();
    }
}

sh::actions::ActionsManager::ManagedAction_ToolbarItem::~ManagedAction_ToolbarItem()
{
    if (menuhelper)
        menuhelper->takeStrongQMenuReference();
    if (_button)
        delete _button;
    if (_widgetaction)
        delete _widgetaction;
}

void sh::actions::ActionsManager::ManagedAction_ToolbarItem::set(std::shared_ptr<sh::actions::ActionInstantiation> acs)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _acs = acs;
    std::shared_ptr<sh::actions::AbstractActionItem> action = acs->createdAction;
    std::shared_ptr<sh::actions::ActionActionItem> itmAction = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(action);
    std::shared_ptr<sh::actions::SubmenuActionItem> itmSubmenu = std::dynamic_pointer_cast<sh::actions::SubmenuActionItem>(action);
    if (itmSubmenu.get())
    {
        if (!menuhelper)
        {
            menuhelper = new ManagedAction_Menu(acs, this);
            _button->unsetClickAction();
            _button->setSubmenu(menuhelper->menu.lock().get());
        }
        menuhelper->setActionItem(acs);
    }
    else if (itmAction.get())
    {
        if (menuhelper) {
            delete menuhelper;
            menuhelper = 0;
            _button->setSubmenu(0);
        }
        _button->setClickAction( [=](){ itmAction->execute(); } );
    }
    mgr->actionitem2managedaction[action] = this;
    if (_button) {
        _widgetaction->setShortcut(QKeySequence::mnemonic(action->text()));
        _widgetaction->setShortcutContext(Qt::WindowShortcut);
        QObject::disconnect(_widgetaction, &QAction::triggered, _button, 0);
        QObject::connect(_widgetaction, &QAction::triggered, _button, [=](){
            _button->trigger();
        });
        QObject::connect(action.get(), &sh::actions::AbstractActionItem::changed, _button, [=](){
            mgr->_apply_baseproperties_to_button(this);
        });
        mgr->_apply_baseproperties_to_button(this);
    }
}

//###### ManagedAction_Menu ##

sh::actions::ActionsManager::ManagedAction_Menu::ManagedAction_Menu(std::shared_ptr<sh::actions::ActionInstantiation> acs, ManagedAction *companion, QString head)
        : ManagedAction(acs), _companion(companion), _head(head)
{
    menu = _strong_menu = std::make_shared<sh::actions::ActionsManager::MyContextMenu>();
    QObject::connect(_strong_menu.get(), &QObject::destroyed, mgr, [=](){
        mgr->unregisterManagedAction(this);
    });
    setActionItem(acs);
}

sh::actions::ActionsManager::ManagedAction_Menu::ManagedAction_Menu(QString head) : sh::actions::ActionsManager::ManagedAction_Menu((std::shared_ptr<sh::actions::ActionInstantiation>)0, 0, head)
{
}

std::shared_ptr<QMenu> sh::actions::ActionsManager::ManagedAction_Menu::takeStrongQMenuReference()
{
    std::shared_ptr<QMenu> result = _strong_menu;
    _strong_menu = 0;
    return result;
}

void sh::actions::ActionsManager::ManagedAction_Menu::setFixedActionItems(QList<std::shared_ptr<ActionInstantiation> > list)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    _fixedactionitems = list;
    populateUI();
}

void sh::actions::ActionsManager::ManagedAction_Menu::populateUI()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<QMenu> _menu = menu.lock();
    foreach(ManagedAction_MenuItem* ma, _items)
        mgr->unregisterManagedAction(ma);
    _items.clear();
    if (_menu->actions().empty()) {  // TODO # this is done a bit weird due to a qt bug (completely clearing a visible menu makes it disappear)
        QLabel* _buddy = new QLabel(_menu.get());
        _buddy->setText("");
        _buddy->setFixedHeight(1);
        QWidgetAction* wa = new QWidgetAction(_menu.get());
        wa->setDefaultWidget(_buddy);
        _menu->addAction(wa);
    }
    while (_menu->actions().count() > 1)
        _menu->removeAction(_menu->actions().last());
    if (!_head.isEmpty()) {
        QAction* hact = _menu->addAction(_head);
        hact->setEnabled(false);
        QFont hdfont = hact->font();
        hdfont.setItalic(true);
        hdfont.setPointSizeF(hdfont.pointSizeF()*1.1);
        hact->setFont(hdfont);
    }
    QList<std::shared_ptr<sh::actions::ActionInstantiation> > itemlist;
    std::shared_ptr<sh::actions::AbstractActionItem> _actionitem = actionitem();
    if (_actionitem.get()) {
        for(auto sitm : ((sh::actions::SubmenuActionItem*) _actionitem.get())->subitems()) {
            auto ssitm = std::make_shared<sh::actions::ActionInstantiation>(*_acs, sitm);
            auto ssa = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(sitm);
            if (ssa && _acs->isLookupOnCurrentDirectoryLevel == ssa->shortcuthintTriggersOnCurrentDirectory())
                ssitm->shortcut = ssa->shortcuthint();
            itemlist.append(ssitm);
        }
    }
    else
        itemlist = _fixedactionitems;
    QList<std::shared_ptr<sh::actions::AbstractActionItem> > aclist;
    for(auto xacs : itemlist)
        aclist.append(xacs->createdAction);
    QFont normalfont = _menu->font();
    QFont boldfont = _menu->font();
    boldfont.setBold(true);
    auto highlightdefaultaction = [=](){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        std::shared_ptr<sh::actions::AbstractActionItem> defaultAction = mgr->getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem> >(aclist));
        for(ManagedAction_MenuItem* itm: _items) {
            if (!std::dynamic_pointer_cast<sh::actions::HeaderActionItem>(itm->actintst()->createdAction)) {
                bool isdefault = itm->actintst()->createdAction == defaultAction;
                itm->widgetaction()->setFont(isdefault ? boldfont : normalfont);
            }
        }
    };
    for(std::shared_ptr<sh::actions::ActionInstantiation> acs: itemlist) {
        ManagedAction_MenuItem* helper = new ManagedAction_MenuItem(acs, _menu.get());
        _items.append(helper);
        QAction* action = helper->widgetaction();
        sh::actions::AbstractActionItem* itm = acs->createdAction.get();
        connect(itm, &sh::actions::AbstractActionItem::changed, action, [=](){
            mgr->_apply_baseproperties_to_qaction(itm, action);
            correctmenuposition();
            highlightdefaultaction();
            if (_companion)
                _companion->populateUI();
        });
        mgr->_apply_baseproperties_to_qaction(itm, action);
    }
    correctmenuposition();
    highlightdefaultaction();
    if (_companion)
        _companion->populateUI();
}

void sh::actions::ActionsManager::ManagedAction_Menu::correctmenuposition()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<QMenu> _menu = menu.lock();
    if (_origMenuPosX == -1 && _menu->isVisible())
    {
        _origMenuPosX = _menu->pos().x();
        _origMenuPosY = _menu->pos().y();
    }
    int x = _origMenuPosX;
    int y = _origMenuPosY;
    bool changed = false;
    QRect screen = QApplication::desktop()->availableGeometry(QApplication::desktop()->screenNumber(QPoint(x, y)));
    int maxX = screen.x() + screen.width();
    int maxY = screen.y() + screen.height();
    if (x + _menu->width() > maxX)
    {
        x = maxX - _menu->width();
        changed = true;
    }
    if (y + _menu->height() > maxY)
    {
        y = maxY - _menu->height();
        changed = true;
    }
    if (changed)
        _menu->move(x, y);
}

void sh::actions::ActionsManager::ManagedAction_Menu::setActionItem(std::shared_ptr<sh::actions::ActionInstantiation> acs)
{
    _acs = acs;
    if (acs.get())
        connect((sh::actions::SubmenuActionItem*) acs->createdAction.get(), &sh::actions::SubmenuActionItem::subitemsChanged, menu.lock().get(), [=](){populateUI();});
    populateUI();
}

sh::actions::ActionsManager::ManagedAction_Menu::~ManagedAction_Menu()
{
    foreach(ManagedAction_MenuItem* ma, _items)
        mgr->unregisterManagedAction(ma);
}

//###### ManagedAction_MenuItem ##

sh::actions::ActionsManager::ManagedAction_MenuItem::ManagedAction_MenuItem(std::shared_ptr<ActionInstantiation> acs, QMenu* parentmenu) : ManagedAction(acs)
{
    std::shared_ptr<sh::actions::ActionActionItem> itmAction = std::dynamic_pointer_cast<sh::actions::ActionActionItem>(acs->createdAction);
    std::shared_ptr<sh::actions::SeparatorActionItem> itmSeparator = std::dynamic_pointer_cast<sh::actions::SeparatorActionItem>(acs->createdAction);
    std::shared_ptr<sh::actions::HeaderActionItem> itmHeader = std::dynamic_pointer_cast<sh::actions::HeaderActionItem>(acs->createdAction);
    std::shared_ptr<sh::actions::SubmenuActionItem> itmSubmenu = std::dynamic_pointer_cast<sh::actions::SubmenuActionItem>(acs->createdAction);
    if (itmAction)
    {
        _widgetaction = parentmenu->addAction(sh::tools::IconManager::instance()->getIconByFullname(itmAction->icon()), itmAction->text());
        if (!acs->shortcut.isEmpty())
        {
            _widgetaction->setShortcut(acs->shortcut);
            _widgetaction->setShortcutContext(Qt::WidgetShortcut); // we just want to display that
        }
        QObject::connect(_widgetaction, SIGNAL(triggered()), mgr, SLOT(slot_actionactiontriggered()));
    }
    else if (itmSeparator)
    {
        _widgetaction = parentmenu->addSeparator();
    }
    else if (itmHeader)
    {
        parentmenu->addSeparator();
        _widgetaction = parentmenu->addAction(sh::tools::IconManager::instance()->getIconByFullname(itmHeader->icon()), itmHeader->text());
        _widgetaction->setEnabled(false);
        _widgetaction->setText("fries");
        QFont hdfont = _widgetaction->font();
        hdfont.setItalic(true);
        hdfont.setBold(true);
        hdfont.setPointSizeF(hdfont.pointSizeF()*1.1);
        _widgetaction->setFont(hdfont);
        parentmenu->addSeparator();
    }
    else if (itmSubmenu)
    {
        _managedSubmenu = new ManagedAction_Menu(acs, this);
        _widgetaction = parentmenu->addMenu(_managedSubmenu->menu.lock().get());
    }
    mgr->registerManagedAction(this);
}

sh::actions::ActionsManager::ManagedAction_MenuItem::~ManagedAction_MenuItem()
{
    if (_managedSubmenu)
        _managedSubmenu->takeStrongQMenuReference();
}

//###### MyContextMenu ##

sh::actions::ActionsManager::MyContextMenu::MyContextMenu(QWidget* parent) : QMenu(parent)
{
    QColor c = sh::configuration::ConfigurationManager::instance()->style()->brandingColor();
    brandingcolor = QColor::fromHsl(c.hue(), c.saturation(), 30);
}

void sh::actions::ActionsManager::MyContextMenu::paintEvent(QPaintEvent *e)
{
    QMenu::paintEvent(e);
    QPainter painter(this);
    QList<QAction*> lstactions = actions();
    int actioncount = lstactions.count();
    for (int iaction=0; iaction<actioncount; iaction++) {
        QAction* action = lstactions[iaction];
        if (action->font().italic()) {  // it's a header
            QRect r = this->actionGeometry(action);
            int iactiona = iaction;
            QRect ra;
            while (ra.isNull()) {
                iactiona++;
                QAction* actiona = lstactions.value(iactiona);
                if (!actiona || !actiona->isSeparator())
                    break;  // just to be sure
                ra = this->actionGeometry(actiona);
            }
            int iactionb = iaction;
            QRect rb;
            while (rb.isEmpty()) {
                iactionb--;
                QAction* actionb = lstactions.value(iactionb);
                if (!actionb || !actionb->isSeparator())
                    break;  // just to be sure
                rb = this->actionGeometry(actionb);
            }
            r.adjust(0, -rb.height()/(rb.top()>5?2:1), 0, ra.height()/2);
            QImage back(r.size(), QImage::Format_ARGB32_Premultiplied);
            back.fill(brandingcolor);
            painter.setCompositionMode(QPainter::RasterOp_NotDestination);
            painter.drawImage(r.x(), r.y(), back);
            painter.setCompositionMode(QPainter::CompositionMode_Screen);
            painter.drawImage(r.x(), r.y(), back);
        }
    }
}
