/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef ACTIONSMANAGER_H
#define ACTIONSMANAGER_H

#include <functional>
#include <limits>
#include <memory>
#include <QMap>
#include <QMutex>
#include <QSet>
#include <QToolBar>
#include <QMenu>
#include "actions/actionfactory.h"
#include "tools/atomiccounter.h"


namespace sh{
    namespace base{class FilesystemNode;}
    namespace ui {class ToolbarButton;}

    namespace actions{
        class AbstractActionItem;
        class ActionActionItem;
        class SubmenuActionItem;
        class AbstractActionFactory;

        /** \brief
         * Factory for actions which are valid for a given list of FileSystemNode and tools for placing actions in some gui elements.
         */
        class ActionsManager : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static ActionsManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** \brief
             * Sets the toolbar to be managed by the action manager.
             */
            void setToolbar(QToolBar* toolbar);

            /** \brief
             * Asychonously gets actions for a list of sh::base::FilesystemNode.
             * \param nodes The list of nodes.
             * \param callback This callback is called once the list is fetched.
             * \param initializedcallback This callback is called for each result action after initialization.
             */
            void getActions(std::shared_ptr<ActionInstantiation> instantiation, std::function<void(QList<std::shared_ptr<ActionInstantiation>>, QList<std::shared_ptr<sh::actions::ActionInstantiation>>)> callback, std::function<void(std::shared_ptr<ActionInstantiation>)> initializedcallback = [](std::shared_ptr<ActionInstantiation>){});

            /** \brief
             * Registers an action factory.
             *
             * This makes an action implementation available in the toolbar and context menus.
             */
            void registerActionFactory(std::shared_ptr<sh::actions::AbstractActionFactory> actionFactory);

            /** \brief
             * Asychonously gets the action for a shortcut.
             * \param selectedNodes The nodes which are currently selected in the active fileview.
             * \param folderNode The folder which is shown by the active fileview.
             * \param key The keypress character code.
             * \param modifiers The keypress modifiers.
             * \param callback This callback is called when an action was found (with an already initialized action).
             */
            void getActionForShortcut(QList<std::shared_ptr<base::FilesystemNode> > selectedNodes, std::shared_ptr<base::FilesystemNode> folderNode,
                                      int key, Qt::KeyboardModifiers modifiers, std::function<void(std::shared_ptr<sh::actions::AbstractActionItem>)> callback);

            /** \brief
             * Asynchronously mounts exacly the actions for a given list of nodes into the toolbar. Removes all the other ones.
             */
            void refreshSituation(std::shared_ptr<ActionInstantiation> instantiation);

            /** \brief
             * Asynchronously mounts exacly the actions for a given list of nodes into a QMenu and returns it.
             */
            std::shared_ptr<QMenu> mountActionItemsInMenu(std::shared_ptr<ActionInstantiation> instantiation);

            /** \brief
             * Returns a short header label text for a node selection.
             */
            static QString selectionLabel(QList<std::shared_ptr<sh::base::FilesystemNode>> selection);
            /** \brief
             * Returns a short header label text for a current directory.
             */
            static QString directoryLabel(std::shared_ptr<sh::base::FilesystemNode> directory);
            /** \brief
             * Sorts a lists of actionitems and adds separators between groups. Modifies the given list.
             */
            static void sortAndSeparateActions(QList<std::shared_ptr<ActionInstantiation> > *actions);

            /** \brief
             * Determines the default action for a given list of actions.
             * \param actionList The list of all available actions.
             * \param callback This callback is called when the default action was found (with an already initialized action).
             */
            void getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback);

            virtual ~ActionsManager();

        private slots:
            /** \brief
             * Is triggered when a action execution is triggered from the ui (e.g. by a click on a button).
             */
            void slot_actionactiontriggered();

        private:
            explicit ActionsManager(QObject *parent = 0);
            static bool _isdeleted;
            class ManagedAction;

            /** \brief
             * Helper for getActions and others.
             */
            QList<std::shared_ptr<ActionInstantiation> > _getActions_raw(std::shared_ptr<ActionInstantiation> instantiation);
            /** \brief
             * Recursively initializes a list of action items and calls a callback for each action item recursively afterwards.
             * \param actions List of all action items to traverse.
             * \param callback This callback is called for each action item (even indirectly in a subtree hierarchy).
             */
            void recursivelyInitializeAction(std::shared_ptr<sh::actions::ActionInstantiation> actsit, std::function<void(std::shared_ptr<sh::actions::ActionInstantiation>)> callback, QList<std::shared_ptr<sh::actions::ActionInstantiation> > sits = {});
            /** \brief
             * Applies basic properties to a button (name, enabled state, icon, ...).
             */
            void _apply_baseproperties_to_button(ManagedAction* managedaction);
            /** \brief
             * Applies basic properties to a qaction (name, enabled state, icon, ...).
             */
            void _apply_baseproperties_to_qaction(sh::actions::AbstractActionItem* itm, QAction* _widgetaction);
            /** \brief
             * Sets a global shortcut for a (typically permanent) action item.
             */
            void _setupGlobalShortcut(std::shared_ptr<sh::actions::AbstractActionItem> action);
            /** \brief
             * Sets global shortcuts for (typically permanent) action items.
             * It iterates one submenu recursively.
             */
            void _setupGlobalShortcut_recursive(std::shared_ptr<sh::actions::AbstractActionItem> submenu);
            /** \brief
             * Determines the default action for a given list of actions.
             *
             * Note: This does not initialize the actions!
             */
            std::shared_ptr<sh::actions::AbstractActionItem> getDefaultAction(QList<std::shared_ptr<sh::actions::AbstractActionItem> > actionList);

            /** \brief
             * Internal data structure used for _getDefaultAction_helper();
             */
            struct GDAHstruct
            {
                GDAHstruct() = default;
                GDAHstruct(const GDAHstruct& o) = default;
                sh::tools::AtomicCounter counter;
                std::function<void(std::shared_ptr<sh::actions::ActionActionItem>)> callback;
                int currentValue = 0;
                QList<std::shared_ptr<sh::actions::AbstractActionItem> > initialActionList;  // just for preserving all the parents until the bitter end
                std::shared_ptr<sh::actions::AbstractActionItem> currentAction = 0;
            };

            /** \brief
             * Internal helper for the public getDefaultAction() method.
             */
            void _getDefaultAction_helper(QList<std::shared_ptr<sh::actions::AbstractActionItem>> actionList, std::shared_ptr<GDAHstruct> gdah);


            class ManagedAction_Toolbar;
            class ManagedAction_ToolbarItem;
            class ManagedAction_Menu;
            class ManagedAction_MenuItem;

            /** \brief
             * Adds an action item to the toolbar permanently. Unlike the typical ones, it will
             * not be deleted under typical circumstances.
             */
            ManagedAction_ToolbarItem* addPermanentActionToToolbar(std::shared_ptr<sh::actions::AbstractActionItem> item, bool rightSide=false, bool preventDefaultAction=false);

            QList<std::shared_ptr<sh::actions::AbstractActionFactory>> _actionfactories;
            QMutex _actionfactoriesMutex;
            QSet<QAction*> toolbarPermanentActions;
            QMap<QString, ManagedAction_ToolbarItem*> actionGroupActions;
            QToolBar* toolbar = 0;

            /** \brief
             * A managed action exists for each action item which is presented by the action manager somehow (toolbar, menu, ...).
             * It contains basic information about the action. It may also control the lifecycle for various other objects (depending on the subclass).
             */
            class ManagedAction
            {
            public:
                ManagedAction(std::shared_ptr<ActionInstantiation> acs);
                virtual ~ManagedAction();
                /** \brief
                 * Gets the associated action item.
                 */
                std::shared_ptr<sh::actions::AbstractActionItem> actionitem();
                /** \brief
                 * Gets the associated qaction.
                 */
                QAction* widgetaction();
                /** \brief
                 * Gets the associated toolbar button (if any).
                 */
                sh::ui::ToolbarButton* button();
                /** \brief
                 * Gets the associated ActionInstantiation.
                 */
                std::shared_ptr<ActionInstantiation> actintst();
                /** \brief
                 * Updates the ui.
                 */
                virtual void populateUI();
            protected:
                std::shared_ptr<ActionInstantiation> _acs;
                QAction* _widgetaction = 0;
                sh::ui::ToolbarButton* _button = 0;
                ActionsManager* mgr;
            };

            class ManagedAction_Toolbar : public ManagedAction
            {
            public:
                /** \brief
                 * Managed action for the entire toolbar. Exists only once for the complete process lifetime.
                 */
                ManagedAction_Toolbar(QToolBar* toolbar);
                /** \brief
                 * Asynchonously populates the toolbar with all the actions (when visiblle in toolbar) for the given node list.
                 */
                void set(std::shared_ptr<ActionInstantiation> instantiation);
                virtual ~ManagedAction_Toolbar();
            protected:
                QToolBar* toolbar;
                QMap<QString, ManagedAction*> keyname2managedaction;
            };

            class ManagedAction_ToolbarItem : public ManagedAction
            {
            public:
                /** \brief
                 * Managed action for an item in a toolbar (e.g. button, separator).
                 * Lifetime: This instance is deleted when the toolbar gets a new content list. On deletion, it also deletes ui representation and the submenu structure (if any).
                 * \param actionitem The actionitem for the new toolbar item.
                 */
                ManagedAction_ToolbarItem(std::shared_ptr<ActionInstantiation> actionitem, QToolBar* toolbar, bool preventDefaultAction);
                /** \brief
                 * Sets this managed action to host a new action item from now on.
                 * Use with care and only if you know what you do.
                 */
                void set(std::shared_ptr<ActionInstantiation> acs);
                virtual ~ManagedAction_ToolbarItem();
                /** \brief
                 * Populates the ui menu.
                 */
                void populateUI();
            private:
                ManagedAction_Menu* menuhelper = 0;
                QToolBar* toolbar;
                bool preventDefaultAction;
            };

            class ManagedAction_Menu : public ManagedAction
            {
            public:
                /** \brief
                 * Managed action for a menu.
                 * Lifetime: This instance becomes deleted when the ui menu (QMenu) becomes unused. On deletion, it deletes the directly contained managed handlers as well.
                 * \param actionitem The submenu actionitem as content the new menu.
                 * \param companion The corresponding managed action which controls the 'front part' of the submenu (the button or menu item which own the submenu)
                 */
                ManagedAction_Menu(std::shared_ptr<ActionInstantiation> acs, ManagedAction* companion = 0, QString head = QString());
                /** \brief
                 * Managed action for a menu.
                 * Lifetime: This instance becomes deleted when the ui menu (QMenu) becomes unused. On deletion, it deletes the directly contained managed handlers as well.
                 * This creates a menu for a fixed list of action items.
                 * See setFixedActionItems.
                 */
                ManagedAction_Menu(QString head = QString());
                /** \brief
                 * Sets the list of action items as content for this menu.
                 * Use this for instances created with ManagedAction_Menu().
                 */
                void setFixedActionItems(QList<std::shared_ptr<sh::actions::ActionInstantiation> > list);
                virtual ~ManagedAction_Menu();
                std::weak_ptr<QMenu> menu;
                std::shared_ptr<QMenu> _strong_menu = 0;
                /** \brief
                 * Takes a shared_ptr to the QMenu and resets it internally.
                 * From the time of calling, you are the only owner of a (strong) reference.
                 */
                std::shared_ptr<QMenu> takeStrongQMenuReference();
                /** \brief
                 * Polulates the ui menu.
                 */
                void populateUI();
                /** \brief
                 * Sets this managed action to refer to another action item.
                 */
                void setActionItem(std::shared_ptr<ActionInstantiation> acs);
            protected:
                ManagedAction* _companion;
                QList<ManagedAction_MenuItem*> _items;
                QList<std::shared_ptr<sh::actions::ActionInstantiation> > _fixedactionitems;
                void correctmenuposition();
                int _origMenuPosX = -1, _origMenuPosY = -1;
                QString _head;
            };

            class ManagedAction_MenuItem : public ManagedAction
            {
            public:
                /** \brief
                 * Managed action for a menu item.
                 * Lifetime: This instance becomes deleted when the ui menu (QMenu) becomes unused.
                 * \param actionitem The actionitem associated with this menu item.
                 * \param parentmenu The parent in ui.
                 */
                ManagedAction_MenuItem(std::shared_ptr<sh::actions::ActionInstantiation> acs, QMenu* parentmenu);
                virtual ~ManagedAction_MenuItem();
            protected:
                ManagedAction_Menu* _managedSubmenu = 0;
            };

            class MyContextMenu : public QMenu
            {
            public:
                /** \brief
                 * A special QMenu implementation which paints slightly differently (e.g. other header styles).
                 */
                explicit MyContextMenu(QWidget* parent=0);
            protected:
                void paintEvent(QPaintEvent *e) override;
                QColor brandingcolor;
            };

            ManagedAction_Toolbar* toolbarManagedAction = 0;

            QMap<QAction*, ManagedAction*> qaction2managedaction;
            QMap<std::shared_ptr<sh::actions::AbstractActionItem>, ManagedAction*> actionitem2managedaction;

            /** \brief
             * Registers a managed action in some data structures. This increments the reference count at some places (enforcing the designed lifetimes)
             */
            void registerManagedAction(ManagedAction* ma);
            /** \brief
             * Unregisters a managed action in some data structures. This decrements the reference count at some places (enforcing the designed lifetimes) and deletes the managed handler itself.
             */
            void unregisterManagedAction(ManagedAction *ma);
        };

    }
}

#endif // ACTIONSMANAGER_H
