/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <memory>
#include <QStack>
#include "actions/separatoractionitem.h"
#include "actions/actionsmanager.h"
#include "actions/actionfactory.h"
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "base/eurl.h"
#include "base/filesystemnode.h"
#include "tools/threaddispatcher.h"
#include "tools/bookmarkmanager.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "ui/managebookmarksdialog.h"
#include "ui/filesystempanel.h"
#include "actionbookmarks.h"


namespace sh{ namespace actions{ namespace common{INITIALIZE_STATIC_IN_MAINTHREAD(10000, ActionAddToBookmarks)}}}

void sh::actions::common::ActionAddToBookmarks::_initialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionAddToBookmarks>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-100000),
    })));
}

void sh::actions::common::ActionAddToBookmarks::_shutdown()
{
}

sh::actions::common::ActionBookmarks::ActionBookmarks()
        : sh::actions::common::ActionBookmarkFolder("&M")
{
    setIcon("bookmarks-organize");
    setVisible(sh::tools::BookmarkManager::instance()->hasBookmarks());

    connect(sh::tools::BookmarkManager::instance(), &sh::tools::BookmarkManager::changed, this, [=](){
        sh::tools::MainThread::dispatcher()->invokeAsync([=](){
            initialize();
        });
    });
}

void sh::actions::common::ActionBookmarks::initialize()
{
    setVisible(sh::tools::BookmarkManager::instance()->hasBookmarks());
    _subs.clear();
    if (sh::tools::BookmarkManager::instance()->hasBookmarks())
    {
        QList<std::shared_ptr<sh::actions::AbstractActionItem>> subitems;
        QList<std::shared_ptr<sh::tools::Bookmark>> bookmarks = sh::tools::BookmarkManager::instance()->getBookmarks();
        QHash<QString, std::shared_ptr<sh::actions::common::ActionBookmarkFolder>> items;
        items[""] = std::dynamic_pointer_cast<sh::actions::common::ActionBookmarkFolder>(this->shared_from_this());
        QStringList folders;
        foreach(std::shared_ptr<sh::tools::Bookmark> bookmark, bookmarks)
            folders.append(bookmark->folder().join('/'));
        qSort(folders.begin(), folders.end());
        foreach(QString folder, folders)
        {
            if (!folder.isEmpty())
            {
                std::shared_ptr<sh::actions::common::ActionBookmarkFolder> parent = std::dynamic_pointer_cast<sh::actions::common::ActionBookmarkFolder>(this->shared_from_this());
                QString fsgm;
                foreach(QString sgm, folder.split('/'))
                {
                    if (!fsgm.isEmpty())
                        fsgm += "/";
                    fsgm += sgm;
                    auto itr = items.find(fsgm);
                    if (itr != items.end())
                    {
                        parent = *itr;
                    }
                    else
                    {
                        std::shared_ptr<sh::actions::common::ActionBookmarkFolder> newmenu = std::make_shared<sh::actions::common::ActionBookmarkFolder>(sgm);
                        items[fsgm] = newmenu;
                        parent->addSubitem(newmenu);
                        parent = newmenu;
                    }
                }
            }
        }
        foreach(std::shared_ptr<sh::tools::Bookmark> bookmark, bookmarks)
        {
            std::shared_ptr<sh::actions::AbstractActionItem> act = std::make_shared<sh::actions::common::ActionBookmark>(bookmark->label().isEmpty()?bookmark->eurl()->asString():bookmark->label(), bookmark->eurl());
            items[bookmark->folder().join('/')]->addSubitem(act);
        }
        foreach(std::shared_ptr<sh::actions::AbstractActionItem> sub, _subs)
            subitems.append(sub);
        subitems.append(std::make_shared<sh::actions::SeparatorActionItem>());
        subitems.append(std::make_shared<sh::actions::common::ActionManageBookmarks>());
        setSubitems(subitems);
    }
}

sh::actions::common::ActionBookmark::ActionBookmark(QString label, std::shared_ptr<const sh::base::Eurl> eurl)
        : sh::actions::ActionActionItem(label), eurl(eurl)
{
}

void sh::actions::common::ActionBookmark::action(sh::actions::ActionExecutionInfo*)
{
    IGNORE_FORBIDDEN_ASYNC_CALLS
    sh::ui::MainWindow::mainWindow()->jumpToEurl(eurl);
}

sh::actions::common::ActionBookmarkFolder::ActionBookmarkFolder(QString label)
        : sh::actions::SubmenuActionItem(label, QList<std::shared_ptr<sh::actions::AbstractActionItem>>(), true, "folder-bookmark")
{
}

void sh::actions::common::ActionBookmarkFolder::addSubitem(std::shared_ptr<sh::actions::AbstractActionItem> item)
{
    _subs.append(item);
}

void sh::actions::common::ActionBookmarkFolder::clearSubitems()
{
    _subs.clear();
}

void sh::actions::common::ActionBookmarkFolder::initialize()
{
    setSubitems(_subs);
}

sh::actions::common::ActionManageBookmarks::ActionManageBookmarks()
        : sh::actions::ActionActionItem(tr("Manage bookmarks..."), true, "configure")
{
}

void sh::actions::common::ActionManageBookmarks::action(sh::actions::ActionExecutionInfo* info)
{
    info->setVisualProcessFeedbackActive(false);
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        sh::ui::ManageBookmarksDialog* dlg = new sh::ui::ManageBookmarksDialog(sh::ui::MainWindow::mainWindow());
        connect(dlg, &QDialog::finished, [=](){
            dlg->deleteLater();
        });
        dlg->show();
    });
}

sh::actions::common::ActionAddToBookmarks::ActionAddToBookmarks(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
        : sh::actions::ActionActionItem(tr("Add to bookmarks"), true, QString())
        , nodes(nodes)
{
}

void sh::actions::common::ActionAddToBookmarks::action(sh::actions::ActionExecutionInfo*)
{
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodes)
        sh::tools::BookmarkManager::instance()->addBookmark(QStringList(), node->displayName(), node->eurl());
}
