/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QBuffer>
#include "actions/actionexecutioninfo.h"
#include "actions/actionsmanager.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "base/filesystemhandler.h"
#include "base/filesystemoperation.h"
#include "base/filesystemnode.h"
#include "exceptions/runtimeexception.h"
#include "tools/iconmanager.h"
#include "tools/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "actioncreatefile.h"


namespace sh{ namespace actions{ namespace common{INITIALIZE_STATIC_IN_MAINTHREAD(10000, ActionCreateFile)}}}

void sh::actions::common::ActionCreateFile::_initialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionCreateFile>>(
                    sh::actions::ActionCategory::getByName("create"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::HideOnSelectionLevelPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-210000),
    })));
}

void sh::actions::common::ActionCreateFile::_shutdown()
{
}

sh::actions::common::ActionCreateFile::ActionCreateFile(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem( tr("Create file..."), true, "text-x-generic+list-add")
{
    _nodes = nodes;
}

void sh::actions::common::ActionCreateFile::initialize()
{
    sh::base::Operation op;
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, _nodes)
    {
        if (!node->handler() || !node->handler()->canCreateFile(&op, node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionCreateFile::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead("Creating file...");
    int i=0;
    info->setVisualProcessFeedbackActive(true);
    bool dobreak = false;
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, _nodes)
    {
        info->setProgress(i++, _nodes.length(), "");
        sh::exceptions::Exception::executeGuarded([&](){
            QString newname = info->userfeedback()->simpleInputBox(tr("Please enter the name for the new file."),
                                                                   tr("New File"));
            if (!newname.isNull())
            {
                sh::base::Eurl::filenameCheck(newname);
                std::shared_ptr<const sh::base::Eurl> neweurl = node->eurl()->withAppendedSegment(newname);
                if (node->handler()->getType(info->operation(), neweurl) != sh::base::FilesystemNodeType::NONE)
                    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("There already is an item with that name.")));
                QBuffer emptycontent;
                emptycontent.open(QFile::ReadOnly);
                info->operation()->filesystem()->createFile(neweurl, &emptycontent);
            }
            else
                dobreak = true;
        }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
        if (dobreak)
            break;
    }
}
