/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "base/filesystemoperation.h"
#include "base/filesystemhandler.h"
#include "base/filesystemnode.h"
#include "exceptions/permissiondeniedexception.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "actionrenameitem.h"


namespace sh{ namespace actions{ namespace common{INITIALIZE_STATIC_IN_MAINTHREAD(10000, ActionRenameItem)}}}

void sh::actions::common::ActionRenameItem::_initialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::actions::common::ActionRenameItem>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<Predicate>>({
                    std::make_shared<sh::actions::DontResolveLinksPredicate>(),
                    std::make_shared<sh::actions::HideOnCurrentDirectoryLevelPredicate>(),
                    std::make_shared<sh::actions::KeyShortcutPredicate>(QKeySequence("F2"), false),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-300000)
    })));
}

void sh::actions::common::ActionRenameItem::_shutdown()
{
}

sh::actions::common::ActionRenameItem::ActionRenameItem(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
    :sh::actions::ActionActionItem(tr("Rename..."))
{
    _nodes = nodes;
}

void sh::actions::common::ActionRenameItem::initialize()
{
    sh::base::Operation op;
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, _nodes)
    {
        if (!op.filesystem()->canMoveItem(node->eurl()))
        {
            setEnabled(false);
            break;
        }
    }
}

void sh::actions::common::ActionRenameItem::action(sh::actions::ActionExecutionInfo* info)
{
    int i = 0;
    info->setHead(tr("Renaming items..."));
    bool dobreak = false;
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, _nodes)
    {
        info->setProgress(i, _nodes.length(), tr("%1 of %2").arg(i).arg(_nodes.length()));
        i++;
        sh::exceptions::Exception::executeGuarded([&](){
            QString oldname;
            sh::tools::MainThread::dispatcher()->invokeSync([&]() {
                oldname = node->eurl()->basename();
            });
            QString newname = info->userfeedback()->simpleInputBox(tr("Please enter the new name."), oldname);
            if (!newname.isNull())
            {
                sh::base::Eurl::filenameCheck(newname);
                info->setVisualProcessFeedbackActive(true);
                std::shared_ptr<const sh::base::Eurl> nodeeurl = node->eurl();
                std::shared_ptr<const sh::base::Eurl> desteurl = nodeeurl->parentSegment()->withAppendedSegment(newname);
                if (node->handler()->getType(info->operation(), desteurl) != sh::base::FilesystemNodeType::NONE)
                    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("There already is an item with that name.")));
                info->setDetails(tr("renaming"), nodeeurl->asString(), tr("to"), desteurl->asString());
                info->operation()->filesystem()->moveItem(nodeeurl, desteurl, std::make_shared<sh::base::FilesystemOperationProgressMonitor>(info));
            }
            else
                dobreak = true;
        }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
        if (dobreak)
            break;
    }
}
