/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/filesystemnode.h"
#include "base/filesystemhandler.h"
#include "base/operation.h"
#include "exceptions/programexception.h"
#include "tools/benchmarking.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "ui/mainwindow.h"
#include "detailcolumn.h"


const uint sh::base::DetailColumn::DISPLAYINDEX_CORE = 10000;
const uint sh::base::DetailColumn::DISPLAYINDEX_INTERESTING = 20000;
const uint sh::base::DetailColumn::DISPLAYINDEX_EXOTIC = 30000;

sh::base::DetailColumn::DetailColumn(QString name, QString displayName, uint displayIndex, bool sort_doTypediff, int defaultWidth, bool isRightAligned)
    :_displayName(displayName), _name(name), _displayIndex(displayIndex), _sort_doTypediff(sort_doTypediff), _defaultWidth((defaultWidth==-1)?140:defaultWidth), _isRightAligned(isRightAligned)
{
}

QVariant sh::base::DetailColumn::value(std::shared_ptr<sh::base::FilesystemNode> node, bool ignoreAged)
{
    QMutexLocker l1(&sh::base::FilesystemModel::instance()->_nodeDataMutex);
    auto itr = node->_detailvalues.find(this);
    if (itr != node->_detailvalues.end())
        return *itr;
    if (!ignoreAged)
    {
        auto itr2 = node->_last_detailvalues.find(this);
        if (itr2 != node->_last_detailvalues.end())
            return *itr2;
    }
    return VALUE_UNAVAILABLE();
}

bool sh::base::DetailColumn::isValueAvailable(std::shared_ptr<sh::base::FilesystemNode> node)
{
    return value(node) != VALUE_UNAVAILABLE();
}

QString sh::base::DetailColumn::displayValue(std::shared_ptr<sh::base::FilesystemNode> node, const sh::base::FilesystemModelFileviewProxy*)
{
    return value(node).toString();
}

bool sh::base::DetailColumn::isVisible()
{
    return true;
}

QVariant sh::base::DetailColumn::computeValue(std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation* op)
{
    BENCHMARK_SCOPE("DetailColumn::computeValue")
    return determineValue(node, op);
}

void sh::base::DetailColumn::applyValueByEurl(std::shared_ptr<const sh::base::Eurl> eurl, sh::base::Operation* op, QVariant value)
{
    applyValueByNode(FilesystemModel::instance()->findNodesForEurl(eurl).first(), op, value);
}

void sh::base::DetailColumn::applyValueByNode(std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation* op, QVariant value)
{
    applyValueByEurl(node->eurl(), op, value);
}

QVariant sh::base::DetailColumn::requestValue(std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation* op, bool withNodeValues, bool withOperationsCache)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    BENCHMARK_SCOPE("DetailColumn::requestValue")

    // we can reuse an existing Operation or get a fresh one here
    sh::base::Operation* _op;
    if (op == 0)
        _op = new sh::base::Operation();
    else
        _op = op;

    QVariant v;

    // do we have something cached?
    if (_op->isDetailsCacheEnabled() && withOperationsCache)
        v = _op->getDetailFromCache(node->eurl(), this);
    else
        v = VALUE_UNAVAILABLE();

    if (v == VALUE_UNAVAILABLE() && withNodeValues)
        v = value(node, true);

    sh::exceptions::Exception exception;

    // call the retrieval logic
    if (v == VALUE_UNAVAILABLE())
    {
        try
        {
            v = computeValue(node, _op);
        }
        catch(sh::exceptions::Exception& e)
        {
            exception = e;
        }
    }

    // cache the value
    if (_op->isDetailsCacheEnabled())
        _op->storeDetailInCache(node->eurl(), this, v);

    // put the value into the storage
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        node->setDetail(this->shared_from_this(), v);
    });

    // remove a freshly created Operation
    if (op == 0)
    {
        _op->abort();
        delete _op;
    }

    if (!exception.data().isEmpty())
        throw exception;

    return v;
}

bool sh::base::DetailColumn::sort_doTypediff()
{
    return _sort_doTypediff;
}

int sh::base::DetailColumn::sort_order(std::shared_ptr<sh::base::FilesystemNode> n1, std::shared_ptr<sh::base::FilesystemNode> n2)
{
    QVariant v1 = value(n1);
    QVariant v2 = value(n2);
    if (v1 != v2)
        return v1<v2?-1:1;
    else
        return 0;
}

uint sh::base::DetailColumn::displayIndex()
{
    return _displayIndex;
}

int sh::base::DetailColumn::defaultWidth()
{
    return _defaultWidth;
}

bool sh::base::DetailColumn::isRightAligned()
{
    return _isRightAligned;
}

void sh::base::DetailColumn::deleteLater()
{
    delete this;
}

void sh::base::DetailColumn::registerKnownDetailColumn(QString name, std::shared_ptr<DetailColumn> column)
{
    QMutexLocker _l1(&_knownDetailColumnsMutex);
    _knownDetailColumns[name] = column;
}

std::shared_ptr<sh::base::DetailColumn> sh::base::DetailColumn::findKnownDetailColumn(QString name)
{
    QMutexLocker _l1(&_knownDetailColumnsMutex);
    return _knownDetailColumns.value(name);
}

QMap<QString, std::shared_ptr<sh::base::DetailColumn>> sh::base::DetailColumn::_knownDetailColumns;

QMutex sh::base::DetailColumn::_knownDetailColumnsMutex;
