/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMutex>
#include <QObject>
#include <QRegularExpression>
#include "exceptions/argumentexception.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "tools/singletoninitializer.h"
#include "tools/logger.h"
#include "tools/benchmarking.h"
#include "eurl.h"


const QChar sh::base::Eurl::WRAPPER_BEGIN = '[';

const QChar sh::base::Eurl::WRAPPER_END   = ']';

const QString sh::base::Eurl::FORBIDDEN_FILENAME_CHARACTERS = QString("/")
#ifdef Q_OS_WIN
        + "\\"
#endif
        ;

namespace sh{namespace base{INITIALIZE_STATIC_IN_MAINTHREAD(10300, Eurl)}}

void sh::base::Eurl::_initialize()
{
#ifdef BENCHMARKING
    sh::tools::Benchmarking::instance()->addReporter([=](){
        QMap<QString,QString> result;
        QMutexLocker _locker(&_mutex);
        result.insert("Eurl count", QString("%1").arg(_eurluniverse.count()));
        return result;
    });
#endif
}

void sh::base::Eurl::_shutdown()
{
}

sh::base::EurlMisformattedException::EurlMisformattedException(sh::exceptions::ExceptionData data)
            : sh::exceptions::ArgumentException(data.name("EurlMisformattedException").message(QObject::tr("The address is invalid.")).detailsAreInteresting())
{
}

QString sh::base::Eurl::check_scheme(QString scheme)
{
    BENCHMARK_SCOPE("Eurl checks")
    if (scheme.contains("/"))
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Invalid scheme '%1'.").arg(scheme)));
    return scheme;
}

QString sh::base::Eurl::check_inner(QString inner)
{
    BENCHMARK_SCOPE("Eurl checks")
    if (inner.isEmpty())
        return "";
    if (!inner.startsWith(WRAPPER_BEGIN))
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Invalid inner part '%1'.").arg(inner)));
    if (!inner.endsWith(WRAPPER_END))
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Invalid inner part '%1'.").arg(inner)));
    return WRAPPER_BEGIN+Eurl::fromString(inner.mid(1,inner.length()-2))->asString()+WRAPPER_END;
}

QString sh::base::Eurl::check_hostname(QString hostname)
{
    BENCHMARK_SCOPE("Eurl checks")
    if (hostname.contains("/"))
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Invalid hostname '%1'.").arg(hostname)));
    return hostname;
}

QString sh::base::Eurl::check_path(QString path)
{
    BENCHMARK_SCOPE("Eurl checks")
    QStringList result;
#ifdef Q_OS_WIN
    path = path.replace("\\", "/");
#endif
    foreach(QString splitted, path.split("/"))
    {
        if (splitted.isEmpty())
            continue;
        else if (splitted==".")
            continue;
        else if (splitted=="..")
        {
            if (!result.isEmpty())
                result.removeLast();
        }
        else
            result.append(check_filename(splitted));
    }
    return "/" + result.join("/");
}

QString sh::base::Eurl::check_filename(QString filename)
{
    BENCHMARK_SCOPE("Eurl checks")
    filenameCheck(filename);
    return filename;
}

QHash<QChar, QString> sh::base::Eurl::_escapemap;

QMutex sh::base::Eurl::_escapemapmutex;

QString sh::base::Eurl::_escape(QString s)
{
    BENCHMARK_SCOPE("Eurl escape")
    QMutexLocker _l1(&_escapemapmutex);
    if (_escapemap.isEmpty())
    {
        _escapemap[Eurl::WRAPPER_BEGIN] = "%" + QString::number(Eurl::WRAPPER_BEGIN.cell(), 16);
        _escapemap[Eurl::WRAPPER_END] = "%" + QString::number(Eurl::WRAPPER_END.cell(), 16);
        _escapemap['%'] = "%" + QString::number('%', 16);
    }
    QStringList result;
    foreach(QChar c, s)
    {
        auto itr = _escapemap.find(c);
        if (itr != _escapemap.end())
            result << (*itr);
        else
            result << c;
    }
    return result.join("");
}

QString sh::base::Eurl::_unescape(QString s)
{
    BENCHMARK_SCOPE("Eurl escape")
    QByteArray bs = s.toUtf8();
    return QUrl::fromPercentEncoding(bs);
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::fromString(QString eurlstring)
{
    BENCHMARK_SCOPE("Eurl creation")
    QString innerpart;
    QString outerpart;
    QString leftouterpart;
    QString rightouterpart_host;
    QString rightouterpart_path;
    {// determine innerpart and outerpart
        int i1 = eurlstring.indexOf(WRAPPER_BEGIN);
        if (i1 != -1)
        {
            int i2 = eurlstring.lastIndexOf(WRAPPER_END);
            if (i2 == -1)
                throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Missing '%1'.").arg(WRAPPER_END)));
            innerpart = eurlstring.mid(i1, i2-i1+1);
            outerpart = eurlstring.mid(0, i1) + eurlstring.mid(i2+1);
        }
        else
        {
            outerpart = eurlstring;
            innerpart = "";
        }
    }
    {// determine left and right outer parts
        outerpart = outerpart.replace("://", ":/");
        int i=outerpart.indexOf(":/", 2);
        if (i < 0)
            return sh::filesystemhandlers::LocalFilesystemHandler::instance()->localToEurl(eurlstring);
        int j=outerpart.indexOf("/", i+2);
        leftouterpart = outerpart.mid(0, i);
        if (j >= 0)
        {
            rightouterpart_host = outerpart.mid(i+2, j-i-2);
            rightouterpart_path = outerpart.mid(j);
        }
        else
        {
            rightouterpart_host = outerpart.mid(i+2);
            rightouterpart_path = "";
        }
    }
    return fromStringNOCHECK(
                _escape(check_scheme(_unescape(leftouterpart)))+
                ":/"+
                check_inner(innerpart)+
                "/"+
                _escape(check_hostname(_unescape(rightouterpart_host)))+
                _escape(check_path(_unescape(rightouterpart_path)))
                );
}

QHash<QString, std::weak_ptr<const sh::base::Eurl>> sh::base::Eurl::_eurluniverse;

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::fromStringNOCHECK(QString eurlstring)
{
    QMutexLocker _locker(&_mutex);
    std::shared_ptr<const sh::base::Eurl> result = 0;
    auto it = _eurluniverse.find(eurlstring);
    if (it != _eurluniverse.end())
        result = it->lock();
    if (result == 0)
    {
        result = std::make_shared<Eurl>(eurlstring);
        _eurluniverse.insert(eurlstring, result);
    }
    return result;
}

sh::base::Eurl::~Eurl()
{
    QMutexLocker _locker(&_mutex);
    _eurluniverse.remove(_eurlstring);
}

sh::base::Eurl::Eurl(QString eurlstring) :
    _eurlstring(eurlstring)
{
}

QMutex sh::base::Eurl::_mutex(QMutex::Recursive);

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::create(QString scheme, QString hostname, QString path)
{
    BENCHMARK_SCOPE("Eurl creation")
    return fromStringNOCHECK(_escape(check_scheme(scheme))+"://"+_escape(check_hostname(hostname))+_escape(check_path(path)));
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::create(QString scheme, const Eurl* inner, QString hostname, QString path)
{
    BENCHMARK_SCOPE("Eurl creation")
    return fromStringNOCHECK(_escape(check_scheme(scheme))+":/"
                             +(inner?(WRAPPER_BEGIN+inner->asString()+WRAPPER_END):"")
                             +"/"+_escape(check_hostname(hostname))+_escape(check_path(path)));
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::createNOCHECK(QString scheme, const Eurl* inner, QString hostname, QString path)
{
    return fromStringNOCHECK(_escape(scheme)+":/"
                             +(inner?(WRAPPER_BEGIN+inner->asString()+WRAPPER_END):"")
                             +"/"+_escape(hostname)+_escape(path));
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::outerUrl() const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    if (_cache_outerurl != 0)
        return _cache_outerurl;
    if (_cache_outerurl_isthis)
        return shared_from_this();
    int i = _eurlstring.indexOf(WRAPPER_BEGIN);
    if (i==-1) {
        _cache_outerurl_isthis = true;
        return shared_from_this();
    } else {
        int j = _eurlstring.lastIndexOf(WRAPPER_END);
        _cache_outerurl = fromStringNOCHECK( _eurlstring.mid(0,i) + _eurlstring.mid(j+1) );
        return _cache_outerurl;
    }
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::outermostInnerEurl() const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    if (_cache_outermostinnereurl != 0)
        return _cache_outermostinnereurl;
    int i = _eurlstring.indexOf(WRAPPER_BEGIN);
    int j = _eurlstring.lastIndexOf(WRAPPER_END);
    if (i == -1)
        return 0;
    _cache_outermostinnereurl = fromStringNOCHECK( _eurlstring.mid(i+1, j-i-1) );
    return _cache_outermostinnereurl;
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::enwrapWithOuterUrl(QString scheme, QString hostname, QString path) const
{
    BENCHMARK_SCOPE("Eurl creation")
    return createNOCHECK(check_scheme(scheme), this, check_hostname(hostname), check_path(path));
}

bool sh::base::Eurl::outerUrlIsRootDirectory() const
{
    return path() == "/";
}

bool sh::base::Eurl::hasParentSegment() const
{
    return hasInnerUrls() || !outerUrlIsRootDirectory();
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::parentSegment() const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    if (_cache_parentsegment != 0)
        return _cache_parentsegment;
    if (!hasParentSegment())
        return 0;
    if (outerUrlIsRootDirectory())
        _cache_parentsegment = outermostInnerEurl();
    else
        _cache_parentsegment = createNOCHECK(scheme(), outermostInnerEurl().get(), hostname(), check_path(path()+"/.."));
    return _cache_parentsegment;
}

QString sh::base::Eurl::basename() const
{
    BENCHMARK_SCOPE("Eurl other")
    QMutexLocker mutexlocker(&_mutex);
    if (!_cache_basename.isNull())
        return _cache_basename;
    _cache_basename = _unescape(_eurlstring.mid( _eurlstring.lastIndexOf("/")+1));
    if(_cache_basename.isNull())
        _cache_basename = "";
    return _cache_basename;
}

QString sh::base::Eurl::hostname() const
{
    BENCHMARK_SCOPE("Eurl other")
    QMutexLocker mutexlocker(&_mutex);
    if (!_cache_hostname.isNull())
        return _cache_hostname;
    QString oeurl = outerUrl()->_eurlstring;
    int i = oeurl.indexOf("://");
    _cache_hostname = _unescape( oeurl.mid(i+3, oeurl.indexOf("/", i+3)-i-3 ));
    return _cache_hostname;
}

QString sh::base::Eurl::path() const
{
    BENCHMARK_SCOPE("Eurl other")
    QMutexLocker mutexlocker(&_mutex);
    if (!_cache_path.isNull())
        return _cache_path;
    QString oeurl = outerUrl()->_eurlstring;
    int i = oeurl.indexOf("://");
    _cache_path = _unescape(oeurl.mid( oeurl.indexOf("/",i+3) ));
    return _cache_path;
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::root() const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    if (_cache_root)
        return _cache_root;
    else if (_cache_root_isthis)
        return shared_from_this();
    if (path() == "/")
    {
        _cache_root_isthis = true;
        return shared_from_this();
    }
    else
    {
        _cache_root = createNOCHECK(scheme(), outermostInnerEurl().get(), hostname(), "/");
        return _cache_root;
    }
}

QString sh::base::Eurl::scheme() const
{
    BENCHMARK_SCOPE("Eurl other")
    QMutexLocker mutexlocker(&_mutex);
    if (!_cache_scheme.isEmpty())
        return _cache_scheme;
    _cache_scheme = _unescape( _eurlstring.mid(0, _eurlstring.indexOf(":/") ) );
    return _cache_scheme;
}

bool sh::base::Eurl::hasInnerUrls() const
{
    QMutexLocker mutexlocker(&_mutex);
    bool res = _eurlstring.contains(WRAPPER_BEGIN);
    return res;
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::withAppendedSegment(QString s) const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    return fromStringNOCHECK(_eurlstring + (_eurlstring.endsWith("/")?"":"/") + _escape(check_filename(s)));
}

std::shared_ptr<const sh::base::Eurl> sh::base::Eurl::withAppendedSegments(QString s) const
{
    BENCHMARK_SCOPE("Eurl creation")
    QMutexLocker mutexlocker(&_mutex);
    return createNOCHECK(scheme(), outermostInnerEurl().get(), hostname(), check_path(path()+"/"+s));
}

QString sh::base::Eurl::asString() const
{
    BENCHMARK_SCOPE("Eurl other")
    QMutexLocker mutexlocker(&_mutex);
    QString res = _eurlstring;
    return res;
}

bool sh::base::Eurl::isPrefixOf(std::shared_ptr<const Eurl> longer) const
{
    BENCHMARK_SCOPE("Eurl other")
    while (longer.get()!=this && longer->hasParentSegment())
        longer=longer->parentSegment();
    return (longer.get()==this);
}

void sh::base::Eurl::filenameCheck(QString filename)
{
    if (filename.trimmed().isEmpty())
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Empty names are not allowed.")));
    if (filename == "." || filename == "..")
        throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("This reserved name is not allowed.")));
    foreach(QChar forbiddenchar, Eurl::FORBIDDEN_FILENAME_CHARACTERS)
        if (filename.contains(forbiddenchar))
            throw EurlMisformattedException(sh::exceptions::ExceptionData().details(QObject::tr("Character '%1' is not allowed.").arg(forbiddenchar)));
}

