/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFile>
#include "exceptions/ioexception.h"
#include "search/searchcriterion.h"
#include "tools/mainthread.h"
#include "filesystemhandler.h"
#include "filesystemhandlerregister.h"
#include "filesystemmodel.h"
#include "filesystemoperation.h"
#include "operation.h"


sh::base::FilesystemHandler::FilesystemHandler(sh::base::FilesystemModel *model) :
    QObject()
{
    this->_model = model;
}

void sh::base::FilesystemHandler::deleteDirectoryIfEmpty(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (getType(op, eurl) != sh::base::FilesystemNodeType::Directory)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("'%1' is not a directory.").arg(eurl->asString())));
    if (!op->filesystem()->itemlist(eurl).isEmpty())
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("'%1' is not empty.").arg(eurl->asString())));
    deleteItem(op, eurl);
}

QList<QString> sh::base::FilesystemHandler::listExtendedAttributes(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return QList<QString>();
}

quint64 sh::base::FilesystemHandler::getExtendedAttributeSize(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
    return 0;
}

QByteArray sh::base::FilesystemHandler::getExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
    return QByteArray();
}

void sh::base::FilesystemHandler::setExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString, QByteArray)
{
}

void sh::base::FilesystemHandler::removeExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
}

sh::base::FilesystemModel* sh::base::FilesystemHandler::model()
{
    return _model;
}

void sh::base::FilesystemHandler::configureItems(sh::base::Operation*, QList<std::shared_ptr<sh::base::FilesystemNode>>)
{
}

QMap<QString, QString> sh::base::FilesystemHandler::getCustomAttributes(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return QMap<QString, QString>();
}

void sh::base::FilesystemHandler::setCustomAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString, QString)
{
}

bool sh::base::FilesystemHandler::isWellKnownScheme()
{
    return false;
}

void sh::base::FilesystemHandler::search(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, std::function<void(std::shared_ptr<const sh::base::Eurl>, QList<std::shared_ptr<sh::search::SearchCriterion>>, sh::base::FilesystemNodeType ftype)> addfile, std::function<void(std::shared_ptr<const sh::base::Eurl>)> visitdir, QList<std::shared_ptr<sh::search::SearchCriterion>> criteria)
{
    foreach(std::shared_ptr<const sh::base::Eurl> item, op->filesystem()->itemlist(eurl))
    {
        sh::base::FilesystemNodeType t = op->filesystem()->getType(item);
        if (t == sh::base::FilesystemNodeType::Directory)
            visitdir(item);
        else
            addfile(item, criteria, t);
    }
}

void sh::base::FilesystemHandler::getCustomUi(std::shared_ptr<FilesystemNode>, QString*, std::shared_ptr<QWidget>*)
{
}

void sh::base::FilesystemHandler::HandlerTransfer::respectCancel()
{
}

void sh::base::FilesystemHandler::HandlerTransfer::incrementTransferredBytes(qint64)
{
}

sh::base::FilesystemHandler::HandlerTransfer::~HandlerTransfer()
{
}
