/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMHANDLER_H
#define FILESYSTEMHANDLER_H

#include <functional>
#include <memory>
#include <QDateTime>
#include <QMap>
#include "filesystemmodel.h"
#include "filesystemnodelist.h"


namespace sh {
    namespace actions{class AbstractActionItem;}
    namespace search{class SearchCriterion;}

    namespace base{

        class Operation;

        /** \brief
         * Abstract base class for a custom filesystem handler.
         *
         * Subclass and register it for implementing a new virtual filesystem,
         * which lets new nodes appear somewhere in the filesystem tree and
         * controls how to handle them.
         *
         * Use sh::base::FilesystemHandlerRegister for registration.
         *
         * For executing some actions or checks on a filesystem, you should not use those
         * handlers directly, but the higher-level sh::base::FilesystemOperation class.
         */
        class FilesystemHandler : public QObject
        {
            Q_OBJECT
        public:
            class HandlerTransfer;

            /** \brief Is (for subclasses) intended to be directly constructed and registered once.
             */
            explicit FilesystemHandler(sh::base::FilesystemModel *model);
            virtual ~FilesystemHandler() {}

            /** \brief
             * Determine a list of subelements in a certain directory.
             */
            virtual void itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list) = 0;
            /** \brief
             * Configure newly created nodes (e.g. setting another icon or changing the display name).
             */
            virtual void configureItems(sh::base::Operation* op, QList<std::shared_ptr<FilesystemNode> > nodes);
            /** \brief
             * Determine if a file is regular, or a dir, or a link, ...
             */
            virtual sh::base::FilesystemNodeType getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Determine the size of a file.
             */
            virtual qint64 getSize(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Determine the mtime of a file.
             */
            virtual QDateTime getMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Set the mtime of a file.
             */
            virtual void setMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QDateTime time) = 0;

            /** \brief
             * Determine mime type of a file.
             */
            virtual QString getMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Resolve a link.
             */
            virtual QString getLinkTarget(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Fetches the list of available extended attributes for an entry.
             */
            virtual QList<QString> listExtendedAttributes(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Returns the site of value for one extended attribute for an entry.
             */
            virtual quint64 getExtendedAttributeSize(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);
            /** \brief
             * Returns the value for one extended attribute for an entry.
             */
            virtual QByteArray getExtendedAttribute(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);
            /** \brief
             * Sets the value for one extended attribute for an entry.
             */
            virtual void setExtendedAttribute(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QByteArray value);
            /** \brief
             * Removes one extended attributes for an entry.
             */
            virtual void removeExtendedAttribute(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);

            /** \brief
             * Returns the custom attributes for an entry.
             *
             * In contrast to extended attributes, the custom attributes are not directly stored in the filesystem this way, but handle
             * implementation specific aspects (like permissions).
             */
            virtual QMap<QString, QString> getCustomAttributes(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Sets the value for one custom attribute for an entry.
             *
             * See also getCustomAttributes.
             */
            virtual void setCustomAttribute(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QString value);

            /** \brief
             * Returns if it is allowed to get the content of a certain file.
             */
            virtual bool canGetFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Get the content of a file.
             */
            virtual std::shared_ptr<QIODevice> getFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Returns if it is allowed to create subdirectories in a certain directory.
             */
            virtual bool canCreateDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Create a directory.
             */
            virtual void createDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;

            /** \brief
             * Returns if it is allowed to create a link in a certain directory.
             */
            virtual bool canCreateLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Create a link.
             */
            virtual void createLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString target) = 0;

            /** \brief
             * Returns if it is allowed to create files in a certain directory.
             */
            virtual bool canCreateFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Creates a file with some content.
             *
             * It may use `handlertransfer` for some better integration, like cancel support and
             * progress monitoring.
             */
            virtual void createFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QIODevice* content, HandlerTransfer* handlertransfer) = 0;

            /** \brief
             * Returns if it is allowed to delete a certain file/directory/link/....
             */
            virtual bool canDeleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Delete a file/directory/link/...
             */
            virtual void deleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl) = 0;
            /** \brief
             * Deletes a directory only of it is empty.
             */
            virtual void deleteDirectoryIfEmpty(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            /** \brief
             * Returns if it is allowed to rename a certain file, directory, link, ... (see renameItem).
             */
            virtual bool canRenameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src) = 0;
            /** \brief
             * Renames an item to another path.
             *
             * It can be a simple filename change or also changes in the deeper path segments.
             */
            virtual void renameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src, QString destpath) = 0;

            /** \brief
             * Returns a list of actions (see former example) for showing for certain files.
             */
            virtual QList< std::shared_ptr<sh::actions::AbstractActionItem> > getActions(const QList<std::shared_ptr<const sh::base::Eurl>> eurls) = 0;

            /** \brief
             * Returns if this handler requires to be used by the engine with resolved links.
             */
            virtual bool requiresResolvedLinks() { return true; }

            /** \brief
             * Callback for preparing stuff when the view entered some directory (e.g. for enabling watchers).
             * See also viewLeftDirectory.
             */
            virtual void viewEnteredDirectory(std::shared_ptr<const sh::base::Eurl>) {}
            /** \brief
             * Callback for preparing stuff when the view left some directory.
             * See also viewEnteredDirectory.
             */
            virtual void viewLeftDirectory(std::shared_ptr<const sh::base::Eurl>) {}

            /**
             * \brief Returns if the scheme for this handler is a well known one (which external programs typically would understand).
             */
            virtual bool isWellKnownScheme();

            /**
             * \brief Helps searching for files.
             *
             * Override this method with a behavior identical to the default, if a specialized implementation can be much faster than
             * the default one.
             */
            virtual void search(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, std::function<void(std::shared_ptr<const sh::base::Eurl>, QList<std::shared_ptr<sh::search::SearchCriterion>>, sh::base::FilesystemNodeType ftype)> addfile, std::function<void(std::shared_ptr<const sh::base::Eurl>)> visitdir, QList<std::shared_ptr<sh::search::SearchCriterion>> criteria);

            /**
             * \brief Creates a custom gui, which becomes part of the fileview.
             */
            virtual void getCustomUi(std::shared_ptr<sh::base::FilesystemNode> node, QString* header, std::shared_ptr<QWidget>* customui);

            /** \brief
             * A convenience shortcut to the sh::base::FilesystemModel (a singleton).
             */
            sh::base::FilesystemModel* model();

            class HandlerTransfer
            {
            public:
                virtual void respectCancel();
                virtual void incrementTransferredBytes(qint64 donebytes);
                virtual ~HandlerTransfer();
            };

        signals:

        public slots:

        protected:
            sh::base::FilesystemModel *_model;
        };

    }
}

#endif // FILESYSTEMHANDLER_H
