/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMHANDLERREGISTER_H
#define FILESYSTEMHANDLERREGISTER_H

#include <memory>
#include <QObject>
#include <QMap>
#include <QMutex>


namespace sh {
    namespace base{

        class FilesystemHandler;

        /** \brief
         * A register of filesystem handlers.
         *
         * Each active (i.e. referred to by existing nodes) instance of sh::base::FilesystemHandler must be registered here.
         */
        class FilesystemHandlerRegister : public QObject
        {
            Q_OBJECT

        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static FilesystemHandlerRegister* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** \brief
             * Registers a filesystem handler.
             * \param scheme The scheme (very first part of a sh::base::Eurl) for which the handler is responsible for.
             * \param handler The filesystem handler.
             */
            void addHandler(QString scheme, std::shared_ptr<sh::base::FilesystemHandler> handler);
            /** \brief
             * Finds a filesystem handler by scheme.
             */
            std::shared_ptr<sh::base::FilesystemHandler> findHandler(QString scheme);

        private:
            explicit FilesystemHandlerRegister(QObject *parent = 0);
            QMap<QString, std::shared_ptr<sh::base::FilesystemHandler>> handlers;
            QMutex mutex;

        signals:

        public slots:

        };

    }
}

#endif // FILESYSTEMHANDLERREGISTER_H
