/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMMODEL_H
#define FILESYSTEMMODEL_H

#include "filesystemnode.h"
#include <functional>
#include <QAbstractItemModel>


namespace sh {
    namespace tools{class ThreadDispatcher;}

    namespace base{

        class FilesystemNode;
        class FilesystemModelDirectoryTreeProxy;
        class FilesystemModelFileviewProxy;
        class LoadOnDemandPlaceholderFilesystemNode;
        class FilesystemHandler;
        class ModelBackedFilesystemNodeList;

        /** \brief
         * The filesystem model.
         *
         * This is the engine which creates and manages
         * sh::base::FilesystemNode nodes. Filesystem nodes are used on many
         * places for all kinds of operations.
         */
        class FilesystemModel : public QAbstractItemModel
        {
            friend class FilesystemNode;
            friend class LoadOnDemandPlaceholderFilesystemNode;
            friend class ModelBackedFilesystemNodeList;
            friend class ::sh::base::DetailColumn;
            Q_OBJECT

        //### MISC ###//
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static FilesystemModel* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            virtual ~FilesystemModel();
            /** \brief
             * The sh::base::FilesystemNode which is the root node of the entire model.
             * It is the  parent for all toplevel nodes.
             */
            std::shared_ptr<sh::base::FilesystemNode> rootNode();

        //### QABSTRACTITEMMODEL ###//

            QVariant data(const QModelIndex &index, int role) const;
            Qt::ItemFlags flags(const QModelIndex &index) const;
            QVariant headerData(int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;
            QModelIndex index(int row, int column, const QModelIndex &parent = QModelIndex()) const;
            QModelIndex parent(const QModelIndex &index) const;
            int rowCount(const QModelIndex &parent = QModelIndex()) const;
            int columnCount(const QModelIndex &parent = QModelIndex()) const;

        //### PROXIES ###//

            /** \brief
             * Creates a sh::base::FilesystemModelFileviewProxy for presenting the content of a directory.
             */
            sh::base::FilesystemModelFileviewProxy* createFileviewProxy(QModelIndex root, bool withtooltip, std::function<void(std::shared_ptr<sh::base::FilesystemNode>)> onBecomesInvalid);

        //### FIND FOR EURL ###//

            /** \brief
             * Get a list of qt model indexes for a sh::base::Eurl.
             *
             * If nodes for this eurl are unknown to the model so far, it tries to build them.
             * In typical cases, this list either contains one element, or
             * is empty if the filesystem handlers decide that this file
             * does not exist. But in some cases, there is also more than
             * one index for one sh::base::Eurl (when the sh::base::Eurl appears on more
             * than one place in the tree).
             */
            QList<QPersistentModelIndex> findIndexesForEurl(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Get a list of sh::base::FilesystemNode for a sh::base::Eurl.
             *
             * If it is unknown to the model so far, it tries to build them.
             * In typical cases, this list either contains one element, or
             * is empty if the filesystem handlers decide that this file
             * does not exist. But in some cases, there is also more than
             * one node for one sh::base::Eurl (when the sh::base::Eurl appears on more
             * than one place in the tree).
             * It only returns nodes, which are 'alive', i.e. which have a living parent
             * and which are a child of this parent.
             */
            QList<std::shared_ptr<sh::base::FilesystemNode> > findNodesForEurl(std::shared_ptr<const sh::base::Eurl> eurl);

        //### DATA TRANSLATION ###//

            /** \brief
             * Returns a list of sh::base::FilesystemNode for a sh::base::Eurl.
             * It only considers the current state of the in-memory model. It will
             * only return nodes which are already known to the model so far.
             * This operation is cheaper and handling only the known nodes is sufficient
             * in many situations.
             * In typical cases, this list either contains one element, or
             * is empty. But in some cases, there is also more than
             * one node for one sh::base::Eurl (when the sh::base::Eurl appears on more
             * than one place in the tree).
             * It only returns nodes, which are 'alive', i.e. which have a living parent
             * and which are a child of this parent.
             */
            QList<std::shared_ptr<sh::base::FilesystemNode> > tryGetNodesForEurl(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Returns the sh::base::FilesystemNode for a qt model index (in the main model).
             */
            std::shared_ptr<sh::base::FilesystemNode> getNodeForIndex(const QModelIndex index) const;
            /** \brief
             * Returns a qt model index (in the main model) for a sh::base::FilesystemNode.
             */
            QModelIndex getIndexForNode(std::shared_ptr<sh::base::FilesystemNode> node);

        //### CREATION ###//

            /** \brief
             * Creates a new sh::base::FilesystemNode for placing it into the model.
             *
             * If such a node (with the same eurl for the same parent node) does not exist,
             * it generates a new one. If there already is such a node alive,
             * but currently not placed in the model, it recycles this one.
             * This can happen when references exist to a node which is not yet inserted
             * or which is removed meanwhile.
             * It is not allowed
             * to call this method when such a node already exists in the model.
             *
             * Use this function for getting a sh::base::FilesystemNode, which is
             * to be added to the model now or later.
             * Depending on some parameter values, a call directly adds the node
             * to the filesystem model (not e.g. when doinsert is `false` or parentnode
             * is 0)
             * It is typically used within a sh::base::FilesystemHandler implementation.
             */
            std::shared_ptr<sh::base::FilesystemNode> createFilesystemNode(std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemHandler* handler,
                                                 sh::base::FilesystemNodeType nodetype, bool isHidden, std::shared_ptr<sh::base::FilesystemNode> parent, bool doinsert = true, bool showInitialLoadingLabel = true);

            /** \brief
             * Returns a sh::base::FilesystemNode for using it as a child node in the model.
             *
             * It either creates a new one, if there currently is no node for this
             * eurl in this parentnode, or returns the existing one.
             * Even for existing ones, this call can change the nodetype of that node.
             *
             * Depending on some parameter values, a call directly adds the node
             * to the filesystem model (not e.g. when doinsert is `false` or parent
             * is 0).
             */
            std::shared_ptr<sh::base::FilesystemNode> getOrCreateFilesystemNode(std::shared_ptr<const sh::base::Eurl> neweurl, sh::base::FilesystemHandler* handler,
                                                      sh::base::FilesystemNodeType nodetype, bool isHidden, std::shared_ptr<sh::base::FilesystemNode> parent,
                                                                                bool doinsert = true, bool showInitialLoadingLabel = true, bool* pIsNew = 0);

        //### EXTENDED SERVICES ###//

            /** \brief
             * Request to refresh the internal model information for a sh::base::Eurl.
             * This may be a place which is already known (then a change of some metadata
             * or the removal is detected) or a formerly unknown place (then new nodes
             * get inserted in the model).
             */
            void refreshData(std::shared_ptr<const sh::base::Eurl> eurl, bool forceFindParent=false, bool withDetails=true);

            /** \brief
             * Registers a helper method for 'opening' (mounting, activating, ...) locations on demand.
             *
             * Only used in very rare cases.
             */
            void addOpenNodeHelper(int index, std::function<QList<std::shared_ptr<sh::base::FilesystemNode>>(std::shared_ptr<const sh::base::Eurl>)> openNodeHelper);

        private:
            explicit FilesystemModel(QObject *parent = 0);
            QList<std::shared_ptr<sh::base::FilesystemNode> > _findNodesForEurl_helper(std::shared_ptr<const sh::base::Eurl> eurl);
            QList<std::shared_ptr<sh::base::FilesystemNode> > openRootEurl(std::shared_ptr<const sh::base::Eurl> eurl);
            void _subitemFetchingStateChanged(std::shared_ptr<sh::base::FilesystemNode> node, bool value);
            QList<std::function<QList<std::shared_ptr<sh::base::FilesystemNode>>(std::shared_ptr<const sh::base::Eurl>)> > _openNodeHelpers;
            QMap<int, std::function<QList<std::shared_ptr<sh::base::FilesystemNode>>(std::shared_ptr<const sh::base::Eurl>)> > _openNodeHelpersMap;
            std::shared_ptr<sh::base::FilesystemNode> rootnode;
            QMutex mutex;
            QHash<std::shared_ptr<const sh::base::Eurl>, std::weak_ptr<sh::base::FilesystemNode>> eurl2node;
            QMutex eurl2nodemutex;
            QMutex _nodeDataMutex;
            QMutex _openNodeHelpersMutex;
        };

    }
}

#endif // FILESYSTEMMODEL_H
