/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "tools/visibleviews.h"
#include "filesystemmodeldirectorytreeproxy.h"
#include "filesystemmodeldirectorytreeproxyvisibilityenforcements.h"


namespace sh{namespace base{INITIALIZE_SINGLETON_IN_MAINTHREAD(30056, FilesystemModelDirectoryTreeProxyVisibilityEnforcements)}}

std::shared_ptr<sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements> sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::instance()
{
    static auto _instance = std::make_shared<sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements>();
    return _instance;
}

void sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::_initialize()
{
    sh::tools::VisibleViews::instance()->registerOnEnteredHandlers([=](std::shared_ptr<sh::base::FilesystemNode> dir){
        instance()->nodeEnteredView(dir);
    });
    sh::tools::VisibleViews::instance()->registerOnLeftHandlers([=](std::shared_ptr<sh::base::FilesystemNode> dir){
        instance()->nodeLeftView(dir);
    });
}

void sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::_shutdown()
{
}

sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::FilesystemModelDirectoryTreeProxyVisibilityEnforcements()
{
}

bool sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::_addenforcement(std::shared_ptr<FilesystemNode> node, std::shared_ptr<FilesystemNode> hnode) {
    if (hnode->isHidden()) {
        auto tex = hiddenForcedVisibles.end();
        auto i = hiddenForcedVisibles.begin();
        while(i != hiddenForcedVisibles.end()) {
            std::shared_ptr<FilesystemNode> hf = std::get<0>(*i).lock();
            if (!hf)
                i = hiddenForcedVisibles.erase(i);
            else {
                if (hf == hnode) {
                    tex = i;
                    break;
                }
                i++;
            }
        }
        if (tex == hiddenForcedVisibles.end()) {
            hiddenForcedVisibles.append(std::make_tuple(hnode, QList<std::shared_ptr<FilesystemNode>>()));
            tex = hiddenForcedVisibles.end() - 1;
            FilesystemModelDirectoryTreeProxy::instance()->enforceVisibility(hnode);
        }
        std::get<1>(*tex).append(node);
        return true;
    }
    return false;
}

void sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::nodeEnteredView(std::shared_ptr<FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    std::shared_ptr<FilesystemNode> hnode = node;
    QObject* repr = new QObject();
    _enterednodesrepresentatives.insert(node, repr);
    while(hnode) {
        if (!_addenforcement(node, hnode)) {  // either it's directly hidden (then we add a visibility enforcement), or we register an ishidden-changed handler for doing that maybe later
            std::function<void()> h = [=](){
                if (this->_addenforcement(node, hnode))
                    QObject::disconnect(hnode.get(), &sh::base::FilesystemNode::isHiddenChanged, repr, 0);  // 'repr' is for removing it after leaving
            };
            QObject::connect(hnode.get(), &sh::base::FilesystemNode::isHiddenChanged, repr, h);
        }
        hnode = hnode->parentnode();
    }
}

void sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::nodeLeftView(std::shared_ptr<FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    auto repr = _enterednodesrepresentatives.find(node);
    if (repr != _enterednodesrepresentatives.end())  // just to be sure
        delete *repr;
    auto i=hiddenForcedVisibles.begin();
    while(i!=hiddenForcedVisibles.end()) {
        std::shared_ptr<FilesystemNode> hf = std::get<0>(*i).lock();
        if (!hf)
            i = hiddenForcedVisibles.erase(i);
        else {
            std::get<1>(*i).removeOne(node);
            i++;
        }
    }
}

void sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::nodeCollapsedInTree(std::shared_ptr<FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QList<std::shared_ptr<FilesystemNode>> ldeenfhf;
    auto ihfvis=hiddenForcedVisibles.begin();
    while(ihfvis!=hiddenForcedVisibles.end()) {
        std::shared_ptr<FilesystemNode> hf = std::get<0>(*ihfvis).lock();
        bool deenfhf = false;
        if (!hf)
            ihfvis = hiddenForcedVisibles.erase(ihfvis);
        else if (std::get<1>(*ihfvis).isEmpty()) {
            std::shared_ptr<FilesystemNode> xnode = hf;
            while (xnode) {
                if (xnode == node) {
                    deenfhf = true;
                    break;
                }
                xnode = xnode->parentnode();
            }
            if (deenfhf) {
                ldeenfhf.append(hf);
                ihfvis = hiddenForcedVisibles.erase(ihfvis);
            }
        }
        if (!deenfhf)
            ihfvis++;
    }
    foreach(auto n, ldeenfhf)
        FilesystemModelDirectoryTreeProxy::instance()->deEnforceVisibility(n);
}
