/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "tools/threaddispatcher.h"
#include "tools/thumbnailmanager.h"
#include "detailcolumn.h"
#include "filesystemmodelfileviewproxy.h"
#include "filesystemmodelsubtreeproxy.h"
#include "filesystemmodel.h"
#include "filesystemnode.h"


sh::base::FilesystemModelFileviewProxy::FilesystemModelFileviewProxy(QModelIndex root, bool withtooltip, QObject *parent) :
    QSortFilterProxyModel(parent), _sizeFormattingMode(SizeFormattingModePrefixes), _root(root), _withtooltip(withtooltip)
{
    QObject::connect(sh::tools::ThumbnailManager::instance(), &sh::tools::ThumbnailManager::thumbnailAvailable, this, [=](std::shared_ptr<sh::base::FilesystemNode> _n){
        QModelIndex si = sh::base::FilesystemModel::instance()->getIndexForNode(_n);
        if (si.isValid())
        {
            QSortFilterProxyModel* sopromo = qobject_cast<QSortFilterProxyModel*>(sourceModel());
            sh::base::FilesystemModelSubtreeProxy* fimosupro = qobject_cast<sh::base::FilesystemModelSubtreeProxy*>(sourceModel());
            QModelIndex interm = sopromo ? sopromo->mapFromSource(si) : (fimosupro ? fimosupro->mapFromSource(si) : si);
            QModelIndex di = mapFromSource(interm);
            if (di.isValid())
            {
                emit dataChanged(di, di);
            }
        }
    });

}

void sh::base::FilesystemModelFileviewProxy::triggerReloadData()
{
    CHECK_CORRECT_CALLER_THREAD
    emit dataChanged(index(0, 0), index(rowCount(), columnCount()));
}

QVariant sh::base::FilesystemModelFileviewProxy::data(const QModelIndex &index, int role) const
{
    CHECK_CORRECT_CALLER_THREAD
    QModelIndex mi = mapToSource(index);
    if (role == Qt::DecorationRole)
    {
        if (index.isValid() && index.column() == 0)
        {
            sh::base::FilesystemNode* item = static_cast<sh::base::FilesystemNode*>(mi.internalPointer());
            if (_thumbnailmode)
            {
                QIcon result;
                bool outdated;
                bool reordered;
                std::shared_ptr<sh::base::FilesystemNode> sitem = item->shared_from_this();
                if (!sh::tools::ThumbnailManager::instance()->getThumbnail(sitem, _thumbnailsize, _thumbnailsize, &result, &outdated, &reordered) || (outdated&&!reordered))
                    sh::tools::ThumbnailManager::instance()->requestThumbnail(sitem, _thumbnailsize, _thumbnailsize);
                if (result.isNull())
                    result = item->icon();
                return result;
            }
            else
                return item->icon();
        }
        return QVariant();
    }
    else if (_withtooltip && role == Qt::ToolTipRole)
        return sourceModel()->data(mi, Qt::DisplayRole);
    else
        return sourceModel()->data(mi, role);
}

void sh::base::FilesystemModelFileviewProxy::setThumbnail(bool enabled, int size)
{
    CHECK_CORRECT_CALLER_THREAD
    _thumbnailmode = enabled;
    _thumbnailsize = size;
}

bool sh::base::FilesystemModelFileviewProxy::getThumbnailEnabled()
{
    CHECK_CORRECT_CALLER_THREAD
    return _thumbnailmode;
}

int sh::base::FilesystemModelFileviewProxy::getThumbnailSize()
{
    CHECK_CORRECT_CALLER_THREAD
    return _thumbnailsize;
}

void sh::base::FilesystemModelFileviewProxy::setSizeFormattingMode(SizeFormatting mode)
{
    CHECK_CORRECT_CALLER_THREAD
    _sizeFormattingMode = mode;
    triggerReloadData();
}

sh::base::SizeFormatting sh::base::FilesystemModelFileviewProxy::getSizeFormattingMode() const
{
    return _sizeFormattingMode;
}

void sh::base::FilesystemModelFileviewProxy::setHiddenFilesVisible(bool v)
{
    _hiddenFilesVisible = v;
    invalidateFilter();
}

bool sh::base::FilesystemModelFileviewProxy::getHiddenFilesVisible() const
{
    return _hiddenFilesVisible;
}

bool sh::base::FilesystemModelFileviewProxy::filterAcceptsRow (int row, const QModelIndex & parent) const
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::base::FilesystemNode> node;
    if (parent.isValid())
        node = sh::base::FilesystemModel::instance()->getNodeForIndex(parent);
    else
        // yes, this is a dirty hack :-/
        node = sh::base::FilesystemModel::instance()->getNodeForIndex(_root);
    if (row < node->childnodeCount())
    {
        std::shared_ptr<sh::base::FilesystemNode> cn = node->childnode(row);
        if (cn->isHidden() && !_hiddenFilesVisible)
            return false;
        if (cn->nodetype() == sh::base::FilesystemNodeType::SpecialTreeOnlyDirectory)
            return false;
        return true;
    }
    else
        return true;
}

void sh::base::FilesystemModelFileviewProxy::setSourceModel(QAbstractItemModel * sourceModel)
{
    QSortFilterProxyModel::setSourceModel(sourceModel);
}

bool sh::base::FilesystemModelFileviewProxy::lessThan (const QModelIndex & left, const QModelIndex & right ) const
{
    CHECK_CORRECT_CALLER_THREAD
    if (!left.isValid()) return true;
    if (!right.isValid()) return false;
    std::shared_ptr<sh::base::FilesystemNode> nleft = sh::base::FilesystemModel::instance()->getNodeForIndex(left);
    std::shared_ptr<sh::base::FilesystemNode> nright = sh::base::FilesystemModel::instance()->getNodeForIndex(right);
    int col = sortColumn();
    std::shared_ptr<sh::base::DetailColumn> detcol = 0;
    bool doTypediff=true;
    if (col>0)
    {
        int idx = col-1;
        if (nleft->parentnode()->_childrenDetailCols.length()>idx) {
            detcol = nleft->parentnode()->_childrenDetailCols[idx];
            doTypediff = detcol->sort_doTypediff();
        }
    }
    if (doTypediff) {
        std::shared_ptr<sh::base::FilesystemNode> absright = nright;
        std::shared_ptr<sh::base::FilesystemNode> absleft = nleft;
        if (absright->nodetype()==sh::base::FilesystemNodeType::Link && !absright->linkDestinationNodes().isEmpty())
            absright = absright->linkDestinationNodes().first();
        if (absleft->nodetype()==sh::base::FilesystemNodeType::Link && !absleft->linkDestinationNodes().isEmpty())
            absleft = absleft->linkDestinationNodes().first();
        sh::base::FilesystemNodeType righttype = absright->nodetype();
        sh::base::FilesystemNodeType lefttype = absleft->nodetype();
        if (righttype==sh::base::FilesystemNodeType::Link)
            righttype=sh::base::FilesystemNodeType::File;
        if (lefttype==sh::base::FilesystemNodeType::Link)
            lefttype=sh::base::FilesystemNodeType::File;
        int typediff = (int)righttype - (int)lefttype;
        if (sortOrder()==Qt::DescendingOrder)
            typediff *= -1;
        if (typediff!=0) return (typediff<0);
    }
    if (col == 0)
        return nleft->displayName() < nright->displayName();
    else if (detcol != 0)
    {
        int comp = detcol->sort_order(nleft, nright);
        if (comp==0)
        {
            QString leftdn = nleft->displayName();
            QString rightdn = nright->displayName();
            if (leftdn!=rightdn)
                comp = (leftdn < rightdn) ? -1 : 1;
            else
                comp = nleft->index() - nright->index();
            if (sortOrder()==Qt::DescendingOrder)
                comp *= -1;
        }
        return comp < 0;
    }
    return QSortFilterProxyModel::lessThan(left, right);
}
