/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <QCoreApplication>
#include "detailcolumns/detailcolumnmimetype.h"
#include "detailcolumns/detailcolumndirectsymlinktarget.h"
#include "detailcolumns/detailcolumnresolvedsymlink.h"
#include "detailcolumns/detailcolumnfilesize.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "tools/mainthread.h"
#include "tools/threadpool.h"
#include "tools/iconmanager.h"
#include "ui/mainwindow.h"
#include "detailcolumn.h"
#include "filesystemhandler.h"
#include "filesystemmodel.h"
#include "filesystemnode.h"
#include "filesystemnodelist.h"
#include "filesystemoperation.h"
#include "operation.h"


sh::base::FilesystemNode::FilesystemNode(sh::base::FilesystemModel *model) :
     QObject(), _model(model), _ishidden(false), _childnodes(new sh::base::ModelBackedFilesystemNodeList(this))
{
    // it should be assumed that at the first time only one thread accesses it
    static std::shared_ptr<const sh::base::Eurl> _nulleurl = sh::base::Eurl::create("shallot.root", "", "/");
    _eurl = _nulleurl;
}

sh::base::FilesystemNode::FilesystemNode(sh::base::FilesystemModel* model, std::shared_ptr<const sh::base::Eurl>eurl,
            QString displayname, sh::base::FilesystemHandler *handler,
            sh::base::FilesystemNodeType nodetype, bool ishidden)
    : FilesystemNode(model)
{
    _eurl = eurl;
    _displayname = displayname;
    _nodetype = nodetype;
    _ishidden = ishidden;
    _handler = handler;
    _initialize(true);
}

void sh::base::FilesystemNode::_initialize(bool addDetails)
{
    if (qobject_cast<LoadOnDemandPlaceholderFilesystemNode*>(this))
        return;
    if (_nodetype == sh::base::FilesystemNodeType::Directory)
        setIcon(sh::tools::IconManager::instance()->getIcon("folder"));
    else if (_nodetype == sh::base::FilesystemNodeType::SpecialTreeOnlyDirectory)
        setIcon(sh::tools::IconManager::instance()->getIcon("folder", "package-x-generic"));
    else if (_nodetype == sh::base::FilesystemNodeType::File)
        setIcon(sh::tools::IconManager::instance()->getIcon("text-x-generic"));
    else if (_nodetype == sh::base::FilesystemNodeType::Link)
        setIcon(sh::tools::IconManager::instance()->getIcon("", "islink"));
    else
        setIcon(sh::tools::IconManager::instance()->getIcon(""));
    if (addDetails)
    {
        addDetail(sh::detailcolumns::DetailColumnResolvedSymlink::instance());
        addDetail(sh::detailcolumns::DetailColumnMimetype::instance());
        addDetail(sh::detailcolumns::DetailColumnDirectSymlinkTarget::instance());
        addDetail(sh::detailcolumns::DetailColumnFilesize::instance());
        addDetail(sh::detailcolumns::DetailColumnMtime::instance());
    }
}

sh::base::FilesystemNode::~FilesystemNode()
{
    delete _childnodes;
}

void sh::base::FilesystemNode::unmount()
{
    CHECK_CORRECT_CALLER_THREAD
    {
        QMutexLocker l(&_model->_nodeDataMutex);
        _parent = 0;
        _linkdestinations.clear();
        _detailvalues.clear();
        _last_detailvalues.clear();
        _detailsRequested = false;
    }
    _loadnode = 0;
    _isConfigured = false;
    foreach(std::shared_ptr<sh::base::FilesystemNode> n, *_childnodes->nodes())
        n->unmount();
    emit removed();
    _childnodes->clear();
}

bool sh::base::FilesystemNode::isAlive() const
{
    return (parentnode()) || (this==sh::base::FilesystemModel::instance()->rootnode.get());
}

QList<std::shared_ptr<sh::base::FilesystemNode>> sh::base::FilesystemNode::linkDestinationNodes() const
{
    QMutexLocker l(&_model->_nodeDataMutex);
    QList<std::shared_ptr<sh::base::FilesystemNode>> result;
    foreach(std::weak_ptr<sh::base::FilesystemNode> wx, _linkdestinations)
    {
        std::shared_ptr<sh::base::FilesystemNode> x = wx.lock();
        if (x)
            result << x;
    }
    return result;
}

void sh::base::FilesystemNode::slot_updateinfofromlinkdestination()
{
    if (nodetype()==sh::base::FilesystemNodeType::Link)
        sh::tools::MainThread::dispatcher()->invokeAsync([=](){
            if (!isAlive())
                return;
            QMutexLocker l(&_model->_nodeDataMutex);
            QList<std::shared_ptr<sh::base::FilesystemNode>> lbuddynodes = linkDestinationNodes();
            if (!lbuddynodes.isEmpty())
                setIcon(sh::tools::IconManager::instance()->getIcon(lbuddynodes.first()->icon(), "islink", QString(), QStringList()));
            else
                setIcon(sh::tools::IconManager::instance()->getIcon("", "islink"));
            QModelIndex pind = _model->getIndexForNode(_parent);
            QModelIndex tl = _model->index(index(), 0, pind);
            QModelIndex br = _model->index(index(), _parent->_childrenDetailCols.count(), pind);
            emit _model->dataChanged(tl, br);
        }, shared_from_this());
}

void sh::base::FilesystemNode::setLinkDestinationNodes(QList<std::shared_ptr<sh::base::FilesystemNode>> linkdestinations)
{
    QList<std::weak_ptr<sh::base::FilesystemNode>> wbuddies;
    foreach(std::shared_ptr<sh::base::FilesystemNode> x, linkdestinations)
        wbuddies << x;
    sh::tools::MainThread::dispatcher()->invokeAsync([=](){
        QMutexLocker l(&_model->_nodeDataMutex);
        foreach(std::weak_ptr<sh::base::FilesystemNode> oldlinkdestination, _linkdestinations)
        {
            std::shared_ptr<sh::base::FilesystemNode> soldlinkdestination = oldlinkdestination.lock();
            if (soldlinkdestination)
                disconnect(soldlinkdestination.get(), 0, this, 0);
        }
        _linkdestinations = wbuddies;
        foreach(std::shared_ptr<sh::base::FilesystemNode> linkdestination, linkdestinations)
            connect(linkdestination.get(), SIGNAL(iconChanged()), this, SLOT(slot_updateinfofromlinkdestination()));
        slot_updateinfofromlinkdestination();
    }, shared_from_this());
}

std::shared_ptr<const sh::base::Eurl> sh::base::FilesystemNode::eurl()
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _eurl;
}

sh::base::FilesystemModel* sh::base::FilesystemNode::model()
{
    return _model;
}

void sh::base::FilesystemNode::addChild(std::shared_ptr<sh::base::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    _childnodes->addItem(node);
}

void sh::base::FilesystemNode::removeChild(std::shared_ptr<sh::base::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    _childnodes->removeItem(node);
}

sh::base::ModelBackedFilesystemNodeList *sh::base::FilesystemNode::childnodes()
{
    CHECK_CORRECT_CALLER_THREAD
    return _childnodes;
}

int sh::base::FilesystemNode::childnodeCount()
{
    CHECK_CORRECT_CALLER_THREAD
    return _childnodes->nodes()->count();
}

std::shared_ptr<sh::base::FilesystemNode> sh::base::FilesystemNode::childnode(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    return _childnodes->nodes()->at(i);
}

QString sh::base::FilesystemNode::displayName()
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _displayname;
}

void sh::base::FilesystemNode::setDisplayName(QString displayname)
{
    CHECK_CORRECT_CALLER_THREAD
    QMutexLocker l(&_model->_nodeDataMutex);
    _displayname = displayname;
    if (_parent)
    {
        QModelIndex tl = _model->index(index(), 0, _model->getIndexForNode(_parent));
        emit _model->dataChanged(tl, tl);
    }
}

sh::base::FilesystemHandler *sh::base::FilesystemNode::handler()
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _handler;
}

sh::base::FilesystemNodeType sh::base::FilesystemNode::nodetype()
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _nodetype;
}

void sh::base::FilesystemNode::requestDetails(bool force)
{
    CHECK_CORRECT_CALLER_THREAD
    if (!_detailsRequested || force)
    {
        _detailsRequested = true;
        _detailsRequestor.requestDetails(shared_from_this());
    }
}

sh::base::FilesystemNode::DetailsRequestor::DetailsRequestor()
{
    availthreads = std::max(sh::tools::ThreadPool::getThreadCount()/2, 1);
}

void sh::base::FilesystemNode::DetailsRequestor::requestDetails(std::shared_ptr<sh::base::FilesystemNode> node)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    static QSet<sh::base::FilesystemNode*> inprocess;
    static QSet<sh::base::FilesystemNode*> doagain;
    bool newthread;
    {
        QMutexLocker _l1(&nodesmutex);
        nodes.enqueue(node);
        newthread = availthreads>0;
        if (newthread)
            availthreads--;
    }
    if (newthread)
    {
        std::shared_ptr<sh::base::Operation> operation(new sh::base::Operation());
        operation->enableDetailsCache();
        sh::tools::ThreadPool::enqueueForce([=]() {
            sh::exceptions::Exception::executeGuarded([&]() {
                while(!sh::tools::ThreadPool::isShuttingDown()) {
                    std::shared_ptr<sh::base::FilesystemNode> node = 0;
                    {
                        QMutexLocker _l1(&nodesmutex);
                        if (!nodes.isEmpty())
                            node = nodes.dequeue();
                        else
                            availthreads++;
                    }
                    if (!node)
                        break;
                    QList<sh::base::DetailColumn*> cols;
                    {
                        QMutexLocker l1(&sh::base::FilesystemModel::instance()->_nodeDataMutex);
                        if (inprocess.contains(node.get()))
                        {
                            doagain.insert(node.get());
                            break;
                        }
                        inprocess.insert(node.get());
                        cols = node->_detailvalues.keys();
                        node->_last_detailvalues = node->_detailvalues;
                        node->_detailvalues.clear();
                    }
                    foreach (sh::base::DetailColumn* col, cols)
                    {
                        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                            sh::exceptions::Exception::executeGuarded([&](){
                                col->requestValue(node, operation.get());
                            }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED);
                        });
                    }
                    {
                        QMutexLocker l1(&sh::base::FilesystemModel::instance()->_nodeDataMutex);
                        inprocess.remove(node.get());
                        if (doagain.remove(node.get()))
                        {
                            QMutexLocker _l1(&nodesmutex);
                            nodes.enqueue(node);
                        }
                        node->_last_detailvalues.clear();
                    }
                }
           });
        });
    }
}

sh::base::FilesystemNode::DetailsRequestor sh::base::FilesystemNode::_detailsRequestor;

void sh::base::FilesystemNode::requestContainedItems(sh::base::FilesystemNodeType type)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::base::FilesystemHandler* _handler = handler();
    if (_handler) {
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            Operation operation;
            std::shared_ptr<const sh::base::Eurl> reurl = operation.filesystem()->resolveEurlLink(eurl());
            if (type==sh::base::FilesystemNodeType::NONE)
                for(int i=(int)sh::base::FilesystemNodeType::FIRSTTYPE; i<=(int)sh::base::FilesystemNodeType::LASTPHYSICALTYPE; i++)
                {
                    sh::base::FilesystemNodeListEditor editor(_childnodes, reurl, (sh::base::FilesystemNodeType)i);
                    _handler->itemlist(&operation, reurl, (sh::base::FilesystemNodeType)i, &editor);
                }
            else
            {
                sh::base::FilesystemNodeListEditor editor(_childnodes, reurl, type);
                _handler->itemlist(&operation, reurl, type, &editor);
            }
        });
    }
}

std::shared_ptr<sh::base::FilesystemNode> sh::base::FilesystemNode::parentnode() const
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _parent;
}

int sh::base::FilesystemNode::index()
{
    QMutexLocker l(&_model->_nodeDataMutex);
    return _index;
}

bool sh::base::FilesystemNode::isRootNode() const
{
    CHECK_CORRECT_CALLER_THREAD
    return _isRootNode;
}

bool sh::base::FilesystemNode::isConfigured() const
{
    CHECK_CORRECT_CALLER_THREAD
    return _isConfigured;
}

void sh::base::FilesystemNode::addDetail(std::shared_ptr<sh::base::DetailColumn> column)
{
    CHECK_CORRECT_CALLER_THREAD
    QMutexLocker l1(&_model->_nodeDataMutex);
    if (!_detailvalues.contains(column.get()))
    {
        _detailsRequested = false;
        setDetail(column, sh::base::DetailColumn::VALUE_UNAVAILABLE());
    }
}

int sh::base::FilesystemNode::getInsertPositionForDetailColumn(std::shared_ptr<sh::base::DetailColumn> newCol)
{
    CHECK_CORRECT_CALLER_THREAD
    int result = 0;
    QList<std::shared_ptr<sh::base::DetailColumn>> cols = _parent->_childrenDetailCols;
    while (result < cols.length() && newCol->displayIndex() > cols[result]->displayIndex())
        result++;
    if (result < cols.length() && newCol->displayIndex() == cols[result]->displayIndex())
        while (result < cols.length() && newCol->name() > cols[result]->name())
            result++;
    return result;
}

void sh::base::FilesystemNode::setDetail(std::shared_ptr<sh::base::DetailColumn> column, QVariant value)
{
    CHECK_CORRECT_CALLER_THREAD
    QMutexLocker l1(&_model->_nodeDataMutex);
    _detailvalues[column.get()]=value;
    if (_parent && column->isVisible())
    {
        int intidx;
        if ((intidx=_parent->_childrenDetailCols.indexOf(column))==-1)
        {
            int idx = getInsertPositionForDetailColumn(column);
            _model->beginInsertColumns( _model->getIndexForNode(_parent), idx+1, idx+1 );
            _parent->_childrenDetailCols.insert(idx, column);
            _model->endInsertColumns();
        }
        else
        {
            QModelIndex tl = _model->index(index(), intidx+1,_model->getIndexForNode( _parent ) );
            QModelIndex br = tl;
            emit _model->dataChanged(tl, br);
        }
    }
}

std::shared_ptr<sh::base::DetailColumn> sh::base::FilesystemNode::getDetailColumnByIndex(int index)
{
    CHECK_CORRECT_CALLER_THREAD
    return _childrenDetailCols[index];
}

int sh::base::FilesystemNode::getDetailColumnsCount() const
{
    CHECK_CORRECT_CALLER_THREAD
    return _childrenDetailCols.length();
}

bool sh::base::FilesystemNode::isFetchingSubitems()
{
    CHECK_CORRECT_CALLER_THREAD
    return _loadnode != 0;
}

QIcon sh::base::FilesystemNode::icon()
{
    CHECK_CORRECT_CALLER_THREAD
    return _icon;
}

void sh::base::FilesystemNode::setIcon(QIcon icon)
{
    CHECK_CORRECT_CALLER_THREAD
    _icon = icon;
    emit iconChanged();
}

bool sh::base::FilesystemNode::isHidden()
{
    CHECK_CORRECT_CALLER_THREAD
    return _ishidden;
}

void sh::base::FilesystemNode::setHidden(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    _ishidden = v;
    emit isHiddenChanged();
    if (_parent)
    {  // this index is for the name; it just triggers re-evaluation of filters
        QModelIndex tl = _model->index(index(), 0, _model->getIndexForNode(_parent));
        emit _model->dataChanged(tl, tl);
    }
}

sh::base::LoadOnDemandPlaceholderFilesystemNode::LoadOnDemandPlaceholderFilesystemNode(std::shared_ptr<sh::base::FilesystemNode> parent) :
    sh::base::FilesystemNode(parent->model(), 0, QString(), 0, sh::base::FilesystemNodeType::NONE, false)
{
    loading = false;
    _parent = _origparent = parent;
}

QString sh::base::LoadOnDemandPlaceholderFilesystemNode::displayName()
{
    CHECK_CORRECT_CALLER_THREAD
    if (!loading)
    {
        loading = true; // we do not lock the _nodeDataMutex here since parentnode should remain constant
        sh::tools::ThreadPool::enqueueForce([=](){
            std::shared_ptr<sh::base::FilesystemNode> pnode = parentnode();
            if (pnode)
            {
                pnode->requestContainedItems(sh::base::FilesystemNodeType::Directory);
                sh::tools::MainThread::dispatcher()->invokeSync([&](){
                    pnode->model()->_subitemFetchingStateChanged(pnode, false);
                });
            }
        }, shared_from_this());
    }
    return tr("loading...");
}

void sh::base::LoadOnDemandPlaceholderFilesystemNode::requestDetails(bool)
{
}

unsigned int qHash(const std::shared_ptr<sh::base::FilesystemNode>& ptr, unsigned int seed)
{
    return qHash(ptr.get(), seed);
}

unsigned int qHash(const std::shared_ptr<const sh::base::Eurl>& ptr, unsigned int seed)
{
    return qHash(ptr.get(), seed);
}
