/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMNODE_H
#define FILESYSTEMNODE_H

#include <functional>
#include <memory>

namespace sh{
    namespace base{
        class FilesystemNode;
        class Eurl;
    }
}
unsigned int qHash(const std::shared_ptr<sh::base::FilesystemNode>& ptr, unsigned int seed = 0);
unsigned int qHash(const std::shared_ptr<const sh::base::Eurl>& ptr, unsigned int seed = 0);

#include <QHash>
#include <QIcon>
#include <QQueue>
#include "eurl.h"


namespace sh{
    namespace base{

        class DetailColumn;
        class FilesystemHandler;
        class FilesystemModel;
        class FilesystemModelFileviewProxy;
        class FilesystemModelSubtreeProxy;
        class ModelBackedFilesystemNodeList;
        class FilesystemNodeList;

        /** \brief
         * Enumeration of types a sh::base::FilesystemNode can have.
         */
        enum class FilesystemNodeType
        {
            /** \brief
             * No type specified or does not exist.
             */
            NONE = 0,
            /** \brief
             * Usual file.
             */
            File = 1,
            /** \brief
             * The first type (used for generating int ranges).
             */
            FIRSTTYPE = File,
            /** \brief
             * Directory.
             */
            Directory,
            /** \brief
             * Link.
             */
            Link,
            /** \brief
             * Unknown.
             */
            Unknown,
            /** \brief
             * The last physical type, excluding internal magic types (used for generating int ranges).
             */
            LASTPHYSICALTYPE = Unknown,
            /** \brief
             * Special kind of directory which behaves differently.
             */
            SpecialTreeOnlyDirectory,
            /** \brief
             * This return type means that the type can't be determined. It should not occur in typical
             * situations.
             */
            Invalid,
            /** \brief
             * The last type (used for generating int ranges).
             */
            LASTTYPE = SpecialTreeOnlyDirectory
        };

        /** \brief
         * A representation of a location in the filesystem tree.
         *
         * It represents filesystem nodes like a file or a directory in sh::base::FilesystemModel.
         *
         * Please note: Instances can be associated with any kind of elements in
         * the filesystem (files, directories, links, ...). It might also point to something
         * which does not exist at all (see parentnode).
         * The documentation sometimes explicitly makes a difference
         * between those kinds (files, directories, links, ...; often called 'node type').
         * But it often uses the term 'file' implicitly while meaning all kinds of elements;
         * assuming that e.g. a directory is just a special kind of a file. It should be clear from
         * the particular context which meaning applies.
         */
        class FilesystemNode : public QObject, public std::enable_shared_from_this<FilesystemNode>
        {
            friend class sh::base::FilesystemModel;
            friend class sh::base::DetailColumn;
            friend class LoadOnDemandPlaceholderFilesystemNode;
            friend class sh::base::FilesystemModelFileviewProxy;
            friend class sh::base::FilesystemModelSubtreeProxy;
            friend class sh::base::ModelBackedFilesystemNodeList;
            Q_OBJECT
        public:
            virtual ~FilesystemNode();
            /** \brief
             * Returns the sh::base::Eurl entry address.
             */
            std::shared_ptr<const sh::base::Eurl> eurl();
            /** \brief
             * Convenience shortcut to the sh::base::FilesystemModel.
             */
            sh::base::FilesystemModel* model();
            /** \brief
             * Returns the display name (what the user interface displays).
             */
            virtual QString displayName();
            /** \brief
             * Sets the display name (what the user interface displays).
             */
            virtual void setDisplayName(QString displayname);
            /** \brief
             * Returns the handler which is responsible for this node.
             * This should be the same as `sh::base::FilesystemHandlerRegister.findHandler(eurl.scheme)`.
             */
            sh::base::FilesystemHandler* handler();
            /** \brief
             * Returns the FilesystemNodeType node type.
             */
            FilesystemNodeType nodetype();
            /** \brief
             * Returns the number of children nodes.
             */
            int childnodeCount();
            /** \brief
             * Returns i-th child node.
             */
            std::shared_ptr<sh::base::FilesystemNode> childnode(int i);
            /** \brief
             * Returns the list of children.
             */
            sh::base::ModelBackedFilesystemNodeList* childnodes();
            /** \brief
             * Returns the parent node.
             *
             * This is 0 if the node does not exist in the model (i.e. not yet inserted or removed afterwards).
             * In most cases, this can be interpreted as 'file currently does not exist'.
             */
            std::shared_ptr<sh::base::FilesystemNode> parentnode() const;
            /** \brief
             * Returns the index of this node in the parent's list of children.
             */
            int index();
            /** \brief
             * Adds a new child. Not allowed to call more than once for a node.
             */
            void addChild(std::shared_ptr<sh::base::FilesystemNode> node);
            /** \brief
             * Removes a new child. Not allowed to call more than once for a node.
             */
            void removeChild(std::shared_ptr<sh::base::FilesystemNode> node);
            /** \brief
             * Checks if currently subitems are fetched (i.e. the 'loading' node is
             * shown).
             */
            bool isFetchingSubitems();
            /** \brief
             * Returns the node icon.
             */
            QIcon icon();
            /** \brief
             * Sets the node icon.
             */
            void setIcon(QIcon icon);
            /** \brief
             * Returns if the node is hidden.
             *
             * Note: It's not difficult for the user to also show the hidden items.
             */
            bool isHidden();
            /** \brief
             * Sets if the node is hidden.
             *
             * See also isHidden().
             */
            void setHidden(bool v);
            /** \brief
             * Adds a detail to this node.
             */
            void addDetail(std::shared_ptr<sh::base::DetailColumn> column);
            /** \brief
             * Returns the i-th detail column registered to this node.
             */
            std::shared_ptr<sh::base::DetailColumn> getDetailColumnByIndex(int index);
            /** \brief
             * Returns the number of detail columns registered to this node.
             */
            int getDetailColumnsCount() const;
            /** \brief
             * Checks if this is the root node of the model.
             */
            bool isRootNode() const;
            /** \brief
             * Checks if this node was already configured by a handler.
             */
            bool isConfigured() const;
            /** \brief
             * Requests to fetch details for this node.
             */
            virtual void requestDetails(bool force = true);
            /** \brief
             * Checks if this node currently exists in the model.
             */
            bool isAlive() const;
            /** \brief
             * Returns the list of link source nodes. If this node is a link,
             * the link destination nodes may influence the behavior and appearance of this node.
             *
             * This information does not come from inside but must be set from outside the model implementation.
             */
            QList<std::shared_ptr<sh::base::FilesystemNode> > linkDestinationNodes() const;
            /** \brief
             * Sets the list of link destination nodes. If this node is a link,
             * the link destination nodes may influence the behavior and appearance of this node.
             */
            void setLinkDestinationNodes(QList<std::shared_ptr<sh::base::FilesystemNode> > linkdestinations);
            /** \brief
             * Requests to fetch the children of this node with help of the filesystem handler.
             */
            void requestContainedItems(sh::base::FilesystemNodeType type = FilesystemNodeType::NONE);

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            FilesystemNode(sh::base::FilesystemModel *model, std::shared_ptr<const sh::base::Eurl> eurl, QString displayname,
                           sh::base::FilesystemHandler* handler, FilesystemNodeType nodetype, bool ishidden);

        signals:
            /** \brief
             * Is emitted when this node is not placed in the
             * tree anymore. This does not mean the physical deletion
             * of the instance, but just means it is not alive from now on.
             */
            void removed();
            /** \brief
             * Is emitted when values for detail columns arrived.
             */
            void _detailsAvailable();
            /** \brief
             * Is emitted when the icon changed.
             */
            void iconChanged();
            /** \brief
             * Is emitted when the hidden flag changed.
             */
            void isHiddenChanged();

        protected slots:
            /** \brief
             * Called when some metadata of the link sources change.
             * Used for updating own stuff.
             */
            void slot_updateinfofromlinkdestination();

        private:
            void setDetail(std::shared_ptr<sh::base::DetailColumn> column, QVariant value);
            int getInsertPositionForDetailColumn(std::shared_ptr<sh::base::DetailColumn> newCol);

        protected:
            FilesystemNode(sh::base::FilesystemModel *model);
            /** \brief
             * Unmounts the node from the model. This is more or less like removing the node.
             * Afterwards, the node is not listed as child somewhere and does not have a parent.
             * Becomes recycled when a node with that eurl on that place reappears.
             */
            void unmount();
            void _initialize(bool addDetails=false);
            QString _displayname;
            FilesystemHandler* _handler = 0;
            FilesystemNodeType _nodetype = FilesystemNodeType::Directory;
            bool _detailsRequested = false;
            bool _isRootNode = false;
            bool _isConfigured = false;
            std::shared_ptr<sh::base::FilesystemNode> _parent = 0;
            std::shared_ptr<sh::base::FilesystemNode> _origparent = 0;
            std::shared_ptr<sh::base::FilesystemNode> _loadnode = 0;
            int _index = -1;
            FilesystemModel* _model = 0;
            QIcon _icon;
            bool _ishidden;
            QList<std::shared_ptr<sh::base::DetailColumn>> _childrenDetailCols;
            QHash<sh::base::DetailColumn*, QVariant> _detailvalues;
            QHash<sh::base::DetailColumn*, QVariant> _last_detailvalues;
            sh::base::ModelBackedFilesystemNodeList* _childnodes;
            std::shared_ptr<const sh::base::Eurl> _eurl;
            QList<std::weak_ptr<sh::base::FilesystemNode>> _linkdestinations;

            /** \brief
             * Internal engine for multithreaded requesting of
             * fetching all details for a node.
             */
            class DetailsRequestor
            {
            public:
                DetailsRequestor();
                QQueue<std::shared_ptr<sh::base::FilesystemNode>> nodes;
                QMutex nodesmutex;
                /** \brief
                 * Fetch details for a node asynchronously in background.
                 */
                void requestDetails(std::shared_ptr<sh::base::FilesystemNode> node);
            private:
                int availthreads;
            };
            static DetailsRequestor _detailsRequestor;

        };

        /** \brief
         * A special node, which is unselectable and shows a 'loading' label.
         *
         * It does not correspond to a file, directory or similar which actually exists.
         * It is also never parent or a child of another node.
         *
         * They are used internally by the filesystem model.
         */
        class LoadOnDemandPlaceholderFilesystemNode : public sh::base::FilesystemNode
        {
            friend class sh::base::FilesystemModel;
            Q_OBJECT
        public:
            virtual QString displayName();
            virtual void requestDetails(bool force);
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            explicit LoadOnDemandPlaceholderFilesystemNode(std::shared_ptr<sh::base::FilesystemNode> parent);
        private:
            bool loading;
        };

    }
}

#endif // FILESYSTEMNODE_H
