/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef FILESYSTEMNODELIST_H
#define FILESYSTEMNODELIST_H

#include <QMutex>
#include <QWaitCondition>
#include "filesystemnode.h"


namespace sh{
    namespace base{

        /** \brief
         * A data structure for children lists of filesystem nodes.
         *
         * Different subclasses exist with different behaviors and use cases.
         */
        class FilesystemNodeList : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Adds items to this list. In a model-backed list, this triggers the internal
             * model mounting calls.
             * It is not allowed to call this method twice with the same node.
             */
            virtual void addItems(QSet<std::shared_ptr<sh::base::FilesystemNode>> nodes) =0;
            /** \brief
             * Adds an item to this list. In a model-backed list, this triggers the internal
             * model mounting calls.
             * It is not allowed to call this method twice with the same node.
             */
            virtual void addItem(std::shared_ptr<sh::base::FilesystemNode> node) =0;
            /** \brief
             * Removes nodes from this list. In a model-backed list, this triggers the internal
             * model unmounting calls.
             * It is not allowed to call this method with a node, which is not contained in that list.
             */
            virtual void removeItems(QSet<std::shared_ptr<sh::base::FilesystemNode>> nodes) =0;
            /** \brief
             * Removes a node from this list. In a model-backed list, this triggers the internal
             * model unmounting calls.
             * It is not allowed to call this method with a node, which is not contained in that list.
             */
            virtual void removeItem(std::shared_ptr<sh::base::FilesystemNode> node) =0;
            /** \brief
             * Resets the list for a given node type to a given new list of children nodes.
             */
            virtual void resetItems(sh::base::FilesystemNodeType type, QSet<std::shared_ptr<sh::base::FilesystemNode>> nodes) =0;
            /** \brief
             * Returns the node which owns this children list. The result may be `0`
             * for non model-backed lists.
             */
            virtual std::shared_ptr<sh::base::FilesystemNode> mynode() =0;
            /** \brief
             * Checks if this list contains a given node.
             */
            bool contains(std::shared_ptr<FilesystemNode> node);
            /** \brief
             * Returns the list of nodes currently stored in this instance.
             */
            const QList<std::shared_ptr<sh::base::FilesystemNode>>* nodes();
        protected:
            explicit FilesystemNodeList();
            QList<std::shared_ptr<sh::base::FilesystemNode>> _nodes;
        };


        /** \brief
         * This FilesystemNodeList subclass is used for spontanously fetching a fresh list of node children.
         *
         * Node additions and removals will not touch the model.
         *
         * Node instances inside it will be deleted together with this list.
         */
        class AdhocFilesystemNodeList : public FilesystemNodeList
        {
            Q_OBJECT
        public:
            /** \brief Is intended to be directly constructed from everywhere.
             */
            explicit AdhocFilesystemNodeList();
            ~AdhocFilesystemNodeList();
            void addItems(QSet<std::shared_ptr<FilesystemNode> > nodes);
            void addItem(std::shared_ptr<FilesystemNode> node);
            void removeItems(QSet<std::shared_ptr<FilesystemNode> > nodes);
            void removeItem(std::shared_ptr<FilesystemNode> node);
            void resetItems(sh::base::FilesystemNodeType type, QSet<std::shared_ptr<FilesystemNode> > nodes);
            std::shared_ptr<FilesystemNode> mynode();
        };


        /** \brief
         * This FilesystemNodeList subclass builds a part of the filesystem model.
         *
         * The listed nodes are actually owned and handled by a sh::base::FilesystemNode.
         * Node additions and removals will directly influence
         * the model.
         */
        class ModelBackedFilesystemNodeList : public FilesystemNodeList
        {
            Q_OBJECT
        public:
            /** \brief Constructed only internally.
             */
            explicit ModelBackedFilesystemNodeList(sh::base::FilesystemNode* node);
            void addItems(QSet<std::shared_ptr<FilesystemNode> > nodes);
            void addItem(std::shared_ptr<FilesystemNode> node);
            void removeItems(QSet<std::shared_ptr<FilesystemNode> > nodes);
            void removeItem(std::shared_ptr<FilesystemNode> node);
            void resetItems(sh::base::FilesystemNodeType type, QSet<std::shared_ptr<FilesystemNode> > nodes);
            std::shared_ptr<FilesystemNode> mynode();
            /** \brief
             * Clears the entire list.
             * This is not a high-level operation, but something only used internally (not part of the interface).
             */
            void clear();
            /** \brief
             * Adds an item to this list. In a model-backed list, this triggers the internal
             * model mounting calls.
             * It is not allowed to call this method twice with the same node.
             *
             * This list is safe against removing from later item lists. Only removePermanentItem removes it. See also addItem.
             */
            virtual void addPermanentItem(std::shared_ptr<sh::base::FilesystemNode> node);
            /** \brief
             * Removes a permanent node from this list. In a model-backed list, this triggers the internal
             * model unmounting calls.
             * It is not allowed to call this method with a node, which is not contained in that list.
             */
            virtual void removePermanentItem(std::shared_ptr<sh::base::FilesystemNode> node);

        private:
            QSet<sh::base::FilesystemNode*> _permanentnodes;
            sh::base::FilesystemNode* _mynode;
            sh::base::FilesystemModel* mymodel;
            void addItem_helper(std::shared_ptr<sh::base::FilesystemNode> node);

        #ifdef QT_DEBUG
            QList<std::shared_ptr<const sh::base::Eurl>> _childreneurls;
        #endif

        };

        /** \brief
         * A list editor for easily modifying a FilesystemNodeList.
         *
         * Mainly used for listing child nodes in sh::base::FilesystemHandler.itemlist.
         * In most cases, you should just use setItems.
         */
        class FilesystemNodeListEditor
        {
        public:
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            FilesystemNodeListEditor(FilesystemNodeList* list, std::shared_ptr<const sh::base::Eurl> parenteurl, sh::base::FilesystemNodeType nodestype);
            ~FilesystemNodeListEditor();
            /** \brief
             * Sets a new list content. This function is all you need in most situations. For iteratively
             * adding nodes, which makes the intermediate results visible in the user interface, see addItem.
             */
            void setItems(QList<QString> items);
            /** \brief
             * Must be called before you begin to iteratively fill the children list (e.g. with addItem).
             * You must also call endIterativeAdding afterwards.
             */
            void beginIterativeAdding();
            /** \brief
             * Adds a new children to the list, directly showing that in the user interface, while you can proceed
             * filling the list. Please read beginIterativeAdding as well! In most cases, you
             * don't need this function.
             */
            void addItem(QString item);
            /** \brief
             * Adds a new children to the list, directly showing that in the user interface, while you can proceed
             * filling the list. Please read beginIterativeAdding as well! In most cases, you
             * don't need this function.
             *
             * Note: If you specify a `handler`, it must be the one matching to `eurl`'s scheme.
             */
            std::shared_ptr<FilesystemNode> addCustomItem(std::shared_ptr<const sh::base::Eurl> eurl, QString displayname = QString(), sh::base::FilesystemHandler *handler = 0);
            /** \brief
             * Adds a new children to the list, directly showing that in the user interface, while you can proceed
             * filling the list. Please read beginIterativeAdding as well! In most cases, you
             * don't need this function.
             */
            void addExistingNodeItem(std::shared_ptr<sh::base::FilesystemNode> node);
            /** \brief
             * Must be called after you iteratively filled the children list with addItem. This automatically
             * removes all the old nodes, which you haven't added in this session.
             */
            void endIterativeAdding();
            /** \brief
             * After this call, the addItem method will consider the given names as hostnames instead of new
             * path segments. The resulting sh::base::Eurl will differ accordingly. This only makes sense as children
             * for root eurls.
             */
            void setItemsAreHosts();
            /** \brief
             * Returns the FilesystemNodeList backend. You should rarely need it.
             */
            FilesystemNodeList* rawlist();
        private:
            FilesystemNodeList* _list;
            std::shared_ptr<const sh::base::Eurl> _parenteurl;
            sh::base::FilesystemNodeType _nodestype;
            std::shared_ptr<sh::base::FilesystemNode> _parentnode;
            sh::base::FilesystemModel* _model;
            sh::base::FilesystemHandler* _handler;
            bool _itemsAreHosts = false;
            bool _beganAddingIteratively = false;
            QSet<std::shared_ptr<sh::base::FilesystemNode>> _iterativeAddingsBeforeState;
            static QSet<FilesystemNodeList*> _pendingIterativeAddings;
            static QMutex _pendingMutex;
            static QWaitCondition _iterativeAddingPossibleCondition;
            std::shared_ptr<const sh::base::Eurl> getChildEurl(std::shared_ptr<const sh::base::Eurl> eurl, QString item);
        };

    }
}

#endif // FILESYSTEMNODELIST_H
