/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "filesystemoperation.h"
#include "base/operation.h"
#include "base/filesystemhandler.h"
#include "base/filesystemnode.h"
#include "exceptions/ioexception.h"
#include "tools/mainthread.h"
#include "tools/threadpool.h"


sh::base::FilesystemOperation::FilesystemOperation(sh::base::Operation* operation)
            : _operation(operation)
{
}

QList<std::shared_ptr<const sh::base::Eurl>> sh::base::FilesystemOperation::itemlist(std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type)
{
    QList<std::shared_ptr<sh::base::FilesystemNode>> nodes = sh::base::FilesystemModel::instance()->tryGetNodesForEurl(eurl);
    if (nodes.isEmpty())
    {
        int i_t1, i_t2;
        QSet<std::shared_ptr<const sh::base::Eurl>> result;
        if (type==sh::base::FilesystemNodeType::NONE)
        {
            i_t1 = (int)sh::base::FilesystemNodeType::FIRSTTYPE;
            i_t2 = (int)sh::base::FilesystemNodeType::LASTPHYSICALTYPE;
        }
        else
            i_t1 = i_t2 = (int)type;
        for(int i=i_t1; i<=i_t2; i++)
        {
            sh::base::AdhocFilesystemNodeList list;
            std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
            std::shared_ptr<const sh::base::Eurl> resdeurl = _resolveIfNeeded(handler, eurl);
            sh::base::FilesystemNodeListEditor editor(&list, resdeurl, (sh::base::FilesystemNodeType)i);
            handler->itemlist(_operation, resdeurl, (sh::base::FilesystemNodeType)i, &editor);
            foreach(std::shared_ptr<sh::base::FilesystemNode> n, *(list.nodes()))
                result.insert(n->eurl());
        }
        return result.toList();
    }
    else
    {
        std::shared_ptr<sh::base::FilesystemNode> node = nodes.first();
        node->requestContainedItems();
        QList<std::shared_ptr<const sh::base::Eurl>> result;
        QList<std::shared_ptr<sh::base::FilesystemNode>> cnodes;
        sh::tools::MainThread::dispatcher()->invokeSync([&](){
            cnodes = *(node->childnodes()->nodes());
        });
        foreach(std::shared_ptr<sh::base::FilesystemNode> cnode, cnodes)
            result << cnode->eurl();
        return result;
    }
}

sh::base::FilesystemNodeType sh::base::FilesystemOperation::getType(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->getType(_operation, eurl);
}

qint64 sh::base::FilesystemOperation::getFileSize(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->getSize(_operation, eurl);
}

QString sh::base::FilesystemOperation::getLinkTarget(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->getLinkTarget(_operation, eurl);
}

bool sh::base::FilesystemOperation::canGetFileContent(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    if (!handler)
        return false;
    std::shared_ptr<const sh::base::Eurl> resdeurl = _resolveIfNeeded(handler, eurl);
    return handler->canGetFileContent(_operation, resdeurl);
}

std::shared_ptr<QIODevice> sh::base::FilesystemOperation::getFileContent(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    std::shared_ptr<const sh::base::Eurl> resdeurl = _resolveIfNeeded(handler, eurl);
    std::shared_ptr<QIODevice> result = handler->getFileContent(_operation, resdeurl);
    if (!result->isOpen())
        result->open(QFile::ReadOnly);
    return result;
}

bool sh::base::FilesystemOperation::canCreateDirectory(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    if (!handler)
        return false;
    return handler->canCreateDirectory(_operation, eurl);
}

void sh::base::FilesystemOperation::createDirectory(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->createDirectory(_operation, eurl);
    progressmon->addChangedEurl(eurl);
}

bool sh::base::FilesystemOperation::canCreateLink(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    if (!handler)
        return false;
    return handler->canCreateLink(_operation, eurl);
}

void sh::base::FilesystemOperation::createLink(std::shared_ptr<const sh::base::Eurl> eurl, QString target, std::shared_ptr<FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->createLink(_operation, eurl, target);
    progressmon->addChangedEurl(eurl);
}

bool sh::base::FilesystemOperation::canCreateFile(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    if (!handler)
        return false;
    return handler->canCreateFile(_operation, eurl);
}

void sh::base::FilesystemOperation::createFile(std::shared_ptr<const sh::base::Eurl> eurl, QIODevice* content, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    progressmon->_enablestatistics();
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    sh::exceptions::Exception::executeGuarded([&](){
        if (getType(eurl) != sh::base::FilesystemNodeType::NONE)
            deleteItem(eurl);
    }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED | sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
    MyHandlerTransfer myhandlertransfer(progressmon.get());
    handler->createFile(_operation, eurl, content, &myhandlertransfer);
    progressmon->addChangedEurl(eurl);
}

bool sh::base::FilesystemOperation::canDeleteItem(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    if (!handler)
        return false;
    return handler->canDeleteItem(_operation, eurl);
}

void sh::base::FilesystemOperation::deleteItem(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->deleteItem(_operation, eurl);
    progressmon->addChangedEurl(eurl);
}

void sh::base::FilesystemOperation::deleteDirectoryIfEmpty(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->deleteDirectoryIfEmpty(_operation, eurl);
    progressmon->addChangedEurl(eurl);
}

bool sh::base::FilesystemOperation::canMoveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest)
{
    std::shared_ptr<sh::base::FilesystemHandler> srchandler = _handler(src);
    if (!srchandler)
        return false;
    if (!dest)
        dest = src;
    bool renameallowed = src->root()==dest->root() && srchandler->canRenameItem(_operation, src);
    _operation->setCustomData(src, "_canRename", renameallowed);
    if (renameallowed)
        return true;
    else
    {
        std::shared_ptr<sh::base::FilesystemHandler> desthandler = _handler(dest);
        if (!desthandler)
            return false;
        sh::base::FilesystemNodeType stype = srchandler->getType(_operation, src);
        if (stype == sh::base::FilesystemNodeType::File)
            return srchandler->canGetFileContent(_operation, src) && srchandler->canDeleteItem(_operation, src) && desthandler->canCreateFile(_operation, dest);
        else if (stype == sh::base::FilesystemNodeType::Directory)
            return srchandler->canDeleteItem(_operation, src) && desthandler->canCreateDirectory(_operation, dest);
        else if (stype == sh::base::FilesystemNodeType::Link)
            return srchandler->canDeleteItem(_operation, src) && desthandler->canCreateLink(_operation, dest);
        else
            return false;
    }
}

void sh::base::FilesystemOperation::moveItems(QList<std::shared_ptr<const Eurl>> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    sh::base::FilesystemOperationTransfers transfers(this, progressmon);
    foreach(std::shared_ptr<const Eurl> _src, src)
        transfers.addMoveItem(_src, dest->withAppendedSegment(_src->basename()));
    transfers.executestepqueue();
}

void sh::base::FilesystemOperation::moveItem(std::shared_ptr<const Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    sh::base::FilesystemOperationTransfers transfers(this, progressmon);
    transfers.addMoveItem(src, dest);
    transfers.executestepqueue();
}

bool sh::base::FilesystemOperation::canCopyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest)
{
    std::shared_ptr<sh::base::FilesystemHandler> srchandler = _handler(src);
    if (!srchandler)
        return false;
    if (!dest)
        dest = src;
    std::shared_ptr<sh::base::FilesystemHandler> desthandler = _handler(dest);
    if (!desthandler)
        return false;
    sh::base::FilesystemNodeType stype = srchandler->getType(_operation, src);
    if (stype == sh::base::FilesystemNodeType::File)
        return srchandler->canGetFileContent(_operation, src) && desthandler->canCreateFile(_operation, dest);
    else if (stype == sh::base::FilesystemNodeType::Directory)
        return desthandler->canCreateDirectory(_operation, dest);
    else if (stype == sh::base::FilesystemNodeType::Link)
        return desthandler->canCreateLink(_operation, dest);
    else
        return false;
}

void sh::base::FilesystemOperation::copyItems(QList<std::shared_ptr<const Eurl>> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    sh::base::FilesystemOperationTransfers transfers(this, progressmon);
    foreach(std::shared_ptr<const Eurl> _src, src)
        transfers.addCopyItem(_src, dest->withAppendedSegment(_src->basename()));
    transfers.executestepqueue();
}

void sh::base::FilesystemOperation::copyItem(std::shared_ptr<const Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon)
{
    progressmon = _monitor(progressmon);
    sh::base::FilesystemOperationTransfers transfers(this, progressmon);
    transfers.addCopyItem(src, dest);
    transfers.executestepqueue();
}

QList<QString> sh::base::FilesystemOperation::listExtendedAttributes(std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->listExtendedAttributes(_operation, eurl);
}

quint64 sh::base::FilesystemOperation::getExtendedAttributeSize(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->getExtendedAttributeSize(_operation, eurl, attribute);
}

QByteArray sh::base::FilesystemOperation::getExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    return handler->getExtendedAttribute(_operation, eurl, attribute);
}

void sh::base::FilesystemOperation::setExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QByteArray value)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->setExtendedAttribute(_operation, eurl, attribute, value);
}

void sh::base::FilesystemOperation::removeExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
    std::shared_ptr<sh::base::FilesystemHandler> handler = _handler(eurl);
    handler->removeExtendedAttribute(_operation, eurl, attribute);
}

QList<std::shared_ptr<sh::base::FilesystemNode>> sh::base::FilesystemOperation::resolveNodeLink(std::shared_ptr<sh::base::FilesystemNode> node, bool recursive, bool excludeOwn)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // nearly the same happens in the other resolveLink
    if (node->nodetype() != sh::base::FilesystemNodeType::Link) // this is for better performance on non-links; a similar (but heavier) test will come again later for links
    {
        if (excludeOwn)
            return QList<std::shared_ptr<sh::base::FilesystemNode>>();
        else
            return QList<std::shared_ptr<sh::base::FilesystemNode>>({node});
    }
    else
    {
        std::shared_ptr<const sh::base::Eurl> tgteurl = resolveEurlLink(node->eurl(), recursive, excludeOwn);
        if (tgteurl)
            return sh::base::FilesystemModel::instance()->findNodesForEurl(tgteurl);
        else
            return QList<std::shared_ptr<sh::base::FilesystemNode>>();
    }
}

std::shared_ptr<const sh::base::Eurl> sh::base::FilesystemOperation::resolveEurlLink(std::shared_ptr<const sh::base::Eurl> eurl, bool recursive, bool excludeOwn, int tries)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    // nearly the same happens in the other resolveLink
    if (!tries)
        return 0;
    if (!eurl->outerUrlIsRootDirectory())
    {
        std::shared_ptr<const sh::base::Eurl> rpeurl = resolveEurlLink(eurl->parentSegment());
        if (!rpeurl)
            return 0;
        eurl = rpeurl->withAppendedSegment(eurl->basename());
    }
    std::shared_ptr<sh::base::FilesystemHandler> handler = sh::base::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme());
    if (handler->getType(_operation, eurl) != sh::base::FilesystemNodeType::Link)
        return excludeOwn ? 0 : eurl;
    else
    {
        QString rawtgt = handler->getLinkTarget(_operation, eurl);
        std::shared_ptr<const sh::base::Eurl> tgteurl = 0;
        try{
            if (rawtgt.startsWith("/"))
                tgteurl = sh::base::Eurl::fromString(eurl->root()->asString()+rawtgt);
            else
                tgteurl = sh::base::Eurl::fromString(eurl->parentSegment()->asString()+"/"+rawtgt);
        }
        catch (sh::exceptions::Exception& e)
        {
            if (!e.isClass(sh::exceptions::ArgumentException(sh::exceptions::ExceptionData()).name()))
                throw;
        }
        if ((!recursive) || (!tgteurl))
            return tgteurl;
        else
            return resolveEurlLink(tgteurl, true, false, tries-1);
    }
}

void sh::base::FilesystemOperation::addTransferrableDetailColumn(int index, std::shared_ptr<sh::base::DetailColumn> detailColumn)
{
    QMutexLocker _l1(&_detailColumnsMutex);
    _detailColumnsMap.insertMulti(index, detailColumn);
    QList<int> indexes = _detailColumnsMap.keys();
    qSort(indexes);
    _detailColumns.clear();
    foreach (int index, indexes)
        _detailColumns.append(_detailColumnsMap.values(index));
}

QList<std::shared_ptr<sh::base::DetailColumn> > sh::base::FilesystemOperation::transferrableDetailColumns()
{
    QMutexLocker _l1(&_detailColumnsMutex);
    return _detailColumns;
}

QList<std::shared_ptr<sh::base::DetailColumn> > sh::base::FilesystemOperation::_detailColumns;

QMap<int, std::shared_ptr<sh::base::DetailColumn> > sh::base::FilesystemOperation::_detailColumnsMap;

QMutex sh::base::FilesystemOperation::_detailColumnsMutex;

sh::base::FilesystemHandlerRegister* sh::base::FilesystemOperation::_handlerRegister = sh::base::FilesystemHandlerRegister::instance();

sh::base::FilesystemOperation::MyHandlerTransfer::MyHandlerTransfer(sh::base::FilesystemOperationProgressMonitor* progressmon)
            : _progressmon(progressmon)
{
    if (_progressmon)
        _progressmon->incProgress(0, 1, 0, -1);
}

sh::base::FilesystemOperation::MyHandlerTransfer::~MyHandlerTransfer()
{
    if (_progressmon)
        _progressmon->incProgress(1, 0, 0, _progressmon->doneBytes() + 1);
}

void sh::base::FilesystemOperation::MyHandlerTransfer::respectCancel()
{
    if (_progressmon)
        _progressmon->respectCancel();
}

void sh::base::FilesystemOperation::MyHandlerTransfer::incrementTransferredBytes(qint64 donebytes)
{
    if (_progressmon)
        _progressmon->incProgress(0, 0, donebytes, 0);
}


