/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMOPERATION_H
#define FILESYSTEMOPERATION_H

#include <functional>
#include "base/filesystemnode.h"
#include <QIODevice>
#include "base/filesystemhandlerregister.h"
#include "exceptions/ioexception.h"
#include "filesystemoperationtransfers.h"


namespace sh{
    namespace actions {
        class ActionExecutionInfo;
    }

    namespace base{

        class Operation;
        class DetailColumn;
        class FilesystemHandler;

        /** \brief
         * A high-level interface for filesystem operations.
         *
         * It is always based on the transaction of a Operation instance (which also gives you access to a FilesystemOperation object).
         *
         * Some calls optionally allow to give an own instance of sh::base::FilesystemOperationProgressMonitor
         * for some additional functionality. Please note that not all calls provide all those functionality (some do not use e.g. the
         * conflict resolution at all). If not provided, a default behavior is applied.
         */
        class FilesystemOperation
        {
            friend class FilesystemOperationTransfers;
        public:

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            FilesystemOperation(sh::base::Operation* operation);
            /** \brief
             * Gets a list of shallot.sh::base::FilesystemNode nodes in a directory (optionally filter by given types).
             */
            QList<std::shared_ptr<const sh::base::Eurl>> itemlist(std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type=sh::base::FilesystemNodeType::NONE);
            /** \brief
             * Gets the shallot.sh::base::FilesystemNodeType node type for an entry.
             */
            sh::base::FilesystemNodeType getType(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Gets the file size for an entry.
             */
            qint64 getFileSize(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Gets the link target for an entry (if it is a link).
             */
            QString getLinkTarget(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Can we get the file content for an entry?
             */
            bool canGetFileContent(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Gets the file content for an entry as QIODevice.
             */
            std::shared_ptr<QIODevice> getFileContent(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Can we create a given directory?
             */
            bool canCreateDirectory(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Creates a directory.
             * \param progressmon An optional progress monitor.
             */
            void createDirectory(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Can we create a given link?
             */
            bool canCreateLink(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Creates a link.
             * \param progressmon An optional progress monitor.
             */
            void createLink(std::shared_ptr<const sh::base::Eurl> eurl, QString target, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Can we create a given file?
             */
            bool canCreateFile(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Creates a file.
             * \param progressmon An optional progress monitor.
             */
            void createFile(std::shared_ptr<const sh::base::Eurl> eurl, QIODevice* content, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Can we delete a given entry?
             */
            bool canDeleteItem(std::shared_ptr<const sh::base::Eurl> eurl);
            /** \brief
             * Delete an entry.
             * \param progressmon An optional progress monitor.
             */
            void deleteItem(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Delete a directory entry if empty.
             * \param progressmon An optional progress monitor.
             */
            void deleteDirectoryIfEmpty(std::shared_ptr<const sh::base::Eurl> eurl, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Checks if it is allowed to move a certain item.
             *
             * If the destination is known as well, it might help to pass it as well.
             *
             * This does not guarantee success in the transfer, but is just a cheap early check.
             */
            bool canMoveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest = 0);
            /** \brief
             * Moves entries.
             * \param dest The requested new common parent directory.
             * \param progressmon An optional progress monitor.
             */
            void moveItems(QList<std::shared_ptr<const sh::base::Eurl>> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Moves an entry.
             * \param dest The new location (not the new parent).
             * \param progressmon An optional progress monitor.
             */
            void moveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Checks if it is allowed to copy a certain item.
             *
             * If the destination is known as well, it might help to pass it as well.
             *
             * This does not guarantee success in the transfer, but is just a cheap early check.
             */
            bool canCopyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const Eurl> dest = 0);
            /** \brief
             * Copies entries.
             * \param dest The requested new common parent directory.
             * \param progressmon An optional progress monitor.
             */
            void copyItems(QList<std::shared_ptr<const sh::base::Eurl>> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Copies an entry.
             * \param dest The new location (not the new parent).
             * \param progressmon An optional progress monitor.
             */
            void copyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> progressmon = 0);
            /** \brief
             * Resolve a link as node with or without recursion.
             */
            QList<std::shared_ptr<FilesystemNode> > resolveNodeLink(std::shared_ptr<FilesystemNode> node, bool recursive=true, bool excludeOwn=false);
            /** \brief
             * Resolve a link as sh::base::Eurl with or without recursion.
             */
            std::shared_ptr<const sh::base::Eurl> resolveEurlLink(std::shared_ptr<const sh::base::Eurl> eurl, bool recursive=true, bool excludeOwn=false, int tries=100);

            QList<QString> listExtendedAttributes(std::shared_ptr<const sh::base::Eurl> eurl);
            quint64 getExtendedAttributeSize(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);
            QByteArray getExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);
            void setExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QByteArray value);
            void removeExtendedAttribute(std::shared_ptr<const sh::base::Eurl> eurl, QString attribute);

            /**
             * @brief Registers a detail column for transferring it when file transfer take place.
             * \param index An integer which controls the order of transferring.
             */
            static void addTransferrableDetailColumn(int index, std::shared_ptr<sh::base::DetailColumn> detailColumn);
            /**
             * @brief A list of all detail columns which are registered becoming transferred in file transfers.
             */
            static QList<std::shared_ptr<sh::base::DetailColumn> > transferrableDetailColumns();

        private:
            Operation* _operation;
            static sh::base::FilesystemHandlerRegister* _handlerRegister;
            static QMap<int, std::shared_ptr<sh::base::DetailColumn> > _detailColumnsMap;
            static QList<std::shared_ptr<sh::base::DetailColumn> > _detailColumns;
            static QMutex _detailColumnsMutex;

            inline std::shared_ptr<sh::base::FilesystemHandler> _handler(std::shared_ptr<const sh::base::Eurl> eurl)
            {
                return _handlerRegister->findHandler(eurl->scheme());
            }

            inline std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> _monitor(std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> m)
            {
                if (m)
                    return m;
                else
                    return std::make_shared<sh::base::FilesystemOperationProgressMonitor>();
            }

            inline std::shared_ptr<const sh::base::Eurl> _resolveIfNeeded(std::shared_ptr<sh::base::FilesystemHandler> handler, std::shared_ptr<const sh::base::Eurl> eurl)
            {
                if (handler->requiresResolvedLinks()) {
                    std::shared_ptr<const sh::base::Eurl> resolved = resolveEurlLink(eurl);
                    if (!resolved)
                        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unable to resolve '%1'.").arg(eurl->asString())));
                    return resolved;
                }
                else
                    return eurl;
            }

            class MyHandlerTransfer : public sh::base::FilesystemHandler::HandlerTransfer
            {
            public:
                MyHandlerTransfer(sh::base::FilesystemOperationProgressMonitor* progressmon = 0);
                ~MyHandlerTransfer();
                virtual void respectCancel();
                virtual void incrementTransferredBytes(qint64 donebytes);
            private:
                sh::base::FilesystemOperationProgressMonitor* _progressmon;
            };

        };

    }
}

#endif // FILESYSTEMOPERATION_H
