/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/actionexecutioninfo.h"
#include "base/detailcolumn.h"
#include "tools/threadpool.h"
#include "filesystemoperation.h"
#include "filesystemoperationtransfers.h"
#include "eurl.h"
#include "operation.h"


sh::base::FilesystemOperationTransfers::FilesystemOperationTransfers(FilesystemOperation *filesystem, std::shared_ptr<FilesystemOperationProgressMonitor> progressmon) : filesystem(filesystem), progressmon(progressmon)
{
}

void sh::base::FilesystemOperationTransfers::_compile_moveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*>* result, QList<OperationStepClass*> withExpansion)
{
    QVariant _renameallowed_cached = filesystem->_operation->getCustomData(src, "_canRename");
    bool renameallowed = src->root()==dest->root() && (_renameallowed_cached.isNull() ? filesystem->_handler(src)->canRenameItem(filesystem->_operation, src) : _renameallowed_cached.toBool());
    if (renameallowed)
        result->append(new OperationStep_RenameItem(this, src, dest->path(), withExpansion));
    else
        _compile_copyItem(src, dest, result, QList<OperationStepClass*>({new OperationStep_DeleteItem(this, src, dest, withExpansion)}));
}

void sh::base::FilesystemOperationTransfers::_compile_copyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*>* result, QList<OperationStepClass*> withExpansion)
{
    std::shared_ptr<sh::base::FilesystemHandler> srchandler = filesystem->_handler(src);
    sh::base::FilesystemNodeType stype = srchandler->getType(filesystem->_operation, src);
    OperationStepClass* s;
    if (stype == sh::base::FilesystemNodeType::File)
        s = new OperationStep_CopyFile(this, src, dest, withExpansion);
    else if (stype == sh::base::FilesystemNodeType::Directory)
        s = new OperationStep_CopyDirectory(this, src, dest, withExpansion);
    else if (stype == sh::base::FilesystemNodeType::Link)
        s = new OperationStep_CopyLink(this, src, dest, withExpansion);
    else
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Unknown file type: %1").arg((int)stype)).retryable(false));
    result->append(s);
}

void sh::base::FilesystemOperationTransfers::addMoveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest)
{
    _compile_moveItem(src, dest, &stepqueue, QList<OperationStepClass*>());
}

void sh::base::FilesystemOperationTransfers::addCopyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest)
{
    _compile_copyItem(src, dest, &stepqueue, QList<OperationStepClass*>());
}

void sh::base::FilesystemOperationTransfers::computeProgress()
{
    quint64 _allitems = 0, _doneitems = 0;
    quint64 _allbytes = 0, _donebytes = 0;
    QList<OperationStepClass*> _mystepqueue = stepqueue;
    if (currentstep)
        _mystepqueue.append(currentstep); // this is required since in some situation, a logical member is technically not in the queue anymore (but something like 'in process')
    foreach(OperationStepClass* s, _mystepqueue)
    {
        if (s->conflictResolution() != OperationStepClass::ConflictResolution::Skip)
        {
            _allitems += s->_cntItems;
            _allbytes += s->_cntBytes;
        }
    }
    foreach(OperationStepClass* s, donesteps)
    {
        _allitems += s->_cntItems;
        _allbytes += s->_cntBytes;
        _doneitems += s->_cntItems;
        _donebytes += s->_cntBytes;
    }
    progressmon->setProgress(_doneitems, _allitems, _donebytes, _allbytes);
}

void sh::base::FilesystemOperationTransfers::executestepqueue()
{
    doreview(stepqueue);
    computeProgress();
    progressmon->_enablestatistics();
    while(!stepqueue.isEmpty())
    {
        OperationStepClass* step = currentstep = stepqueue.takeFirst();
        if (step->conflictResolution() != OperationStep::ConflictResolution::Skip)
        {
            bool _isretry = false;
            sh::exceptions::Exception::executeGuarded([&](){
                if (_isretry)
                    computeProgress();
                _isretry = true;
                if (doreview(QList<OperationStepClass*>({step})))
                    computeProgress();
                if (step->conflictResolution() != OperationStep::ConflictResolution::Skip)
                {
                    QString sourcestring = (step->source()) ? step->source()->asString() : "";
                    QString destinationstring = (step->effectiveDestination()) ? step->effectiveDestination()->asString() : "";
                    progressmon->setItemInfo(sourcestring, destinationstring);
                    if (step->conflictResolution() == OperationStep::ConflictResolution::OverwriteDestination)
                        sh::exceptions::Exception::executeGuarded([&](){
                            filesystem->deleteItem(step->effectiveDestination());
                        }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED | sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
                    else if (step->conflictResolution() == OperationStep::ConflictResolution::RenameDestinationBefore)
                        filesystem->moveItem(step->effectiveDestination(), step->effectiveDestination()->parentSegment()->withAppendedSegment(step->conflictResolution_renameDestinationBeforeTo()), std::make_shared<FilesystemOperationProgressMonitor>(progressmon->_actionExecution));
                    step->execute(filesystem->_operation);
                    if (!step->hasOwnProgressIncreaseHandling())
                        progressmon->incProgress(step->_cntItems, 0, step->_cntBytes, 0);
                    donesteps.append(step);
                } else
                    delete step;
            }, sh::exceptions::ExecuteGuardFlag::MANUAL_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED);
        } else
            delete step;
        progressmon->respectCancel();
        currentstep = 0;
    }
}

bool sh::base::FilesystemOperationTransfers::doreview(QList<OperationStepClass*> reviewsteps)
{
    QList<OperationStep*> conflictsteps;
    QList<OperationStepClass*> cconflictsteps;
    for(int i=0; i<reviewsteps.length(); i++) // this list can increase during the iteration and we iterate over all new steps as well.
    {
        OperationStepClass* reviewstep = reviewsteps[i];
        if (reviewstep->checkconflicts(filesystem->_operation))
        {
            // conflict occurred
            bool realconflict = true;
            if (reviewstep->_parentnode)
            {
                OperationStep::ConflictResolution presolution = reviewstep->_parentnode->conflictResolution();
                if ((presolution==OperationStep::ConflictResolution::MergeDirectories)
                                && reviewstep->setConflictResolve_MergeDirectories_checkAllowed())
                {
                    reviewstep->setConflictResolve_MergeDirectories();
                    realconflict = false;
                }
                else if ((presolution==OperationStep::ConflictResolution::Indirect)
                                || (presolution==OperationStep::ConflictResolution::OverwriteDestination)
                                || (presolution==OperationStep::ConflictResolution::RenameDestinationBefore))
                {
                    reviewstep->_conflictResolution = OperationStep::ConflictResolution::Indirect;
                    realconflict = false;
                }
            }
            if (realconflict)
            {
                conflictsteps.append(reviewstep);
                reviewstep->_unexpand(&reviewsteps);
            }
            cconflictsteps.append(reviewstep);
        }
        else
        {
            // no conflict occurred
            if (reviewstep->conflictResolution() != OperationStep::ConflictResolution::Skip)
                reviewstep->_expand(&reviewsteps);
        }
    }
    if (!conflictsteps.isEmpty())
        progressmon->resolveConflicts(conflictsteps);
    if (!cconflictsteps.isEmpty())
        doreview(cconflictsteps);
    return !cconflictsteps.isEmpty();
}

sh::base::FilesystemOperationTransfers::~FilesystemOperationTransfers()
{
    progressmon->_stoptriggerchanged();
    foreach(OperationStepClass* sc, stepqueue)
        delete sc;
    foreach(OperationStepClass* sc, donesteps)
        delete sc;
    if (currentstep)
        delete currentstep;
}

sh::base::FilesystemOperationTransfers::OperationStep::OperationStep(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> source, std::shared_ptr<const Eurl> destination) : _transfers(transfers), _source(source), _destination(destination)
{
}

sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution sh::base::FilesystemOperationTransfers::OperationStep::conflictResolution()
{
    return _conflictResolution;
}

QString sh::base::FilesystemOperationTransfers::OperationStep::conflictDescription()
{
    return _conflictDescription;
}

QString sh::base::FilesystemOperationTransfers::OperationStep::conflictResolution_renameDestinationBeforeTo()
{
    return _conflictResolution_renameDestinationBefore;
}

QString sh::base::FilesystemOperationTransfers::OperationStep::conflictResolution_differentDestinationNameTo()
{
    return _conflictResolution_differentDestinationName;
}

std::shared_ptr<const sh::base::Eurl> sh::base::FilesystemOperationTransfers::OperationStep::source()
{
    return _source;
}

std::shared_ptr<const sh::base::Eurl> sh::base::FilesystemOperationTransfers::OperationStep::originalDestination()
{
    return _destination;
}

std::shared_ptr<const sh::base::Eurl> sh::base::FilesystemOperationTransfers::OperationStep::effectiveDestination()
{
    if (_conflictResolution == OperationStep::ConflictResolution::UseDifferentDestinationName)
        return _destination->parentSegment()->withAppendedSegment(_conflictResolution_differentDestinationName);
    else
        return _destination;
}

sh::base::FilesystemNodeType sh::base::FilesystemOperationTransfers::OperationStep::sourcetype()
{
    return _transfers->filesystem->getType(_source);
}

void sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_Skip()
{
    _conflictResolution = sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::Skip;
}

void sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_OverwriteDestination()
{
    _conflictResolution = sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::OverwriteDestination;
}

void sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_MergeDirectories()
{
    _conflictResolution = sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::MergeDirectories;
}

bool sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_MergeDirectories_checkAllowed()
{
    return (sourcetype() == sh::base::FilesystemNodeType::Directory)
                    && (_transfers->filesystem->getType(originalDestination()) == sh::base::FilesystemNodeType::Directory);
}

void sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_RenameDestinationBefore(QString newname)
{
    _conflictResolution = sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::RenameDestinationBefore;
    _conflictResolution_renameDestinationBefore = newname;
}

void sh::base::FilesystemOperationTransfers::OperationStep::setConflictResolve_UseDifferentDestinationName(QString newname)
{
    _conflictResolution = sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::UseDifferentDestinationName;
    _conflictResolution_differentDestinationName = newname;
}

sh::base::FilesystemOperationTransfers::OperationStep::~OperationStep()
{
}

sh::base::FilesystemOperationTransfers::OperationStepClass::OperationStepClass(FilesystemOperationTransfers *transfers, std::shared_ptr<const Eurl> source, std::shared_ptr<const Eurl> destination, QList<OperationStepClass *> withExpansion) :
        sh::base::FilesystemOperationTransfers::OperationStep::OperationStep(transfers, source, destination), _withExpansion(withExpansion)
{
    foreach(OperationStepClass* step, _withExpansion)
        step->_deleteonunexpand = false;
}

bool sh::base::FilesystemOperationTransfers::OperationStepClass::checkconflicts(Operation *op)
{
    _conflictDescription = QString();
    if (_conflictResolution == sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::Skip
                    || _conflictResolution == sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::Indirect
                    || _conflictResolution == sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::OverwriteDestination)
        return false;
    else if (_conflictResolution == sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::MergeDirectories)
    {
        if (setConflictResolve_MergeDirectories_checkAllowed())
            return false;
        else
        {
            std::shared_ptr<const sh::base::Eurl> effectiveTarget = effectiveDestination();
            if (op->filesystem()->getType(effectiveTarget) == sh::base::FilesystemNodeType::NONE)
                return false;
            else
            {
                _conflictDescription = QObject::tr("Merging '%1' into '%2' is not allowed.").arg(effectiveTarget->basename()).arg(effectiveTarget->parentSegment()->asString());
                return true;
            }
        }
    }
    else if (_conflictResolution == sh::base::FilesystemOperationTransfers::OperationStep::ConflictResolution::RenameDestinationBefore)
    {
        std::shared_ptr<const sh::base::Eurl> newDestinationName = this->_destination->parentSegment()->withAppendedSegment(_conflictResolution_renameDestinationBefore);
        if (op->filesystem()->getType(newDestinationName) == sh::base::FilesystemNodeType::NONE)
            return false;
        else
        {
            _conflictDescription = QObject::tr("'%1' already exists in '%2'.").arg(newDestinationName->basename()).arg(newDestinationName->parentSegment()->asString());
            return true;
        }
    }
    else
    {
        std::shared_ptr<const sh::base::Eurl> effectiveTarget = effectiveDestination();
        if (op->filesystem()->getType(effectiveTarget) == sh::base::FilesystemNodeType::NONE)
            return false;
        else
        {
            _conflictDescription = QObject::tr("'%1' already exists in '%2'.").arg(effectiveTarget->basename()).arg(effectiveTarget->parentSegment()->asString());
            return true;
        }
    }
}

bool sh::base::FilesystemOperationTransfers::OperationStepClass::hasOwnProgressIncreaseHandling()
{
    return false;
}

QList<QPair<sh::base::DetailColumn*, QVariant>> sh::base::FilesystemOperationTransfers::OperationStepClass::getFileDetails(Operation* op, std::shared_ptr<const sh::base::Eurl> source)
{
    QList<QPair<sh::base::DetailColumn*, QVariant>> result;
    foreach(std::shared_ptr<sh::base::DetailColumn> detailColumn, _transfers->filesystem->transferrableDetailColumns())
    {
        QVariant value = detailColumn->computeValue(sh::base::FilesystemModel::instance()->findNodesForEurl(source).first(), op);
        result.append(QPair<sh::base::DetailColumn*, QVariant>(detailColumn.get(), value));
    }
    return result;
}

void sh::base::FilesystemOperationTransfers::OperationStepClass::applyFileDetails(Operation* op, QList<QPair<sh::base::DetailColumn*, QVariant>> values, std::shared_ptr<const sh::base::Eurl> destination)
{
    auto itr = values.begin();
    while (itr != values.end())
    {
        itr->first->applyValueByEurl(destination, op, itr->second);
        itr++;
    }
}

QList<sh::base::FilesystemOperationTransfers::OperationStepClass *> sh::base::FilesystemOperationTransfers::OperationStepClass::expand()
{
    return _withExpansion;
}

void sh::base::FilesystemOperationTransfers::OperationStepClass::_unexpand(QList<OperationStepClass *> *intlist)
{
    if (_isExpanded)
    {
        foreach(OperationStepClass* e, _expandnodes)
        {
            e->_unexpand(intlist);
            intlist->removeOne(e);
            _transfers->stepqueue.removeOne(e);
            if (e->_deleteonunexpand)
                delete e;
        }
        _expandnodes.clear();
        _isExpanded = false;
    }
}

void sh::base::FilesystemOperationTransfers::OperationStepClass::_expand(QList<OperationStepClass *> *intlist)
{
    if (!_isExpanded)
    {
        int i = _transfers->stepqueue.indexOf(this);
        QList<OperationStepClass *> expanded = expand();
        foreach(OperationStepClass* e, expanded)
        {
            intlist->append(e);
            _transfers->stepqueue.insert(++i, e);
            _expandnodes.append(e);
            e->_parentnode = this;
        }
        _isExpanded = true;
        if (!expanded.isEmpty())
            _transfers->computeProgress();
    }
}

sh::base::FilesystemOperationTransfers::OperationStepClass::~OperationStepClass()
{
    if (!_isExpanded)
    {
        foreach(OperationStepClass* e, _withExpansion)
            delete e;
    }
}

sh::base::FilesystemOperationTransfers::OperationStep_CopyFile::OperationStep_CopyFile(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion) :
         sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, src, dest, withExpansion)
{
    _cntItems = 1;
    _cntBytes = transfers->filesystem->_handler(src)->getSize(transfers->filesystem->_operation, src);
}

sh::base::FilesystemNodeType sh::base::FilesystemOperationTransfers::OperationStep_CopyFile::sourcetype()
{
    return sh::base::FilesystemNodeType::File;
}

void sh::base::FilesystemOperationTransfers::OperationStep_CopyFile::execute(Operation *op)
{
    std::shared_ptr<QIODevice> srcstream = op->filesystem()->getFileContent(_source);
    if (!srcstream->isOpen() && !srcstream->open(QFile::ReadOnly))
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Error while opening: ") + _source->asString()).autoRetryRecommended(100));
    SingleStepMonitor handlertransfer(this);
    std::shared_ptr<const Eurl> effectiveDest = effectiveDestination();
    op->filesystem()->_handler(effectiveDest)->createFile(op, effectiveDest, srcstream.get(), &handlertransfer);
    applyFileDetails(op, getFileDetails(op, _source), effectiveDest);
    _transfers->progressmon->addChangedEurl(effectiveDest);
}


bool sh::base::FilesystemOperationTransfers::OperationStep_CopyFile::hasOwnProgressIncreaseHandling()
{
    return true;
}

sh::base::FilesystemOperationTransfers::OperationStep_CopyDirectory::OperationStep_CopyDirectory(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass *> withExpansion):
    sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, src, dest, withExpansion)
{
    _cntItems = 1;
    _cntBytes = 0;
    FilesystemModel::instance()->findNodesForEurl(src).first()->requestContainedItems(); // without this call, a big performance penalty would occur.
}

sh::base::FilesystemNodeType sh::base::FilesystemOperationTransfers::OperationStep_CopyDirectory::sourcetype()
{
    return sh::base::FilesystemNodeType::Directory;
}

QList<sh::base::FilesystemOperationTransfers::OperationStepClass*> sh::base::FilesystemOperationTransfers::OperationStep_CopyDirectory::expand()
{
    QList<QPair<sh::base::DetailColumn*, QVariant>> detailvals = getFileDetails(_transfers->filesystem->_operation, _source);
    QList<OperationStepClass*> expanded;
    std::shared_ptr<const Eurl> effectiveDest = effectiveDestination();
    foreach(std::shared_ptr<const Eurl> n, _transfers->filesystem->itemlist(_source))
        _transfers->_compile_copyItem(n, effectiveDest->withAppendedSegment(n->basename()), &expanded, QList<OperationStepClass*>());
    expanded.append(new OperationStep_ApplyDirectoryDetails(_transfers, effectiveDest, detailvals, QList<OperationStepClass*>()));
    expanded.append(OperationStepClass::expand());
    return expanded;
}

void sh::base::FilesystemOperationTransfers::OperationStep_CopyDirectory::execute(Operation *op)
{
    if (conflictResolution() != OperationStep::ConflictResolution::MergeDirectories)
    {
        std::shared_ptr<const sh::base::Eurl> effectiveDest = effectiveDestination();
        op->filesystem()->createDirectory(effectiveDest);
        _transfers->progressmon->addChangedEurl(effectiveDest);
    }_transfers->doreview(_expandnodes);
}

sh::base::FilesystemOperationTransfers::OperationStep_CopyLink::OperationStep_CopyLink(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion):
    sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, src, dest, withExpansion)
{
    _cntItems = 1;
    _cntBytes = 0;
}

sh::base::FilesystemNodeType sh::base::FilesystemOperationTransfers::OperationStep_CopyLink::sourcetype()
{
    return sh::base::FilesystemNodeType::Link;
}

void sh::base::FilesystemOperationTransfers::OperationStep_CopyLink::execute(Operation *op)
{
    std::shared_ptr<const Eurl> effectiveDest = effectiveDestination();
    QString linktgt = op->filesystem()->getLinkTarget(_source);
    op->filesystem()->createLink(effectiveDest, linktgt);
    applyFileDetails(op, getFileDetails(op, _source), effectiveDest);
}

sh::base::FilesystemOperationTransfers::OperationStep_RenameItem::OperationStep_RenameItem(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> src, QString destpath, QList<OperationStepClass*> withExpansion):
    sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, src, src->root()->withAppendedSegments(destpath), withExpansion)
{
    _cntItems = 1;
    _cntBytes = 0;
}

void sh::base::FilesystemOperationTransfers::OperationStep_RenameItem::execute(Operation *op)
{
    if (conflictResolution() != OperationStepClass::ConflictResolution::MergeDirectories)
    {
        FilesystemHandlerRegister::instance()->findHandler(_source->scheme())->renameItem(op, _source, effectiveDestination()->path());
        _transfers->progressmon->addChangedEurl(_source);
        _transfers->progressmon->addChangedEurl(effectiveDestination());
    }
}

QList<sh::base::FilesystemOperationTransfers::OperationStepClass*> sh::base::FilesystemOperationTransfers::OperationStep_RenameItem::expand()
{
    QList<OperationStepClass*> expanded;
    if (conflictResolution() == OperationStepClass::ConflictResolution::MergeDirectories)
    {
        foreach(std::shared_ptr<const Eurl> n, _transfers->filesystem->itemlist(_source))
            _transfers->_compile_moveItem(n, effectiveDestination()->withAppendedSegment(n->basename()), &expanded, QList<OperationStepClass*>());
        expanded.append(new OperationStep_DeleteItem(_transfers, _source, _source, QList<OperationStepClass*>()));
    }
    expanded.append(OperationStepClass::expand());
    return expanded;
}

sh::base::FilesystemOperationTransfers::OperationStep_DeleteItem::OperationStep_DeleteItem(FilesystemOperationTransfers *transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion):
    sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, src, dest, withExpansion)
{
}

sh::base::FilesystemNodeType sh::base::FilesystemOperationTransfers::OperationStep_DeleteItem::sourcetype()
{
    return sh::base::FilesystemNodeType::NONE;
}

void sh::base::FilesystemOperationTransfers::OperationStep_DeleteItem::execute(Operation *op)
{
    op->filesystem()->deleteItem(source());
}

bool sh::base::FilesystemOperationTransfers::OperationStep_DeleteItem::checkconflicts(Operation*)
{
    return false;
}

sh::base::FilesystemOperationTransfers::OperationStep_ApplyDirectoryDetails::OperationStep_ApplyDirectoryDetails(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> item, QList<QPair<sh::base::DetailColumn*, QVariant>> values, QList<OperationStepClass*> withExpansion) :
    sh::base::FilesystemOperationTransfers::OperationStepClass(transfers, 0, item, withExpansion), _detailvals(values)
{
}

void sh::base::FilesystemOperationTransfers::OperationStep_ApplyDirectoryDetails::execute(Operation* op)
{
    applyFileDetails(op, _detailvals, _destination);
}

bool sh::base::FilesystemOperationTransfers::OperationStep_ApplyDirectoryDetails::checkconflicts(Operation*)
{
    return false;
}

sh::base::FilesystemOperationProgressMonitor::FilesystemOperationProgressMonitor(sh::actions::ActionExecutionInfo* actionExecution)
            : _actionExecution(actionExecution)
{
}

void sh::base::FilesystemOperationProgressMonitor::_enablestatistics()
{
    {
        QMutexLocker _l1(&_mutex);
        if (_begancomputestatistics)
            return;
        _begancomputestatistics = true;
    }
    _laststatistictime = QDateTime::currentDateTime();
    _computestatistics();
}

void sh::base::FilesystemOperationProgressMonitor::_computestatistics()
{
    std::weak_ptr<sh::base::FilesystemOperationProgressMonitor> _wptr = shared_from_this();
    sh::tools::ThreadPool::enqueueForceBeyondAsyncCallBarrier([=](){
        std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> _ptr = _wptr.lock();
        {
            if (!_ptr)
                return;
            QMutexLocker _l1(&_triggermutex);
            if (_triggerchanged_stopped)
                return;
        }
        bool makestatistics;
        if (_statisticcountdown)
        {
            _statisticcountdown--;
            makestatistics = false;
        }
        else
            makestatistics = true;
        // compute statistics
        quint64 doneitems, allitems, donebytes, allbytes;
        quint64 laststatisticdoneitems, laststatisticdonebytes;
        QDateTime laststatistictime, thisstatistictime;
        {
            QMutexLocker _l1(&_mutex);
            doneitems = _doneitems; allitems = _allitems; donebytes = _donebytes; allbytes = _allbytes;
            laststatistictime = _laststatistictime;
            laststatisticdonebytes = _laststatisticdonebytes;
            laststatisticdoneitems = _laststatisticdoneitems;
            _laststatisticdonebytes = _donebytes;
            _laststatisticdoneitems = _doneitems;
            _laststatistictime = thisstatistictime = QDateTime::currentDateTime();
        }
        bool bytemode = hasBytesInfo();
        if (bytemode || hasItemInfo())
        {
            double duration = (double)(laststatistictime.msecsTo(thisstatistictime));
            double currspeed = (1000.0 * (double)(bytemode ? (donebytes - laststatisticdonebytes) : (doneitems - laststatisticdoneitems) )) / duration;
            if (std::isfinite(currspeed))
            {
                if (bytemode)
                    currspeed = _statisticbytespeed = (0.7 *_statisticbytespeed) + (0.3 * currspeed);
                else
                    currspeed = _statisticitemspeed = (0.7 *_statisticitemspeed) + (0.3 * currspeed);
                QString estimation;
                if (bytemode)
                {
                    QString lunit = QObject::tr("byte");
                    if (currspeed>1024)
                    {
                        lunit = QObject::tr("kilobyte");
                        currspeed = currspeed / 1024.0;
                        if (currspeed>1024)
                        {
                            lunit = QObject::tr("megabyte");
                            currspeed = currspeed / 1024.0;
                            if (currspeed>1024)
                            {
                                lunit = QObject::tr("gigabyte");
                                currspeed = currspeed / 1024.0;
                                if (currspeed>1024)
                                {
                                    lunit = QObject::tr("terabyte");
                                    currspeed = currspeed / 1024.0;
                                }
                            }
                        }
                    }
                    estimation = QObject::tr("%1 %2/second").arg(currspeed, 0, 'f', 2).arg(lunit);
                }
                else
                {
                    QString tunit = QObject::tr("second");
                    if (currspeed<10)
                    {
                        tunit = QObject::tr("minute");
                        currspeed *= 60;
                        if (currspeed<10)
                        {
                            tunit = QObject::tr("hour");
                            currspeed *= 60;
                            if (currspeed<10)
                            {
                                tunit = QObject::tr("day");
                                currspeed *= 24;
                            }
                        }
                    }
                    estimation = QObject::tr("%1 items/%2").arg((qint64)currspeed).arg(tunit);
                }
                if (makestatistics)
                {
                    double remaining;
                    if (bytemode)
                        remaining = (double)(allbytes-donebytes) / _statisticbytespeed;
                    else
                        remaining = (double)(allitems-doneitems) / _statisticitemspeed;
                    QString runit = QObject::tr("seconds");
                    if (remaining > 100)
                    {
                        remaining /= 60.0;
                        runit = QObject::tr("minutes");
                        if (remaining > 100)
                        {
                            remaining /= 60.0;
                            runit = QObject::tr("hours");
                            if (remaining > 30)
                            {
                                remaining /= 24.0;
                                runit = QObject::tr("days");
                            }
                        }
                    }
                    setEstimation(estimation + " (" + QObject::tr("%1 %2 remaining").arg((int)remaining).arg(runit) + ")");
                }
            }
        }
        // end compute statistics
        QThread::msleep(2000);
        _computestatistics();
    }, std::shared_ptr<void>());
}

void sh::base::FilesystemOperationProgressMonitor::setProgress(quint64 doneitems, quint64 allitems, quint64 donebytes, quint64 allbytes)
{
    QMutexLocker _l1(&_mutex);
    _doneitems = doneitems;
    _allitems = allitems;
    _donebytes = donebytes;
    _allbytes = allbytes;
    _triggerchanged();
}

void sh::base::FilesystemOperationProgressMonitor::incProgress(quint64 doneitems, quint64 allitems, quint64 donebytes, quint64 allbytes)
{
    QMutexLocker _l1(&_mutex);
    _doneitems += doneitems;
    _allitems += allitems;
    _donebytes += donebytes;
    _allbytes += allbytes;
    _triggerchanged();
}

void sh::base::FilesystemOperationProgressMonitor::setItemInfo(QString from, QString to)
{
    QMutexLocker _l1(&_mutex);
    _itemfrom = from;
    _itemto = to;
    _triggerchanged();
}

void sh::base::FilesystemOperationProgressMonitor::setEstimation(QString estimation)
{
    QMutexLocker _l1(&_mutex);
    _estimation = estimation;
    _triggerchanged();
}

bool sh::base::FilesystemOperationProgressMonitor::hasItemInfo()
{
    QMutexLocker _l1(&_mutex);
    return _allitems > 0;
}

quint64 sh::base::FilesystemOperationProgressMonitor::allItems()
{
    QMutexLocker _l1(&_mutex);
    return _allitems;
}

quint64 sh::base::FilesystemOperationProgressMonitor::doneItems()
{
    QMutexLocker _l1(&_mutex);
    return _doneitems;
}

bool sh::base::FilesystemOperationProgressMonitor::hasBytesInfo()
{
    QMutexLocker _l1(&_mutex);
    return _allbytes > 0;
}

quint64 sh::base::FilesystemOperationProgressMonitor::allBytes()
{
    QMutexLocker _l1(&_mutex);
    return _allbytes;
}

quint64 sh::base::FilesystemOperationProgressMonitor::doneBytes()
{
    QMutexLocker _l1(&_mutex);
    return _donebytes;
}

QString sh::base::FilesystemOperationProgressMonitor::getItemInfoFrom()
{
    QMutexLocker _l1(&_mutex);
    return _itemfrom;
}

QString sh::base::FilesystemOperationProgressMonitor::getItemInfoTo()
{
    QMutexLocker _l1(&_mutex);
    return _itemto;
}

QString sh::base::FilesystemOperationProgressMonitor::estimation()
{
    QMutexLocker _l1(&_mutex);
    return _estimation;
}

void sh::base::FilesystemOperationProgressMonitor::changed()
{
}

void sh::base::FilesystemOperationProgressMonitor::resolveConflicts(QList<sh::base::FilesystemOperationTransfers::OperationStep*> conflictsteps)
{
    foreach(sh::base::FilesystemOperationTransfers::OperationStep* step, conflictsteps)
        step->setConflictResolve_OverwriteDestination();
}

void sh::base::FilesystemOperationProgressMonitor::respectCancel()
{
    if (_actionExecution)
        _actionExecution->respectCancel();
}

void sh::base::FilesystemOperationProgressMonitor::addChangedEurl(std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (_actionExecution)
        _actionExecution->addChangedEurl(eurl);
    else
        changes.append(eurl);
}

void sh::base::FilesystemOperationProgressMonitor::_triggerchanged()
{
    QMutexLocker _l1(&_triggermutex);
    if (_triggerchanged_stopped)
        return;
    if (_triggerchanged_changedrunning)
        _triggerchanged_changedrunagain = true;
    else
    {
        _triggerchanged_changedrunning = true;
        std::weak_ptr<sh::base::FilesystemOperationProgressMonitor> _wptr = shared_from_this();
        sh::tools::ThreadPool::enqueueForceBeyondAsyncCallBarrier([=](){
            bool _exectc = false;
            std::shared_ptr<sh::base::FilesystemOperationProgressMonitor> _ptr = _wptr.lock();
            {
                if (!_ptr)
                    return;
                QMutexLocker _l1(&_triggermutex);
                if (_triggerchanged_stopped)
                    return;
            }
            changed();
            {
                QMutexLocker _l1(&_triggermutex);
                if (_triggerchanged_stopped)
                    return;
                _triggerchanged_changedrunning = false;
                if (_triggerchanged_changedrunagain)
                {
                    _exectc = true;
                    _triggerchanged_changedrunagain = false;
                }
            }
            if (_exectc)
            {
                IGNORE_FORBIDDEN_ASYNC_CALLS
                _triggerchanged();
            }
        }, std::shared_ptr<void>());
    }
}

void sh::base::FilesystemOperationProgressMonitor::_stoptriggerchanged()
{
    QMutexLocker _l1(&_triggermutex);
    _triggerchanged_stopped = true;
}

sh::base::FilesystemOperationProgressMonitor::~FilesystemOperationProgressMonitor()
{
    foreach(std::shared_ptr<const sh::base::Eurl> changedEurl, changes)
        sh::base::FilesystemModel::instance()->refreshData(changedEurl);
}

sh::base::FilesystemOperationTransfers::SingleStepMonitor::SingleStepMonitor(OperationStepClass* step) : step(step)
{
}

void sh::base::FilesystemOperationTransfers::SingleStepMonitor::respectCancel()
{
    step->_transfers->progressmon->respectCancel();
}

void sh::base::FilesystemOperationTransfers::SingleStepMonitor::incrementTransferredBytes(qint64 donebytes)
{
    qint64 olddonebytes = _donebytes;
    _donebytes += donebytes;
    if (_donebytes > (qint64)step->_cntBytes)
        _donebytes = step->_cntBytes;
    if (_donebytes < 0)
        _donebytes = 0;
    int deltadonebytes = _donebytes - olddonebytes;
    step->_transfers->progressmon->incProgress(0, 0, deltadonebytes, 0);
}

sh::base::FilesystemOperationTransfers::SingleStepMonitor::~SingleStepMonitor()
{
    step->_transfers->progressmon->incProgress(step->_cntItems, 0, step->_cntBytes - _donebytes, 0);
}
