/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILESYSTEMOPERATIONTRANSFERS_H
#define FILESYSTEMOPERATIONTRANSFERS_H

#include <cmath>
#include <memory>
#include <QMutex>
#include <QObject>
#include "base/filesystemhandler.h"
#include "base/filesystemnode.h"


namespace sh{
    namespace actions {
        class ActionExecutionInfo;
    }
    namespace base{

        class FilesystemOperationProgressMonitor;
        class FilesystemOperation;
        class Operation;
        class Eurl;

        /**
         * @brief Used by FilesystemOperation for managing the execution of
         *        transfer operations.
         *
         * This class is only used directly by FilesystemOperation (but some inner class might be interesting).
         */
        class FilesystemOperationTransfers
        {
        private:
            class SingleStepMonitor;
        public:

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            FilesystemOperationTransfers(FilesystemOperation* filesystem, std::shared_ptr<FilesystemOperationProgressMonitor> progressmon);

            /**
             * @brief One step of execution in a transfer operation by FilesystemOperation.
             *
             * It is typically about a certain source and destination in the filesystem.
             * It also can stay in conflict (due to checks from FilesystemOperationTransfers::doreview)
             * and provides methods for resolving them (in FilesystemOperationTransfers::FilesystemOperationProgressMonitor::resolveConflicts).
             *
             * Different subclasses implement distinct kinds of operations (e.g. copying
             * or deleting).
             */
            class OperationStep
            {
                friend class FilesystemOperationTransfers;
             public:
                /**
                 * @brief Enumeration of ways how to resolve a filesystem conflict.
                 */
                enum class ConflictResolution
                {
                    /** \brief Skip this element.
                      */
                    Skip,
                    /** \brief Overwrite the destination.
                      */
                    OverwriteDestination,
                    /** \brief Rename the file at the destination before transferring.
                      */
                    RenameDestinationBefore,
                    /** \brief Transfer to another destination filename.
                      */
                    UseDifferentDestinationName,
                    /** \brief Merge source into destination directory (recursively).
                      */
                    MergeDirectories,
                    /** \brief No strategy.
                      */
                    Unresolved,
                    /** \brief Indirectly solved. Only set by the engine.
                      */
                    Indirect

                };

                /** \brief Constructed only by the infrastructure and made available otherwise.
                 */
                OperationStep(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> source, std::shared_ptr<const sh::base::Eurl> destination);
                /** \brief The description of the conflict (if any) as text.
                 */
                QString conflictDescription();
                /** \brief The currently chosen conflict resolution.
                 */
                ConflictResolution conflictResolution();
                /** \brief Returns new destination name (if current conflict resolution is ConflictResolution::RenameDestinationBefore).
                 */
                QString conflictResolution_renameDestinationBeforeTo();
                /** \brief Returns new destination name (if current conflict resolution is ConflictResolution::UseDifferentDestinationName).
                 */
                QString conflictResolution_differentDestinationNameTo();
                /** \brief Returns the source location to be transferred (if specified).
                 */
                std::shared_ptr<const sh::base::Eurl> source();
                /** \brief Returns the destination location (if specified).
                 *
                 * This is the complete target path, never the parent directory of the new element.
                 *
                 * This is the original destination location. See also effectiveDestination.
                 */
                std::shared_ptr<const sh::base::Eurl> originalDestination();
                /** \brief Returns the real destination location with conflict resolution applied.
                 */
                std::shared_ptr<const sh::base::Eurl> effectiveDestination();
                /** \brief Returns the node type of the source.
                 *
                 * Should be overwritten in subclasses whenever it can determine the source type faster than the infrastructure.
                 */
                virtual sh::base::FilesystemNodeType sourcetype();
                /** \brief Set conflict resolution to ConflictResolution::Skip.
                 */
                void setConflictResolve_Skip();
                /** \brief Set conflict resolution to ConflictResolution::OverwriteDestination.
                 */
                void setConflictResolve_OverwriteDestination();
                /** \brief Set conflict resolution to ConflictResolution::RenameDestinationBefore.
                 */
                void setConflictResolve_RenameDestinationBefore(QString newname);
                /** \brief Set conflict resolution to ConflictResolution::UseDifferentDestinationName.
                 */
                void setConflictResolve_UseDifferentDestinationName(QString newname);
                /** \brief Set conflict resolution to ConflictResolution::MergeDirectories.
                 *
                 * See also setConflictResolve_MergeDirectories_isAllowed.
                 */
                void setConflictResolve_MergeDirectories();
                /** \brief Checks if conflict resolution ConflictResolution::MergeDirectories is allowed.
                 */
                bool setConflictResolve_MergeDirectories_checkAllowed();
                virtual ~OperationStep();
            protected:
                FilesystemOperationTransfers* _transfers;
                std::shared_ptr<const sh::base::Eurl> _source;
                std::shared_ptr<const sh::base::Eurl> _destination;
                QString _conflictDescription;
                ConflictResolution _conflictResolution = ConflictResolution::Unresolved;
                QString _conflictResolution_renameDestinationBefore;
                QString _conflictResolution_differentDestinationName;
            };

            /**
             * @brief Executes the queue.
             *
             * The execution model is as follows:
             * The complete queue of OperationStepClass steps becomes expanded,
             * filesystem conflicts gets resolved and then all steps become executed.
             *
             * In detail:
             *
             * At first there is an interleaved process of expanding all steps
             * (see OperationStepClass::expand) and checking/resolving filesystem conflicts in the fresh tree
             * (see doreview). It checks all steps in the queue for conflicts (e.g. destination already exists). It expands each step
             * which has no conflicts (by enqueueing it to the stepqueue). After those checks, it asks
             * FilesystemOperationProgressMonitor::resolveConflicts for a resolution
             * for all outstanding conflicts. With all the steps, which were (or are) in conflict, the entire process
             * becomes repeated. This ends when all steps are expanded and no open conflicts exist anymore.
             *
             * When this process is done, the execution phase begins. It iterates the queue.
             * It takes the (in FIFO manner) first element (if one is there, otherwise stops). For that step, it makes a
             * late-time conflict check at first. If a conflict appears, it becomes temporarily unexpanded.
             * A resolution loop as above will run for this element, eventually leading to re-expansion.
             * After the check, when no conflicts are open, it executes the step.
             */
            void executestepqueue();

            /**
             * @brief adds one move transfer into the queue.
             */
            void addMoveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest);

            /**
             * @brief adds one copy transfer into the queue.
             */
            void addCopyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest);

            ~FilesystemOperationTransfers();

        private:

            /** @brief A base class for all implementations.
             *
             * They are used from inside FilesystemOperationTransfers::executestepqueue.
             */
            class OperationStepClass : public OperationStep
            {
            public:
                OperationStepClass(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> source, std::shared_ptr<const Eurl> destination, QList<OperationStepClass*> withExpansion);
                /**
                 * @brief Implement this and handle the execution part of this step here.
                 */
                virtual void execute(Operation* op) = 0;
                /**
                 * @brief Implement this and handle the expansion part of this step here.
                 */
                virtual QList<OperationStepClass*> expand();
                /**
                 * @brief Implement this for customizing the conflict checking.
                 */
                virtual bool checkconflicts(Operation* op);
                /** @brief Number of items to be transferred in this step (used for progress monitoring).
                 */
                quint64 _cntItems = 0;
                /** @brief Number of byte to be transferred in this step (used for progress monitoring).
                 */
                quint64 _cntBytes = 0;
                /** @brief If it takes care on its own to signal increases in the transfer progress.
                 */
                virtual bool hasOwnProgressIncreaseHandling();
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                bool _isExpanded = false;
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                QList<OperationStepClass*> _expandnodes;
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                OperationStepClass* _parentnode = 0;
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                void _unexpand(QList<OperationStepClass*>* intlist);
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                void _expand(QList<OperationStepClass*>* intlist);
                /** @brief Bookkeeping in FilesystemOperationTransfers.
                 */
                bool _deleteonunexpand = true;

                QList<OperationStepClass*> _withExpansion;
                ~OperationStepClass();

                QList<QPair<sh::base::DetailColumn*, QVariant>> getFileDetails(sh::base::Operation *op, std::shared_ptr<const sh::base::Eurl> source);
                void applyFileDetails(sh::base::Operation *op, QList<QPair<sh::base::DetailColumn*, QVariant>> values, std::shared_ptr<const sh::base::Eurl> destination);
            };

            /**
             * @brief Used for managing detailled progress changes about a single step.
             */
            class SingleStepMonitor : public sh::base::FilesystemHandler::HandlerTransfer
            {
            public:
                SingleStepMonitor(OperationStepClass* step);
                virtual void respectCancel();
                virtual void incrementTransferredBytes(qint64 donebytes);
                virtual ~SingleStepMonitor();
            private:
                OperationStepClass* step;
                qint64 _donebytes = 0;
            };

            /** @brief Adds a move transfer step into given list.
             */
            void _compile_moveItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass *> *result, QList<OperationStepClass *> withExpansion);

            /** @brief Adds a copy transfer step into given list.
             */
            void _compile_copyItem(std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass *> *result, QList<OperationStepClass *> withExpansion);

            /**
             * @brief Computes the current progress (how many items/bytes are transferred from which total?)
             *        and notifies the progress monitor.
             */
            void computeProgress();

            /**
             * @brief Makes checks if the given queue has conflicts
             *        and for asks `resolveConflicts` how to handle them.
             *
             * For each non-conflicting step, it triggers its expansion,
             * thereby modifying the stepqueue.
             *
             * It returns when no open conflicts exist.
             *
             * @return if there were any conflicts.
             */
            bool doreview(QList<OperationStepClass *> reviewsteps);

            /** @brief The queue of steps for execution.
             */
            QList<OperationStepClass *> stepqueue;

            /** @brief The step which is currently in execution (or 0).
             */
            OperationStepClass* currentstep = 0;

            /** @brief Steps which were already executed.
             */
            QList<OperationStepClass *> donesteps;

            FilesystemOperation* filesystem;

            std::shared_ptr<FilesystemOperationProgressMonitor> progressmon;

            class OperationStep_CopyFile : public OperationStepClass
            {
             public:
                OperationStep_CopyFile(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion);
                virtual void execute(Operation* op);
                virtual sh::base::FilesystemNodeType sourcetype();
                virtual bool hasOwnProgressIncreaseHandling();
            };

            class OperationStep_CopyDirectory : public OperationStepClass
            {
             public:
                OperationStep_CopyDirectory(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion);
                virtual void execute(Operation* op);
                virtual QList<OperationStepClass*> expand();
                virtual sh::base::FilesystemNodeType sourcetype();
            };

            class OperationStep_CopyLink : public OperationStepClass
            {
             public:
                OperationStep_CopyLink(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion);
                virtual void execute(Operation* op);
                virtual sh::base::FilesystemNodeType sourcetype();
            };

            class OperationStep_RenameItem : public OperationStepClass
            {
             public:
                OperationStep_RenameItem(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> src, QString destpath, QList<OperationStepClass*> withExpansion);
                virtual QList<OperationStepClass*> expand();
                virtual void execute(Operation* op);
            };

            class OperationStep_DeleteItem : public OperationStepClass
            {
             public:
                OperationStep_DeleteItem(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> src, std::shared_ptr<const sh::base::Eurl> dest, QList<OperationStepClass*> withExpansion);
                virtual void execute(Operation* op);
                virtual bool checkconflicts(Operation* op);
                virtual sh::base::FilesystemNodeType sourcetype();
            };

            class OperationStep_ApplyDirectoryDetails : public OperationStepClass
            {
             public:
                OperationStep_ApplyDirectoryDetails(FilesystemOperationTransfers* transfers, std::shared_ptr<const sh::base::Eurl> item, QList<QPair<sh::base::DetailColumn*, QVariant>> values, QList<OperationStepClass*> withExpansion);
                virtual void execute(Operation* op);
                virtual bool checkconflicts(Operation* op);
                QList<QPair<sh::base::DetailColumn*, QVariant>> _detailvals;
            };

        };

        /**
         * @brief Implement this class and use an instance of it as parameter in some methods of
         * sh::base::FilesystemOperation for some additional functionality.
         *
         * This includes monitoring the progress, specifying a resolution
         * when conflicts in the filesystem would occur and more.
         *
         * It also makes sense to directly instantiate this class in some cases.
         */
        class FilesystemOperationProgressMonitor : public std::enable_shared_from_this<FilesystemOperationProgressMonitor>
        {
            friend class FilesystemOperationTransfers;
            friend class FilesystemOperation;
        public:
            /** \brief Is intended to be directly constructed from everywhere.
             *
             * @param actionExecution An optional sh::actions::ActionExecutionInfo which is used in some places in the default implementation,
             *                        when available.
             */
            explicit FilesystemOperationProgressMonitor(sh::actions::ActionExecutionInfo* actionExecution = 0);
            /**
             * @brief Checks if this progress monitor provides information about how many items of a certain total number
             *        are transferred so far.
             */
            bool hasItemInfo();
            /**
             * @brief Checks how many items are transferred so far.
             */
            quint64 doneItems();
            /**
             * @brief Checks how many items are to be transferred in total (as predicted in the current moment).
             */
            quint64 allItems();
            /**
             * @brief Checks if this progress monitor provides information about how many byte of a certain total number
             *        are transferred so far.
             */
            bool hasBytesInfo();
            /**
             * @brief Checks how many bytes are transferred so far.
             */
            quint64 doneBytes();
            /**
             * @brief Checks how many bytes are to be transferred in total (as predicted in the current moment).
             */
            quint64 allBytes();
            /**
             * @brief Returns the current source of transfer (as textual information).
             */
            QString getItemInfoFrom();
            /**
             * @brief Returns the current destination of transfer (as textual information).
             */
            QString getItemInfoTo();
            /**
             * @brief Returns the current performance and time estimation (as textual information).
             */
            QString estimation();
            virtual ~FilesystemOperationProgressMonitor();

        protected:
            /**
             * @brief Implement this for reacting on changes in some of the state infos.
             *
             * Might be used e.g. for refreshing an UI.
             */
            virtual void changed();
            /**
             * @brief Implement this for specifying how to resolve an upcoming conflict
             * in the filesystem.
             *
             * You must specify how to resolve the conflict for
             * each entry at some time (maybe by set it to 'skip' as last resort).
             * This method will be called until no conflicts are unresolved.
             *
             * The default implementation always decides to overwrite the destination.
             */
            virtual void resolveConflicts(QList<FilesystemOperationTransfers::OperationStep*> conflictsteps);
            /**
             * @brief Implement this for reacting on user cancellation requests, if any, in a custom way.
             *
             * The default implementation uses the associated sh::actions::ActionExecutionInfo::respectCancel, if any.
             */
            virtual void respectCancel();

            /**
             * @brief Called by the operations for notifying that a certain eurl was changed (e.g. was created,
             *        removed, updated, ...).
             *
             * If you overwrite it, you must call the base implementation inside your one.
             */
            virtual void addChangedEurl(std::shared_ptr<const sh::base::Eurl> eurl);

            sh::actions::ActionExecutionInfo* _actionExecution;
        private:
            /**
             * @brief Used by FilesystemOperationTransfers for setting status changes.
             */
            void setProgress(quint64 doneitems, quint64 allitems, quint64 donebytes, quint64 allbytes);
            /**
             * @brief Used by FilesystemOperationTransfers for setting status changes.
             */
            void incProgress(quint64 doneitems, quint64 allitems, quint64 donebytes, quint64 allbytes);
            /**
             * @brief Used by FilesystemOperationTransfers for setting status changes.
             */
            void setItemInfo(QString from, QString to);
            /**
             * @brief Used by FilesystemOperationTransfers for setting status changes.
             */
            void setEstimation(QString estimation);

            void _enablestatistics();
            void _computestatistics();
            QDateTime _laststatistictime;
            quint64 _laststatisticdonebytes = 0;
            quint64 _laststatisticdoneitems = 0;
            double _statisticbytespeed = 0.0;
            double _statisticitemspeed = 0.0;
            int _statisticcountdown = 4;

            void _triggerchanged();
            void _stoptriggerchanged();
            QMutex _mutex;
            QMutex _triggermutex;
            quint64 _allitems = 0, _doneitems = 0, _allbytes = 0, _donebytes = 0;
            bool _triggerchanged_changedrunning = false, _triggerchanged_changedrunagain = false, _triggerchanged_stopped = false;
            QString _itemfrom, _itemto;
            QString _estimation;
            QList<std::shared_ptr<const sh::base::Eurl>> changes;
            bool _begancomputestatistics = false;
        };

    }
}

#endif // FILESYSTEMOPERATIONTRANSFERS_H
