/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QApplication>
#include <QDataStream>
#include <QDir>
#include <QFile>
#include <QStorageInfo>
#include "base/eurl.h"
#include "base/filesystemhandler.h"
#include "base/filesystemhandlerregister.h"
#include "base/detailcolumn.h"
#include "base/filesystemoperation.h"
#include "exceptions/ioexception.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "tools/datastream.h"
#include "tools/mainthread.h"
#include "operation.h"


QString sh::base::Operation::_tempdir(QDir::tempPath());

sh::base::Operation::Operation(QObject *parent) :
    QObject(parent), _maxAllowedSizeRatioPerPart(-1), operationlock(QMutex::Recursive)
{
    _filesystemOperation = new sh::base::FilesystemOperation(this);
}

sh::base::Operation::~Operation()
{
    abort();
    delete _filesystemOperation;
}

QString sh::base::Operation::getTempDir()
{
    QString file;
    QFile tempfile;
    QString fullpath;
    {
        QMutexLocker _l1(&operationlock);
        QMutexLocker _l2(&_tempfilemutex);
        while (file.isNull() || QFile(_tempdir + "/" + file).exists())
            file = QString("shallot.%1.%2").arg(QCoreApplication::applicationPid()).arg(std::rand());
        fullpath = _tempdir + "/" + file;
        tempfile.setFileName(fullpath);
        if (!QDir(_tempdir).mkdir(file))
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while creating temporary dir '%1'").arg(fullpath)));
        tempDirs.append(fullpath);
    }
    return fullpath;
}

quint64 sh::base::Operation::getFreeCacheSpace()
{
    qint64 res = QStorageInfo(_tempdir).bytesFree();
    return (res > 0) ? res : 0;
}

QString sh::base::Operation::fetchContainerFile(std::shared_ptr<const sh::base::Eurl> eurl, QString namehint)
{
    std::shared_ptr<const sh::base::Eurl> containerEurl = eurl->outermostInnerEurl();
    return containerEurl ? fetchFile(containerEurl, namehint) : QString();
}

QString sh::base::Operation::fetchFile(std::shared_ptr<const sh::base::Eurl> eurl, QString namehint)
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QMutexLocker _l1(&operationlock);
    qint64 pid = QCoreApplication::applicationPid();
    QString result;
    if (files.contains(eurl))
    {
        result = files[eurl];
    }
    else
    {
        if (_maxAllowedSizeRatioPerPart >= 0) {
            if (!qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(sh::base::FilesystemHandlerRegister::instance()->findHandler(eurl->scheme()).get())) {
                qint64 contentsize = filesystem()->getFileSize(eurl);
                quint64 free = getFreeCacheSpace();
                if (!free || ((double)contentsize/free) > _maxAllowedSizeRatioPerPart)
                    throw MaxAllowedSizeRatioPerPartExceededException();
            }
        }
        std::shared_ptr<QIODevice> content = filesystem()->getFileContent(eurl);
        sh::tools::LocalFile* localContent = dynamic_cast<sh::tools::LocalFile*>(content.get());
        if (localContent)
        {
            result = QFileInfo(*localContent).absoluteFilePath();
            files.insert(eurl, result);
            fileIsTemporary.insert(eurl, localContent->isTemp());
        }
        else
        {
            QString file;
            {
                QMutexLocker _l(&_tempfilemutex);
                while ( file.isNull() || QFile(file).exists() )
                    file = _tempdir + QString("/shallot.%1.%2.%3").arg(pid).arg(std::rand()).arg(namehint);
            }
            if (!content->open(QFile::ReadOnly))
                throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to open '%1': %2").arg(eurl->asString()).arg(content->errorString())));
            try
            {
                writeIODeviceToFile(content.get(), file);
            }
            catch (sh::exceptions::Exception&)
            {
                content->close();
                throw;
            }
            content->close();
            files.insert(eurl, file);
            fileIsTemporary.insert(eurl, true);
            result = file;
        }
    }
    return result;
}

void sh::base::Operation::abort()
{
    QMutexLocker _l1(&operationlock);
    foreach(std::shared_ptr<const sh::base::Eurl> eurl, files.keys())
    {
        if ( fileIsTemporary[eurl] == true )
        {
            QString file = files[eurl];
            QFile(file).remove();
        }
    }
    files.clear();
    foreach(QString tempdir, tempDirs)
        QDir(tempdir).removeRecursively();
    tempDirs.clear();
}

void sh::base::Operation::commit()
{
    NOT_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QMutexLocker _l1(&operationlock);
    foreach(std::shared_ptr<const sh::base::Eurl> commitnow, pendingCommits())
    {
        QString filename = files[commitnow];
        QFile file( filename );
        file.open(QIODevice::ReadOnly);
        filesystem()->createFile(commitnow, &file);
        file.close();
    }
    abort();
}

QList<std::shared_ptr<const sh::base::Eurl>> sh::base::Operation::pendingCommits()
{
    QMutexLocker _l1(&operationlock);
    QList<std::shared_ptr<const sh::base::Eurl>> tocommit(files.keys());
    QList<std::shared_ptr<const sh::base::Eurl>> result;
    while (!tocommit.isEmpty())
    {
        std::shared_ptr<const sh::base::Eurl> commitnow = 0;
        foreach (std::shared_ptr<const sh::base::Eurl> eurl, tocommit)
        {
            bool isFree = true;
            foreach (std::shared_ptr<const sh::base::Eurl> eurl2, tocommit)
            {
                if (eurl2->outermostInnerEurl() == eurl)
                {
                    isFree = false;
                    break;
                }
            }
            if (isFree) {
                commitnow = eurl;
                break;
            }
        }
        if (fileIsTemporary[commitnow])
            result.append(commitnow);
        tocommit.removeOne(commitnow);
    }
    return result;
}

void sh::base::Operation::enableDetailsCache()
{
    QMutexLocker _l1(&operationlock);
    _detailsCacheEnabled = true;
}

bool sh::base::Operation::isDetailsCacheEnabled()
{
    QMutexLocker _l1(&operationlock);
    return _detailsCacheEnabled;
}

void sh::base::Operation::storeDetailInCache(std::shared_ptr<const sh::base::Eurl>eurl, sh::base::DetailColumn* column, QVariant value)
{
    QMutexLocker _l1(&operationlock);
    if (!_detailsCacheEnabled)
        return;
    _detailsCache[column][eurl] = value;
}

QVariant sh::base::Operation::getDetailFromCache(std::shared_ptr<const sh::base::Eurl>eurl, const sh::base::DetailColumn* column)
{
    QMutexLocker _l1(&operationlock);
    if (!_detailsCacheEnabled)
        return sh::base::DetailColumn::VALUE_UNAVAILABLE();
    auto i1 = _detailsCache.find(column);
    if (i1 == _detailsCache.end())
        return sh::base::DetailColumn::VALUE_UNAVAILABLE();
    QMap<std::shared_ptr<const sh::base::Eurl>, QVariant> innermap = *i1;
    auto i2 = innermap.find(eurl);
    if (i2 == innermap.end())
        return sh::base::DetailColumn::VALUE_UNAVAILABLE();
    return *i2;
}

void sh::base::Operation::writeIODeviceToFile(QIODevice* content, QString filepath)
{
    const int BUFFER_SIZE = 4096;
    char buffer[BUFFER_SIZE];
    QFile tfile(filepath);
    if (!tfile.open(QIODevice::WriteOnly))
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while creating file '%1': %2").arg(filepath).arg(tfile.errorString())));
    QString _cnt_noerror = content->errorString();
    while (qint64 l = content->read(buffer, BUFFER_SIZE))
    {
        if (l == -1)
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to read from a stream while writing '%1': %2").arg(filepath).arg(content->errorString())));
        char* mbuffer = buffer;
        while (l) {
            qint64 w = tfile.write(mbuffer, l);
            l-=w;
            mbuffer+=w;
        }
    }
    if (tfile.error() != QFile::NoError)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to write data to temporary file '%1': %2").arg(filepath).arg(tfile.errorString())));
    if (content->errorString() != _cnt_noerror)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to read: %1").arg(content->errorString())));
}

void sh::base::Operation::setCustomData(std::shared_ptr<const Eurl> eurl, QString key, QVariant data)
{
    QMutexLocker _l1(&operationlock);
    _customData[eurl][key] = data;
}

QVariant sh::base::Operation::getCustomData(std::shared_ptr<const Eurl> eurl, QString key)
{
    QMutexLocker _l1(&operationlock);
    auto itr = _customData.find(eurl);
    if (itr != _customData.end())
    {
        QMap<QString, QVariant> innmap = *itr;
        return innmap.value(key);
    }
    return QVariant();
}

void sh::base::Operation::setMaxAllowedSizeRatioPerPart(double v)
{
    QMutexLocker _l1(&operationlock);
    this->_maxAllowedSizeRatioPerPart = v;
}

sh::base::FilesystemOperation* sh::base::Operation::filesystem()
{
    return _filesystemOperation;
}

sh::base::Operation::MaxAllowedSizeRatioPerPartExceededException::MaxAllowedSizeRatioPerPartExceededException() :
    sh::exceptions::IOException(sh::exceptions::ExceptionData().retryable(false).details("exceeded maximum allowed size"))
{
}

QMutex sh::base::Operation::_tempfilemutex;
