/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef OPERATION_H
#define OPERATION_H

#include <memory>
#include <QIODevice>
#include <QList>
#include <QObject>
#include <QMap>
#include <QMutex>
#include "exceptions/ioexception.h"


namespace sh{
    namespace base{

        class DetailColumn;
        class Eurl;
        class FilesystemOperation;

        /** \brief
         * A operation surrounds a completed series of steps in the filesystem.
         *
         * For some steps to execute in the filesystem (i.e. not only the local one but the entire Eurl universe),
         * it fetches intermediate files, caches them and writes them back to their real location afterwards.
         *
         * It provides transaction-like operations for committing/dropping those intermediate files.
         * A high-level interface for file management is avaiable in `filesystem`.
         *
         * It is allowed to create new instances from scratch, but you should check if you implicitely have
         * an instance available in your situation, or if sh::tools::OperationsCache is an option.
         */
        class Operation : public QObject
        {
            Q_OBJECT
        public:
            /** \brief Often constructed by the infrastructure and made available, but can also be constructed directly.
             */
            explicit Operation(QObject *parent = 0);
            virtual ~Operation();

            /** \brief
             * Returns the free disk space (in bytes) available for operations like fetchContainerFile() or fetchFile().
             */
            quint64 getFreeCacheSpace();
            /** \brief
             * Fetches the container file for a eurl locally and returns the local path.
             * Use this with care and only if needed. In many cases working with streams is the better way!
             */
            QString fetchContainerFile(std::shared_ptr<const sh::base::Eurl> eurl, QString namehint = "tmp");
            /** \brief
             * Fetches a file locally and returns the local path.
             * Use this with care and only if needed. In many cases working with streams is the better way!
             */
            QString fetchFile(std::shared_ptr<const sh::base::Eurl> eurl, QString namehint = "tmp");
            /** \brief
             * Aborts transaction dropping all pending changes.
             */
            void abort();
            /** \brief
             * Commits transaction applying all pending changes.
             */
            void commit();

            /** \brief
             * Enable details caching.
             */
            void enableDetailsCache();
            /** \brief
             * Is details caching enabled?
             */
            bool isDetailsCacheEnabled();
            /** \brief
             * Stores a column detail in cache.
             */
            void storeDetailInCache(std::shared_ptr<const sh::base::Eurl> eurl, sh::base::DetailColumn* column, QVariant value);
            /** \brief
             * Gets a column value from cache.
             */
            QVariant getDetailFromCache(std::shared_ptr<const sh::base::Eurl> eurl, const sh::base::DetailColumn* column);

            /** \brief
             * Returns a list of sh::base::Eurl items which are marked for commit.
             */
            QList<std::shared_ptr<const sh::base::Eurl>> pendingCommits();

            /** \brief
             * Creates a fresh temporary folder and returns the path to it.
             */
            QString getTempDir();

            /** \brief
             * Stores custom data.
             */
            void setCustomData(std::shared_ptr<const sh::base::Eurl> eurl, QString key, QVariant data);
            /** \brief
             * Gets stored custom data.
             */
            QVariant getCustomData(std::shared_ptr<const sh::base::Eurl> eurl, QString key);

            /** \brief
             * Sets a maximum part of available disk space for usage.
             * This is exotic functionality you typically don't need.
             * @a_infrastructure
             * @param v: A ratio (between 0 and 1) of the available disk space which one
             *           part maximally may cost.
             */
            void setMaxAllowedSizeRatioPerPart(double v);

            /** \brief
             * Low-level function for writing a QIODevice content to a local path.
             */
            static void writeIODeviceToFile(QIODevice* content, QString filepath);

            /** \brief
             * Gets the sh::base::FilesystemOperation filesystem operation object.
             */
            sh::base::FilesystemOperation* filesystem();

            /** \brief
             * IO exception raised when the value in setMaxAllowedSizeRatioPerPart() was exceeded.
             */
            class MaxAllowedSizeRatioPerPartExceededException : public sh::exceptions::IOException
            {
            public:
                MaxAllowedSizeRatioPerPartExceededException();
            };

        signals:

        public slots:

        private:
            QMap<std::shared_ptr<const sh::base::Eurl>, QString> files;
            QMap<std::shared_ptr<const sh::base::Eurl>, bool> fileIsTemporary;
            QList<QString> tempDirs;
            bool _detailsCacheEnabled = false;
            QMap<std::shared_ptr<const sh::base::Eurl>, QMap<QString, QVariant>> _customData;
            QMap<const sh::base::DetailColumn*, QMap<std::shared_ptr<const sh::base::Eurl>, QVariant> > _detailsCache;
            sh::base::FilesystemOperation* _filesystemOperation;
            static QMutex _tempfilemutex;
            static QString _tempdir;
            double _maxAllowedSizeRatioPerPart;
            QMutex operationlock;
        };

    }
}

#endif // OPERATION_H
