/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QObject>
#ifdef Q_OS_UNIX
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#endif

#include "base/eurl.h"
#include "exceptions/ioexception.h"
#include "exceptions/runtimeexception.h"
#include "exceptions/argumentexception.h"
#include "exceptions/programexception.h"
#include "ui/mainwindow.h"
#include "ui/fileview/fileview.h"
#include "tools/benchmarking.h"
#include "tools/singletoninitializer.h"
#include "tools/mainthread.h"
#include "tools/logger.h"
#include "configurationmanager.h"
#include "configurationvalue.h"


namespace sh{namespace configuration{INITIALIZE_SINGLETON_IN_MAINTHREAD(10070, ConfigurationManager)}}

struct _ProjectInfo {
#include "_projectinfo.h"
};

bool sh::configuration::ConfigurationManager::mkpath(QDir dir)
{// a more straight-forward implementation did not work (at least for win)
    if (dir.exists())
        return false;
    else
    {
        QDir parent(QDir::cleanPath(dir.absolutePath() + "/.."));
        mkpath(parent);
        parent.mkdir(dir.dirName());
        return true;
    }
}

sh::configuration::ConfigurationManager::ConfigurationManager() :
    QObject()
{
    mkpath(QDir(mypath() + ".lock"));
}

sh::configuration::ConfigurationManager* sh::configuration::ConfigurationManager::instance()
{
    static sh::configuration::ConfigurationManager* _instance = new sh::configuration::ConfigurationManager();
    return _instance;
}

void sh::configuration::ConfigurationManager::_initialize()
{
}

void sh::configuration::ConfigurationManager::_shutdown()
{
}

QString sh::configuration::ConfigurationManager::mypath()
{
    return QDir::homePath()+"/.shallot/";
}

sh::configuration::UIStyle* sh::configuration::ConfigurationManager::style()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    if (_uistyle==0)
        _uistyle = new sh::configuration::UIStyle(sh::ui::MainWindow::mainWindow());
    return _uistyle;
}

sh::configuration::ConfigurationManager::ConfigurationValueImpl::ConfigurationValueImpl(QString name, QVariant deflt, QString desc, QString longdesc, sh::configuration::ConfigurationCategory cat, sh::configuration::ConfigurationValueType* valuetype, QString changehint, sh::configuration::ConfigurationManager *mgr)
   :sh::configuration::ConfigurationValue(name, deflt, desc, longdesc, cat, valuetype, changehint), mgr(mgr)
{
}

QVariant sh::configuration::ConfigurationManager::ConfigurationValueImpl::_getConfigValueVariant()
{
    ConfigFilesLock sfl;
    QSettings settings(mypath()+"cfg", QSettings::IniFormat);
    QVariant result = settings.value(name, deflt);
    return result;
}

void sh::configuration::ConfigurationManager::ConfigurationValueImpl::setConfigValue(QVariant value)
{
    {
        ConfigFilesLock sfl;
        QSettings settings(mypath()+"cfg", QSettings::IniFormat);
        settings.setValue(name, value);
    }
    foreach(std::function<void()> f, callbacks())
        f();
}

QList<std::shared_ptr<sh::configuration::ConfigurationValue> > sh::configuration::ConfigurationManager::getAllConfigurationValues()
{
    return _configvalues;
}

std::shared_ptr<sh::configuration::ConfigurationValue> sh::configuration::ConfigurationManager::registerConfigValue(QString name, QVariant deflt, sh::configuration::ConfigurationValueType* valuetype, QString desc, sh::configuration::ConfigurationCategory cat, QString longdesc, QString changehint)
{
    // ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD - technically impossible here due to initialization stuff
    if (valuetype==0)
        valuetype = sh::configuration::ConfigurationValue::ValueTypeString();
    std::shared_ptr<sh::configuration::ConfigurationValue> result = std::make_shared<ConfigurationValueImpl>(name, deflt, desc, longdesc, cat, valuetype, changehint, this);
    _configvalues.append(result);
    return result;
}

QString sh::configuration::ConfigurationManager::revisionString()
{
    return _ProjectInfo().version;
}

QString sh::configuration::ConfigurationManager::homepage()
{
    return _ProjectInfo().homepage;
}

QString sh::configuration::ConfigurationManager::shallotDataDir()
{
#ifdef SHALLOT_DATA_DIR
#define _TO_STR(x) #x
#define TO_STR(x) _TO_STR(x)
    return TO_STR(SHALLOT_DATA_DIR);
#else
    return QCoreApplication::applicationDirPath();
#endif
}

QDateTime sh::configuration::ConfigurationManager::buildtime()
{
    QDateTime result = QDateTime::fromString(_ProjectInfo().buildtimeutc, "yyyy-MM-dd hh:mm:ss");
    result.setTimeSpec(Qt::UTC);
    return result.toLocalTime();
}

sh::configuration::ConfigurationManager::ConfigFilesLock::ConfigFilesLock()
{
    BENCHMARK_SCOPE("ConfigurationManager::ConfigFilesLock")
    ConfigurationManager::instance(); // ensure that it is created.
    QMutexLocker _l1(&mutex);
    if (QThread::currentThread() == currentThread)
        dounlock = false;
    else
    {
        dounlock = true;
#ifdef Q_OS_UNIX
        static char* cmypath = mypath().toLocal8Bit().data();
        static key_t key = ftok(cmypath, 's');
        static bool _inited = false;
        if (!_inited) {
            if ((semset_id = semget(key, 1, IPC_CREAT|IPC_EXCL|0666)) == -1)
                semset_id = semget(key, 0, 0666);
            else
            {
                union semun semopts;
                semopts.val = 1;
                semctl(semset_id, 0, SETVAL, semopts);
            }
        }
        mutex.unlock();
        struct sembuf sem_lock = {0, -1, SEM_UNDO};
        sem_lock.sem_num = 0;
        int res = semop(semset_id, &sem_lock, 1);
        mutex.lock();
        if (res == -1)
            throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while acquiring the config files lock.")));
#else
        slockpref = mypath() + ".lock";
        bool succ = false;
        while(!succ)
        {
            int i = 0;
            qint64 nowms = QDateTime::currentMSecsSinceEpoch();
            QString _slockfile;
            while(!succ && (QDateTime::currentMSecsSinceEpoch() < nowms+5000))
            {
                _slockfile = slockpref+QString(".%1").arg(i);
                succ = QFile::rename(slockpref, _slockfile);
                i++;
                if (!succ)
                {
                    mutex.unlock();
                    QDateTime now = QDateTime::currentDateTime();
                    foreach(QString e, QDir(mypath()).entryList(QDir::Dirs))
                    {
                        if (e.startsWith(".lock.") && QFileInfo(mypath()+e).lastModified().secsTo(now) > 30)
                            QFile::remove(mypath()+e);
                    }
                    QThread::msleep(5);
                    mutex.lock();
                }
            }
            if (succ)
                slockfile = _slockfile;
            else
            {
                foreach(QString e, QDir(mypath()).entryList(QDir::Dirs))
                {
                    if (e.startsWith(".lock."))
                        QFile::remove(mypath()+e);
                }
                QDir(mypath()).mkdir(".lock");
            }
        }
#endif
        currentThread = QThread::currentThread();
    }
}

QMutex sh::configuration::ConfigurationManager::ConfigFilesLock::mutex;

QThread* sh::configuration::ConfigurationManager::ConfigFilesLock::currentThread = 0;

sh::configuration::ConfigurationManager::ConfigFilesLock::~ConfigFilesLock()
{
    QMutexLocker _l1(&mutex);
    BENCHMARK_SCOPE("ConfigurationManager::ConfigFilesLock")
    if (dounlock)
    {
        currentThread = 0;
        sh::exceptions::Exception::executeGuarded([&](){
#ifdef Q_OS_UNIX
            struct sembuf sem_unlock = {0, 1, SEM_UNDO};
            sem_unlock.sem_num = 0;
            if ((semop(semset_id, &sem_unlock, 1)) == -1)
                throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while releasing the config files lock.")));
#else
            if (!QFile::rename(slockfile, slockpref))
                throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("The lock was stolen while saving settings.")));
#endif
        }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
    }
}
