/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CONFIGURATIONMANAGER_H
#define CONFIGURATIONMANAGER_H

#include <memory>
#include <QDir>
#include <QMap>
#include <QMutex>
#include <QVariant>
#include "configurationvalue.h"
#include "uistyle.h"


namespace sh{
    namespace base{ class Eurl; }

    namespace configuration {

        /** \brief
         * Manages the Shallot configuration.
         *
         * This is the more static part of Shallot.
         * Much more stuff, i.e. everything you see in 'Manage saved settings', is in
         * sh::settings::SettingsManager.
         */
        class ConfigurationManager : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static ConfigurationManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** \brief
             * Creates and registers a new ConfigurationValue.
             * This is typically done once at the beginning.
             * Each registered instance is shown in the 'Tuning' dialog and can be set (for making changes) and observed (for applying changes) in code as well.
             * Useful e.g. for some internal bookkeeping or for exotic machine-wide configuration values (path to some tool, ...).
             */
            std::shared_ptr<sh::configuration::ConfigurationValue> registerConfigValue(QString name, QVariant deflt, sh::configuration::ConfigurationValueType* valuetype=0, QString desc="", sh::configuration::ConfigurationCategory cat=CategoryNone, QString longdesc="", QString changehint="");
            /** \brief
             * Gets a list of all registered configuration value instances.
             */
            QList<std::shared_ptr<sh::configuration::ConfigurationValue> > getAllConfigurationValues();

            /** \brief
             * Gets a UIStyle instance, which helps for common ui styling.
             */
            sh::configuration::UIStyle* style();
            /** \brief
             * Returns the path to a storage for per-user data. It resides somewhere within the user home directory.
             */
            static QString mypath();

            /** \brief
             * Returns the Shallot revision string, i.e. the exact version number.
             * Note: This is only updated by a special build tool (anise) and not by directly compiling via qmake.
             */
            QString revisionString();
            /** \brief
             * Returns the url to the Shallot homepage.
             */
            QString homepage();

            /** \brief
             * Returns the directory of the Shallot program data. It contains the static files which are part
             * of the Shallot program.
             */
            static QString shallotDataDir();

            /** \brief
             * Returns the time when this Shallot build was created.
             * Note: This is only updated by a special build tool (anise) and not by directly compiling via qmake.
             */
            QDateTime buildtime();

            /** \brief Locks the Shallot config files during exclusive usage.
             *
             * This lock is rather heavy-weight and should be used only if necessary!
             */
            class ConfigFilesLock
            {
            public:
                /** \brief Is intended to be directly constructed from everywhere.
                 */
                ConfigFilesLock();
                ~ConfigFilesLock();
            private:
                static QThread* currentThread;
                static QMutex mutex;
                bool dounlock;
#ifdef Q_OS_UNIX
                int semset_id;
#else
                QString slockpref;
                QString slockfile;
#endif
            };

        private:

            class ConfigurationValueImpl : public sh::configuration::ConfigurationValue
            {
            public:
                void setConfigValue(QVariant value);
                ConfigurationValueImpl(QString name, QVariant deflt, QString desc, QString longdesc, sh::configuration::ConfigurationCategory cat, sh::configuration::ConfigurationValueType* valuetype, QString changehint, ConfigurationManager* mgr);
            protected:
                QVariant _getConfigValueVariant();
            private:
                ConfigurationManager* mgr;
            };

            explicit ConfigurationManager();

            bool mkpath(QDir dir);

            QList<std::shared_ptr<sh::configuration::ConfigurationValue>> _configvalues;
            sh::configuration::UIStyle* _uistyle=0;

        };

#ifdef Q_OS_UNIX

        union semun {
            int val; /* Value for SETVAL */
            struct semid_ds *buf; /* Buffer for IPC_STAT, IPC_SET */
            unsigned short *array; /* Array for GETALL, SETALL */
            struct seminfo *__buf; /* Buffer for IPC_INFO (Linux-specific) */
        };

#endif

    }
}

#endif // CONFIGURATIONMANAGER_H
