/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CONFIGURATIONVALUE_H
#define CONFIGURATIONVALUE_H

#include <functional>
#include <QMutex>
#include <QVariant>


namespace sh{
    namespace configuration {

        /** \brief
         * Categories of Configursh::configuration::ConfigurationValuentations.
         * They are used for grouping them in the dialogs. There is no
         * difference in behavior implied by this choice.
         */
        enum ConfigurationCategory
        {
            /** \brief
             * No category.
             */
            CategoryNone=0,
            /** \brief
             * User interface category.
             */
            CategoryGUI=1,
            /** \brief
             * Behavioral configuration category.
             */
            CategoryBehavior=2,
            /** \brief
             * External tools category.
             */
            CategoryExternalTools=3
        };

        class ConfigurationValueType;

        /** \brief
         * Abstract base class for a configuration value which is managed in the *Tuning* dialog.
         * The implementation can get and set this value.
         * See Shallot documentation for details.
         */
        class ConfigurationValue
        {
            friend class ConfigurationValueType;
        public:
            /** \brief Constructed only indirectly.
             */
            ConfigurationValue(QString name, QVariant deflt, QString desc, QString longdesc, ConfigurationCategory cat, ConfigurationValueType* valuetype, QString changehint);

            virtual int getConfigValueInt();
            virtual bool getConfigValueBool();
            virtual QString getConfigValueString();
            virtual void setConfigValue(QVariant value) = 0;
            virtual void consumeValue(QObject* o, std::function<void()> h);

            virtual QString getDescription();
            virtual QString getLongDescription();
            virtual ConfigurationCategory getCategory();
            virtual QString changeHint();

            virtual ConfigurationValueType* getValueType();

            static ConfigurationValueType* ValueTypeString();
            static ConfigurationValueType* ValueTypeInteger();
            static ConfigurationValueType* ValueTypeBoolean();
            static ConfigurationValueType* ValueTypeLocalFilePath();

            QList<std::function<void()>> callbacks();

        protected:
            virtual QVariant _getConfigValueVariant() = 0;
            QString name;
            QString desc;
            QString longdesc;
            QString changehint;
            ConfigurationCategory cat;
            ConfigurationValueType* valuetype;
            QVariant deflt;
            QMap<int, std::function<void()>> _callbacks;
            QMutex _callbacksmutex;
            static QMutex* _typeinitmutex;
            static QMutex* _gettypeinitmutex();

        };

    }
}

#endif // CONFIGURATIONVALUE_H
