/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration/configurationmanager.h"
#include "configuration/configurationvalue.h"
#include "ui/mainwindow.h"
#include "uistyle.h"


sh::configuration::UIStyle::UIStyle(sh::ui::MainWindow *parent) :
    QObject(parent)
{
    _mainWindow = parent;
    cfgvalBrandingColor = sh::configuration::ConfigurationManager::instance()->registerConfigValue(
                "sh::configuration::UiStyle.BrandingColor", "#3c001e");
    _brandingcolor = QColor(cfgvalBrandingColor->getConfigValueString());
}

QColor sh::configuration::UIStyle::windowColor()
{
    return _mainWindow->palette().color(QPalette::Active, QPalette::Window);
}

QColor sh::configuration::UIStyle::backgroundColor()
{
    return _mainWindow->palette().color(QPalette::Active, QPalette::Base);
}

QColor sh::configuration::UIStyle::inactiveBackgroundColor()
{
    return backgroundColor().darker(110);
}

QColor sh::configuration::UIStyle::windowColorHighlighted()
{
    return mix(0.5, windowColor(), backgroundColor());
}

QColor sh::configuration::UIStyle::windowColorHalfHighlighted()
{
    return mix(0.85, windowColor(), backgroundColor());
}

QColor sh::configuration::UIStyle::windowColorDarkened()
{
    return mix(0.85, windowColor(), foregroundColor());
}

QColor sh::configuration::UIStyle::windowColorLikeTitlebar()
{
    return mix(0.7, windowColorHighlighted(), titlebarColor());
}

QColor sh::configuration::UIStyle::titlebarColor()
{
    return _mainWindow->palette().color(QPalette::Active, QPalette::Highlight);
}

QColor sh::configuration::UIStyle::brandingColor()
{
    return _brandingcolor;
}

QColor sh::configuration::UIStyle::linkColor()
{
    return _mainWindow->palette().color(QPalette::Active, QPalette::Link);
}

QColor sh::configuration::UIStyle::foregroundColor()
{
    return _mainWindow->palette().color(QPalette::Active, QPalette::Text);
}

QColor sh::configuration::UIStyle::foregroundColorLighter1()
{
    return mix(0.4, windowColor(), foregroundColor());
}

QColor sh::configuration::UIStyle::foregroundColorLighter2()
{
    return mix(0.7, windowColor(), foregroundColor());
}


int sh::configuration::UIStyle::fontSizeInPt(int r)
{
    return _mainWindow->font().pointSize()*r /100;
}

QColor sh::configuration::UIStyle::mix(float n1, QColor c1, QColor c2)
{
    int r = c1.red() * n1 + c2.red() * (1.0-n1);
    int g = c1.green() * n1 + c2.green() * (1.0-n1);
    int b = c1.blue() * n1 + c2.blue() * (1.0-n1);
    return QColor(std::min(r, 255), std::min(g, 255), std::min(b, 255));
}


QString sh::configuration::UIStyle::Sheet::sheet()
{
    return _sheet;
}

sh::configuration::UIStyle::Sheet::Sheet(sh::configuration::UIStyle* style, QString sheet)
    :_sheet(sheet), _style(style)
{
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::customcss(QString css)
{
    return Sheet(_style, _sheet + css + ";");
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::fontsize_byFactor(int f)
{
    return Sheet(_style, _sheet + QString("font-size:%1pt;").arg(_style->fontSizeInPt(f)));
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::fontsize_header()
{
    return fontsize_byFactor(130);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::fontsize_smaller()
{
    return fontsize_byFactor(90);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::bold()
{
    return Sheet(_style, _sheet + "font-weigth:bold;");
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::textcolor(QColor color)
{
    return Sheet(_style, _sheet + QString("color:%1;").arg(color.name()));
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::textcolor_lighter1()
{
    return textcolor(_style->foregroundColorLighter1());
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::textcolor_lighter2()
{
    return textcolor(_style->foregroundColorLighter2());
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::background(QColor color, bool restrictQWidget)
{
    QString n = QString("background:%1;").arg(color.name());
    if (restrictQWidget)
        n = ".QWidget{"+n+"}\n";
    return Sheet(_style, _sheet + n);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::background_highlighted(bool restrictQWidget)
{
    return background(_style->windowColorHighlighted(), restrictQWidget);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::background_halfhighlighted(bool restrictQWidget)
{
    return background(_style->windowColorHalfHighlighted(), restrictQWidget);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::background_darkened(bool restrictQWidget)
{
    return background(_style->windowColorDarkened(), restrictQWidget);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::background_liketitlebar(bool restrictQWidget)
{
    return background(_style->windowColorLikeTitlebar(), restrictQWidget);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::Sheet::applyTo(QWidget* widget)
{
    widget->setStyleSheet(_sheet);
    return Sheet(_style,_sheet);
}

sh::configuration::UIStyle::Sheet sh::configuration::UIStyle::createSheet()
{
    return Sheet(this, "");
}
