/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <qsystemdetection.h>
#ifdef Q_OS_UNIX
#define HAVE_EXECINFO_H
#endif

#ifdef HAVE_EXECINFO_H
#include "execinfo.h"
#include <cxxabi.h>
#endif

#include <QApplication>
#include <QThread>
#include "tools/singletoninitializer.h"
#include "tools/logger.h"
#include "tools/mainthread.h"
#include "tools/threadpool.h"
#include "ui/exceptiondialog.h"
#include "ui/mainwindow.h"
#include "exception.h"
#include "programexception.h"


const QString sh::exceptions::Exception::SHALLOT_MUST_CLOSE_TEXT=QObject::tr("Your Shallot process is disturbed by an error and needs to be closed now. You can try if starting Shallot again solves the issue. If you are interested in technical details, please take a look at the log.");

const QString sh::exceptions::Exception::UNDEFINED_ERROR_OCCURRED=QObject::tr("An unspecified error occurred.");

const QString sh::exceptions::Exception::Value_isShallotException="_isShallotException";

namespace sh{namespace exceptions{INITIALIZE_SINGLETON_IN_MAINTHREAD(900, ExceptionHandlerSettingsManager)}}

QString sh::exceptions::Exception::_demangle(QString l)
{
    QString result = l;
#ifdef HAVE_EXECINFO_H
    int i1 = l.indexOf("(");
    int i2 = l.indexOf("+0x",i1);
    int i3 = l.indexOf(")", i1);
    if (i1!=-1 && i2!=-1 && i3!=-1 && i1<i2 && i2<i3)
    {
        QString mangledname=l.mid(i1+1, i2-i1-1);
        if (mangledname.length()>0) {
            int status = -1;
            QByteArray rawmangledname;
            rawmangledname = mangledname.toLocal8Bit();
            char *demangledName = abi::__cxa_demangle( rawmangledname.data(), NULL, NULL, &status );
            if ( status == 0 ) {
                result = result.mid(0,i1+1) + demangledName + QString(" ") + result.mid(i2, i3-i2);
            }
            free(demangledName);
        }
    }
#endif
    int i4 = result.indexOf("/Shallot(");
    if (i4!=-1)
        result = result.mid(i4+1);
    return result;
}

sh::exceptions::Exception::Exception() : _data(sh::exceptions::ExceptionData(QMap<QString, QString>()))
{
}

sh::exceptions::Exception::Exception(sh::exceptions::ExceptionData data)
    : _data(data.message(QObject::tr("An error occurred.")).details("").program().resumeable(false).retryable(false).autoRetryRecommended(-1))
{
}

sh::exceptions::ExceptionData::ExceptionData(const ExceptionData& o) : QMap<QString, QString>()
{
    for(auto itr=o.begin();itr!=o.end();itr++)
        insert(itr.key(), itr.value());
}

sh::exceptions::ExceptionData::ExceptionData(const QMap<QString, QString> &o) : QMap<QString, QString>()
{
    for(auto itr=o.begin();itr!=o.end();itr++)
        insert(itr.key(), itr.value());
}

sh::exceptions::ExceptionData::ExceptionData() : QMap<QString, QString>()
{
    QString _callstack = "unavailable";
    #ifdef HAVE_EXECINFO_H
    void *array[50];
    size_t size;
    size = backtrace(array, 50);
    char **strings;
    strings = backtrace_symbols(array, size);
    if (strings != 0)
    {
        _callstack = "";
        for(size_t i=0; i< size; i++)
        {
            QString l = strings[i];
            l=Exception::_demangle(l);
            _callstack += l + QString("\n");
        }
        free(strings);
    }
    #endif
    insert(Exception::Value_isShallotException, "1");
    insert("callstack", _callstack);
    insert("classes", "|");
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::details(QString details)
{
    insertIfUnset("details", details);
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::message(QString message)
{
    insertIfUnset("message", message);
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::runtime()
{
    insertIfUnset("isprogram", "0");
    insertIfUnset("isruntime", "1");
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::program()
{
    insertIfUnset("isprogram", "1");
    insertIfUnset("isruntime", "0");
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::retryable(bool v)
{
    insertIfUnset("isretryable", v?"1":"0");
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::resumeable(bool v)
{
    insertIfUnset("isresumeable", v?"1":"0");
    if (value("isresumeable")=="0")
    {
        insert("isretryable", "0");
        insert("autoretryrecommendedin", "-1");
    }
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::autoRetryRecommended(int v)
{
    insertIfUnset("autoretryrecommendedin", QString::number(v));
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::detailsAreInteresting(bool v)
{
    insertIfUnset("detailsareinteresting", v?"1":"0");
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::name(QString name)
{
    insertIfUnset("name", name);
    return (*this).classes(name);
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::classes(QString classes)
{
    insert("classes", value("classes") + classes + "|");
    return *this;
}

sh::exceptions::ExceptionData sh::exceptions::ExceptionData::aux(QString aux)
{
    insertIfUnset("aux", aux);
    return *this;
}

void sh::exceptions::ExceptionData::insertIfUnset(QString key, QString value)
{
    if (!contains(key))
        insert(key, value);
}

sh::exceptions::ExceptionData sh::exceptions::Exception::data()
{
    return _data;
}

QString sh::exceptions::Exception::message() const
{
    return _data["message"];
}

QString sh::exceptions::Exception::details() const
{
    return _data["details"];
}

QString sh::exceptions::Exception::classes() const
{
    return _data["classes"];
}

QString sh::exceptions::Exception::name() const
{
    return _data["name"];
}

QString sh::exceptions::Exception::callstack() const
{
    return _data["callstack"];
}

QString sh::exceptions::Exception::auxiliary() const
{
    return _data["aux"];
}

bool sh::exceptions::Exception::isRuntimeException() const
{
    return _data["isruntime"]=="1";
}

bool sh::exceptions::Exception::isProgramException() const
{
    return _data["isprogram"]=="1";
}

bool sh::exceptions::Exception::isRetryable() const
{
    return _data["isretryable"]=="1";
}

bool sh::exceptions::Exception::isResumeable() const
{
    return _data["isresumeable"]=="1";
}

bool sh::exceptions::Exception::isDetailsAreInteresting() const
{
    return _data["detailsareinteresting"]=="1";
}

int sh::exceptions::Exception::autoRetryRecommendedIn() const
{
    return _data["autoretryrecommendedin"].toInt();
}

QString sh::exceptions::Exception::customValue(QString key) const
{
    return _data.value("custom."+key);
}

void sh::exceptions::Exception::setCustomValue(QString key, QString value)
{
    _data["custom."+key] = value;
}

void sh::exceptions::Exception::setResumeable(bool v)
{
    _data.insert("isresumeable", v?"1":"0");
}

void sh::exceptions::Exception::setReTryable(bool v)
{
    _data.insert("isretryable", v?"1":"0");
}

bool sh::exceptions::Exception::isClass(QString classname)
{
    return _data["classes"].contains("|"+classname+"|");
}

void sh::exceptions::Exception::executeGuarded(std::function<void()> fct, int flags)
{
    std::function<void()> _fct = fct;
    // next following variable is only valid in the original caller thread
    Exception::HandlerSettings* handlerSettings = ExceptionHandlerSettingsManager::instance()->handlerSettings();
    _fct = [=](){
        try
        {
            _fct();
        }
        catch(sh::exceptions::Exception&)
        {
            throw;
        }
        catch(CancelException&)
        {
            throw;
        }
        catch(ThreadAbortException&)
        {
            throw;
        }
        catch(...)
        {
            sh::exceptions::ProgramException ex(sh::exceptions::ExceptionData().details(sh::exceptions::Exception::UNDEFINED_ERROR_OCCURRED));
            throw ex;
        }
    };

    _fct = [=](){
        try
        {
            _fct();
        }
        catch(CancelException&)
        {
            if (!handlerSettings->exceptionHandler_cancel.isEmpty()) {
                if (flags & ExecuteGuardFlag::CANCELABLE_UP_TO_HERE)
                    handlerSettings->exceptionHandler_cancel.last()();
                else
                    throw;
            }
            else
                throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Some uncancellable program parts were cancelled."));
        }
    };
    if (flags & ExecuteGuardFlag::CANCELABLE_UP_TO_HERE)
    {
        _fct = [=](){
            auto x1 = Exception::createRegisterHandler<std::function<void()> >([=]()
            {
            }, &handlerSettings->exceptionHandler_cancel);
            _fct();
        };
    }

    if (!(flags & ExecuteGuardFlag::LOGGING_DISABLED))
    {
        _fct = [=](){
            try
            {
                _fct();
            }
            catch(sh::exceptions::Exception& ex)
            {
                if (ex.customValue("islogged") != "1")
                {
                    sh::tools::Logger::instance()->logException(ex);
                    ex.setCustomValue("islogged", "1");
                }
                throw ex;
            }
        };
    }

    if (flags & ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY)
    {
        _fct = [=](){
            try
            {
                _fct();
            }
            catch(sh::exceptions::Exception& ex)
            {
                if (!ex.isResumeable())
                    throw ex;
            }
        };
    }

    if (flags & ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED)
    {
        _fct = [=](){
            bool allowNewTry=true;
            bool success=false;
            int tries=3;
            while (!success && allowNewTry) {
                try
                {
                    _fct();
                    success=true;
                }
                catch(sh::exceptions::Exception &ex)
                {
                    if (ex.isResumeable() && ex.isRetryable())
                    {
                        int waitingtime = ex.autoRetryRecommendedIn();
                        allowNewTry = (--tries)>0 && waitingtime>0;
                        if (!allowNewTry)
                            throw ex;
                        QThread::msleep(waitingtime);
                    }
                    else
                        throw ex;
                }
            }
        };
    }

    if (flags & ExecuteGuardFlag::MANUAL_RETRY_ENABLED)
    {
        _fct = [=](){
            bool allowNewTry=true;
            bool success=false;
            while (!success && allowNewTry) {
                try
                {
                    _fct();
                    success=true;
                }
                catch(sh::exceptions::Exception &ex)
                {
                    if (ex.isResumeable() && ex.isRetryable())
                    {
                        std::function<bool(sh::exceptions::Exception& ex)> exceptionhandler =
                                !handlerSettings->exceptionHandler_retryable.isEmpty()? handlerSettings->exceptionHandler_retryable.last() :
                                [=](sh::exceptions::Exception& ex){
                            return sh::ui::ExceptionDialog::exceptionDialog(ex);
                        };
                        allowNewTry=exceptionhandler(ex);
                        if (!allowNewTry)
                            throw ex;
                    }
                    else
                        throw ex;
                }
            }
        };
    }

    if (flags & ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT)
    {
        _fct = [=](){
            try
            {
                _fct();
            }
            catch(sh::exceptions::Exception& ex)
            {
                ex.setResumeable(false);
                throw ex;
            }
        };
    }

    if (flags & ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE)
    {
        _fct = [=](){
            try
            {
                _fct();
            }
            catch(sh::exceptions::Exception& ex)
            {
                bool skipHandler = ex.isRetryable() && (flags & ExecuteGuardFlag::MANUAL_RETRY_ENABLED); // the user has already seen that
                ex.setReTryable(false);
                if (ex.isResumeable())
                {
                    if (!skipHandler)
                    {
                        if (!handlerSettings->exceptionHandler_resumeable.isEmpty())
                            handlerSettings->exceptionHandler_resumeable.last()(ex);
                        else
                        {
                            sh::ui::ExceptionDialog::exceptionDialog(ex);
                        }
                    }
                }
                else
                    throw ex;
            }
        };
    }

    _fct = [=](){
        try
        {
            _fct();
        }
        catch(sh::exceptions::Exception& ex)
        {
            if (!ex.isResumeable())
            {
                IGNORE_FORBIDDEN_ASYNC_CALLS
                if (!handlerSettings->exceptionHandler_hard.isEmpty())
                    handlerSettings->exceptionHandler_hard.last()(ex);
                else
                {
                    sh::ui::ExceptionDialog::exceptionDialog(ex);
                }
                sh::tools::MainThread::dispatcher()->invokeSync([](){
                    sh::ui::MainWindow::mainWindow()->quit();
                });
                if (dynamic_cast<sh::tools::ThreadPoolThread*>(QThread::currentThread()) || sh::tools::SingletonInitializer::initializerThread()==QThread::currentThread())
                    throw sh::exceptions::ThreadAbortException();
            }
            else
                throw ex;
        }
    };
    _fct();
}

sh::exceptions::CancelException::CancelException()
{
}

sh::exceptions::ThreadAbortException::ThreadAbortException()
{
}

sh::exceptions::Exception::HandlerSettings::HandlerSettings()
{
}

sh::exceptions::ExceptionHandlerSettingsManager* sh::exceptions::ExceptionHandlerSettingsManager::instance()
{
    static ExceptionHandlerSettingsManager* _inst = new ExceptionHandlerSettingsManager();
    return _inst;
}

void sh::exceptions::ExceptionHandlerSettingsManager::_initialize()
{
}

void sh::exceptions::ExceptionHandlerSettingsManager::_shutdown()
{
}

sh::exceptions::ExceptionHandlerSettingsManager::~ExceptionHandlerSettingsManager()
{
    foreach (sh::exceptions::Exception::HandlerSettings* hs, _handlerSettings)
        delete hs;
}

sh::exceptions::ExceptionHandlerSettingsManager::ExceptionHandlerSettingsManager()
{
}

sh::exceptions::Exception::HandlerSettings* sh::exceptions::ExceptionHandlerSettingsManager::handlerSettings()
{
    // safely initialized from the main thread at the beginning
    Exception::HandlerSettings* result;
    _mutex.lock();
    auto itr = _handlerSettings.constFind(QThread::currentThread());
    if (itr == _handlerSettings.constEnd())
    {
        result = new Exception::HandlerSettings();
        _handlerSettings.insert(QThread::currentThread(), result);
    }
    else
        result = (*itr);
    _mutex.unlock();
    return result;
}

void sh::exceptions::Exception::_executeGuarded_errorpanel_helper()
{
    QMutexLocker l1(&errorpanelmutex);
    if (!errorpanel)
        errorpanel = sh::ui::ActionExecutionInfoPanel::showErrorPanel();
    QObject::connect(errorpanel, &QObject::destroyed, [](){
        errorpanel = 0;
    });
}

void sh::exceptions::Exception::executeGuarded_errorpanel(std::function<void()> fct, int flags)
{
    Exception::HandlerSettings* handlerSettings = ExceptionHandlerSettingsManager::instance()->handlerSettings();
    auto x1 = Exception::createRegisterHandler<std::function<void(sh::exceptions::Exception&)> >([&](sh::exceptions::Exception&)
    {
        _executeGuarded_errorpanel_helper();
    }, &handlerSettings->exceptionHandler_resumeable);
    auto x2 = Exception::createRegisterHandler<std::function<bool(sh::exceptions::Exception&)> >([&](sh::exceptions::Exception&)
    {
        _executeGuarded_errorpanel_helper();
        return false;
    }, &handlerSettings->exceptionHandler_retryable);
    Exception::executeGuarded(fct, flags | ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

sh::ui::ActionExecutionInfoPanel* sh::exceptions::Exception::errorpanel = 0;
QMutex sh::exceptions::Exception::errorpanelmutex;
