/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXCEPTION_H
#define EXCEPTION_H

#include <functional>
#include <memory>
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QStack>
#include <QThread>
#include "ui/actionexecutioninfopanel.h"


namespace sh{
    namespace exceptions {

        enum ExecuteGuardFlag {
            MANUAL_RETRY_ENABLED = 1 << 0,
            AUTOMATIC_RETRY_ENABLED = 1 << 1,
            LOGGING_DISABLED = 1 << 2,
            RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE = 1 << 3,
            CANCELABLE_UP_TO_HERE = 1 << 4,
            ALL_ERRORS_KILL_SHALLOT = 1 << 5,
            IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY = 1 << 6,
        };

        /**
         * @brief Used for specifying metadata for a sh::exceptions::Exception.
         */
        class ExceptionData : public QMap<QString, QString>
        {
        public:
            ExceptionData(const ExceptionData& o);
            ExceptionData(const QMap<QString, QString>& o);
            ExceptionData();
            ExceptionData name(QString name);
            ExceptionData classes(QString classes);
            ExceptionData aux(QString aux);
            ExceptionData details(QString details);
            ExceptionData message(QString message);
            ExceptionData runtime();
            ExceptionData program();
            ExceptionData retryable(bool v=true);
            ExceptionData resumeable(bool v=true);
            ExceptionData autoRetryRecommended(int v=-1);
            ExceptionData detailsAreInteresting(bool v=true);
        protected:
            void insertIfUnset(QString key, QString value);
        };

        /** \brief
         * Shallot exception base class. Also contains some static methods for general work with exceptions.
         */
        class Exception
        {
        public:

            template<class Handler>
            class RegisterHandler
            {
            public:
                RegisterHandler(Handler handler, QStack<Handler>* stack)
                {
                    stack->push(handler);
                    _stack = stack;
                }

                ~RegisterHandler()
                {
                    _stack->pop();
                }
            private:
                QStack<Handler>* _stack;
            };

            template<class Handler>
            static
            std::shared_ptr<RegisterHandler<Handler> > createRegisterHandler(Handler handler, QStack<Handler>* stack)
            {
                return std::shared_ptr<RegisterHandler<Handler> >(new RegisterHandler<Handler>(handler, stack));
            }

            class HandlerSettings
            {
                friend class ExceptionHandlerSettingsManager;
            public:
                QString cancelText;
                QStack<std::function<bool (sh::exceptions::Exception&)> > exceptionHandler_retryable;
                QStack<std::function<void (sh::exceptions::Exception&)> > exceptionHandler_resumeable;
                QStack<std::function<void (sh::exceptions::Exception&)> > exceptionHandler_hard;
                QStack<std::function<void ()> > exceptionHandler_cancel;
            private:
                HandlerSettings();
            };

            static const QString UNDEFINED_ERROR_OCCURRED;
            static const QString SHALLOT_MUST_CLOSE_TEXT;
            static const QString Value_isShallotException;
            QString message() const ;
            QString name() const;
            QString classes() const;
            QString details() const;
            QString callstack() const;
            QString auxiliary() const;
            QString customValue(QString key) const;
            bool isRuntimeException() const;
            bool isProgramException() const;
            bool isRetryable() const;
            bool isResumeable() const;
            bool isDetailsAreInteresting() const;
            int autoRetryRecommendedIn() const;
            void setResumeable(bool v);
            void setReTryable(bool v);
            void setCustomValue(QString key, QString value);
            bool isClass(QString classname);

            Exception(ExceptionData data);
            Exception();

            /** \brief
             * Executes some code with some standard exection handling around.
             * \param fct The code to execute guarded.
             * \param flags Flags of #ExecuteGuardFlag for choosing the behavior.
             */
            static void executeGuarded(std::function<void()> fct, int flags=0);
            static void executeGuarded_errorpanel(std::function<void()> fct, int flags=0);

            sh::exceptions::ExceptionData data();

            static QString _demangle(QString l);

        protected:
            sh::exceptions::ExceptionData _data;
            static void _executeGuarded_errorpanel_helper();
            static sh::ui::ActionExecutionInfoPanel* errorpanel;
            static QMutex errorpanelmutex;

        };

        /** \brief A special exception for cancellation of some action on user behalf.
         *
         * A very special exception outside of the hierarchy.
         * It is only used by the infrastructure; never throw it directly.
         */
        class CancelException
        {
        public:
            CancelException();
        };

        /** \brief
         * Only used by the infrastructure; never throw it directly.
         */
        class ThreadAbortException
        {
        public:
            ThreadAbortException();
        };

        /** \brief Managing how to default-handle unhandled exceptions.
         */
        class ExceptionHandlerSettingsManager
        {
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static ExceptionHandlerSettingsManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            Exception::HandlerSettings* handlerSettings();
            virtual ~ExceptionHandlerSettingsManager();
        private:
            ExceptionHandlerSettingsManager();
            QMap<QThread*, Exception::HandlerSettings*> _handlerSettings;
            QMutex _mutex;
        };

    }
}

#endif // EXCEPTION_H
