/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFile>
#include <QDir>
#include <QBuffer>
#include <QProcess>
#include "actions/actionsmanager.h"
#include "base/filesystemmodel.h"
#include "base/filesystemhandlerregister.h"
#include "base/operation.h"
#include "exceptions/ioexception.h"
#include "tools/datastream.h"
#include "tools/mainthread.h"
#include "tools/filetypes/filetypemanager.h"
#include "tools/singletoninitializer.h"
#include "archivefilesystemhandler.h"


namespace sh{namespace filesystemhandlers{INITIALIZE_STATIC_IN_MAINTHREAD(100410, ArchiveFilesystemHandler)}}

sh::filesystemhandlers::ArchiveFilesystemHandler::ArchiveFilesystemHandler(sh::base::FilesystemModel *model) :
    sh::base::FilesystemHandler(model)
{
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list)
{
    QStringList ql;
    QSet<QString> implicitdirs;
    QSet<QString> explicitdirs;
    QString slashedpath = eurl->path();
    if (!slashedpath.endsWith("/"))
        slashedpath += "/";
    std::function<void(qint64, QDateTime, QString)> onitem = [&](qint64, QDateTime, QString fpath) {
        QString unslashedfpath = fpath;
        if (unslashedfpath.endsWith("/") && unslashedfpath.length()>1)
            unslashedfpath = unslashedfpath.mid(0, unslashedfpath.length()-1);
        if (type==sh::base::FilesystemNodeType::Directory)
        {
            int idx;
            QString usfpath = unslashedfpath;
            while ((idx = usfpath.lastIndexOf("/")) >= 0)
            {
                usfpath = usfpath.mid(0, idx);
                if (!usfpath.isEmpty())
                    implicitdirs.insert(usfpath);
            }
        }
        if ((fpath.endsWith("/") && (type==sh::base::FilesystemNodeType::Directory))
                || (!fpath.endsWith("/") && (type==sh::base::FilesystemNodeType::File)))
        {
            if (fpath.startsWith(slashedpath))
            {
                QString relname = fpath.mid(slashedpath.length());
                if (!relname.isEmpty())
                {
                    if (relname.endsWith("/"))
                        relname = relname.mid(0, relname.length()-1);
                    if (!relname.contains("/")) {
                        if (type==sh::base::FilesystemNodeType::Directory)
                            explicitdirs.insert(relname);
                        ql << relname;
                    }
                }
            }
        }
    };
    if (eurl->scheme() == "zip")
    {
        if (type!=sh::base::FilesystemNodeType::File && type!=sh::base::FilesystemNodeType::Directory)
            return;
        _parseZipEntryListOutput(op, eurl, onitem);
    }
    else if (eurl->scheme() == "tar")
    {
        if (type!=sh::base::FilesystemNodeType::File && type!=sh::base::FilesystemNodeType::Directory)
            return;
        _parseTarEntryListOutput(op, eurl, onitem);
    }
    if (type==sh::base::FilesystemNodeType::Directory)
    {
        foreach(QString fid, implicitdirs)
        {
            if (fid.startsWith(slashedpath))
            {
                QString relname = fid.mid(eurl->path().length());
                if (!relname.contains("/") && !relname.isEmpty() && !explicitdirs.contains(relname))
                    ql << relname;
            }
        }
    }
    list->setItems(ql);
}

QString sh::filesystemhandlers::ArchiveFilesystemHandler::getLinkTarget(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return QString();
}

sh::base::FilesystemNodeType sh::filesystemhandlers::ArchiveFilesystemHandler::getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QSet<QString> implicitdirs;
    sh::base::FilesystemNodeType result = sh::base::FilesystemNodeType::NONE;
    if (eurl->outerUrlIsRootDirectory())
        return sh::base::FilesystemNodeType::Directory;
    std::function<void(qint64, QDateTime, QString)> onitem = [&](qint64, QDateTime, QString fpath) {
        QString slashedpath = eurl->path();
        if (!slashedpath.endsWith("/"))
            slashedpath += "/";
        QString slashedfpath = fpath;
        if (!slashedfpath.endsWith("/"))
            slashedfpath += "/";
        QString unslashedfpath = fpath;
        if (unslashedfpath.endsWith("/") && unslashedfpath.length()>1)
            unslashedfpath = unslashedfpath.mid(0, unslashedfpath.length()-1);
        int idx;
        QString usfpath = unslashedfpath;
        while ((idx = usfpath.lastIndexOf("/")) >= 0)
        {
            usfpath = usfpath.mid(0, idx);
            if (!usfpath.isEmpty())
            {
                if (!implicitdirs.contains(usfpath))
                {
                    implicitdirs.insert(usfpath);
                    if(usfpath == eurl->path())
                        result = sh::base::FilesystemNodeType::Directory;
                }
            }
        }
        if (slashedfpath == slashedpath)
        {
            if (fpath.endsWith("/"))
                result = sh::base::FilesystemNodeType::Directory;
            else
                result = sh::base::FilesystemNodeType::File;
        }
    };
    if (eurl->scheme() == "zip")
    {
        _parseZipEntryListOutput(op, eurl, onitem);
    }
    else if (eurl->scheme() == "tar")
    {
        _parseTarEntryListOutput(op, eurl, onitem);
    }
    return result;
}

qint64 sh::filesystemhandlers::ArchiveFilesystemHandler::getSize(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    qint64 result = 0;
    if (eurl->scheme() == "zip")
    {
        _parseZipEntryListOutput(op, eurl, [&](int fsize, QDateTime, QString fpath){
            QString slashedpath = eurl->path();
            if (!slashedpath.endsWith("/"))
                slashedpath += "/";
            QString slashedfpath = fpath;
            if (!slashedfpath.endsWith("/"))
                slashedfpath += "/";
            if (slashedfpath == slashedpath)
            {
                result = fsize;
            }
        });
    }
    else if (eurl->scheme() == "tar")
    {
        QString container = op->fetchContainerFile(eurl);
        QString ans = execute("tar", QStringList({"tfv", container, "--no-recursion", eurl->path().mid(1)}));
        QStringList sans2 = ans.split(" ");
        QStringList sans;
        foreach(QString s, sans2)
            if (!s.isEmpty())
                sans.append(s.trimmed());
        if (sans.length() >= 3)
        {
            bool ok = true;
            qlonglong ll = sans[2].toLongLong(&ok);
            if (ok)
                result = (qint64) ll;
        }
    }
    return result;
}

QString sh::filesystemhandlers::ArchiveFilesystemHandler::getMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    return sh::tools::filetypes::FileTypeManager::instance()->determineMimetype(op, eurl);
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canGetFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

std::shared_ptr<QIODevice> sh::filesystemhandlers::ArchiveFilesystemHandler::getFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString contentfile;
    QString container = op->fetchContainerFile(eurl);
    if (eurl->scheme() == "zip")
    {
        QString tempdir = op->getTempDir();
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        execute("unzip", QStringList({"-j", container, eurl->path().mid(1)}));
        QDir::setCurrent(cwd.path());
        contentfile = tempdir+"/"+eurl->basename();
    }
    else if (eurl->scheme() == "tar")
    {
        QString tempdir = op->getTempDir();
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        execute("tar", QStringList({"xf", container, eurl->path().mid(1)}));
        QDir::setCurrent(cwd.path());
        contentfile = tempdir+"/"+eurl->path();
    }
    if (!contentfile.isEmpty())
        return std::make_shared<sh::tools::LocalFile>(contentfile, true);
    else
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to fetch: ") + eurl->asString()).retryable(false));
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canCreateDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canCreateFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::createDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString container = op->fetchContainerFile(eurl);
    if (eurl->scheme() == "zip")
    {
        QString tempdir = op->getTempDir();
        QDir(tempdir).mkpath(eurl->path().mid(1));
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        QByteArray bfilelist = eurl->path().mid(1).toUtf8();
        execute("zip", QStringList({container, "-@"}), true, 0, &bfilelist);
        QDir::setCurrent(cwd.path());
    }
    else if (eurl->scheme() == "tar")
    {
        QString tempdir = op->getTempDir();
        QDir(tempdir).mkpath(eurl->path().mid(1));
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        _tarOperationAfterCompression(op, container, [&](QString tarfile){
            execute("tar", QStringList({"r", "--file="+tarfile, eurl->path().mid(1)}));
        });
        QDir::setCurrent(cwd.path());
    }
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canCreateLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::createLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to create link")).retryable(false));
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::createFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QIODevice *content, HandlerTransfer*)
{
    QString container = op->fetchContainerFile(eurl);
    if (eurl->scheme() == "zip")
    {
        QString tempdir = op->getTempDir();
        QDir(tempdir).mkpath(eurl->parentSegment()->path().mid(1));
        sh::base::Operation::writeIODeviceToFile(content, tempdir + eurl->path());
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        QByteArray bfilelist = (eurl->path().mid(1)).toUtf8();
        execute("zip", QStringList({container, "-@"}), true, 0, &bfilelist);
        QDir::setCurrent(cwd.path());
    }
    else if (eurl->scheme() == "tar")
    {
        QString tempdir = op->getTempDir();
        QDir(tempdir).mkpath(eurl->parentSegment()->path().mid(1));
        sh::base::Operation::writeIODeviceToFile(content, tempdir + eurl->path());
        if (getType(op, eurl) != sh::base::FilesystemNodeType::NONE)
            deleteItem(op, eurl);
        QDir cwd = QDir::current();
        QDir::setCurrent(tempdir);
        _tarOperationAfterCompression(op, container, [&](QString tarfile){
            execute("tar", QStringList({"r", "--file="+tarfile, eurl->path().mid(1)}));
        });
        QDir::setCurrent(cwd.path());
    }
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::ArchiveFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::base::Eurl>>)
{
    return QList<std::shared_ptr<sh::actions::AbstractActionItem> >();
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canDeleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::canRenameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::renameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src, QString destpath)
{
    if (src->scheme() == "zip" || src->scheme() == "tar")
    {
        std::shared_ptr<QIODevice> cnt = getFileContent(op, src);
        if (!cnt->isOpen() && !cnt->open(QFile::ReadOnly))
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while opening: ") + src->asString()));
        HandlerTransfer _handlertransfer;
        createFile(op, src->root()->withAppendedSegments(destpath), cnt.get(), &_handlertransfer);
        deleteItem(op, src);
    }
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::deleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString container = op->fetchContainerFile(eurl);
    if (eurl->scheme() == "zip")
    {
        bool isdir = getType(op, eurl) == sh::base::FilesystemNodeType::Directory;
        QByteArray bfilelist = (_zipMaskFile(eurl->path().mid(1))+(isdir?"/*":"")).toUtf8();
        execute("zip", QStringList({"-d", container, "-@"}), true, 0, &bfilelist);
    }
    else if (eurl->scheme() == "tar")
    {
        _tarOperationAfterCompression(op, container, [&](QString tarfile){
            execute("tar", QStringList({"--delete", "--file="+tarfile, eurl->path().mid(1)}));
        });
    }
}

QDateTime sh::filesystemhandlers::ArchiveFilesystemHandler::getMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QDateTime result;
    if (eurl->scheme() == "zip")
    {
        _parseZipEntryListOutput(op, eurl, [&](qint64, QDateTime mtime, QString fpath){
            QString slashedpath = eurl->path();
            if (!slashedpath.endsWith("/"))
                slashedpath += "/";
            QString slashedfpath = fpath;
            if (!slashedfpath.endsWith("/"))
                slashedfpath += "/";
            if (slashedfpath == slashedpath)
            {
                result = mtime;
            }
        });
    }
    else if (eurl->scheme() == "tar")
    {
        QString container = op->fetchContainerFile(eurl);
        QString ans = execute("tar", QStringList({"tfv", container, "--no-recursion", eurl->path().mid(1), "--full-time"}));
        QStringList sans2 = ans.split(" ");
        QStringList sans;
        foreach(QString s, sans2)
            if (!s.isEmpty())
                sans.append(s.trimmed());
        if (sans.length() >= 5)
        {
            result = QDateTime::fromString(sans[3] + "T" + sans[4], Qt::ISODate);
        }
    }
    return result;
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::setMtime(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QDateTime)
{
}

QString sh::filesystemhandlers::ArchiveFilesystemHandler::_zipMaskFile(QString path)
{
    return path.replace("[", "[[]").replace("*", "[*]").replace(".", "[.]").replace("?", "[?]");
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::_parseZipEntryListOutput(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, std::function<void(qint64, QDateTime, QString)> onitem)
{
    QString container = op->fetchContainerFile(eurl);
    QString custdatakey = "_archive_unzipout_" + container;
    QString output;
    QVariant unziplistout = op->getCustomData(0, custdatakey);
    if (unziplistout.isNull())
    {
        output = execute("unzip", QStringList({"-l", container}));
        op->setCustomData(0, custdatakey, output);
    }
    else
    {
        output = unziplistout.toString();
    }
    foreach(QString l, output.split("\n"))
    {
        QRegularExpressionMatch m = parsere.match(l);
        if (m.hasMatch())
        {
            bool sizeok = true;
            qulonglong size = m.captured(1).toULongLong(&sizeok);
            if (sizeok)
            {
                QDateTime mtime = QDateTime::fromString(m.captured(2).replace(" ", "T"), Qt::ISODate);
                if (mtime.isValid())
                {
                    onitem(size, mtime, "/"+m.captured(3));
                }
            }
        }
    }
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::_parseTarEntryListOutput(sh::base::Operation *op, std::shared_ptr<const sh::base::Eurl>eurl, std::function<void(qint64, QDateTime, QString)> onitem)
{
    QString container = op->fetchContainerFile(eurl);
    QString custdatakey = "_archive_tarout_" + container;
    QString output;
    QVariant tarlistout = op->getCustomData(0, custdatakey);
    if (tarlistout.isNull())
    {
        output = execute("tar", QStringList({"tf", container, "--quoting-style=c"}));
        op->setCustomData(0, custdatakey, output);
    }
    else
    {
        output = tarlistout.toString();
    }
    foreach(QString l, output.split("\n"))
    {
        l = l.trimmed();
        l = l.mid(1, l.length()-2);
        l = l.replace("\\a", "\a")
                .replace("\\b", "\b")
                .replace("\\f", "\f")
                .replace("\\n", "\n")
                .replace("\\r", "\r")
                .replace("\\t", "\t")
                .replace("\\v", "\v")
                .replace("\\\"", "\"")
                .replace("\\\\", "\\");
        onitem(0, QDateTime(), "/"+l);
    }
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::_tarOperationAfterCompression(sh::base::Operation *op, QString container, std::function<void(QString)> ontarfile)
{
    QString compr;
    QString tarfile = container;
    if (container.endsWith(".tar.gz") || container.endsWith(".tgz"))
        compr = "gzip";
    if (container.endsWith(".tar.bz2") || container.endsWith(".tbz2"))
        compr = "bzip2";
    if (!compr.isNull())
    {
        QString tempdir = op->getTempDir();
        tarfile = tempdir + "/x.tar";
        QByteArray rawanswer;
        execute(compr, QStringList({"-d", "-k", "-c", container}), true, &rawanswer);
        QBuffer buf(&rawanswer);
        buf.open(QFile::ReadOnly);
        sh::base::Operation::writeIODeviceToFile(&buf, tarfile);
    }
    ontarfile(tarfile);
    if (!compr.isNull())
    {
        QByteArray rawanswer;
        execute(compr, QStringList({"-c", tarfile}), true, &rawanswer);
        QBuffer buf(&rawanswer);
        buf.open(QFile::ReadOnly);
        sh::base::Operation::writeIODeviceToFile(&buf, container);
    }
}

bool sh::filesystemhandlers::ArchiveFilesystemHandler::testexecute(QString cmd)
{
    QProcess fileprocess;
    fileprocess.setProgram(cmd);
    fileprocess.start();
    fileprocess.waitForStarted();
    bool result = fileprocess.error() != QProcess::FailedToStart;
    fileprocess.close();
    return result;
}

QString sh::filesystemhandlers::ArchiveFilesystemHandler::execute(QString cmd, QStringList args, bool throwOnBadExitCode, QByteArray* rawanswer, QByteArray* bstdin)
{
    QProcess fileprocess;
    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    fileprocess.setProcessEnvironment(env);
    fileprocess.setProgram(cmd);
    fileprocess.setArguments(args);
    fileprocess.start();
    if (fileprocess.waitForStarted())
    {
        if (bstdin)
            fileprocess.write(*bstdin);
        fileprocess.closeWriteChannel();
    }
    if (fileprocess.waitForFinished())
    {
        QByteArray ba = fileprocess.readAll();
        if (rawanswer)
            *rawanswer = ba;
        QString outp = QString::fromUtf8(ba);
        if (throwOnBadExitCode && (fileprocess.exitStatus() != QProcess::NormalExit || fileprocess.exitCode() != 0))
            throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().details(tr("The command '%1 %2' led to an error: %3")
                                   .arg(cmd).arg(args.join(" ")).arg(outp)));
        return outp;
    }
    else
        throw sh::exceptions::RuntimeException(sh::exceptions::ExceptionData().details(QString("The '%1' command reacted in an unexpected way.").arg(cmd)));
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::_initialize()
{
    static std::shared_ptr<sh::filesystemhandlers::ArchiveFilesystemHandler> _handler = std::make_shared<sh::filesystemhandlers::ArchiveFilesystemHandler>(sh::base::FilesystemModel::instance());
    if (testexecute("zip") && testexecute("unzip"))
    {
        sh::base::FilesystemHandlerRegister::instance()->addHandler("zip", _handler);
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenZipArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.zip")
        })));
    }
    if (testexecute("tar"))
    {
        sh::base::FilesystemHandlerRegister::instance()->addHandler("tar", _handler);
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenTarArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.tar")
        })));
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenTarArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.tar.gz")
        })));
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenTarArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.tgz")
        })));
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenTarArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.tar.bz2")
        })));
        sh::actions::ActionsManager::instance()->registerActionFactory(
                    std::make_shared<sh::actions::ActionFactory<ActionOpenTarArchive>>(
                        sh::actions::ActionCategory::getByName("open"), QList<std::shared_ptr<sh::actions::Predicate>>({
                        std::make_shared<sh::actions::OnFilesPredicate>(),
                        std::make_shared<sh::actions::ByRegExpPredicate>(".*\\.tbz2")
        })));
    }
}

void sh::filesystemhandlers::ArchiveFilesystemHandler::_shutdown()
{
    foreach(QObject* afr, actionfactoryregistrations)
        delete afr;
}

QRegularExpression sh::filesystemhandlers::ArchiveFilesystemHandler::parsere("^ *(\\d+)  +([^ ]+ ?[^ ]+)  +(.+)$");

QList<QObject*> sh::filesystemhandlers::ArchiveFilesystemHandler::actionfactoryregistrations;
