/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef GIO

#include <gio/gio.h>
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "actions/actionfactory.h"
#include "actions/actionsmanager.h"
#include "base/filesystemhandlerregister.h"
#include "exceptions/ioexception.h"
#include "tools/mainthread.h"
#include "tools/logger.h"
#include "tools/iconmanager.h"
#include "tools/singletoninitializer.h"
#include "tools/accounts/accountsmanager.h"
#include "ui/mainwindow.h"
#include "gnomeiofilesystemhandler.h"
#include "gnomeionetworkfilesystemhandler.h"
#include "gnomeiodevicesfilesystemhandler.h"


namespace sh{ namespace filesystemhandlers{INITIALIZE_STATIC_IN_MAINTHREAD(10000, ActionMountGnomeIOLocation)}}

namespace sh{ namespace filesystemhandlers{INITIALIZE_STATIC_IN_MAINTHREAD(10000, ActionUnmountGnomeIOLocation)}}

sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOFilesystemHandler(sh::base::FilesystemModel *model) :
    sh::base::FilesystemHandler(model)
{
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::logIfError(void* error)
{
    GError** _e = (GError**) error;
    if (*_e) {
        LOG_DEBUG(QString(((GError*)*_e)->message))
        *_e = 0;
    }
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::throwIfError(void* error)
{
    GError** _e = (GError**) error;
    if (*_e)
    {
        QString details = QObject::tr("Something bad happened in the Gnome IO subsystem: ") + QString(((GError*)*_e)->message);
        logIfError(error);
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(details));
    }
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::itemlist(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list)
{
    QByteArray cguri = eurl2gurlb(eurl);
    GFile* d = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    GFileEnumerator* direnum = g_file_enumerate_children(d, "standard::target-uri,standard::name,standard::display-name,standard::type,mountable::can-mount,mountable::can-unmount",
                                                         G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    list->beginIterativeAdding();
    if (direnum)
    {
        while (TRUE)
        {
            GFileInfo *info = g_file_enumerator_next_file(direnum, NULL, &error);
            logIfError(&error);
            if (!info)
                break;
            quint32 giotype = getFileInfoUInt32(G_FILE_ATTRIBUTE_STANDARD_TYPE, info);
            sh::base::FilesystemNodeType ntype = gfiletypetonodetype(giotype);
            if (type == ntype)
            {
                QString url = getFileInfoString(G_FILE_ATTRIBUTE_STANDARD_TARGET_URI, info);
                QString name = g_file_info_get_name(info);
                QString displayname = getFileInfoString(G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME, info);
                std::shared_ptr<const sh::base::Eurl> neweurl = (url.isEmpty()) ? eurl->withAppendedSegment(name) : gurl2eurl(url);
                std::shared_ptr<sh::base::FilesystemNode> node = list->addCustomItem(neweurl, displayname, sh::base::FilesystemHandlerRegister::instance()->findHandler(neweurl->scheme()).get());
                bool canmount = getFileInfoBool(G_FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT, info);
                bool canunmount = getFileInfoBool(G_FILE_ATTRIBUTE_MOUNTABLE_CAN_UNMOUNT, info);
                // hack: for some locations (e.g. for smb://somehost/), gio requires to mount it, but does not tell us this way :-/
                if (giotype == G_FILE_TYPE_SHORTCUT && !canmount && !canunmount)
                {
                    GError* _error = 0;
                    QByteArray _cguri = eurl2gurlb(neweurl);
                    GFile* _f = g_file_new_for_uri(_cguri);
                    GFileInfo* _finf = g_file_query_info(_f, "", G_FILE_QUERY_INFO_NONE, NULL, &_error);
                    canmount = (_error &&_error->code == G_IO_ERROR_NOT_MOUNTED);
                    g_object_unref(_f);
                    g_object_unref(_finf);
                }
                // end hack
                if (canmount)
                    sh::filesystemhandlers::ActionMountGnomeIOLocation::offerMountLocation(node, url);
                else
                    sh::filesystemhandlers::ActionMountGnomeIOLocation::unofferMount(node);
                if (canunmount)
                    sh::filesystemhandlers::ActionUnmountGnomeIOLocation::offerUnmountLocation(node, url);
                else
                    sh::filesystemhandlers::ActionUnmountGnomeIOLocation::unofferUnmount(node);
            }
            g_object_unref(info);
        }
        g_object_unref(direnum);
    }
    list->endIterativeAdding();
    g_object_unref(d);
}

qint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::getSize(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    return getFileInfoUInt(G_FILE_ATTRIBUTE_STANDARD_SIZE, eurl2gurl(eurl));
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::getLinkTarget(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QByteArray cguri = eurl2gurlb(eurl);
    GFile* f = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET, G_FILE_QUERY_INFO_NONE, NULL, &error);
    throwIfError(&error);
    QString result = g_file_info_get_symlink_target(finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::getMimetype(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    return getFileInfoString(G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE, eurl2gurl(eurl));
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canCreateDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canCreateLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canCreateFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

sh::base::FilesystemNodeType sh::filesystemhandlers::GnomeIOFilesystemHandler::getType(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    return gfiletypetonodetype(getFileInfoUInt32(G_FILE_ATTRIBUTE_STANDARD_TYPE, eurl2gurl(eurl)));
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canGetFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

std::shared_ptr<QIODevice> sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    return std::make_shared<sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOReadDataDevice>(eurl2gurl(eurl));
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::createDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    QByteArray cguri = eurl2gurlb(eurl);
    GFile* d = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    g_file_make_directory(d, NULL, &error);
    throwIfError(&error);
    g_object_unref(d);
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::createLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString target)
{
    QByteArray cguri = eurl2gurlb(eurl);
    QByteArray tgtba = target.toUtf8();
    char* tgtcguri = tgtba.data();
    GFile* d = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    g_file_make_symbolic_link (d, tgtcguri, NULL, &error);
    throwIfError(&error);
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::createFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl, QIODevice *content, HandlerTransfer *handlertransfer)
{
    QByteArray cguri = eurl2gurlb(eurl);
    GFile* d = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    GFileOutputStream* ostream = g_file_create(d, G_FILE_CREATE_NONE, NULL, &error);
    throwIfError(&error);
    const int BUFFER_SIZE = 4096;
    char buffer[BUFFER_SIZE];
    while (qint64 l = content->read(buffer, BUFFER_SIZE))
    {
        if (l == -1)
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to read from a stream.")));
        qint64 origl = l;
        char* mbuffer = buffer;
        while (l) {
            gssize w = g_output_stream_write ((GOutputStream*)ostream, mbuffer, l, NULL, &error);
            throwIfError(&error);
            l-=w;
            mbuffer+=w;
        }
        handlertransfer->respectCancel();
        handlertransfer->incrementTransferredBytes(origl);
    }
    g_output_stream_close((GOutputStream*)ostream, NULL, &error);
    throwIfError(&error);
    g_object_unref(ostream);
    g_object_unref(d);
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canRenameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::renameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> src, QString destpath)
{
    GError* error = 0;
    QByteArray cgsrc = eurl2gurlb(src);
    QByteArray cgdest = eurl2gurlb(src->root()->withAppendedSegments(destpath));
    GFile* source = g_file_new_for_uri(cgsrc.data());
    GFile* destination = g_file_new_for_uri(cgdest.data());
    g_file_move(source, destination, (GFileCopyFlags)(G_FILE_COPY_ALL_METADATA | G_FILE_COPY_NOFOLLOW_SYMLINKS), NULL, NULL, NULL, &error);
    throwIfError(&error);
    g_object_unref(source);
    g_object_unref(destination);
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::GnomeIOFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::base::Eurl>>)
{
    return QList<std::shared_ptr<sh::actions::AbstractActionItem>>();
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::canDeleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::deleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QByteArray cguri = eurl2gurlb(eurl);
    GFile* d = g_file_new_for_uri(cguri.data());
    GError* error = 0;
    GFileEnumerator* direnum = g_file_enumerate_children(d, G_FILE_ATTRIBUTE_STANDARD_NAME,
                                                         G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    if (direnum)
    {
        while (TRUE)
        {
            GFileInfo *info = g_file_enumerator_next_file(direnum, NULL, &error);
            logIfError(&error);
            if (!info)
                break;
            std::shared_ptr<const sh::base::Eurl> childeurl = eurl->withAppendedSegment(g_file_info_get_name(info));
            deleteItem(op, childeurl);
        }
        g_object_unref(direnum);
    }
    g_file_delete(d, NULL, &error);
    throwIfError(&error);
}

QDateTime sh::filesystemhandlers::GnomeIOFilesystemHandler::getMtime(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString gurl = eurl2gurl(eurl);
    return QDateTime::fromMSecsSinceEpoch(getFileInfoUInt(G_FILE_ATTRIBUTE_TIME_MODIFIED, gurl) * 1000 + getFileInfoUInt32(G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC, gurl));
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::setMtime(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QDateTime time)
{
    QByteArray ba = eurl2gurl(eurl).toUtf8();
    GFile* f = g_file_new_for_uri(ba.data());
    GError* error = 0;
    qint64 itime = time.toMSecsSinceEpoch();
    g_file_set_attribute_uint64(f, G_FILE_ATTRIBUTE_TIME_MODIFIED, itime / 1000, G_FILE_QUERY_INFO_NONE, NULL, &error);
    throwIfError(&error);
    g_file_set_attribute_uint32(f, G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC /*actually msec according to the docs*/, (quint32)(itime % 1000), G_FILE_QUERY_INFO_NONE, NULL, &error); // not always supported
}

QByteArray sh::filesystemhandlers::GnomeIOFilesystemHandler::eurl2gurlb(std::shared_ptr<const sh::base::Eurl> eurl)
{
    return eurl2gurl(eurl).toUtf8();
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::eurl2gurl(std::shared_ptr<const sh::base::Eurl> eurl)
{
    return eurl->asString().replace("{", "[").replace("}", "]");
}

std::shared_ptr<const sh::base::Eurl> sh::filesystemhandlers::GnomeIOFilesystemHandler::gurl2eurl(QString gurl)
{
    std::shared_ptr<const sh::base::Eurl> r = sh::base::Eurl::fromString(gurl.replace("[", "{").replace("]", "}"));
    return sh::base::Eurl::create(r->scheme(), r->outermostInnerEurl().get(), r->hostname().toLower(), r->path());
}

sh::base::FilesystemNodeType sh::filesystemhandlers::GnomeIOFilesystemHandler::gfiletypetonodetype(quint64 gfiletype)
{
    sh::base::FilesystemNodeType nodetype = sh::base::FilesystemNodeType::NONE;
    if (gfiletype == G_FILE_TYPE_DIRECTORY || gfiletype == G_FILE_TYPE_MOUNTABLE || gfiletype == G_FILE_TYPE_SHORTCUT)
        nodetype = sh::base::FilesystemNodeType::Directory;
    else if (gfiletype == G_FILE_TYPE_SYMBOLIC_LINK)
        nodetype = sh::base::FilesystemNodeType::Link;
    else if (gfiletype == G_FILE_TYPE_REGULAR || gfiletype == G_FILE_TYPE_SPECIAL)
        nodetype = sh::base::FilesystemNodeType::File;
    return nodetype;
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoString(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    GFile* f = g_file_new_for_uri(ba.data());
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    QString result = getFileInfoString(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoBool(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    GFile* f = g_file_new_for_uri(cguri);
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    bool result = getFileInfoBool(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

qint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoInt(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    GFile* f = g_file_new_for_uri(cguri);
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    qint64 result = getFileInfoInt(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

quint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoUInt(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    GFile* f = g_file_new_for_uri(cguri);
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    quint64 result = getFileInfoUInt(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

qint32 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoInt32(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    GFile* f = g_file_new_for_uri(cguri);
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    qint64 result = getFileInfoInt32(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

quint32 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoUInt32(const char* key, QString gurl)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    GFile* f = g_file_new_for_uri(cguri);
    GError* error = 0;
    GFileInfo* finf = g_file_query_info(f, key, G_FILE_QUERY_INFO_NONE, NULL, &error);
    logIfError(&error);
    quint64 result = getFileInfoUInt32(key, finf);
    g_object_unref(finf);
    g_object_unref(f);
    return result;
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoString(const char* key, void* fileinfo)
{
    const char* cresult = g_file_info_get_attribute_string((GFileInfo*)fileinfo, key);
    QString result = cresult;
    return result;
}

bool sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoBool(const char* key, void* fileinfo)
{
    return g_file_info_get_attribute_boolean((GFileInfo*)fileinfo, key);
}

qint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoInt(const char* key, void* fileinfo)
{
    return g_file_info_get_attribute_int64((GFileInfo*)fileinfo, key);
}

quint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoUInt(const char* key, void* fileinfo)
{
    return g_file_info_get_attribute_uint64((GFileInfo*)fileinfo, key);
}

qint32 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoInt32(const char* key, void* fileinfo)
{
    return g_file_info_get_attribute_int32((GFileInfo*)fileinfo, key);
}

quint32 sh::filesystemhandlers::GnomeIOFilesystemHandler::getFileInfoUInt32(const char* key, void* fileinfo)
{
    return g_file_info_get_attribute_uint32((GFileInfo*)fileinfo, key);
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::mountDevice(QString gvolumeuuid, sh::actions::ActionExecutionInfo* info)
{
    QString gurl;
    QMutex* mutex = new QMutex();
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* opdata;
    GMountOperation* mountOp = (GMountOperation*) sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::createMountOperation(info, gvolumeuuid, &opdata);
    GVolume* volume = 0;
    GList* vollist;
    GError* innererror = 0;
    QList<void*> userdata({&innererror, mutex});
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        GVolumeMonitor* volmon = g_volume_monitor_get ();
        vollist = g_volume_monitor_get_volumes (volmon);
        GList *vollisti;
        for (vollisti = vollist; vollisti != NULL; vollisti = vollisti->next)
        {
            GVolume* _volume = (GVolume*) vollisti->data;
            if (!volume)
            {
                QString uuid = g_volume_get_identifier(_volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                if (uuid == gvolumeuuid)
                {
                    mutex->lock();
                    g_volume_mount(_volume, G_MOUNT_MOUNT_NONE, mountOp, NULL, [](GObject *source_object,
                                                                                  GAsyncResult *res,
                                                                                  gpointer user_data){
                        GError** error = (GError**)((QList<void*>*)user_data)->at(0);
                        QMutex* mutex = (QMutex*)((QList<void*>*)user_data)->at(1);
                        g_volume_mount_finish((GVolume*)source_object, res, error);
                        g_object_unref(source_object);
                        mutex->unlock();
                    }, &userdata);
                    volume = _volume;
                    continue;
                }
            }
            g_object_unref(_volume);
        }
        g_object_unref(volmon);
    });
    mutex->lock();
    mutex->unlock();
    delete mutex;
    if (volume != 0)
    {
        GMount* mount = g_volume_get_mount(volume);
        GFile* root = g_mount_get_root(mount);
        char* croot = g_file_get_uri(root);
        gurl = QString(croot);
        g_free(croot);
        g_object_unref(root);
        g_object_unref(mount);
    }
    g_list_free(vollist);
    if (innererror && innererror->code == G_IO_ERROR_ALREADY_MOUNTED)
        innererror = 0;
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::deleteMountOperation(opdata);
    throwIfError(&innererror);
    return gurl;
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::mountLocation(QString gurl, sh::actions::ActionExecutionInfo* info)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    QMutex* mutex = new QMutex();
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* opdata;
    GMountOperation* mountOp = (GMountOperation*) sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::createMountOperation(info, gurl, &opdata);
    GError* innererror = 0;
    QList<void*> userdata({&innererror});
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        GFile* file = g_file_new_for_uri(cguri);
        userdata.append(mutex);
        mutex->lock();
        g_file_mount_enclosing_volume(file, G_MOUNT_MOUNT_NONE, mountOp, NULL, [](GObject *source_object,
                                      GAsyncResult *res,
                                      gpointer user_data){
            GError** error = (GError**)((QList<void*>*)user_data)->at(0);
            QMutex* mutex = (QMutex*)((QList<void*>*)user_data)->at(1);
            g_file_mount_enclosing_volume_finish((GFile*)source_object, res, error);
            g_object_unref(source_object);
            mutex->unlock();
        }, &userdata);
        g_object_unref(file);
    });
    mutex->lock();
    mutex->unlock();
    delete mutex;
    if (innererror && innererror->code == G_IO_ERROR_ALREADY_MOUNTED)
        innererror = 0;
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::deleteMountOperation(opdata);
    throwIfError(&innererror);
    return gurl;
}

QString sh::filesystemhandlers::GnomeIOFilesystemHandler::unmountLocation(QString gurl, sh::actions::ActionExecutionInfo* info)
{
    QByteArray ba = gurl.toUtf8();
    char* cguri = ba.data();
    QMutex* mutex = new QMutex();
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* opdata;
    GMountOperation* mountOp = (GMountOperation*) sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::createMountOperation(info, gurl, &opdata);
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        GError* error = 0;
        GFile* file = g_file_new_for_uri(cguri);
        gpointer userdata = mutex = new QMutex();
        GMount* mount = g_file_find_enclosing_mount(file, NULL, &error);
        throwIfError(&error);
        mutex->lock();
        g_mount_unmount_with_operation(mount, G_MOUNT_UNMOUNT_NONE, mountOp, NULL, [](GObject *source_object,
                                       GAsyncResult *res,
                                       gpointer user_data){
            sh::exceptions::Exception::executeGuarded([&](){
                GError* error = 0;
                g_mount_unmount_with_operation_finish((GMount*)source_object, res, &error);
                throwIfError(&error);
            }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
            g_object_unref(source_object);
            ((QMutex*)user_data)->unlock();
        }, userdata);
        g_object_unref(file);
        g_object_unref(mount);
    });
    mutex->lock();
    mutex->unlock();
    delete mutex;
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::deleteMountOperation(opdata);
    return gurl;
}

sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOReadDataDevice::GnomeIOReadDataDevice(QString guri)
{
    QByteArray ba = guri.toUtf8();
    char* cguri = ba.data();
    gfile = g_file_new_for_uri(cguri);
    GError* error = 0;
    gstream = g_file_read((GFile*)gfile, NULL, &error);
    sh::filesystemhandlers::GnomeIOFilesystemHandler::throwIfError(&error);
}

qint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOReadDataDevice::readData(char* data, qint64 maxSize)
{
    GError* error = 0;
    qint64 res = g_input_stream_read((GInputStream*)gstream, (void*)data, maxSize, NULL, &error);
    sh::filesystemhandlers::GnomeIOFilesystemHandler::throwIfError(&error);
    return res;
}

qint64 sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOReadDataDevice::writeData(const char*, qint64)
{
    return -1;
}

sh::filesystemhandlers::GnomeIOFilesystemHandler::GnomeIOReadDataDevice::~GnomeIOReadDataDevice()
{
    g_object_unref(gstream);
    g_object_unref(gfile);
}

sh::filesystemhandlers::ActionMountGnomeIOLocation::ActionMountGnomeIOLocation(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
    : sh::actions::ActionActionItem(tr("Mount"), true, "online"), _nodes(nodes)
{
    initialize();
}

QString sh::filesystemhandlers::ActionMountGnomeIOLocation::mount(sh::actions::ActionExecutionInfo* info, QString gvolumeuuid, QString location, std::shared_ptr<sh::base::FilesystemNode> node)
{
    QString grooturl;
    sh::actions::ActionExecutionInfo* _info;
    if (info)
        _info = info;
    else
        sh::tools::MainThread::dispatcher()->invokeSync([&](){
            _info = new sh::actions::ActionExecutionInfo();
        });
    try{
        if (!gvolumeuuid.isEmpty())
        {
            _info->setHead(tr("Mounting %1").arg(gvolumeuuid));
            grooturl = sh::filesystemhandlers::GnomeIOFilesystemHandler::mountDevice(gvolumeuuid, _info);
        }
        else if (!location.isEmpty())
        {
            _info->setHead(tr("Mounting %1").arg(location));
            grooturl = sh::filesystemhandlers::GnomeIOFilesystemHandler::mountLocation(location, _info);
        }
    }
    catch (...)
    {
        if (!info)
            _info->deleteLater();
        throw;
    }
    if (!info)
        _info->deleteLater();
    if (!grooturl.isNull() && node)
    {
        std::shared_ptr<sh::base::FilesystemNode> parentnode = node->parentnode();
        if (parentnode) {
            parentnode->requestContainedItems(sh::base::FilesystemNodeType::Directory);
            std::shared_ptr<const sh::base::Eurl> eurl = sh::filesystemhandlers::GnomeIOFilesystemHandler::gurl2eurl(grooturl);
            foreach(std::shared_ptr<sh::base::FilesystemNode> _node, sh::base::FilesystemModel::instance()->tryGetNodesForEurl(eurl))
            {
                parentnode->requestContainedItems();
                if (_node->isAlive())
                    _node->requestContainedItems();
            }
        }
    }
    return grooturl;
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::action(sh::actions::ActionExecutionInfo* info)
{
    std::shared_ptr<sh::base::FilesystemNode> node = _nodes[0];
    QString gvolumeuuid;
    {
        QMutexLocker _l(&_nodeToMountMutex);
        auto it = _nodeToMountGvolumeuuid.find(node);
        if (it != _nodeToMountGvolumeuuid.end())
            gvolumeuuid = (*it);
    }
    QString location;
    {
        QMutexLocker _l(&_nodeToMountMutex);
        auto it = _nodeToMountLocations.find(node);
        if (it != _nodeToMountLocations.end())
            location = (*it);
    }
    QString grooturl = mount(info, gvolumeuuid, location, node);
    if (!grooturl.isEmpty())
    {
        sh::tools::MainThread::dispatcher()->invokeSync([](){
            sh::ui::MainWindow::mainWindow()->jumpToNode(sh::ui::MainWindow::mainWindow()->getCurrentDirectory());
        });
    }
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::initialize()
{
    QMutexLocker _l(&_nodeToMountMutex);
    setVisible(_nodes.length()==1 && (_nodeToMountGvolumeuuid.contains(_nodes[0]) || _nodeToMountLocations.contains(_nodes[0])));
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::offerMountGVolumeUuid(std::shared_ptr<sh::base::FilesystemNode> node, QString gvolumeuuid)
{
    QMutexLocker _l(&_nodeToMountMutex);
    _nodeToMountGvolumeuuid[node] = gvolumeuuid;
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::offerMountLocation(std::shared_ptr<sh::base::FilesystemNode> node, QString location)
{
    QMutexLocker _l(&_nodeToMountMutex);
    _nodeToMountLocations[node] = location;
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::unofferMount(std::shared_ptr<sh::base::FilesystemNode> node)
{
    QMutexLocker _l(&_nodeToMountMutex);
    _nodeToMountGvolumeuuid.remove(node);
    _nodeToMountLocations.remove(node);
}

QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> sh::filesystemhandlers::ActionMountGnomeIOLocation::_nodeToMountGvolumeuuid;

QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> sh::filesystemhandlers::ActionMountGnomeIOLocation::_nodeToMountLocations;

QMutex sh::filesystemhandlers::ActionMountGnomeIOLocation::_nodeToMountMutex;

sh::filesystemhandlers::ActionUnmountGnomeIOLocation::ActionUnmountGnomeIOLocation(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
    : sh::actions::ActionActionItem(tr("Un-mount"), true, "offline"), _nodes(nodes)
{
    initialize();
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::action(sh::actions::ActionExecutionInfo* info)
{
    std::shared_ptr<sh::base::FilesystemNode> node = _nodes[0];
    QString location;
    {
        QMutexLocker _l(&_nodeToMountMutex);
        auto it = _nodeToMountLocations.find(node);
        if (it != _nodeToMountLocations.end())
            location = (*it);
    }
    QString grooturl;
    bool staticmode = false;
    if (!location.isEmpty())
    {
        info->setHead(tr("Un-mounting %1").arg(location));
        if (_staticModeNodes.contains(node))
        {
            _staticModeNodes.remove(node);
            ActionUnmountGnomeIOLocation::unofferUnmount(node);
            staticmode = true;
        }
        sh::exceptions::Exception::executeGuarded([&](){
            grooturl = sh::filesystemhandlers::GnomeIOFilesystemHandler::unmountLocation(location, info);
        }, staticmode ? sh::exceptions::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY : 0);
        if (staticmode)
            sh::tools::MainThread::dispatcher()->invokeSync([&](){
                if (node->parentnode())
                    node->parentnode()->childnodes()->removePermanentItem(node);
            });
    }
    if (!grooturl.isEmpty())
    {
        std::shared_ptr<sh::base::FilesystemNode> parentnode = node->parentnode();
        if (parentnode)
        {
            if (!staticmode)
            {
                parentnode->requestContainedItems(sh::base::FilesystemNodeType::Directory);
                std::shared_ptr<const sh::base::Eurl> eurl = sh::filesystemhandlers::GnomeIOFilesystemHandler::gurl2eurl(grooturl);
                foreach(std::shared_ptr<sh::base::FilesystemNode> _node, sh::base::FilesystemModel::instance()->tryGetNodesForEurl(eurl))
                {
                    parentnode->requestContainedItems();
                    if (_node->isAlive())
                       _node->requestContainedItems();
                }
            }
        }
        sh::tools::MainThread::dispatcher()->invokeSync([](){
            sh::ui::MainWindow::mainWindow()->jumpToNode(sh::ui::MainWindow::mainWindow()->getCurrentDirectory());
        });
    }
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::initialize()
{
    QMutexLocker _l(&_nodeToMountMutex);
    setVisible(_nodes.length()==1 && _nodeToMountLocations.contains(_nodes[0]));
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::offerUnmountLocation(std::shared_ptr<sh::base::FilesystemNode> node, QString location, bool staticmode)
{
    QMutexLocker _l(&_nodeToMountMutex);
    _nodeToMountLocations[node] = location;
    if (staticmode)
        _staticModeNodes.insert(node);
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::unofferUnmount(std::shared_ptr<sh::base::FilesystemNode> node)
{
    QMutexLocker _l(&_nodeToMountMutex);
    if (_staticModeNodes.contains(node))
        return;
    if (_nodeToMountLocations.contains(node))
        _nodeToMountLocations.remove(node);
}

QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> sh::filesystemhandlers::ActionUnmountGnomeIOLocation::_nodeToMountLocations;

QSet<std::shared_ptr<sh::base::FilesystemNode>> sh::filesystemhandlers::ActionUnmountGnomeIOLocation::_staticModeNodes;

QMutex sh::filesystemhandlers::ActionUnmountGnomeIOLocation::_nodeToMountMutex;

sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData::MountOperationsCallbackData(sh::actions::ActionExecutionInfo* info, size_t index, void *mountoperation, QString address)  : info(info), index(index), mountoperation(mountoperation), address(address)
{
}

void mountoperationcallback_aborted (GMountOperation *op, gpointer)
{
    g_mount_operation_reply(op, G_MOUNT_OPERATION_HANDLED);
}

void mountoperationcallback_askpassword(GMountOperation  *op, gchar *message, gchar *default_user, gchar *default_domain, GAskPasswordFlags flags, gpointer user_data)
{
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* _callbackdata;
    {
        QMutexLocker _l(&sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdatamutex);
        _callbackdata = sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdata[(size_t)user_data];
    }
    QString domain = default_domain;
    QString username = default_user;
    QString password;
    bool isanonymous;
    bool isremember;
    // check for infos in storage
    std::shared_ptr<const sh::base::Eurl> eurl;
    try {
        eurl = sh::base::Eurl::fromString(_callbackdata->address);
    }
    catch(sh::base::EurlMisformattedException&) {
        eurl = sh::base::Eurl::create("dummy", 0, QString::fromUtf8(_callbackdata->address.toUtf8().toBase64()), "");
    }
    QString lsserver = eurl->hostname();
    int lsport = -1;
    int lsservercl = lsserver.lastIndexOf(":");
    if (lsservercl > -1) {
        lsport = lsserver.mid(lsservercl+1).toInt();
        lsserver = lsserver.left(lsservercl);
    }
    std::shared_ptr<sh::tools::accounts::Account> searchforacct = std::make_shared<sh::tools::accounts::Account>();
    searchforacct->username = username;
    searchforacct->domain = domain;
    searchforacct->server = lsserver;
    searchforacct->serverport = lsport;
    searchforacct->protocol = eurl->scheme();
    searchforacct->authtype = "password";
    QList<std::shared_ptr<sh::tools::accounts::Account>> storedaccounts = sh::tools::accounts::AccountsManager::instance()->findAccounts(searchforacct);
    if (storedaccounts.isEmpty()) {  // anonymous if nothing else there
        searchforacct->username = "";
        storedaccounts = sh::tools::accounts::AccountsManager::instance()->findAccounts(searchforacct);
    }
    foreach (auto storedaccount, storedaccounts) {
        if (username.isEmpty() || storedaccount->username.isEmpty() || !storedaccount->authinfo.isEmpty()) {
            // either we started without even a username (then we take what we get) or we now have even full login info.
            if (storedaccount->username.isEmpty())
                g_mount_operation_set_anonymous(op, true);
            else
            {
                QByteArray pwdbd = storedaccount->domain.toUtf8();
                g_mount_operation_set_domain(op, pwdbd.data());
                QByteArray pwdbu = storedaccount->username.toUtf8();
                g_mount_operation_set_username(op, pwdbu.data());
            }
            g_mount_operation_set_password(op, storedaccount->authinfo.data());
            g_mount_operation_reply(op, G_MOUNT_OPERATION_HANDLED);
            return;
        }
    }
    // populate and execute ui
    bool needsdomain = (flags & G_ASK_PASSWORD_NEED_DOMAIN) > 0;
    bool needsusername = (flags & G_ASK_PASSWORD_NEED_USERNAME) > 0;
    bool needspassword = (flags & G_ASK_PASSWORD_NEED_PASSWORD) > 0;
    bool mayanonymous = (flags & G_ASK_PASSWORD_ANONYMOUS_SUPPORTED) > 0;
    _callbackdata->info->setVisualProcessFeedbackActive(true);  // so we see the dialog while mounting
    bool chosen = _callbackdata->info->userfeedback()->credentialsDialog(message, needsdomain, needsusername, needspassword, mayanonymous, true, &domain, &username, &password, &isanonymous, &isremember);
    if (chosen)
    {
        if (isremember) {
            std::shared_ptr<sh::tools::accounts::Account> storeacct = std::make_shared<sh::tools::accounts::Account>();
            storeacct->username = isanonymous ? "" : username;
            storeacct->domain = domain;
            storeacct->server = lsserver;
            storeacct->serverport = lsport;
            storeacct->protocol = eurl->scheme();
            storeacct->authtype = "password";
            storeacct->authinfo = password.toUtf8();
            sh::tools::accounts::AccountsManager::instance()->storeAccount(storeacct);
        }
        if (isanonymous)
            g_mount_operation_set_anonymous(op, true);
        else
        {
            QByteArray pwdbd = domain.toUtf8();
            g_mount_operation_set_domain(op, pwdbd.data());
            QByteArray pwdbu = username.toUtf8();
            g_mount_operation_set_username(op, pwdbu.data());
        }
        QByteArray pwdbp = password.toUtf8();
        g_mount_operation_set_password(op, pwdbp.data());
        g_mount_operation_reply(op, G_MOUNT_OPERATION_HANDLED);
    }
    else
        g_mount_operation_reply(op, G_MOUNT_OPERATION_ABORTED);
}

void mountoperationcallback_askquestion(GMountOperation *op, gchar *message, GStrv choices, gpointer user_data)
{
    sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* _callbackdata;
    {
        QMutexLocker _l(&sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdatamutex);
        _callbackdata = sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdata[(size_t)user_data];
    }
    QStringList answers;
    char** s = choices;
    while (*s != 0)
    {
        answers.append(*s);
        s++;
    }
    int i = _callbackdata->info->userfeedback()->messageBox(message, answers);
    g_mount_operation_set_choice(op, i);
    g_mount_operation_reply(op, G_MOUNT_OPERATION_HANDLED);
}

void mountoperationcallback_showprocesses(GMountOperation *, gchar *, GArray *, GStrv, gpointer)
{
    // we currently don't handle this
}

void mountoperationcallback_showunmountprogress(GMountOperation *op, gchar, gint64, gint64, gpointer)
{
    g_mount_operation_reply(op, G_MOUNT_OPERATION_HANDLED);
}

void* sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::createMountOperation(sh::actions::ActionExecutionInfo* info, QString address, sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData** opdata)
{
    size_t myid;
    GMountOperation* result = g_mount_operation_new();
    {
        QMutexLocker _l(&callbackdatamutex);
        myid = callbackdataindex++;
        callbackdata[myid] = (*opdata) = new sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData(info, myid, result, address);
    }
    g_signal_connect(result, "aborted", G_CALLBACK(mountoperationcallback_aborted), (gpointer) myid);
    g_signal_connect(result, "ask-password", G_CALLBACK(mountoperationcallback_askpassword), (gpointer) myid);
    g_signal_connect(result, "ask-question", G_CALLBACK(mountoperationcallback_askquestion), (gpointer) myid);
    g_signal_connect(result, "show-processes", G_CALLBACK(mountoperationcallback_showprocesses), (gpointer) myid);
    g_signal_connect(result, "show-unmount-progress", G_CALLBACK(mountoperationcallback_showunmountprogress), (gpointer) myid);
    return result;
}

void sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::deleteMountOperation(sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData* opdata)
{
    QMutexLocker _l(&callbackdatamutex);
    callbackdata.take(opdata->index);
    g_object_unref(opdata->mountoperation);
    delete opdata;
}

QMap<size_t, sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperationsCallbackData*> sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdata;

size_t sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdataindex;

QMutex sh::filesystemhandlers::GnomeIOFilesystemHandler::MountOperations::callbackdatamutex;

void sh::filesystemhandlers::ActionMountGnomeIOLocation::_initialize()
{

    sh::base::FilesystemModel::instance()->addOpenNodeHelper(9000, [=](std::shared_ptr<const sh::base::Eurl> eurl){ // helps opening a new node by mounting a known gnome io location
        QList<std::shared_ptr<sh::base::FilesystemNode>> result;
        if (!eurl->outerUrlIsRootDirectory())
        {
            QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> nodeToMountGvolumeuuid, nodeToMountLocations;
            {
                QMutexLocker _l1(&_nodeToMountMutex);
                nodeToMountGvolumeuuid = _nodeToMountGvolumeuuid;
                nodeToMountLocations = _nodeToMountLocations;
            }
            std::shared_ptr<const sh::base::Eurl> parenteurl = eurl->parentSegment();
            foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodeToMountLocations.keys())
            {
                if (node->eurl() == parenteurl)
                {
                    ActionMountGnomeIOLocation::mount(0, QString(), nodeToMountLocations[node], node);
                    result = QList<std::shared_ptr<sh::base::FilesystemNode>>({node});
                    break;
                }
            }
            if (result.isEmpty())
            {
                foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodeToMountGvolumeuuid.keys())
                {
                    if (node->eurl() == parenteurl)
                    {
                        ActionMountGnomeIOLocation::mount(0, nodeToMountGvolumeuuid[node], QString(), node);
                        break;
                    }
                }
            }
        }
        return result;
    });

    sh::base::FilesystemModel::instance()->addOpenNodeHelper(11000, [=](std::shared_ptr<const sh::base::Eurl> eurl){ // helps opening a new node by trying to mount the url via gio
        QList<std::shared_ptr<sh::base::FilesystemNode>> result;
        QString rooturl;
        if (eurl->outerUrlIsRootDirectory() && !eurl->hasInnerUrls())
        {
            sh::exceptions::Exception::executeGuarded([&](){
                rooturl = ActionMountGnomeIOLocation::mount(0, QString(), eurl->asString(), 0);
            }, sh::exceptions::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
            if (!rooturl.isEmpty())
            {
                sh::tools::MainThread::dispatcher()->invokeSync([&](){
                    std::shared_ptr<sh::base::FilesystemNode> nnode;
                    sh::filesystemhandlers::GnomeIONetworkFilesystemHandler::instance()->requestNewNode(eurl, &nnode);
                    if (!nnode)
                        sh::filesystemhandlers::GnomeIODevicesFilesystemHandler::instance()->requestNewNode(eurl, &nnode);
                    if (nnode)
                    {
                        result.append(nnode);
                        sh::filesystemhandlers::ActionUnmountGnomeIOLocation::offerUnmountLocation(nnode, eurl->asString(), true);
                    }
                });
            }
        }
        return result;
    });

    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::filesystemhandlers::ActionMountGnomeIOLocation>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<sh::actions::Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-800000)
    })));
}

void sh::filesystemhandlers::ActionMountGnomeIOLocation::_shutdown()
{
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::_initialize()
{
    sh::actions::ActionsManager::instance()->registerActionFactory(
                std::make_shared<sh::actions::ActionFactory<sh::filesystemhandlers::ActionUnmountGnomeIOLocation>>(
                    sh::actions::ActionCategory::getByName("manage"), QList<std::shared_ptr<sh::actions::Predicate>>({
                    std::make_shared<sh::actions::OnDirectoriesPredicate>(),
                    std::make_shared<sh::actions::OnSingleEntrySelectionPredicate>(),
                    std::make_shared<sh::actions::PositionIndexPredicate>(-800000)
    })));
}

void sh::filesystemhandlers::ActionUnmountGnomeIOLocation::_shutdown()
{
}

#endif
