/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GNOMEIOFILESYSTEMHANDLER_H
#define GNOMEIOFILESYSTEMHANDLER_H

#include "actions/actionactionitem.h"
#include "base/filesystemhandler.h"


#ifdef GIO

namespace sh{
    namespace filesystemhandlers {

        /** \brief
         * A filesystem handler for GIO filesystem parts.
         *
         * This is the abstract base class.
         */
        class GnomeIOFilesystemHandler : public sh::base::FilesystemHandler
        {
        Q_OBJECT
        protected:

            class GnomeIOReadDataDevice : public QIODevice
            {
            public:
                GnomeIOReadDataDevice(QString guri);
                ~GnomeIOReadDataDevice();
            protected:
                virtual qint64 readData(char* data, qint64 maxSize);
                virtual qint64 writeData(const char* data, qint64 maxSize);
                virtual bool isSequential() const { return true; }
                void* gstream;
                void* gfile;
            };

        public:
            explicit GnomeIOFilesystemHandler(sh::base::FilesystemModel *model);

            virtual void itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list);
            virtual sh::base::FilesystemNodeType getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual qint64 getSize(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual QDateTime getMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void setMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QDateTime time);

            virtual QString getMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual QString getLinkTarget(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canGetFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual std::shared_ptr<QIODevice> getFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canCreateDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canCreateLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString target);

            virtual bool canCreateFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QIODevice* content, HandlerTransfer* handlertransfer);

            virtual bool canDeleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void deleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canRenameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src);
            virtual void renameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src, QString destpath);

            virtual QList<std::shared_ptr<sh::actions::AbstractActionItem> > getActions(const QList<std::shared_ptr<const sh::base::Eurl>> eurls);

            virtual bool requiresResolvedLinks() { return false; }

            virtual bool isWellKnownScheme() { return true; }

            static QString mountDevice(QString gvolumeuuid, sh::actions::ActionExecutionInfo *info);
            static QString mountLocation(QString gurl, sh::actions::ActionExecutionInfo *info);
            static QString unmountLocation(QString gurl, sh::actions::ActionExecutionInfo *info);

            static QByteArray eurl2gurlb(std::shared_ptr<const sh::base::Eurl> eurl);
            static QString eurl2gurl(std::shared_ptr<const sh::base::Eurl> eurl);
            static std::shared_ptr<const sh::base::Eurl> gurl2eurl(QString gurl);

            static void logIfError(void *error);
            static void throwIfError(void *error);

            struct MountOperationsCallbackData
            {
                MountOperationsCallbackData(sh::actions::ActionExecutionInfo* info, size_t index, void* mountoperation, QString address);
                sh::actions::ActionExecutionInfo* info;
                size_t index;
                void* mountoperation;
                QString address;
            };

            class MountOperations
            {
            public:
                static void* createMountOperation(sh::actions::ActionExecutionInfo *info, QString address, MountOperationsCallbackData **opdata);
                static void deleteMountOperation(MountOperationsCallbackData* opdata);
                static QMap<size_t, MountOperationsCallbackData*> callbackdata;
                static size_t callbackdataindex;
                static QMutex callbackdatamutex;
            private:
            };

        protected:
            QString getFileInfoString(const char *key, QString gurl);
            bool getFileInfoBool(const char* key, QString gurl);
            qint64 getFileInfoInt(const char *key, QString gurl);
            quint64 getFileInfoUInt(const char *key, QString gurl);
            qint32 getFileInfoInt32(const char *key, QString gurl);
            quint32 getFileInfoUInt32(const char *key, QString gurl);
            QString getFileInfoString(const char *key, void* fileinfo);
            bool getFileInfoBool(const char* key, void* fileinfo);
            qint64 getFileInfoInt(const char *key, void* fileinfo);
            quint64 getFileInfoUInt(const char *key, void* fileinfo);
            qint32 getFileInfoInt32(const char *key, void* fileinfo);
            quint32 getFileInfoUInt32(const char *key, void* fileinfo);

            sh::base::FilesystemNodeType gfiletypetonodetype(quint64 gfiletype);

        };

        /** \brief Action for mounting a GIO location.
         */
        class ActionMountGnomeIOLocation : public sh::actions::ActionActionItem
        {
            Q_OBJECT
        public:
            /** \brief Constructed only internally.
             */
            explicit ActionMountGnomeIOLocation(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes);
            static void offerMountGVolumeUuid(std::shared_ptr<sh::base::FilesystemNode> node, QString gvolumeuuid);
            static void offerMountLocation(std::shared_ptr<sh::base::FilesystemNode> node, QString location);
            static void unofferMount(std::shared_ptr<sh::base::FilesystemNode> node);

            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            static void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            static void _shutdown();

            static QString mount(sh::actions::ActionExecutionInfo* info, QString gvolumeuuid, QString location, std::shared_ptr<base::FilesystemNode> node);

        protected:
            virtual void action(sh::actions::ActionExecutionInfo* info);
            virtual void initialize();
            QList<std::shared_ptr<sh::base::FilesystemNode>> _nodes;
            static QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> _nodeToMountGvolumeuuid;
            static QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> _nodeToMountLocations;
            static QMutex _nodeToMountMutex;
        };

        /** \brief Action for unmounting a GIO location.
         */
        class ActionUnmountGnomeIOLocation : public sh::actions::ActionActionItem
        {
            Q_OBJECT
        public:
            /** \brief Constructed only internally.
             */
            explicit ActionUnmountGnomeIOLocation(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes);
            static void offerUnmountLocation(std::shared_ptr<sh::base::FilesystemNode> node, QString location, bool staticmode=false);
            static void unofferUnmount(std::shared_ptr<sh::base::FilesystemNode> node);
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            static void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            static void _shutdown();
        protected:
            virtual void action(sh::actions::ActionExecutionInfo* info);
            virtual void initialize();
            QList<std::shared_ptr<sh::base::FilesystemNode>> _nodes;
            static QSet<std::shared_ptr<sh::base::FilesystemNode>> _staticModeNodes;
            static QHash<std::shared_ptr<sh::base::FilesystemNode>, QString> _nodeToMountLocations;
            static QMutex _nodeToMountMutex;
        };

    }
}

#endif

#endif // GNOMEIOFILESYSTEMHANDLER_H
