/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef GIO

#include <gio/gio.h>
#include "base/filesystemmodel.h"
#include "base/operation.h"
#include "base/filesystemoperation.h"
#include "base/filesystemhandlerregister.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "gnomeiosmbfilesystemhandler.h"


namespace sh{namespace filesystemhandlers{INITIALIZE_SINGLETON_IN_MAINTHREAD(40023, GnomeIOSmbFilesystemHandler)}}

sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::GnomeIOSmbFilesystemHandler(sh::base::FilesystemModel *model) :
    sh::filesystemhandlers::GnomeIOFilesystemHandler(model)
{
}

std::shared_ptr<sh::filesystemhandlers::GnomeIOSmbFilesystemHandler> sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::instance()
{
    static std::shared_ptr<sh::filesystemhandlers::GnomeIOSmbFilesystemHandler> _instance = std::make_shared<sh::filesystemhandlers::GnomeIOSmbFilesystemHandler>(sh::base::FilesystemModel::instance());
    return _instance;
}

sh::base::FilesystemNodeType sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl>eurl)
{
    sh::base::FilesystemNodeType result = GnomeIOFilesystemHandler::getType(op, eurl);
    if ((result == sh::base::FilesystemNodeType::NONE) && (!eurl->outerUrlIsRootDirectory()) && eurl->parentSegment()->outerUrlIsRootDirectory())
    {// for this case, gio behaves somewhat ill and we have to get the type the other way around
        QString basename = eurl->basename();
        QByteArray cguri = eurl2gurlb(eurl->parentSegment());
        GFile* d = g_file_new_for_uri(cguri.data());
        GError* error = 0;
        GFileEnumerator* direnum = g_file_enumerate_children(d, "standard::name,standard::display-name,standard::type",
                                                             G_FILE_QUERY_INFO_NONE, NULL, &error);
        logIfError(&error);
        if (direnum)
        {
            bool runloop = true;
            while (runloop)
            {
                GFileInfo *info = g_file_enumerator_next_file(direnum, NULL, &error);
                logIfError(&error);
                if (!info)
                    break;
                QString name = g_file_info_get_name(info);
                if (name == basename)
                {
                    quint32 giotype = getFileInfoUInt32(G_FILE_ATTRIBUTE_STANDARD_TYPE, info);
                    result = gfiletypetonodetype(giotype);
                    runloop = false;
                }
                g_object_unref(info);
            }
            g_object_unref(direnum);
        }
        g_object_unref(d);
    }
    return result;
}

void sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::_initialize()
{
    sh::tools::MainThread::dispatcher()->invokeSync([](){
        sh::base::FilesystemHandlerRegister::instance()->addHandler("smb", sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::instance());
    });

    sh::base::FilesystemModel::instance()->addOpenNodeHelper(5000, [=](std::shared_ptr<const sh::base::Eurl> eurl){ // helps opening a new node by requesting smb://hostname/ subitems when in smb://hostname/dir (without returning anything)
        if (eurl->scheme()=="smb" && !eurl->outerUrlIsRootDirectory() && eurl->parentSegment()->outerUrlIsRootDirectory())
        {
            foreach(std::shared_ptr<sh::base::FilesystemNode> node, sh::base::FilesystemModel::instance()->findNodesForEurl(eurl->parentSegment()))
                node->requestContainedItems();
        }
        return QList<std::shared_ptr<sh::base::FilesystemNode>>();
    });

}

void sh::filesystemhandlers::GnomeIOSmbFilesystemHandler::_shutdown()
{
}

#endif
