/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QObject>
#ifdef Q_OS_UNIX
#include <sys/stat.h>
#include <stdio.h>
#include <time.h>
#include <utime.h>
#include <sys/time.h>
#endif
#ifdef Q_OS_WIN
#define _WIN32_WINNT 0x0601
#include <SdkDdkVer.h>
#include <windows.h>
#include <FileAPI.h>
#endif
#include <QDir>
#include "actions/common/actionopenterminal.h"
#include "base/operation.h"
#include "base/filesystemmodel.h"
#include "base/filesystemnode.h"
#include "base/filesystemhandlerregister.h"
#include "exceptions/ioexception.h"
#include "tools/datastream.h"
#include "tools/filetypes/filetypemanager.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "tools/threadpool.h"
#include "localfilesystemhandler.h"

namespace sh{namespace filesystemhandlers{INITIALIZE_SINGLETON_IN_MAINTHREAD(40000, LocalFilesystemHandler)}}

#ifdef Q_OS_UNIX
#include <unistd.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/xattr.h>
#endif


void sh::filesystemhandlers::LocalFilesystemHandler::createRootNode()
{
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        rootnode = _model->createFilesystemNode(sh::base::Eurl::create("file", "", "/"), this, sh::base::FilesystemNodeType::Directory, false, sh::base::FilesystemModel::instance()->rootNode());
        rootnode->setDisplayName(tr("Local Filesystem"));
    });
}

sh::filesystemhandlers::LocalFilesystemHandler::LocalFilesystemHandler(sh::base::FilesystemModel *model) :
    sh::base::FilesystemHandler(model)
{
}

std::shared_ptr<sh::filesystemhandlers::LocalFilesystemHandler> sh::filesystemhandlers::LocalFilesystemHandler::instance()
{
    static std::shared_ptr<sh::filesystemhandlers::LocalFilesystemHandler> _inst = std::make_shared<sh::filesystemhandlers::LocalFilesystemHandler>(sh::base::FilesystemModel::instance());
    return _inst;
}

void sh::filesystemhandlers::LocalFilesystemHandler::_initialize()
{
    sh::base::FilesystemHandlerRegister::instance()->addHandler("file", sh::filesystemhandlers::LocalFilesystemHandler::instance());
    sh::filesystemhandlers::LocalFilesystemHandler::instance()->createRootNode();
}

void sh::filesystemhandlers::LocalFilesystemHandler::_shutdown()
{
}

void sh::filesystemhandlers::LocalFilesystemHandler::itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list)
{
    QString dir = eurlToLocal(eurl);
    QDir d(dir);
    QStringList ql;
    QDir::Filters filter;
    if (type == sh::base::FilesystemNodeType::Directory)
        filter = QDir::NoDotAndDotDot|QDir::AllDirs|QDir::Hidden|QDir::NoSymLinks;
    else if (type == sh::base::FilesystemNodeType::File)
        filter = QDir::Files|QDir::Hidden|QDir::NoSymLinks;
    else if (type == sh::base::FilesystemNodeType::Link)
        filter = QDir::AllEntries|QDir::Hidden|QDir::System;
    bool fetchentries = true;
#ifdef Q_OS_WIN
    if (eurl->outerUrlIsRootDirectory() && eurl->scheme() == "file")
    {
        fetchentries = false;
        if (type == sh::base::FilesystemNodeType::Directory) {
            foreach(QFileInfo fi, QDir::drives())
                ql << fi.absolutePath().mid(0,2);
        }
    }
#endif
    if (fetchentries)
    {
        foreach(QString s, d.entryList(filter)) {
            QFileInfo finf(dir+QDir::separator()+s);
            if (type != sh::base::FilesystemNodeType::Link || finf.isSymLink())
                ql << s;
        }
    }
    foreach(QString s, ql) // used in listExtendedAttributes
        op->setCustomData(eurl->withAppendedSegment(s), "LocalFilesystemHandler.itemtype", (int)type);
    list->setItems(ql);
}

void sh::filesystemhandlers::LocalFilesystemHandler::configureItems(sh::base::Operation*, QList<std::shared_ptr<sh::base::FilesystemNode> > nodes)
{
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodes) {
        if (node->eurl()->basename().startsWith("."))
            node->setHidden(true);
    }
}

qint64 sh::filesystemhandlers::LocalFilesystemHandler::getSize(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QFile file(eurlToLocal(eurl));
    return file.size();
}

QString sh::filesystemhandlers::LocalFilesystemHandler::getLinkTarget(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
#ifdef Q_OS_UNIX
    char buf[4096];
    ssize_t len;
    QByteArray linkname = eurlToLocal(eurl).toUtf8();
    if ((len = readlink(linkname.data(), buf, sizeof(buf)-1)) != -1)
    {
        buf[len] = '\0';
        QByteArray tgt(buf);
        return QString::fromUtf8(tgt);
    }
    else
        return QString();
#else
    QFileInfo finf(eurlToLocal(eurl));
    return finf.symLinkTarget();
#endif
}

QString sh::filesystemhandlers::LocalFilesystemHandler::getMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    return sh::tools::filetypes::FileTypeManager::instance()->determineMimetype(op, eurl);
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canCreateDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canCreateLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canCreateFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

sh::base::FilesystemNodeType sh::filesystemhandlers::LocalFilesystemHandler::getType(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    QString slocal = eurlToLocal(eurl);
    QFileInfo local(slocal);
    if (local.isSymLink())
        return sh::base::FilesystemNodeType::Link;
    else if (local.isDir())
        return sh::base::FilesystemNodeType::Directory;
    else if (local.isFile())
        return sh::base::FilesystemNodeType::File;
    else if (local.exists())
        return sh::base::FilesystemNodeType::Unknown;
    return sh::base::FilesystemNodeType::NONE;
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canGetFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

std::shared_ptr<QIODevice> sh::filesystemhandlers::LocalFilesystemHandler::getFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    QString filename = eurlToLocal( eurl );
    return std::make_shared<sh::tools::LocalFile>(filename);
}

void sh::filesystemhandlers::LocalFilesystemHandler::createDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl)
{
    QString parentdirname = eurlToLocal( eurl->parentSegment() );
    QDir parentdir(parentdirname);
    QString basename = eurl->basename();
    parentdir.mkdir(basename);
}

void sh::filesystemhandlers::LocalFilesystemHandler::createLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString target)
{
    QString filename = eurlToLocal( eurl );
    QFile::link(target, filename);
}

void sh::filesystemhandlers::LocalFilesystemHandler::createFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>eurl, QIODevice *content, HandlerTransfer *handlertransfer)
{
    QString filename = eurlToLocal( eurl );
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly))
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while opening: ") + eurl->asString()));
    const int BUFFER_SIZE = 4096;
    char buffer[BUFFER_SIZE];
    while (qint64 l = content->read(buffer, BUFFER_SIZE))
    {
        if (l == -1)
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to read from a stream while writing ")+filename));
        qint64 origl = l;
        char* mbuffer = buffer;
        while (l) {
            qint64 w = file.write(mbuffer, l);
            l-=w;
            mbuffer+=w;
        }
        handlertransfer->respectCancel();
        handlertransfer->incrementTransferredBytes(origl);
    }
    if (file.error() != QFile::NoError)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to write to file '%1': %2").arg(eurl->asString()).arg(file.errorString())));
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canRenameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::LocalFilesystemHandler::renameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> src, QString destpath)
{
    QFile f(eurlToLocal(src));
#ifdef Q_OS_WIN
#ifdef QT_DEBUG
    if (destpath.mid(2,2) != ":/")
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details("Bad rename destination path: "+destpath));
#endif
    destpath = destpath.mid(1);
#endif
    f.rename(destpath);
}

QString sh::filesystemhandlers::LocalFilesystemHandler::eurlToLocal(std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (eurl->hasInnerUrls() || eurl->scheme()!="file")
        return QString();
    QString result = eurl->path();
#ifdef Q_OS_WIN
    result = result.mid(1);
    if (result.length()==2 && result.endsWith(":"))
        result += "/";
    result = QDir::toNativeSeparators(result);  // a few things in windows seem to get trouble with "/" in paths (e.g. SHOpenWithDialog)
#endif
    return result;
}

std::shared_ptr<const sh::base::Eurl> sh::filesystemhandlers::LocalFilesystemHandler::localToEurl(const QString local)
{
    return sh::base::Eurl::create("file", "", QDir(local).absolutePath());
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::filesystemhandlers::LocalFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::base::Eurl>> eurls)
{
    QList< std::shared_ptr<sh::actions::AbstractActionItem> > result;
    if (eurls.length() == 1)
    {
        QString locl = eurlToLocal(eurls[0]);
        QFileInfo loclfile(locl);
        if (loclfile.isDir()) {
            result.append(std::make_shared<sh::actions::common::ActionOpenTerminal>(locl));
        }
    }
    return result;
}

bool sh::filesystemhandlers::LocalFilesystemHandler::canDeleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return true;
}

void sh::filesystemhandlers::LocalFilesystemHandler::deleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString locl = eurlToLocal(eurl);
    QFileInfo f(locl);
    if (f.isFile() || f.isSymLink())
    {
        QFile file(locl);
        file.remove();
    }
    else if (f.isDir())
    {
        QDir dir(locl);
        dir.removeRecursively();
    }
    else
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Unable to remove %1.").arg(locl)));
}

QDateTime sh::filesystemhandlers::LocalFilesystemHandler::getMtime(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QString slocal = eurlToLocal(eurl);
    QFileInfo local(slocal);
    return local.lastModified();
}

void sh::filesystemhandlers::LocalFilesystemHandler::setMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QDateTime mtime)
{
    QVariant att = op->getCustomData(eurl, "LocalFilesystemHandler.itemtype");
    if (att.isNull())
    {
        if (getType(op, eurl) == sh::base::FilesystemNodeType::Link)
            return;
    }
    else
    {
        if (att.toInt() == (int)sh::base::FilesystemNodeType::Link)
            return;
    }
#ifdef Q_OS_WIN
    QString slocal = eurlToLocal(eurl);
    wchar_t* filename = new wchar_t[slocal.length()+1];
    *(slocal.toWCharArray(filename) + filename) = 0;
    HANDLE hfile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
    qint64 tstampms = mtime.toMSecsSinceEpoch();
    qint64 tstamps = tstampms / 1000;
    qint64 tstampresms = tstampms % 1000;
    qint64 tswin = (tstamps + 11644473600) * 1000 * 1000 * 10 + tstampresms * 1000 * 10;
    FILETIME fmtime;
    fmtime.dwLowDateTime = (DWORD) tswin;
    fmtime.dwHighDateTime = (DWORD) (tswin >> 32);
    if (!SetFileTime(hfile, NULL, NULL, &fmtime))
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to set mtime for (3,%1): %2").arg(GetLastError()).arg(eurl->asString())));
    CloseHandle(hfile);
    delete[] filename;
#else
    struct stat foo;
    QByteArray rfilename = eurlToLocal(eurl).toUtf8();
    if (stat(rfilename.data(), &foo) < 0) {
        perror(rfilename.data());
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to set mtime for (1,%1): %2").arg(errno).arg(eurl->asString())));
    }
    struct timeval snewtimes[2];
    snewtimes[0].tv_sec = foo.st_atime;
    snewtimes[0].tv_usec = 0;
    snewtimes[1].tv_sec = mtime.toMSecsSinceEpoch()/1000;
    snewtimes[1].tv_usec = (mtime.toMSecsSinceEpoch() % 1000) * 1000;
    if (utimes(rfilename.data(), snewtimes) < 0) {
        perror(rfilename.data());
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to set mtime for (2,%1): %2").arg(errno).arg(eurl->asString())));
    }
#endif
}

QList<QString> sh::filesystemhandlers::LocalFilesystemHandler::listExtendedAttributes(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QList<QString> result;
#ifdef Q_OS_UNIX
    QByteArray path = eurlToLocal(eurl).toUtf8();
    char* listbuf = 0;
    ssize_t sizeneeded = -1;
    ssize_t sizeneededafter = 0;
    while (sizeneeded < sizeneededafter)
    {
        if (listbuf)
            delete[] listbuf;
        sizeneeded = llistxattr(path.data(), 0, 0);
        listbuf = new char[sizeneeded];
        sizeneededafter = llistxattr(path.data(), listbuf, sizeneeded);
    }
    ssize_t namebegin = 0;
    for(ssize_t i=0; i<sizeneededafter; i++)
    {
        if (listbuf[i] == 0)
        {
            result.append(QString::fromUtf8(listbuf+namebegin, (i-namebegin)));
            namebegin = i+1;
        }
    }
    delete[] listbuf;
#endif
    return result;
}

quint64 sh::filesystemhandlers::LocalFilesystemHandler::getExtendedAttributeSize(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
#ifdef Q_OS_UNIX
    QByteArray path = eurlToLocal(eurl).toUtf8();
    QByteArray attname = attribute.toUtf8();
    return lgetxattr(path.data(), attname.data(), 0, 0);
#else
    return 0;
#endif
}

QByteArray sh::filesystemhandlers::LocalFilesystemHandler::getExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
    QByteArray path = eurlToLocal(eurl).toUtf8();
#ifdef Q_OS_UNIX
    QByteArray attname = attribute.toUtf8();
    char* listbuf = 0;
    ssize_t sizeneeded = -1;
    ssize_t sizeneededafter = 0;
    while (sizeneeded < sizeneededafter)
    {
        if (listbuf) {
            delete[] listbuf;
            listbuf = 0;
        }
        sizeneeded = lgetxattr(path.data(), attname.data(), 0, 0);
        if (sizeneeded>0)
        {
            listbuf = new char[sizeneeded];
            sizeneededafter = lgetxattr(path.data(), attname.data(), listbuf, sizeneeded);
        }
        else
            break;
    }
    QByteArray result;
    if (listbuf) {
        result = QByteArray(listbuf, sizeneededafter);
        delete[] listbuf;
    }
    else if (sizeneeded != -1)
    { // the value exists, but is empty
        result = QByteArray(0);
    }
    return result;
#else
    return QByteArray();
#endif
}

void sh::filesystemhandlers::LocalFilesystemHandler::setExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QByteArray value)
{
    QByteArray path = eurlToLocal(eurl).toUtf8();
#ifdef Q_OS_UNIX
    QByteArray attname = attribute.toUtf8();
    if (lsetxattr(path.data(), attname.data(), value.data(), value.length(), 0) != 0)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while setting extended attribute '%1' on '%2'.").arg(attribute).arg(eurl->asString())));
#endif
}

void sh::filesystemhandlers::LocalFilesystemHandler::removeExtendedAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute)
{
#ifdef Q_OS_UNIX
    QByteArray path = eurlToLocal(eurl).toUtf8();
    QByteArray attname = attribute.toUtf8();
    if (lremovexattr(path.data(), attname.data()) != 0)
        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("Error while removing extended attribute '%1' on '%2'.").arg(attribute).arg(eurl->asString())));
#endif
}

QMap<QString, QString> sh::filesystemhandlers::LocalFilesystemHandler::getCustomAttributes(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QMap<QString, QString> result;
    QByteArray path = eurlToLocal(eurl).toUtf8();
#ifdef Q_OS_UNIX
    if (getType(op, eurl) != sh::base::FilesystemNodeType::Link)
    {
        struct stat foo;
        if (stat(path.data(), &foo) < 0) {
            perror(path.data());
            throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to get permissions for (4,%1): %2").arg(errno).arg(eurl->asString())));
        }
        result["unix_permissions"] = QString("%1;%2;%3").arg(foo.st_mode).arg(foo.st_uid).arg(foo.st_gid);
    }
#endif
    return result;
}

void sh::filesystemhandlers::LocalFilesystemHandler::setCustomAttribute(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl, QString attribute, QString value)
{
    QByteArray path = eurlToLocal(eurl).toUtf8();
#ifdef Q_OS_UNIX
    if (attribute == "unix_permissions")
    {
        bool completely = false;
        if (value.endsWith("!"))
        {
            value = value.left(value.length()-1);
            completely = true;
        }
        QStringList vals = value.split(";");
        mode_t mode = (mode_t)vals.at(0).toInt();
        if (completely)
        {
            chown(path.data(), (uid_t)vals.at(1).toInt(), (gid_t)vals.at(2).toInt());
            chmod(path.data(), mode);
        }
        else
        {
            if (mode & (S_IXUSR|S_IXGRP|S_IXOTH))
            {
                struct stat foo;
                if (stat(path.data(), &foo) < 0) {
                    perror(path.data());
                    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("unable to set permissions for (5,%1): %2").arg(errno).arg(eurl->asString())));
                }
                mode_t nmode = foo.st_mode;
                if (nmode & S_IRUSR)
                    nmode |= S_IXUSR;
                if (nmode & S_IRGRP)
                    nmode |= S_IXGRP;
                if (nmode & S_IROTH)
                    nmode |= S_IXOTH;
                chmod(path.data(), nmode);
            }
        }
    }
#endif
}
