/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SHARCFILESYSTEMHANDLER_H
#define SHARCFILESYSTEMHANDLER_H

#include <QBuffer>
#include "base/detailcolumn.h"
#include "base/filesystemhandler.h"


namespace sh{
    namespace filesystemhandlers {

        /** \brief
         * A filesystem handler for the sharc archives.
         *
         * This is a brutally easy and inefficient archive format which exists
         * mostly for testing.
         */
        class SharcFilesystemHandler : public sh::base::FilesystemHandler
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static std::shared_ptr<SharcFilesystemHandler> instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            explicit SharcFilesystemHandler(sh::base::FilesystemModel *model);

            virtual void itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list);
            virtual void configureItems(sh::base::Operation* op, QList<std::shared_ptr<sh::base::FilesystemNode>> nodes);
            virtual sh::base::FilesystemNodeType getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual qint64 getSize(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual QDateTime getMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void setMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QDateTime time);

            virtual QString getMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual QString getLinkTarget(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canGetFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual std::shared_ptr<QIODevice> getFileContent(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canCreateDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createDirectory(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canCreateLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createLink(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QString target);

            virtual bool canCreateFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void createFile(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, QIODevice* content, HandlerTransfer* handlertransfer);

            virtual bool canDeleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);
            virtual void deleteItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl);

            virtual bool canRenameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src);
            virtual void renameItem(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> src, QString destpath);

            virtual QList<std::shared_ptr<sh::actions::AbstractActionItem> > getActions(const QList<std::shared_ptr<const sh::base::Eurl>> eurls);

            static std::shared_ptr<const base::Eurl> localToEurl(const QString local);

        protected:
            static void _parse(sh::base::Operation *op, std::shared_ptr<const sh::base::Eurl>eurl, bool doesOutput, QStringList append, std::function<QString(QString,QString)> _decide, bool *peekData = 0, std::function<void(QString)> _data = 0);

            static QString _eurlToLocal(std::shared_ptr<const sh::base::Eurl> eurl);
            std::shared_ptr<sh::base::DetailColumn> _detailColumnNameArchivedSize;

            class ArchivedSizeDetailColumn : public sh::base::DetailColumn
            {
            public:
                ArchivedSizeDetailColumn();
            protected:
                virtual QVariant determineValue(std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation* op);
            };

            class QBuffer2 : public QBuffer
            {
            public:
                QBuffer2(QByteArray* a) : QBuffer(a), _a(a) {}
                virtual ~QBuffer2() { delete _a; }
            private:
                QByteArray* _a;
            };

        };

    }
}

#endif // SHARCFILESYSTEMHANDLER_H
