/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFileInfo>
#include <QWidget>
#ifdef Q_OS_WIN
#include "base/filesystemmodel.h"
#include "base/filesystemhandlerregister.h"
#include "base/filesystemoperation.h"
#include "base/operation.h"
#include "tools/iconmanager.h"
#include "tools/singletoninitializer.h"
#include "tools/mainthread.h"
#include "tools/logger.h"
#include "windowsnetworkfilesystemhandler.h"
#define _WIN32_WINNT 0x0601
#define INITGUID
#include <SdkDdkVer.h>
#include <windows.h>
#include <objbase.h>
#include <Shobjidl.h>
#include <shlguid.h>
#include <shlobj.h>
#include <shlwapi.h>


namespace sh{namespace filesystemhandlers{INITIALIZE_SINGLETON_IN_MAINTHREAD(40024, WindowsNetworkFilesystemHandler)}}


sh::filesystemhandlers::WindowsNetworkFilesystemHandler::WindowsNetworkFilesystemHandler(sh::base::FilesystemModel *model)
        : sh::filesystemhandlers::LocalFilesystemHandler(model)
{
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::createRootNode()
{
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        rootnode = _model->createFilesystemNode(sh::base::Eurl::create("network", "", "/"), this, sh::base::FilesystemNodeType::Directory, false, sh::base::FilesystemModel::instance()->rootNode());
        rootnode->setDisplayName(tr("Network"));
        rootnode->setIcon(sh::tools::IconManager::instance()->getIcon("network-server"));
    });
}

QString sh::filesystemhandlers::WindowsNetworkFilesystemHandler::eurlToLocal(std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (eurl->hasInnerUrls() || eurl->scheme()!="smb")
        return QString();
    return QString("\\\\%1\%2").arg(eurl->hostname().toLower()).arg(eurl->path().replace("/", "\\"));
}

std::shared_ptr<const sh::base::Eurl> sh::filesystemhandlers::WindowsNetworkFilesystemHandler::localToEurl(const QString local)
{
    std::shared_ptr<const sh::base::Eurl> result;
    if (local.startsWith("\\\\"))
    {
        int ibslash = local.indexOf("\\", 2);
        result = sh::base::Eurl::fromString(QString("smb://%1/%2").arg(local.mid(2, ibslash-2).toLower()).arg(local.mid(ibslash+1).replace("\\", "/")));
    }
    return result;
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::itemlist(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl, sh::base::FilesystemNodeType type, sh::base::FilesystemNodeListEditor *list)
{
    if (eurl->outerUrlIsRootDirectory() && eurl->scheme()=="network")
    {
        list->beginIterativeAdding();
        QStringList names;
        if (type == sh::base::FilesystemNodeType::Directory)
        {
            CoInitializeEx(NULL, COINIT_MULTITHREADED);
            IShellItem* pItem = nullptr;
            HRESULT hr = SHCreateItemInKnownFolder(FOLDERID_NetworkFolder, 0, nullptr, IID_PPV_ARGS(&pItem));
            if (SUCCEEDED(hr))
            {
                IEnumShellItems* pEnum = nullptr;
                hr = pItem->BindToHandler(nullptr, BHID_EnumItems, IID_PPV_ARGS(&pEnum));
                if (SUCCEEDED(hr))
                {
                     IShellItem* pChildItem = nullptr;
                     ULONG ulFetched = 0;
                     do
                     {
                         hr = pEnum->Next(1, & pChildItem, &ulFetched);
                         if (FAILED(hr))
                             break;
                         if (ulFetched != 0)
                         {
                             LPWSTR szChildName = nullptr;
                             pChildItem->GetDisplayName(SIGDN_NORMALDISPLAY, &szChildName);
                             QString name = QString::fromWCharArray(szChildName);
                             names << name;
                             CoTaskMemFree(szChildName);
                             pChildItem ->Release();
                         }
                     } while (hr != S_FALSE);
                     pEnum->Release();
                }
                pItem->Release();
            }
        }
        foreach(QString name, names)
            list->addCustomItem(sh::base::Eurl::create("smb", name, "/"), name, this);
        list->endIterativeAdding();
    }
    else
        sh::filesystemhandlers::LocalFilesystemHandler::itemlist(op, eurl, type, list);
}

sh::base::FilesystemNodeType sh::filesystemhandlers::WindowsNetworkFilesystemHandler::getType(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (!eurl->outerUrlIsRootDirectory() && eurl->parentSegment()->outerUrlIsRootDirectory() && eurl->scheme()=="network")
    {
        QList<std::shared_ptr<const sh::base::Eurl>> l = op->filesystem()->itemlist(eurl->parentSegment());
        foreach(std::shared_ptr<const sh::base::Eurl> e, l)
        {
            if (e->basename().toLower() == eurl->basename().toLower())
                return sh::base::FilesystemNodeType::Directory;
        }
        return sh::base::FilesystemNodeType::NONE;
    }
    else
        return sh::filesystemhandlers::LocalFilesystemHandler::getType(op, eurl);
}

std::shared_ptr<sh::filesystemhandlers::WindowsNetworkFilesystemHandler> sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance()
{
    static std::shared_ptr<sh::filesystemhandlers::WindowsNetworkFilesystemHandler> _inst = std::make_shared<sh::filesystemhandlers::WindowsNetworkFilesystemHandler>(sh::base::FilesystemModel::instance());
    return _inst;
}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::_initialize()
{
    sh::base::FilesystemHandlerRegister::instance()->addHandler("network", sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance());
    sh::base::FilesystemHandlerRegister::instance()->addHandler("smb", sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance());
    sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance()->createRootNode();

    sh::base::FilesystemModel::instance()->addOpenNodeHelper(11000, [=](std::shared_ptr<const sh::base::Eurl> eurl){ // helps opening a new smb host node via windows api
        QList<std::shared_ptr<sh::base::FilesystemNode>> result;
        if (QFileInfo(eurlToLocal(eurl)).exists())
        {
            std::shared_ptr<sh::base::FilesystemNode> nnode = sh::base::FilesystemModel::instance()->createFilesystemNode(eurl,
                                             sh::filesystemhandlers::WindowsNetworkFilesystemHandler::instance().get(), sh::base::FilesystemNodeType::Directory,
                                             false, sh::filesystemhandlers::WindowsNetworkFilesystemHandler::rootnode, false);
            nnode->setDisplayName(eurl->hostname().toUpper());
            nnode->setIcon(sh::tools::IconManager::instance()->getIcon("network-server"));
            qobject_cast<sh::base::ModelBackedFilesystemNodeList*>(sh::filesystemhandlers::WindowsNetworkFilesystemHandler::rootnode->childnodes())->addPermanentItem(nnode);
            result << nnode;
        }
        return result;
    });

}

void sh::filesystemhandlers::WindowsNetworkFilesystemHandler::_shutdown()
{
}

#endif
