/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QLocale>
#include "base/filesystemnode.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "detailcolumns/detailcolumnfilesize.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "paneldetails/paneldetailmanager.h"
#include "tools/singletoninitializer.h"
#include "tools/operationscache.h"
#include "tools/threadpool.h"
#include "commonpaneldetails.h"


namespace sh{namespace paneldetails{INITIALIZE_STATIC_IN_MAINTHREAD(13300, CommonPanelDetails)}}

void sh::paneldetails::CommonPanelDetails::NumberOfSelectedItems(std::shared_ptr<sh::paneldetails::PanelDetail> detail, QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::base::Operation*)
{
    detail->setRows({ sh::paneldetails::PanelDetailRow(
                   QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(QObject::tr("%1 items selected.").arg(nodes.length()))})
                      ) });
}

void sh::paneldetails::CommonPanelDetails::TotalFileSize(std::shared_ptr<PanelDetail> detail, QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::base::Operation*)
{
    sh::tools::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            quint64 totalsize = 0;
            foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodes)
            {
                sh::base::FilesystemNodeType ntype = node->nodetype();
                if (ntype == sh::base::FilesystemNodeType::File || ntype == sh::base::FilesystemNodeType::Link)
                {
                    std::shared_ptr<sh::base::Operation> op = sh::tools::OperationsCache::instance()->getOperationForContainer(node->eurl()->outermostInnerEurl());
                    totalsize += sh::detailcolumns::DetailColumnFilesize::instance()->requestValue(node, op.get()).toLongLong();
                }
            }
            static QLocale l;
            detail->setRows({ sh::paneldetails::PanelDetailRow(QObject::tr("Total size"),
                              QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(l.toString(totalsize) + QObject::tr(" bytes"))})
                              )});
        });
    });
}

void sh::paneldetails::CommonPanelDetails::_ViaDetailColumn(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::base::FilesystemNode> node, std::shared_ptr<sh::base::DetailColumn> column, QString keyname, std::function<QString(QVariant)> valtostring)
{
    detail->setRows({ sh::paneldetails::PanelDetailRow(keyname,
                      QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement()})
                      ) });
    sh::tools::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            std::shared_ptr<sh::base::Operation> op = sh::tools::OperationsCache::instance()->getOperationForContainer(node->eurl()->outermostInnerEurl());
            detail->setRows({ sh::paneldetails::PanelDetailRow(keyname,
                              QList<sh::paneldetails::PanelDetailRowValueElement>({sh::paneldetails::PanelDetailRowValueElement(valtostring(column->requestValue(node, op.get())))})
                              ) });
        });
    });
}

void sh::paneldetails::CommonPanelDetails::FileType(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation*)
{
    if (node->nodetype() == sh::base::FilesystemNodeType::File || node->nodetype() == sh::base::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnMimetype::instance(), QObject::tr("Type"), [](QVariant v){return v.toString();});
}

void sh::paneldetails::CommonPanelDetails::FileSize(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation*)
{
    if (node->nodetype() == sh::base::FilesystemNodeType::File || node->nodetype() == sh::base::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnFilesize::instance(), QObject::tr("Size"), [](QVariant v){ static QLocale l; return l.toString(v.toLongLong()) + QObject::tr(" bytes");});
}

void sh::paneldetails::CommonPanelDetails::FileModificationTime(std::shared_ptr<sh::paneldetails::PanelDetail> detail, std::shared_ptr<sh::base::FilesystemNode> node, sh::base::Operation*)
{
    if (node->nodetype() == sh::base::FilesystemNodeType::File || node->nodetype() == sh::base::FilesystemNodeType::Link)
        _ViaDetailColumn(detail, node, sh::detailcolumns::DetailColumnMtime::instance(), QObject::tr("Last modified"), [](QVariant v){return v.toDateTime().toString(Qt::DefaultLocaleShortDate);});
}

void sh::paneldetails::CommonPanelDetails::_initialize()
{
    sh::paneldetails::PanelDetailManager* dm = sh::paneldetails::PanelDetailManager::instance();
    dm->registerFactoryMulti(NumberOfSelectedItems, [](QList<std::shared_ptr<sh::base::FilesystemNode>>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 100, 5);
    dm->registerFactoryMulti(TotalFileSize, [](QList<std::shared_ptr<sh::base::FilesystemNode>>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 120, 5);
    dm->registerFactorySingle(FileType, [](std::shared_ptr<sh::base::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 120, 4);
    dm->registerFactorySingle(FileSize, [](std::shared_ptr<sh::base::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 140, 5);
    dm->registerFactorySingle(FileModificationTime, [](std::shared_ptr<sh::base::FilesystemNode>,QString){}, sh::paneldetails::PanelDetailPositionIndex::Interesting + 160, 3);
}

void sh::paneldetails::CommonPanelDetails::_shutdown()
{

}
