/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PANELDETAILMANAGER_H
#define PANELDETAILMANAGER_H

#include <functional>
#include <memory>
#include <QIcon>
#include <QMutex>
#include <QObject>


namespace sh{
    namespace base{class FilesystemNode; class Operation;}
    namespace ui {class FileDetailsPanel;}

    namespace paneldetails {

        /**
         * @brief One element of a detail row's value.
         *
         * This can be something like a string, an icon or a link button.
         */
        class PanelDetailRowValueElement
        {
        public:
            /** \brief A placeholder for a value which is not yet arrived. It could be visualized with a loading animation or '...'. Is intended to be directly constructed from everywhere.
             */
            PanelDetailRowValueElement();
            /** \brief A piece of text. Is intended to be directly constructed from everywhere.
             */
            PanelDetailRowValueElement(QString text);
            /** \brief An icon. Is intended to be directly constructed from everywhere.
             */
            PanelDetailRowValueElement(QIcon icon);
            /** \brief A link button. Is intended to be directly constructed from everywhere.
             */
            PanelDetailRowValueElement(QString text, QString linktarget, bool autohide);
            QString text();
            QIcon icon();
            QString linktarget();
            bool linkautohide();
            bool iswaiting();
        private:
            QString _text;
            QIcon _icon;
            QString _linktarget;
            bool _linkautohide = false;
            bool _iswaiting = false;
        };

        /** \brief One pair of label text and value text in a panel detail.
         */
        class PanelDetailRow
        {
        public:
            /** \brief Is intended to be directly constructed from everywhere.
             */
            PanelDetailRow(QString key, QList<PanelDetailRowValueElement> value);
            /** \brief Is intended to be directly constructed from everywhere.
             */
            PanelDetailRow(QList<PanelDetailRowValueElement> value);
            QString key();
            QList<PanelDetailRowValueElement> value();
        private:
            QString _key;
            QList<PanelDetailRowValueElement> _value;
        };

        /** \brief
         * Base values for position indexes of panel details.
         */
        enum PanelDetailPositionIndex
        {
            VeryInteresting = 1000000,
            Interesting = 2000000,
            Exotic = 3000000
        };

        /** \brief
         * A detail panel entry.
         *
         * Is is essentially a list of PanelDetailRow and some layout information.
         */
        class PanelDetail : public QObject
        {
            friend class sh::ui::FileDetailsPanel;
            Q_OBJECT
        public:
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            PanelDetail(int positionIndex, int valueWidthHint);
            /** \brief Returns the list of detail rows stored in this panel detail.
             */
            QList<PanelDetailRow> rows();
            /** \brief Returns a position index.
             *
             * It controls the position of this panel detail in relation to the other ones.
             */
            int positionIndex();
            /** \brief Returns the specified width for the panel detail.
             */
            int valueWidthHint();
            /** \brief Sets the list of detail rows.
             */
            void setRows(QList<PanelDetailRow> rows);
            /** \brief Adds a new detail row.
             */
            void addRow(PanelDetailRow row);
            /** \brief Adds a new detail row (replacing an old one with the same key).
             */
            void addOrReplaceRow(PanelDetailRow row);
        signals:
            /** \brief Signals when data has changed.
             */
            void changed();
            void linkTriggered(QString link);
        private:
            void _emit_linkTriggered(QString link);
            QMutex _rowsmutex;
            QList<PanelDetailRow> _rows;
            int _positionIndex;
            int _valueWidthHint;
        };

        /** \brief Abstract factory for creating panel details for the selected nodes.
         */
        class PanelDetailFactory
        {
        public:
            virtual std::shared_ptr<PanelDetail> construct(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::base::Operation* op) =0;
        };

        /** \brief
         * Manager for details provider as used in the details panel.
         *
         * Register some code and data here for implementing a new provider.
         */
        class PanelDetailManager : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static PanelDetailManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            QList<std::shared_ptr<PanelDetail>> getDetails(QList<std::shared_ptr<sh::base::FilesystemNode> > nodes);
            std::shared_ptr<PanelDetailFactory> registerFactorySingle(std::shared_ptr<PanelDetailFactory> factory);
            std::shared_ptr<PanelDetailFactory> registerFactoryMulti(std::shared_ptr<PanelDetailFactory> factory);
            std::shared_ptr<PanelDetailFactory> registerFactorySingle(std::function<void(std::shared_ptr<PanelDetail>, std::shared_ptr<sh::base::FilesystemNode>, sh::base::Operation*)> fctsetvalue, std::function<void(std::shared_ptr<sh::base::FilesystemNode>, QString link)> fctlinktriggered, int position, int valueWidthHint);
            std::shared_ptr<PanelDetailFactory> registerFactoryMulti(std::function<void(std::shared_ptr<PanelDetail>, QList<std::shared_ptr<sh::base::FilesystemNode>>, sh::base::Operation*)> fctsetvalue, std::function<void(QList<std::shared_ptr<sh::base::FilesystemNode>>, QString link)> fctlinktriggered, int position, int valueWidthHint);

        signals:

        public slots:

        private:
            PanelDetailManager();
            QList<std::shared_ptr<PanelDetailFactory>> _factoriessingle;
            QList<std::shared_ptr<PanelDetailFactory>> _factoriesmulti;
            QMutex _mutex;
        };

        /** \brief A factory for creating panel details for just one selected node.
         */
        class PanelDetailFactoryForFunctionSingle : public PanelDetailFactory
        {
        public:
            /** \brief Constructed only indirectly.
             */
            PanelDetailFactoryForFunctionSingle(std::function<void(std::shared_ptr<PanelDetail>, std::shared_ptr<sh::base::FilesystemNode>, sh::base::Operation* op)> fctsetvalue, std::function<void(std::shared_ptr<sh::base::FilesystemNode>, QString link)> fctlinktriggered, int position, int valueWidthHint);
            std::shared_ptr<PanelDetail> construct(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::base::Operation* op);
        protected:
            std::function<void(std::shared_ptr<PanelDetail>, std::shared_ptr<sh::base::FilesystemNode>, sh::base::Operation* op)> _fctsetvalue;
            std::function<void(std::shared_ptr<sh::base::FilesystemNode>, QString link)> _fctlinktriggered;
            int _position;
            int _valueWidthHint;
        };

        /** \brief A factory for creating panel details for more than one selected node.
         */
        class PanelDetailFactoryForFunctionMulti : public PanelDetailFactory
        {
        public:
            /** \brief Constructed only indirectly.
             */
            PanelDetailFactoryForFunctionMulti(std::function<void(std::shared_ptr<PanelDetail>, QList<std::shared_ptr<sh::base::FilesystemNode>>, sh::base::Operation* op)> fctsetvalue, std::function<void(QList<std::shared_ptr<sh::base::FilesystemNode>>, QString link)> fctlinktriggered, int position, int valueWidthHint);
            std::shared_ptr<PanelDetail> construct(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::base::Operation* op);
        protected:
            std::function<void(std::shared_ptr<PanelDetail>, QList<std::shared_ptr<sh::base::FilesystemNode>>, sh::base::Operation* op)> _fctsetvalue;
            std::function<void(QList<std::shared_ptr<sh::base::FilesystemNode>>, QString link)> _fctlinktriggered;
            int _position;
            int _valueWidthHint;
        };

    }
}

#endif // PANELDETAILMANAGER_H
