/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionabstracttransfertree.h"
#include "base/filesystemnode.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "detailcolumns/detailcolumnextendedattributes.h"
#include "tools/singletoninitializer.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "tools/iconmanager.h"
#include "ui/filepropertydialog/filepropertydialogtabtableview.h"
#include "ui/filepropertydialog/filepropertydialogtabactionsview.h"
#include "ui/icontextbanner.h"
#include "apifilepropertydialogtab.h"


const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_STRING = 1;

const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_STRINGMAP = 2;

const int sh::scripting::api::ApiFilePropertyDialogTab::PROPERTY_TYPE_ICONTEXTBANNER = 3;

sh::scripting::api::ApiFilePropertyDialogTab::ApiFilePropertyDialogTab(QString title, QList<PropertyConfig> properties)
                : _title(title), _properties(properties)
{
}

sh::scripting::api::ApiFilePropertyDialogTab::PropertyConfig::PropertyConfig(QString title, int propertytype, QList<ButtonConfig> buttons)
                : title(title), propertytype(propertytype), buttons(buttons)
{
}

QString sh::scripting::api::ApiFilePropertyDialogTab::title()
{
    return _title;
}

QList<QString> sh::scripting::api::ApiFilePropertyDialogTab::properties()
{
    QList<QString> result;
    foreach(PropertyConfig p, _properties)
        result.append(p.title);
    return result;
}

QWidget* sh::scripting::api::ApiFilePropertyDialogTab::createWidget(int i)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    int ptype = _properties[i].propertytype;
    QWidget* content = 0;
    sh::ui::filepropertydialog::FilePropertyDialogTabActionsView* actionsview = new sh::ui::filepropertydialog::FilePropertyDialogTabActionsView();
    if (ptype == PROPERTY_TYPE_STRING)
    {
        QLabel* lbl = new QLabel(actionsview);
        lbl->setWordWrap(true);
        content = lbl;
    }
    else if (ptype == PROPERTY_TYPE_STRINGMAP)
    {
        content = new sh::ui::filepropertydialog::FilePropertyDialogTabTableView(actionsview);
    }
    else if (ptype == PROPERTY_TYPE_ICONTEXTBANNER)
    {
        content = new sh::ui::IconTextBanner(actionsview);
    }
    actionsview->setContent(content);
    QStringList buttons;
    foreach(PropertyConfig::ButtonConfig btncfg, _properties[i].buttons)
        buttons << btncfg.first;
    actionsview->setButtons(buttons);
    QObject::connect(actionsview, &sh::ui::filepropertydialog::FilePropertyDialogTabActionsView::buttonTriggered, [=](int j){
        sh::tools::ThreadPool::enqueueForce([=](){
            sh::exceptions::Exception::executeGuarded_errorpanel([&](){
                _buttonTriggered( _properties[i].buttons[j].second);
            });
        });
    });
    return actionsview;
}

void sh::scripting::api::ApiFilePropertyDialogTab::updateWidget(int i, QWidget* widget, sh::base::Operation* op)
{
    sh::ui::filepropertydialog::FilePropertyDialogTabActionsView* awidget = qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabActionsView*>(widget);
    QLabel* wlabel = qobject_cast<QLabel*>(awidget->content());
    sh::ui::filepropertydialog::FilePropertyDialogTabTableView* wgrid = qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabTableView*>(awidget->content());
    sh::ui::IconTextBanner* witbanner = qobject_cast<sh::ui::IconTextBanner*>(awidget->content());
    QList<QString> lvals;
    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
        lvals = _updateWidget(i, op, (qintptr)widget);
    });
    QString curkey;
    QList<QPair<QString, QString>> newval;
    double sizefact = awidget->logicalDpiY() / 4;
    foreach(QString s, lvals)
    {
        if (curkey.isNull())
            curkey = s;
        else
        {
            newval.append(QPair<QString,QString>(curkey, s));
            curkey = QString();
        }
    }
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        if (wlabel)
        {
            foreach(StringPair p, newval)
                if (p.first == "text")
                    wlabel->setText(p.second);
        }
        else if (wgrid)
        {
            QList<QPair<QString,QString>> content;
            foreach(StringPair p, newval)
                if (p.first.startsWith("val_"))
                    content.append(QPair<QString,QString>(p.first.mid(4), p.second));
            wgrid->setContent(content);
        }
        else if (witbanner)
        {
            witbanner->clear();
            foreach(StringPair p, newval)
            {
                if (p.first.startsWith("val_"))
                {
                    QString type = p.first.mid(4);
                    QString value = p.second;
                    if (type == "icon")
                    {
                        int iund = value.indexOf(";");
                        int size = value.mid(0, iund).toFloat() * sizefact;
                        value = value.mid(iund+1);
                        witbanner->insertIcon(sh::tools::IconManager::instance()->getIconByFullname(value), size);
                    }
                    else if (type == "text")
                        witbanner->insertText(value);
                }
            }
        }
        bool visible = false;
        foreach(StringPair p, newval)
            if (p.first == "visible")
            {
                visible = p.second == "1";
                break;
            }
        widget->setVisible(visible);
    });
}

sh::scripting::api::ApiFilePropertyDialogTabFactory::ApiFilePropertyDialogTabFactory()
            : sh::ui::filepropertydialog::FilePropertyDialogTabFactory()
{
}

std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> sh::scripting::api::ApiFilePropertyDialogTabFactory::construct()
{
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> result;
    sh::exceptions::Exception::executeGuarded_errorpanel([&](){
        result = _construct();
    });
    return result;
}
