/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath> // just a workaround for a compile issue with python in mingw! - not really needed
#ifdef PYTHONSCRIPT
// python needs to be included before the other stuff
#include <Python.h>
#endif

#include "base/filesystemnode.h"
#include "base/filesystemhandlerregister.h"
#include "configuration/configurationmanager.h"
#include "tools/singletoninitializer.h"
#include "tools/mainthread.h"
#include "tools/logger.h"
#include "api.h"
#include "pythonscriptinterpreter.h"

namespace sh{namespace scripting{INITIALIZE_SINGLETON_IN_MAINTHREAD(500, PythonScriptInterpreter)}}

#ifdef PYTHONSCRIPT

sh::scripting::PythonScriptInterpreter::~PythonScriptInterpreter()
{
}

bool sh::scripting::PythonScriptInterpreter::isAvailable()
{
    return true;
}

QString sh::scripting::PythonScriptInterpreter::filesuffix()
{
    return "py";
}

namespace sh{namespace scripting{
    /** \brief A wrapper around a native object for the Python world.
      */
    typedef struct {
        PyObject_HEAD
        void* nativeObject;
        std::shared_ptr<void> sharedNativeObject;
        bool isReallyShared;
    } shallot_ShallotObject;
}}

QMap<QString, PyTypeObject*> _pyTypes;

QMap<QString, PyCFunction> sh::scripting::PythonScriptInterpreter::_methods;

QMap<void*, PyObject*> sh::scripting::PythonScriptInterpreter::_buddies;

QMap<void*, std::weak_ptr<void>> sh::scripting::PythonScriptInterpreter::_natives;

PyObject* sh::scripting::pythoninterop::getPyNone()
{
    return Py_BuildValue("");
}

bool sh::scripting::pythoninterop::isPyNone(PyObject* o)
{
    return getPyNone()==o;
}

void* sh::scripting::pythoninterop::getPointerFromShallotPyObject(PyObject* o)
{
    PyObject* pystr_native = Converters::getPyObjectFromString("_native");
    PyObject* pyo;
    if (PyObject_HasAttr(o, pystr_native))
        pyo = PyObject_GetAttr(o, pystr_native);
    else
        pyo = o;
    PyDecRef(pystr_native);
    return ((sh::scripting::shallot_ShallotObject*)pyo)->nativeObject;
}

std::shared_ptr<void> sh::scripting::pythoninterop::getSharedPointerFromShallotPyObject(PyObject* o)
{
    PyObject* pystr_native = Converters::getPyObjectFromString("_native");
    PyObject* pyo;
    if (PyObject_HasAttr(o, pystr_native))
        pyo = PyObject_GetAttr(o, pystr_native);
    else
        pyo = o;
    PyDecRef(pystr_native);
    sh::scripting::shallot_ShallotObject* spyo = ((sh::scripting::shallot_ShallotObject*)pyo);
    if (!spyo->sharedNativeObject)
        spyo->sharedNativeObject = std::shared_ptr<void>(spyo->nativeObject, [](void*){});
    return spyo->sharedNativeObject;
}

PyObject* sh::scripting::pythoninterop::getShallotPyObjectFromPointer(void* o, sh::scripting::ApiClass *apiclass)
{
    if (o == 0)
        return getPyNone();
    PyTypeObject* pytype = _pyTypes[apiclass->name()];
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) PyObject_CallObject((PyObject *) pytype, NULL);
    pymember->nativeObject = o;
    pymember->isReallyShared = false;
    return (PyObject*) pymember;
}

PyObject* sh::scripting::pythoninterop::getShallotPyObjectFromSharedPointer(std::shared_ptr<void> o, sh::scripting::ApiClass *apiclass)
{
    if (o == 0)
        return getPyNone();
    PyTypeObject* pytype = _pyTypes[apiclass->name()];
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) PyObject_CallObject((PyObject *) pytype, NULL);
    pymember->nativeObject = o.get();
    pymember->sharedNativeObject = o;
    pymember->isReallyShared = true;
    return (PyObject*) pymember;
}

sh::scripting::ApiClass* sh::scripting::pythoninterop::getApiClass(QString typeIdName)
{
    static sh::scripting::Api* api = new sh::scripting::Api();
    return api->getClassByTypeIdName(typeIdName);
}

PyObject* sh::scripting::pythoninterop::PyDictNew()
{
    return PyDict_New();
}

PyObject* sh::scripting::pythoninterop::PyDictKeys(PyObject *o)
{
    return PyDict_Keys(o);
}

PyObject* sh::scripting::pythoninterop::PyDictGetItem(PyObject* dict, PyObject* key)
{
    return PyDict_GetItem(dict, key);
}

void sh::scripting::pythoninterop::PyDictSetItem(PyObject* dict, PyObject* key, PyObject* val)
{
    PyDict_SetItem(dict, key, val);
}

PyObject* sh::scripting::pythoninterop::PyTupleGetItem(PyObject* o, int itm)
{
    return PyTuple_GetItem(o, itm);
}

PyObject* sh::scripting::pythoninterop::PyListGetItem(PyObject* o, int itm)
{
    return PyList_GetItem(o, itm);
}

int sh::scripting::pythoninterop::PyTupleCount(PyObject* o)
{
    return PyTuple_Size(o);
}

int sh::scripting::pythoninterop::PyListCount(PyObject* o)
{
    return PyList_Size(o);
}

bool sh::scripting::pythoninterop::PyIsDict(PyObject* o)
{
    return PyDict_Check(o);
}

bool sh::scripting::pythoninterop::PyIsTuple(PyObject* o)
{
    return PyTuple_Check(o);
}

bool sh::scripting::pythoninterop::PyIsList(PyObject* o)
{
    return PyList_Check(o);
}

bool sh::scripting::pythoninterop::PyIsUnicode(PyObject* o)
{
    return PyUnicode_Check(o);
}

bool sh::scripting::pythoninterop::PyIsBytes(PyObject* o)
{
    return PyBytes_Check(o);
}

bool sh::scripting::pythoninterop::PyIsBool(PyObject* o)
{
    return PyBool_Check(o);
}

bool sh::scripting::pythoninterop::PyIsFloat(PyObject* o)
{
    return PyFloat_Check(o);
}

bool sh::scripting::pythoninterop::PyIsLong(PyObject* o)
{
    return PyLong_Check(o);
}

void sh::scripting::pythoninterop::PyIncRef(PyObject* o)
{
    Py_IncRef(o);
}

void sh::scripting::pythoninterop::PyDecRef(PyObject* o)
{
    Py_DecRef(o);
}


void sh::scripting::pythoninterop::PyTupleSetItem(PyObject* o, int itm, PyObject* val)
{
    PyTuple_SetItem(o, itm, val);
}

void sh::scripting::pythoninterop::invokePython(std::function<void()> fct)
{
    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();
    try
    {
        sh::exceptions::Exception::executeGuarded([&](){
            fct();
        });
    }
    catch(...)
    {
        PyGILState_Release(gstate);
        throw;
    }
    PyGILState_Release(gstate);
}

PyObject* sh::scripting::pythoninterop::PyObjectCall(PyObject* fct, PyObject* args)
{
    PyIncRef(fct);
    PyObject* result=PyObject_Call(fct, args, 0);
    if (!result)
        sh::scripting::PythonScriptInterpreter::throwNativeExceptionFromPythonTraceback();
    PyDecRef(fct);
    return result;
}

PyObject* sh::scripting::pythoninterop::PyTupleNew(int cnt)
{
    return PyTuple_New(cnt);
}

bool sh::scripting::pythoninterop::PythonToC_ExecuteGuarded(std::function<void()> fct)
{
    bool result = false;
    QMap<QString, QString> _exdata;
    bool cancelexception = false;
    bool threadabortexception = false;
    Py_BEGIN_ALLOW_THREADS
    try
    {
        try
        {
            fct();
            result = true;
        }
        catch(sh::exceptions::CancelException&)
        {
            cancelexception = true;
        }
        catch(sh::exceptions::ThreadAbortException&)
        {
            threadabortexception = true;
        }
        catch(sh::exceptions::Exception&)
        {
            throw;
        }
        catch(...)
        {
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(sh::exceptions::Exception::UNDEFINED_ERROR_OCCURRED));
        }
    }
    catch(sh::exceptions::Exception& ex)
    {
        _exdata = ex.data();
    }
    Py_END_ALLOW_THREADS
    if (!result) {
        if (cancelexception)
            PyErr_SetString(sh::scripting::PythonScriptInterpreter::instance()->_CancelException, "");
        else if (threadabortexception)
            PyErr_SetString(sh::scripting::PythonScriptInterpreter::instance()->_ThreadAbortException, "");
        else
        {
            PyObject* exdata = PyDict_New();
            foreach(QString key, _exdata.keys())
                PyDict_SetItemString(exdata, key.toUtf8().data(), NativeToPyObject<QString>::toPyObject(_exdata[key]));
            Py_IncRef(exdata);
            PyErr_SetObject(sh::scripting::PythonScriptInterpreter::instance()->_Exception, exdata);
        }
    }
    return result;
}

QString sh::scripting::pythoninterop::Converters::getStringFromPyObject(PyObject* o)
{
    if (isPyNone(o))
        return QString();
    if (!PyIsUnicode(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("A string was expected but something else was found.").resumeable());
    return QString(PyUnicode_AsUTF8(o));
}

QVariant sh::scripting::pythoninterop::Converters::getVariantFromPyObject(PyObject* o)
{
    if (isPyNone(o))
        return QVariant();
    else if (PyIsBool(o))
        return QVariant(getBoolFromPyObject(o));
    else if (PyIsLong(o))
        return QVariant(getInt64FromPyObject(o));
    else if (PyIsFloat(o))
        return QVariant(getDoubleFromPyObject(o));
    else if (PyIsList(o)||PyIsDict(o))
        return QVariant(QStringList( PyObjectToNative<QList<QString>>::toNative(0, o) ));
    else
        return QVariant(getStringFromPyObject(o));
}

int sh::scripting::pythoninterop::Converters::getIntFromPyObject(PyObject* o)
{
    if (!pythoninterop::PyIsLong(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("An integer was expected but something else was found.").resumeable());
    return (int) PyLong_AsLong(o);
}

qint64 sh::scripting::pythoninterop::Converters::getInt64FromPyObject(PyObject* o)
{
    if (!pythoninterop::PyIsLong(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("An integer was expected but something else was found.").resumeable());
    return PyLong_AsLongLong(o);
}

quint64 sh::scripting::pythoninterop::Converters::getUint64FromPyObject(PyObject* o)
{
    if (!pythoninterop::PyIsLong(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("An integer was expected but something else was found.").resumeable());
    return PyLong_AsLongLong(o);
}

double sh::scripting::pythoninterop::Converters::getDoubleFromPyObject(PyObject* o)
{
    if (pythoninterop::PyIsLong(o))
        return (double)PyLong_AsLongLong(o);
    if (!pythoninterop::PyIsFloat(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("A floating point number was expected but something else was found.").resumeable());
    return PyFloat_AsDouble(o);
}

bool sh::scripting::pythoninterop::Converters::getBoolFromPyObject(PyObject* o)
{
    return PyObject_IsTrue(o);
}

void* sh::scripting::pythoninterop::Converters::getPointerFromPyObject(PyObject* o)
{
    if (isPyNone(o))
        return 0;
    return getPointerFromShallotPyObject(o);
}

std::shared_ptr<void> sh::scripting::pythoninterop::Converters::getSharedPointerFromPyObject(PyObject* o)
{
    if (isPyNone(o))
        return 0;
    return getSharedPointerFromShallotPyObject(o);
}

QByteArray sh::scripting::pythoninterop::Converters::getByteArrayFromPyObject (PyObject* o)
{
    if (isPyNone(o))
        return QByteArray();
    if (!PyIsBytes(o))
        throw sh::scripting::ScriptingEngine::ScriptedException( sh::exceptions::ExceptionData().details("A bytestring was expected but something else was found.").resumeable());
    QByteArray result("");
    char* buf;
    Py_ssize_t len;
    PyBytes_AsStringAndSize(o, &buf, &len);
    result.append(buf,len);
    return result;
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromByteArray(QByteArray v)
{
    if (v.isNull())
        return getPyNone();
    int len = v.length();
    char* buf = v.data();
    return PyBytes_FromStringAndSize(buf,len);
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromString(QString v)
{
    if (v.isNull())
        return getPyNone();
    QByteArray bav = v.toUtf8();
    const char* cstr = bav.data();
    PyObject* ret = PyUnicode_FromString(cstr);
    return ret;
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromVariant(QVariant v)
{
    QVariant::Type vtype;
    if (v.isNull())
        return getPyNone();
    else if ((vtype = v.type()) == QVariant::Bool)
        return getPyObjectFromBool(v.value<bool>());
    else if (vtype==QVariant::Int||vtype==QVariant::LongLong)
        return getPyObjectFromInt64(v.value<qint64>());
    else if (vtype==QVariant::UInt||vtype==QVariant::ULongLong)
        return getPyObjectFromUint64(v.value<quint64>());
    else if (vtype==QVariant::Double)
        return getPyObjectFromDouble(v.value<double>());
    else if (vtype==QVariant::StringList)
        return NativeToPyObject<QList<QString>>::toPyObject(v.toStringList());
    else
        return getPyObjectFromString(v.value<QString>());
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromInt(int v)
{
    return PyLong_FromLong(v);
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromInt64(qint64 v)
{
    return PyLong_FromLongLong((long long)v);
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromUint64(quint64 v)
{
    return PyLong_FromLongLong((long long)v);
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromDouble(double v)
{
    return PyFloat_FromDouble(v);
}

PyObject* sh::scripting::pythoninterop::Converters::getPyObjectFromBool(bool v)
{
    return PyBool_FromLong(v?1:0);
}

void sh::scripting::PythonScriptInterpreter::registerMethod(sh::scripting::ApiClass *clss, QString name, PyCFunction fctptr)
{
    _methods[ clss->name() + "::" + name] = fctptr;
}

PyObject* new_func( PyTypeObject* subtype, PyObject*, PyObject*)
{
    PyObject* pyob = subtype->tp_alloc(subtype, 0);
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) pyob;
    memset(&(pymember->sharedNativeObject), 0, sizeof(std::shared_ptr<void>));
    return pyob;
}

void dealloc_func( PyObject* pyob )
{
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) pyob;
    pymember->sharedNativeObject = 0;
    pyob->ob_type->tp_free(pyob);
}

PyObject* _member_setbuddy(PyObject* t, PyObject* p)
{
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) t;
    PyObject* b = sh::scripting::pythoninterop::PyTupleGetItem(p, 0);
    Py_IncRef(b);
    sh::scripting::PythonScriptInterpreter::_buddies[pymember->nativeObject] = b;
    if (pymember->isReallyShared)
        sh::scripting::PythonScriptInterpreter::_natives[pymember->nativeObject] = pymember->sharedNativeObject;
    return sh::scripting::pythoninterop::getPyNone();
}

int traverse_func(PyObject* t, visitproc visit, void *arg)
{
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) t;
    auto iwptr = sh::scripting::PythonScriptInterpreter::_natives.find(pymember->nativeObject);
    if (iwptr != sh::scripting::PythonScriptInterpreter::_natives.end())
    {
        std::weak_ptr<void> wptr = *iwptr;
        if (wptr.use_count() <= 1)
        {
            Py_VISIT(sh::scripting::PythonScriptInterpreter::_buddies[pymember->nativeObject]);
        }
    }
    return 0;
}

int clear_func(PyObject* t)
{
    sh::scripting::shallot_ShallotObject* pymember = (sh::scripting::shallot_ShallotObject*) t;
    Py_DecRef(sh::scripting::PythonScriptInterpreter::_buddies.take(pymember->nativeObject));
    sh::scripting::PythonScriptInterpreter::_natives.take(pymember->nativeObject);
    return 0;
}

PyObject* sh::scripting::PythonScriptInterpreter::getApiModuleDef()
{
    static PyTypeObject shallot_ShallotType = {
        PyVarObject_HEAD_INIT(NULL, 0)
        "",                        /*tp_name*/
        sizeof(sh::scripting::shallot_ShallotObject), /*tp_basicsize*/
        0,                         /*tp_itemsize*/
        &dealloc_func,             /*tp_dealloc*/
        0,                         /*tp_print*/
        0,                         /*tp_getattr*/
        0,                         /*tp_setattr*/
        0,                         /*tp_compare*/
        0,                         /*tp_repr*/
        0,                         /*tp_as_number*/
        0,                         /*tp_as_sequence*/
        0,                         /*tp_as_mapping*/
        0,                         /*tp_hash */
        0,                         /*tp_call*/
        0,                         /*tp_str*/
        0,                         /*tp_getattro*/
        0,                         /*tp_setattro*/
        0,                         /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,        /*tp_flags*/
        "Shallot API object",      /* tp_doc */
        &traverse_func,            /* tp_traverse */
        &clear_func,               /* tp_clear */
        0,                         /* tp_richcompare */
        0,                         /* tp_weaklistoffset */
        0,                         /* tp_iter */
        0,                         /* tp_iternext */
        0,                         /* tp_methods */
        0,                         /* tp_members */
        0,                         /* tp_getset */
        0,                         /* tp_base */
        0,                         /* tp_dict */
        0,                         /* tp_descr_get */
        0,                         /* tp_descr_set */
        0,                         /* tp_dictoffset */
        0,                         /* tp_init */
        0,                         /* tp_alloc */
        &new_func,                 /* tp_new */
        0,                         /* tp_free */
        0,                         /* tp_is_gc */
        0,                         /* tp_bases */
        0,                         /* tp_mro */
        0,                         /* tp_cache */
        0,                         /* tp_subclasses */
        0,                         /* tp_weaklist */
        0,                         /* tp_del */
        0,                         /* tp_version_tag */
        0                          /* tp_finalize */
    };

    PyObject* module = PyModule_Create((PyModuleDef*)apiModuleDef);

    sh::scripting::PythonScriptInterpreter *pint = sh::scripting::PythonScriptInterpreter::instance();

    pint->_ThreadAbortException = PyErr_NewException("_shallot.ThreadAbortException", PyExc_BaseException, NULL);
    Py_INCREF(pint->_ThreadAbortException);
    PyModule_AddObject(module, "ThreadAbortException", pint->_ThreadAbortException);

    pint->_CancelException = PyErr_NewException("_shallot.CancelException", PyExc_BaseException, NULL);
    Py_INCREF(pint->_CancelException);
    PyModule_AddObject(module, "CancelException", pint->_CancelException);

    pint->_Exception = PyErr_NewException("_shallot.Exception", NULL, NULL);
    Py_INCREF(pint->_Exception);
    PyModule_AddObject(module, "Exception", pint->_Exception);

    foreach(QString shallottypename, api->classes())
    {
        sh::scripting::ApiClass* mclass = api->getClass(shallottypename);
        PyTypeObject* newpytype = (PyTypeObject*) malloc(sizeof(PyTypeObject));
        memcpy(newpytype, &shallot_ShallotType, sizeof(PyTypeObject));
        QByteArray bnewpytypename = ("_shallot." + mclass->name()).toLatin1();
        const char* newpytypename = bnewpytypename.data();
        char* cnewpytypename = (char*) malloc(sizeof(char) * bnewpytypename.length()+1);
        strcpy(cnewpytypename, newpytypename);
        newpytype->tp_name = cnewpytypename;
        PyMethodDef* pymethods = (PyMethodDef*) malloc(sizeof(PyMethodDef)* (mclass->methods().length()+2)  );
        int i=0;
        pymethods[i].ml_flags = METH_VARARGS;
        pymethods[i].ml_meth = (PyCFunction) &_member_setbuddy;
        pymethods[i].ml_name = "_set_buddy";
        pymethods[i].ml_doc = "_set_buddy";
        i++;
        foreach (QString methodname, mclass->methods()) {
            pymethods[i].ml_flags = METH_VARARGS;
            pymethods[i].ml_meth = (PyCFunction) _methods[mclass->name()+"::"+methodname];
            QByteArray bnewpytypename = methodname.toLatin1();
            const char* newpytypename = bnewpytypename.data();
            char* cnewpytypename = (char*) malloc(sizeof(char) * bnewpytypename.length()+1);
            strcpy(cnewpytypename, newpytypename);
            pymethods[i].ml_name = cnewpytypename;
            pymethods[i].ml_doc = cnewpytypename;
            i++;
        }
        memset(pymethods+i, 0, sizeof(PyMethodDef));
        newpytype->tp_methods = pymethods;
        PyType_Ready(newpytype);
        _pyTypes.insert(shallottypename, newpytype);
    }

    foreach (QString rootmember, api->rootMembers()) {
        void* member = api->rootMember(rootmember);
        sh::scripting::ApiClass* mclass = api->rootMemberClass(rootmember);
        PyObject* pymember = sh::scripting::pythoninterop::getShallotPyObjectFromPointer(member, mclass);
        QByteArray brootmember = rootmember.toLatin1();
        const char* crootmember = brootmember.data();
        PyModule_AddObject(module, crootmember, (PyObject*) pymember);
    }

    return module;
}

sh::scripting::Api* sh::scripting::PythonScriptInterpreter::api = new sh::scripting::Api();

void* sh::scripting::PythonScriptInterpreter::apiModuleDef;

PyObject* sh::scripting::PythonScriptInterpreter::modtraceback;

void sh::scripting::PythonScriptInterpreter::initializeInterpreter()
{
    static PyMethodDef methodTable[] = {
        {NULL, NULL, 0, NULL}
    };
    static PyModuleDef myModule = {PyModuleDef_HEAD_INIT, "_shallot", NULL, -1, methodTable, NULL, NULL, NULL, NULL};
    apiModuleDef = &myModule;
    PyImport_AppendInittab("_shallot", &getApiModuleDef);
    Py_Initialize();
    PyObject * mod = PyImport_ImportModule("threading");  // this is just here because it avoids threading error messages at finalize.
    Py_XDECREF(mod);
    PyEval_InitThreads();
    PyObject *sys = PyImport_ImportModule("sys");
    PyObject *path = PyObject_GetAttrString(sys, "path");
    QByteArray currdir = sh::configuration::ConfigurationManager::instance()->shallotDataDir().toLatin1();
    PyObject* currdirpybytes = PyBytes_FromString(currdir.data());
    PyObject* currdirpystring = PyCodec_Decode( currdirpybytes, "utf8", "strict" );
    PyList_Append(path, currdirpystring);
    QByteArray spluginsdir = QString("/etc/shallot/plugins").toLatin1();
    PyObject* spluginsdirpybytes = PyBytes_FromString(spluginsdir.data());
    PyObject* spluginsdirpystring = PyCodec_Decode(spluginsdirpybytes, "utf8", "strict" );
    QByteArray upluginsdir = (sh::configuration::ConfigurationManager::instance()->mypath() + "/plugins").toLatin1();
    PyObject* upluginsdirpybytes = PyBytes_FromString(upluginsdir.data());
    PyObject* upluginsdirpystring = PyCodec_Decode(upluginsdirpybytes, "utf8", "strict" );
    QByteArray bpluginsdir = (sh::configuration::ConfigurationManager::instance()->shallotDataDir() + "/plugins").toLatin1();
    PyObject* bpluginsdirpybytes = PyBytes_FromString(bpluginsdir.data());
    PyObject* bpluginsdirpystring = PyCodec_Decode(bpluginsdirpybytes, "utf8", "strict" );
    PyList_Append(path, spluginsdirpystring);
    PyList_Append(path, upluginsdirpystring);
    PyList_Append(path, bpluginsdirpystring);
    modtraceback = PyImport_ImportModule("traceback");
    Py_DecRef(currdirpystring);
    Py_DecRef(currdirpybytes);
    Py_DecRef(spluginsdirpystring);
    Py_DecRef(spluginsdirpybytes);
    Py_DecRef(upluginsdirpystring);
    Py_DecRef(upluginsdirpybytes);
    Py_DecRef(bpluginsdirpystring);
    Py_DecRef(bpluginsdirpybytes);
    Py_DecRef(path);
    Py_DecRef(sys);
    PyEval_ReleaseLock();
    sh::tools::Logger::instance()->log("Python scripting engine is initialized.");
}

void sh::scripting::PythonScriptInterpreter::throwNativeExceptionFromPythonTraceback()
{
    bool isRawPythonException=true;
    QMap<QString, QString> excdata;
    excdata["isresumeable"] = "1";
    bool cancelexception = false;
    bool threadabortexception = false;
    PyObject* et;PyObject* ev;PyObject* ec;
    PyErr_Fetch(&et, &ev, &ec);
    PyErr_NormalizeException(&et, &ev, &ec);
    if (et == sh::scripting::PythonScriptInterpreter::instance()->_CancelException)
        cancelexception = true;
    else if (et == sh::scripting::PythonScriptInterpreter::instance()->_ThreadAbortException)
        threadabortexception = true;
    else
    {
        if (ev != NULL)
        {
            PyObject* excs = PyObject_Str(ev);
            excdata["details"] = PyUnicode_AsUTF8(excs);
            if (PyObject_HasAttrString(ev,"args"))
            {
                PyObject* _tuple = PyObject_GetAttrString(ev,"args");
                if (PyTuple_Check(_tuple))
                {
                    if (PyTuple_Size(_tuple)>0)
                    {
                        PyObject *dict = PyTuple_GetItem(_tuple, 0);
                        if (PyDict_Check(dict))
                        {
                            isRawPythonException = false;
                            excdata = sh::scripting::pythoninterop::PyObjectToNative<StringMap<QString>>::toNative(pythoninterop::getPyNone(),  dict);
                        }
                    }
                }
            }
            Py_DecRef(ev);
            Py_DecRef(excs);
        }
        if (ec != NULL) {
            PyObject *fctFormatexc = PyObject_GetAttrString(modtraceback, "format_tb");
            PyObject* exccl = PyObject_CallFunctionObjArgs(fctFormatexc, ec, NULL);

            Py_ssize_t size = PyList_Size(exccl);
            for(Py_ssize_t i=0; i<size; i++)
            {
                PyObject* excc = PyList_GetItem(exccl, i);
                QString l = QString("\n") + PyUnicode_AsUTF8(excc);
                if (isRawPythonException)
                    excdata["details"] += l;
                else
                    excdata["aux"] += l;
                Py_DecRef(excc);
            }
            Py_DecRef(ec);
            Py_DecRef(fctFormatexc);
            Py_DecRef(exccl);
        }
    }
    Py_DecRef(et);
    if (cancelexception)
        throw sh::exceptions::CancelException();
    else if (threadabortexception)
        throw sh::exceptions::ThreadAbortException();
    else if (!excdata.isEmpty())
        throw sh::exceptions::Exception(excdata);
}

void sh::scripting::PythonScriptInterpreter::execute(QString script)
{
    QFileInfo scriptfile(script);
    QString modname = scriptfile.completeBaseName();
    modname = modname.mid(0, modname.lastIndexOf("."));
    QByteArray ba = modname.toLatin1();
    char * cmodname = ba.data();
    pythoninterop::invokePython([&](){
        PyObject* module = PyImport_ImportModule(cmodname);
        if (module == NULL)
            throwNativeExceptionFromPythonTraceback();
    });
}

#endif

sh::scripting::PythonScriptInterpreter::PythonScriptInterpreter()
{
}

sh::scripting::PythonScriptInterpreter * sh::scripting::PythonScriptInterpreter::instance()
{
    static sh::scripting::PythonScriptInterpreter* _instance = new sh::scripting::PythonScriptInterpreter();
    return _instance;
}

void sh::scripting::PythonScriptInterpreter::_initialize()
{
}

void sh::scripting::PythonScriptInterpreter::_shutdown()
{
#ifdef PYTHONSCRIPT
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        PyGILState_Ensure();
        Py_Finalize();
    });
#endif
}

