/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SCRIPTINGENGINE_H
#define SCRIPTINGENGINE_H

#include <QCoreApplication>
#include <QDir>
#include <QFileInfo>
#include <QMap>
#include <QMutex>
#include <QObject>
#include "actions/actionactionitem.h"
#include "exceptions/runtimeexception.h"


namespace sh{
    namespace configuration { class ConfigurationValue; }
    namespace scripting {

        /** \brief
         * The base class for integration of a script language.
         */
        class ScriptInterpreter
        {
        public:
            virtual bool isAvailable() { return false; }
            virtual void initializeInterpreter() { }
            virtual QString filesuffix() { return ""; }
            virtual void execute(QString) { }
            virtual ~ScriptInterpreter() { }
            virtual void deleteLater() { delete this; }
        };

        /** \brief
         * The Scripting engine loads scripts by means of the registered interpreters.
         */
        class ScriptingEngine : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static ScriptingEngine* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            void loadScript(QString script);

            class ScriptedException : public sh::exceptions::Exception
            {
            public:
                ScriptedException(sh::exceptions::ExceptionData data);
            };

        signals:

        public slots:

        private:
            explicit ScriptingEngine(QObject *parent = 0);
            QMap<QString, ScriptInterpreter*> _interpreters;
            void registerInterpreter(ScriptInterpreter* interpreter);
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalLastStartupFailed;

            class ActionPluginLoadCrashedAgain : public sh::actions::ActionActionItem
            {
            public:
                ActionPluginLoadCrashedAgain(QString path, QMutex *mutex);
            protected:
                virtual void action(sh::actions::ActionExecutionInfo* info);
                QString _path;
                QMutex* _mutex;
            };


        };

    }
}

#endif // SCRIPTINGENGINE_H
