import shallot
import os
import shutil

# An operation object can be used for fetching content from a location and
# storing content to some location. Read shallot.Operation.
#
# Most calls in the filesystem api are automatically provided with such an object.
# On other places, like this one, we can also create it (and must finish the transaction).
#
# Find the demos in the context menu.

class FetchingFilesAction(shallot.Actions.ActionAction):

    def __init__(self, nodes):
        shallot.Actions.ActionAction.__init__(self, "Operation: Copy files to a local directory")
        self.nodes = nodes

    def action(self, info):
        operation = shallot.Operations.create()
        destdir = info.userfeedback().simple_inputbox("Local destination directory:", "/tmp/foo")
        if destdir is not None:
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
            for node in self.nodes:
                tmpname = operation.fetch_file2(node.eurl(), "myfiles")
                shutil.copy(tmpname, destdir + "/" + node.eurl().basename())
        operation.abort() # after usage, you should always call abort or commit (abort is cheaper, commit actually transfers data to destinations, if something was written)


class StoringFilesAction(shallot.Actions.ActionAction):

    def __init__(self, nodes):
        shallot.Actions.ActionAction.__init__(self, "Operation: Copy all files from a local directory to here")
        self.node = nodes[0]

    def action(self, info):
        operation = shallot.Operations.create()
        srcdir = info.userfeedback().simple_inputbox("Local source directory:", "/tmp/foo")
        if srcdir is not None:
            for spath in os.listdir(srcdir):
                fspath = srcdir + "/" + spath
                if os.path.isfile(fspath):
                    with open(fspath, "rb") as f:
                        operation.filesystem().create_file(self.node.eurl().with_appended_segment(spath), shallot.Streaming.StreamReadDataDevice(f), None)
        operation.commit()


class CustomDataAction(shallot.Actions.ActionAction):

    def __init__(self, nodes):
        shallot.Actions.ActionAction.__init__(self, "Operation: Custom data")
        self.nodes = nodes

    def action(self, info):
        operation = shallot.Operations.create()
        operation.set_custom_data(None, "foo", "Hello ")
        operation.set_custom_data(None, "bar", "World!") # bools and numbers are also allowed
        txt = operation.get_custom_data(None, "foo") + operation.get_custom_data(None, "bar")
        info.userfeedback().simple_messagebox(txt, shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK, "",  0, 0)
        operation.abort()


shallot.Actions.register(FetchingFilesAction, predicates=[shallot.Actions.OnFilesPredicate()])
shallot.Actions.register(StoringFilesAction, predicates=[shallot.Actions.OnDirectoriesPredicate(), shallot.Actions.OnSingleEntrySelectionPredicate()])
shallot.Actions.register(CustomDataAction, predicates=[shallot.Actions.HideOnSelectionLevelPredicate()])
