import shallot
import datetime

# Write a subclass of shallot.Filesystem.Handler for implementing
# a custom filesystem.
# This example implements a nearly full-featured in-memory filesystem
# with some sample nodes.
# In other words: A rather inefficient Python-based in-Shallot ramdisk.

class MyFSH(shallot.Filesystem.Handler):

    def __init__(self):
        shallot.Filesystem.Handler.__init__(self)

        self.myfiles = {  # just for this implementation; not part of the api
            "/it/.dir": b"",
            "/it/is/.dir": b"",
            "/it/is/magic/.dir": b"",
            "/it/is/magic/everywhere/.dir": b"",
            "/readme 1": b"What you are currently seeing is not a place in your local filesystem.\n\nInstead it is a 'virtual' one, provided by the shallot hello world plugin.\nIt is an example for writing shallot plugins in python with the shallot scripting api.\n\nYou should read the other readme files in the subdirectories as well :)\n",
            "/it/readme 2": b"Shallot has a plugin interface. Shallot plugins are python files in certain directories, which make usage of the shallot scripting interface.\n",
            "/it/is/readme 3": b"The hello world plugin provides this virtual filesystem. It is an example for a scripted filesystem handler, which can provide arbitrary virtual filesystems.\n\nThis example is maybe incomplete in corner cases, but is brutally simple and feature-complete. You can even put some new small files here (which will not permanently be stored somewhere).\n\nAnother big part are the provided actions (find them in the menu) which demonstrates some other scripted things.\n",
            "/it/is/magic/readme 4": b"If you are going to develop a shallot plugin, please find this plugin file and use it as an example how to establish things.\n",
            "/it/is/magic/everywhere/readme 5": b"This plugin is also used for testing the behavior in some situations.\n",
        }

        self.mydetailcolumn=MyDetailColumn(self)  # we add a column to the file list; used below

    def get_mimetype(self, operation, eurl):
        return "text/plain"

    def get_type(self, operation, eurl):
        if eurl.path()=="/":
            return shallot.Filesystem.NodeType.Directory
        elif eurl.path() in self.myfiles.keys():
            if self.myfiles[eurl.path()].startswith(b"!!LINK!!"):
                return shallot.Filesystem.NodeType.Link
            else:
                return shallot.Filesystem.NodeType.File
        else:
            for myfile in self.myfiles:
                if myfile.startswith(eurl.path()+"/"):
                    return shallot.Filesystem.NodeType.Directory
        return shallot.Filesystem.NodeType.NONE

    def get_size(self, operation, eurl):
        for myfile in self.myfiles:
            if eurl.path() == myfile:
                return len(self.myfiles[myfile])
        return 0

    def get_mtime(self, operation, eurl):
        return datetime.datetime(year=2011,month=8,day=22,hour=13,minute=37)

    def set_mtime(self, operation, eurl, mtime):
        pass  # left away for simplicity

    def can_create_directory(self, operation, eurl):
        return True

    def create_directory(self, operation, eurl):
        self.create_file(operation, eurl.with_appended_segment(".dir"), None, None)

    def can_create_file(self, operation, eurl):
        return True

    def can_delete_item(self, operation, eurl):
        return True

    def delete_item(self, operation, eurl):
        for key in list(self.myfiles.keys()):
            if key.startswith(eurl.path()+"/") or key == eurl.path():
                del self.myfiles[key]

    def can_rename_item(self, operation, src):
        return True

    def rename_item(self, operation, src, destpath):
        for key in list(self.myfiles.keys()):
            if key.startswith(src.path()+"/") or key==src.path():
                newkey = destpath + key[len(src.path()):]
                self.myfiles[newkey] = self.myfiles.pop(key)

    def itemlist(self, operation, eurl, nodetype, nodelist):
        result = set()
        path = eurl.path()
        if path == "/":  # this makes some stuff easier
            path = ""
        for myfile in self.myfiles.keys():
            if not myfile.startswith(path):
                continue
            rest = myfile[len(path)+1:]
            if rest == "": # this is more a theoretical case
                continue
            if nodetype == shallot.Filesystem.NodeType.Directory:
                if not "/" in rest:
                    continue
                name = rest[:rest.index("/")]
            elif nodetype == shallot.Filesystem.NodeType.File:
                if "/" in rest:
                    continue
                name = rest
            else:
                continue
            result.add(name)
        nodelist.set_items(list(result))

    def configure_item(self, operation, node):
        node.add_detail(self.mydetailcolumn)
        if node.eurl().basename() == ".dir":
            node.set_hidden(True)

    def can_get_filecontent(self, operation, eurl):
        return True

    def get_file_content(self, operation, eurl):
        return shallot.Streaming.BlobReadDataDevice(self.myfiles[eurl.path()])

    def create_file(self, operation, eurl, outstream, handlertransfer):
        self.myfiles[eurl.path()] = outstream.readall() if outstream else b""  # outstream can be None in this implementation since we use it this way some lines above

    def get_actions(self, eurls):
        return [MyAction(self, eurls)]

    def can_create_link(self, operation, eurl):
        return True

    def create_link(self, operation, eurl, tgt):
        self.myfiles[eurl.path()] = ("!!LINK!!"+tgt).encode()

    def get_linktarget(self, operation, eurl):
        return self.myfiles[eurl.path()].decode()[8:]


# A shallot.DetailColumn implements one column in a file list.
class MyDetailColumn(shallot.DetailColumn):

    def __init__(self, handler):
        shallot.DetailColumn.__init__(self, "mydc", "Number of A's")
        self.handler = handler

    def determine_value(self, node, op):
        if node.nodetype() == shallot.Filesystem.NodeType.File and node.eurl().path() in self.handler.myfiles:
            return str(len([x for x in self.handler.myfiles[node.eurl().path()].decode().lower() if x=="a"]))
        else:
            return ""


# This actions dump the filesystem content to the log.
class MyAction(shallot.Actions.SubmenuAction):

    def __init__(self, handler, eurls):
        self.handler = handler
        self.eurls = eurls
        shallot.Actions.SubmenuAction.__init__(self, "Hello World!", icon="dialog-warning")

    def initialize(self):
        self.set_subitems([LogContentAction(self.handler, self.eurls),])


class LogContentAction(shallot.Actions.ActionAction):

    def __init__(self, handler, eurls):
        self.handler = handler
        self.eurls = eurls
        shallot.Actions.ActionAction.__init__(self, "Log all helloworld filesystem content", icon="text-x-generic")

    def _stringforfile(self,key):
        return "\nFILE: {key}\n-----\n{content}\n\n".format(key=key, content=self.handler.myfiles[key])

    def action(self, info):
        content = "\n".join([self._stringforfile(key) for key in self.handler.myfiles.keys()])
        shallot.Logging.log_info("This is the hello world filesystem content:\n\n"+content)
        info.set_head("Writing content to log...")
        info.set_progress(1,1,"")
        info.userfeedback().simple_messagebox("I've written the content to the shallot log. You find it somewhere in the menu.",
                                              shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK,
                                              "dialog-ok", shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK,
                                              shallot.Actions.ExecutionUserFeedback.MessageBoxButton.OK)


# Register the new own filesystem with a certain url schema
# (i.e. the very first part of a url: pythonhelloworld:///foo/bar/baz).
myhandler = MyFSH()
shallot.Filesystem.Handler.register("pyhelloworld", myhandler)

# Create the root node.
node = shallot.Filesystem.create_node(
        shallot.Eurl.create("pyhelloworld","localhorst","/"),
        myhandler,
        shallot.Filesystem.NodeType.Directory,
        shallot.Filesystem.rootnode)
node.set_displayname("Python Hello World")
