/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QLabel>
#include "actions/actionexecutioninfo.h"
#include "base/operation.h"
#include "base/eurl.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "ui/linkbutton.h"
#include "abstractactiondrivensearchcriterion.h"


sh::search::criteria::AbstractActionDrivenSearchCriterion::AbstractActionDrivenSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
    : sh::search::SearchCriterion(factory)
{
}

bool sh::search::criteria::AbstractActionDrivenSearchCriterion::match(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (searchspec.isEmpty())
        return false;
    return match2(op, eurl);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::valuedescription()
{
    return searchspec.join(" ");
}

QWidget* sh::search::criteria::AbstractActionDrivenSearchCriterion::_createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, sh::search::SearchCriterionFactory::AuxiliaryButtonGenerator makeAuxButton)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    sh::ui::LinkButton* changebtn = makeAuxButton();
    QLabel* result = new QLabel();
    QObject::connect(changebtn, &sh::ui::LinkButton::clicked, [=](){
        QMutex* _execmutex = new QMutex();
        std::shared_ptr<ActionAbstractActionDrivenSearchCriterionConfigure> action = std::make_shared<ActionAbstractActionDrivenSearchCriterionConfigure>(std::dynamic_pointer_cast<AbstractActionDrivenSearchCriterion>(cfg), _execmutex);
        sh::tools::ThreadPool::enqueueForce([=](){
            action->initializeSync();
            action->execute();
            _execmutex->lock();
            _execmutex->unlock();
            delete _execmutex;
            sh::tools::MainThread::dispatcher()->invokeSync([&](){
                updatelabel(cfg, result);
            });
        });
    });
    changebtn->setText(QObject::tr("C&onfigure"));
    _aliveWidgets.insert(result);
    QObject::connect(result, &QObject::destroyed, [=](){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        _aliveWidgets.remove(result);
    });
    updatelabel(cfg, result);
    return result;
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::updatelabel(std::shared_ptr<sh::search::SearchCriterion> cfg, QWidget* w)
{
    if (_aliveWidgets.contains(w))
    {
        sh::search::criteria::AbstractActionDrivenSearchCriterion* wcfg = dynamic_cast<sh::search::criteria::AbstractActionDrivenSearchCriterion*>(cfg.get());
        QString lbltext = "???";
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            lbltext = wcfg->searchspec.isEmpty() ? QObject::tr("Not yet configured") : QObject::tr("Search for: %1").arg(wcfg->getSearchSpecDescription());
        });
        qobject_cast<QLabel*>(w)->setText(lbltext);
    }
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::_editorUpdateConfiguration(QWidget*, std::shared_ptr<sh::search::SearchCriterion>)
{
}

void sh::search::criteria::AbstractActionDrivenSearchCriterion::_deserialize(QString s)
{
    searchspec = sh::search::SearchCriterion::_deserializeStringList(s);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::_serialize()
{
    return sh::search::SearchCriterion::_serializeStringList(searchspec);
}

QString sh::search::criteria::AbstractActionDrivenSearchCriterion::getSearchSpecDescription()
{
    return searchspec.join(" ");
}

sh::search::criteria::ActionAbstractActionDrivenSearchCriterionConfigure::ActionAbstractActionDrivenSearchCriterionConfigure(std::shared_ptr<AbstractActionDrivenSearchCriterion> cfg, QMutex* mutex)
    : sh::actions::ActionActionItem(""), _cfg(cfg), _mutex(mutex)
{
    _mutex->lock();
}

void sh::search::criteria::ActionAbstractActionDrivenSearchCriterionConfigure::action(sh::actions::ActionExecutionInfo* info)
{
    _cfg->configure(info);
    QVariant v = info->operation()->getCustomData(0, "searchconfiguration");
    if (!v.isNull())
        _cfg->searchspec = v.toStringList();
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        _mutex->unlock();
    });
}

QSet<QWidget*> sh::search::criteria::AbstractActionDrivenSearchCriterion::_aliveWidgets;
