/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QLineEdit>
#include <QRegularExpression>
#include "base/eurl.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "ui/linkbutton.h"
#include "abstractstringsearchcriterion.h"


sh::search::criteria::AbstractStringSearchCriterion::AbstractStringSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
                : sh::search::SearchCriterion(factory)
{
}

void sh::search::criteria::AbstractStringSearchCriterion::_deserialize(QString s)
{
    QStringList lst = _deserializeStringList(s);
    mode = (Mode) ModeCodes.indexOf(lst[0]);
    string = lst[1];
}

QString sh::search::criteria::AbstractStringSearchCriterion::_serialize()
{
    QStringList res;
    res << ModeCodes[(int)mode];
    res << string;
    return _serializeStringList(res);
}

bool sh::search::criteria::AbstractStringSearchCriterion::match(base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    if (mode == Mode::Exact)
    {
        QString a = itemstring(eurl, op);
        QString b = string;
        if (forceCaseInsensitive())
        {
            a = a.toLower();
            b = b.toLower();
        }
        return a.contains(b);
    }
    else if (mode == Mode::Fuzzy)
    {
        return relativeLevenshteinPartsSubstringDistance(string, itemstring(eurl, op)) < 0.25;
    }
    else if (mode == Mode::RegExp)
    {
        if (reStringString != string)
        {
            reString.setPattern(string);
            reString.setPatternOptions((forceCaseInsensitive()?QRegularExpression::CaseInsensitiveOption:QRegularExpression::NoPatternOption)
                                       |QRegularExpression::DotMatchesEverythingOption|QRegularExpression::UseUnicodePropertiesOption|QRegularExpression::MultilineOption);
            reStringString = string;
        }
        return reString.match(itemstring(eurl, op)).hasMatch();
    }
    else if (mode == Mode::Simple)
    {
        if (reGlobStringString != string)
        {
            QString _patAstk = QRegularExpression::escape("*");
            QString _patQues = QRegularExpression::escape("?");
            QString pattern = QRegularExpression::escape(string);
            reGlobString.setPattern(pattern.replace(_patAstk, ".*").replace(_patQues, "."));
            reGlobString.setPatternOptions(QRegularExpression::CaseInsensitiveOption|QRegularExpression::DotMatchesEverythingOption|QRegularExpression::UseUnicodePropertiesOption);
            reGlobStringString = string;
        }
        return reGlobString.match(itemstring(eurl, op)).hasMatch();
    }
    else
        return false;
}

QWidget* sh::search::criteria::AbstractStringSearchCriterion::_createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, SearchCriterionFactory::AuxiliaryButtonGenerator makeAuxButton)
{
    QLineEdit* result = new QLineEdit();
    sh::search::criteria::AbstractStringSearchCriterion* me = dynamic_cast<sh::search::criteria::AbstractStringSearchCriterion*>(cfg.get());
    result->setText(me->string);
    result->setPlaceholderText(QObject::tr("Search term"));

    sh::ui::LinkButton* btnMode = makeAuxButton();
    btnMode->setMenu(QStringList({ QObject::tr("Simple: with ?* patterns"), QObject::tr("Exact"), QObject::tr("Regular expression"), QObject::tr("Fuzzy") }), QObject::tr("(c&hange)"));
    QObject::connect(btnMode, &ui::LinkButton::menuItemSelected, btnMode, [=](int i){
        me->mode = (Mode)i;
    });
    btnMode->setSelectedMenuItem((int)(me->mode));

    return result;
}

double sh::search::criteria::AbstractStringSearchCriterion::relativeLevenshteinPartsSubstringDistance(QString needle, QString haystack)
{
    QStringList _parts = needle.toLower().split(" ");
    QStringList parts;
    int charcnt = 0;
    foreach(QString p, _parts)
    {
        QString pt = p.trimmed();
        if (!pt.isEmpty())
            parts << pt;
        charcnt += pt.length();
    }
    double dist = 0;
    foreach(QString part, parts)
    {
        int lvn = levenshteinSubstringDistance(part, haystack.toLower());
        double reldist = (double)lvn/part.length();
        dist += (reldist / charcnt) * part.length();
    }
    return dist;
}

// borrowed from http://ginstrom.com/scribbles/2007/12/01/fuzzy-substring-matching-with-levenshtein-distance-in-python/
int sh::search::criteria::AbstractStringSearchCriterion::levenshteinSubstringDistance(QString needle, QString haystack)
{
    int m = needle.length();
    qint64 n = haystack.length();
    if (m == 1)
        return haystack.contains(needle) ? 0 : 1;
    if (n == 0)
        return m;
    int* row1 = new int[n+1];
    int* row2 = new int[n+1];
    for(int i=0; i<=n; i++)
        row1[i] = 0;
    for(int i=0; i<m; i++)
    {
        row2[0] = i+1;
        for(int j=0; j<n; j++)
        {
            int cost = (needle.at(i)!=haystack.at(j)) ? 1 : 0;
            row2[j+1] = std::min(std::min(row1[j+1]+1, row2[j]+1), row1[j]+cost);
        }
        for(int j=0; j<=n; j++)
            row1[j] = row2[j];
    }
    int res = row1[0];
    for(int i=1; i<=n; i++)
        res = std::min(res, row1[i]);
    delete[] row1;
    delete[] row2;
    return res;
}

void sh::search::criteria::AbstractStringSearchCriterion::_editorUpdateConfiguration(QWidget* w, std::shared_ptr<sh::search::SearchCriterion> c)
{
    dynamic_cast<sh::search::criteria::AbstractStringSearchCriterion*>(c.get())->string = dynamic_cast<QLineEdit*>(w)->text();
}

QString sh::search::criteria::AbstractStringSearchCriterion::valuedescription()
{
    return string.mid(0, 40);
}

const QStringList sh::search::criteria::AbstractStringSearchCriterion::ModeCodes({"simple", "exact", "re", "fuzzy"});
