/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QRegularExpression>
#include "base/eurl.h"
#include "base/filesystemmodel.h"
#include "detailcolumns/detailcolumnextendedattributes.h"
#include "search/criteria/abstractstringsearchcriterion.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "search/searchmanager.h"
#include "tools/singletoninitializer.h"
#include "ui/linkbutton.h"
#include "extendedattributesearchcriterion.h"


namespace sh{namespace search{namespace criteria{INITIALIZE_STATIC_IN_MAINTHREAD(100000, ExtendedAttributeSearchCriterion)}}}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_initialize()
{
    REGISTER_CRITERION_FACTORY(sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_CORE+15000, "xattr", QObject::tr("Filesystem extended attributes"), sh::search::criteria::ExtendedAttributeSearchCriterion)
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_shutdown()
{
}

sh::search::criteria::ExtendedAttributeSearchCriterion::ExtendedAttributeSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
                : sh::search::SearchCriterion(factory)
{
}

QWidget* sh::search::criteria::ExtendedAttributeSearchCriterion::createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, SearchCriterionFactory::AuxiliaryButtonGenerator makeAuxButton)
{
    sh::search::criteria::ExtendedAttributeSearchCriterion* me = dynamic_cast<sh::search::criteria::ExtendedAttributeSearchCriterion*>(cfg.get());
    QWidget* result = new QWidget();
    QHBoxLayout* layout = new QHBoxLayout(result);
    layout->setMargin(0);
    QLabel* lblfoo = new QLabel(result); // this hack lets focus jump to value field
    lblfoo->setText("");
    layout->addWidget(lblfoo);
    QLineEdit* txtKey = new QLineEdit(result);
    txtKey->setPlaceholderText(QObject::tr("Attribute key"));
    layout->addWidget(txtKey);
    txtKey->setText(me->key);
    QLineEdit* txtValue = new QLineEdit(result);
    txtValue->setPlaceholderText(QObject::tr("Attribute value"));
    txtValue->setText(me->value);
    layout->addWidget(txtValue);
    lblfoo->setFocusProxy(txtValue);
    sh::ui::LinkButton* btnKeyMode = makeAuxButton();
    QString keymode = QObject::tr("Key mode: ");
    btnKeyMode->setMenu(QStringList({ QObject::tr("No key search"), keymode + QObject::tr("Simple"), keymode + QObject::tr("Regular expression"), keymode + QObject::tr("Fuzzy") }), QObject::tr("(c&hange)"));
    QObject::connect(btnKeyMode, &ui::LinkButton::menuItemSelected, btnKeyMode, [=](int i){
        me->keymode = (Mode)i;
        _checkEditFieldStates(result, me);
    });
    btnKeyMode->setSelectedMenuItem((int)(me->keymode));
    sh::ui::LinkButton* btnValueMode = makeAuxButton();
    QString valuemode = QObject::tr("Value mode: ");
    btnValueMode->setMenu(QStringList({ QObject::tr("No value search"), valuemode + QObject::tr("Simple"), valuemode + QObject::tr("Regular expression"), valuemode + QObject::tr("Fuzzy") }), QObject::tr("(ch&ange)"));
    QObject::connect(btnValueMode, &ui::LinkButton::menuItemSelected, btnValueMode, [=](int i){
        me->valuemode = (Mode)i;
        _checkEditFieldStates(result, me);
    });
    btnValueMode->setSelectedMenuItem((int)(me->valuemode));
    _checkEditFieldStates(result, me);
    return result;
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_checkEditFieldStates(QWidget *w, sh::search::criteria::ExtendedAttributeSearchCriterion* cfg)
{
    QLineEdit* txtKey = qobject_cast<QLineEdit*>(w->layout()->itemAt(1)->widget());
    QLineEdit* txtValue = qobject_cast<QLineEdit*>(w->layout()->itemAt(2)->widget());
    txtKey->setEnabled(cfg->keymode != Mode::None);
    txtValue->setEnabled(cfg->valuemode != Mode::None);
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::editorUpdateConfiguration(QWidget* w, std::shared_ptr<sh::search::SearchCriterion> c)
{
    QLineEdit* txtKey = qobject_cast<QLineEdit*>(w->layout()->itemAt(1)->widget());
    QLineEdit* txtValue = qobject_cast<QLineEdit*>(w->layout()->itemAt(2)->widget());
    sh::search::criteria::ExtendedAttributeSearchCriterion* me = dynamic_cast<sh::search::criteria::ExtendedAttributeSearchCriterion*>(c.get());
    me->key = (me->keymode==Mode::None) ? "" : txtKey->text();
    me->value = (me->valuemode==Mode::None) ? "" : txtValue->text();
}

bool sh::search::criteria::ExtendedAttributeSearchCriterion::match(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QList<std::shared_ptr<sh::base::FilesystemNode>> nodes = sh::base::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (!nodes.isEmpty())
    {
        QVariant vmap = sh::detailcolumns::DetailColumnExtendedAttributes::instance()->computeValue(nodes.first(), op);
        QMap<QString, QByteArray> map = sh::detailcolumns::DetailColumnExtendedAttributes::getMapByValue(vmap);
        foreach(QString s, map.keys())
        {
            bool match = matchString(keymode, key, s, &reKey, &reKeyKey);
            if (!match)
                continue;
            if (valuemode != Mode::None)
            {
                QString val = QString::fromUtf8(map[s]);
                match = matchString(valuemode, value, val, &reValue, &reValueValue);
                if (!match)
                    continue;
            }
            return true;
        }
    }
    return false;
}

bool sh::search::criteria::ExtendedAttributeSearchCriterion::matchString(Mode mode, QString needle, QString haystack, QRegularExpression* re, QString* restring)
{
    if (mode == Mode::Simple)
    {
        return haystack.toLower().contains(needle.toLower());
    }
    else if (mode == Mode::Fuzzy)
    {
        return sh::search::criteria::AbstractStringSearchCriterion::relativeLevenshteinPartsSubstringDistance(needle, haystack) < 0.25;
    }
    else if (mode == Mode::RegExp)
    {
        if (*restring != needle)
        {
            re->setPattern(needle);
            re->setPatternOptions(QRegularExpression::DotMatchesEverythingOption|QRegularExpression::UseUnicodePropertiesOption|QRegularExpression::MultilineOption);
            *restring = needle;
        }
        return re->match(haystack).hasMatch();
    }
    else if (mode == Mode::None)
    {
        return true;
    }
    else
        return false;
}

QString sh::search::criteria::ExtendedAttributeSearchCriterion::valuedescription()
{
    return (valuemode==Mode::None) ? key : value;
}

void sh::search::criteria::ExtendedAttributeSearchCriterion::_deserialize(QString s)
{
    QStringList sl = SearchCriterion::_deserializeStringList(s);
    keymode = (Mode)ModeCodes.indexOf(sl[0]);
    key = sl[1];
    valuemode = (Mode)ModeCodes.indexOf(sl[2]);
    value = sl[3];
}

QString sh::search::criteria::ExtendedAttributeSearchCriterion::_serialize()
{
    QStringList l;
    l << ModeCodes[(int)keymode];
    l << key;
    l << ModeCodes[(int)valuemode];
    l << value;
    return SearchCriterion::_serializeStringList(l);
}

const QStringList sh::search::criteria::ExtendedAttributeSearchCriterion::ModeCodes({"none", "simple", "re", "fuzzy"});
