/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QHBoxLayout>
#include <QLabel>
#include "base/eurl.h"
#include "base/filesystemmodel.h"
#include "detailcolumns/detailcolumnmtime.h"
#include "search/searchcriterionfactoryfromfunction.h"
#include "search/searchmanager.h"
#include "tools/singletoninitializer.h"
#include "mtimesearchcriterion.h"


namespace sh{namespace search{namespace criteria{INITIALIZE_STATIC_IN_MAINTHREAD(100000, MtimeSearchCriterion)}}}

void sh::search::criteria::MtimeSearchCriterion::_initialize()
{
    REGISTER_CRITERION_FACTORY(sh::search::SearchManager::REGISTER_SEARCHCRITERION_INDEX_CORE+10000, "mtime", QObject::tr("Modification time"), sh::search::criteria::MtimeSearchCriterion)
}

void sh::search::criteria::MtimeSearchCriterion::_shutdown()
{
}

sh::search::criteria::MtimeSearchCriterion::MtimeSearchCriterion(std::shared_ptr<SearchCriterionFactory> factory)
                : sh::search::SearchCriterion(factory)
{
    QDateTime now = QDateTime::currentDateTime();
    from = now;
    to = now;
}

QWidget* sh::search::criteria::MtimeSearchCriterion::createEditor(std::shared_ptr<sh::search::SearchCriterion> cfg, SearchCriterionFactory::AuxiliaryButtonGenerator)
{
    sh::search::criteria::MtimeSearchCriterion* me = dynamic_cast<sh::search::criteria::MtimeSearchCriterion*>(cfg.get());
    QWidget* result = new QWidget();
    QHBoxLayout* layout = new QHBoxLayout(result);
    layout->setMargin(0);
    QLabel* lblfrom = new QLabel(result);
    lblfrom->setText(QObject::tr("between"));
    layout->addWidget(lblfrom);
    QDateTimeEdit* from = new MtimeSearchCriterion_QDateTimeEdit(result);
    from->setCalendarPopup(true);
    from->setDateTime(me->from);
    layout->addWidget(from);
    lblfrom->setFocusProxy(from); // important for our focus magic
    QLabel* lblto = new QLabel(result);
    lblto->setText(QObject::tr("and"));
    layout->addWidget(lblto);
    QDateTimeEdit* to = new MtimeSearchCriterion_QDateTimeEdit(result);
    to->setCalendarPopup(true);
    to->setDateTime(me->to);
    to->updateGeometry();
    layout->addWidget(to);
    layout->addStretch(1);
    return result;
}

void sh::search::criteria::MtimeSearchCriterion::editorUpdateConfiguration(QWidget* w, std::shared_ptr<sh::search::SearchCriterion> c)
{
    sh::search::criteria::MtimeSearchCriterion* me = dynamic_cast<sh::search::criteria::MtimeSearchCriterion*>(c.get());
    QDateTimeEdit* wfrom = qobject_cast<QDateTimeEdit*>(w->layout()->itemAt(1)->widget());
    QDateTimeEdit* wto = qobject_cast<QDateTimeEdit*>(w->layout()->itemAt(3)->widget());
    me->from = wfrom->dateTime();
    me->to = wto->dateTime();
}

bool sh::search::criteria::MtimeSearchCriterion::match(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QList<std::shared_ptr<sh::base::FilesystemNode>> nodes = sh::base::FilesystemModel::instance()->findNodesForEurl(eurl);
    if (!nodes.isEmpty())
    {
        QDateTime t= sh::detailcolumns::DetailColumnMtime::instance()->computeValue(nodes.first(), op).toDateTime();
        return (from <= t) && (t <= to);
    }
    return false;
}

QString sh::search::criteria::MtimeSearchCriterion::valuedescription()
{
    return from.toString(Qt::DefaultLocaleShortDate);
}

void sh::search::criteria::MtimeSearchCriterion::_deserialize(QString s)
{
    QStringList sl = SearchCriterion::_deserializeStringList(s);
    from = QDateTime::fromString(sl[0], "yyyyMMdd.HHmm");
    to = QDateTime::fromString(sl[1], "yyyyMMdd.HHmm");
}

QString sh::search::criteria::MtimeSearchCriterion::_serialize()
{
    QStringList l;
    l << from.toString("yyyyMMdd.HHmm");
    l << to.toString("yyyyMMdd.HHmm");
    return SearchCriterion::_serializeStringList(l);
}

sh::search::criteria::MtimeSearchCriterion_QDateTimeEdit::MtimeSearchCriterion_QDateTimeEdit(QWidget* parent)
                : QDateTimeEdit(parent)
{
    QSize s = sizeHint();
    s.setWidth(s.width() + 50);
    setMinimumSize(s);
}
