/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QHBoxLayout>
#include <QKeyEvent>
#include <QLabel>
#include <QLineEdit>
#include <QShowEvent>
#include <QVBoxLayout>
#include "base/eurl.h"
#include "base/operation.h"
#include "base/filesystemnode.h"
#include "tools/mainthread.h"
#include "tools/iconmanager.h"
#include "ui/linkbutton.h"
#include "ui/mainwindow.h"
#include "search.h"
#include "searchcustomui.h"
#include "searchcriterionfactory.h"
#include "searchconfiguration.h"
#include "searchmanager.h"
#include "searchcriterion.h"


sh::search::SearchCustomUi::SearchCustomUi(std::shared_ptr<base::FilesystemNode> node, QWidget *parent)
            : QWidget(parent), _eurl(node->eurl()), _node(node)
{
    QVBoxLayout* wl = new QVBoxLayout(this);
    wl->setMargin(2);
    wl->setSpacing(0);
    QLabel* lbl = new QLabel(this);
    wl->addWidget(lbl);
    QWidget* buttons = new QWidget(this);
    QHBoxLayout* buttonsl = new QHBoxLayout(buttons);
    buttonsl->setMargin(0);
    wl->addWidget(buttons);
    btnMode = new sh::ui::LinkButton(buttons);
    btnMode->setFocusPolicy(Qt::ClickFocus);
    connect(btnMode, &ui::LinkButton::clicked, this, &SearchCustomUi::modeChangeMenu);
    buttonsl->addWidget(btnMode);
    auxbtnpanel = new QWidget(buttons);
    QHBoxLayout* auxbtnlayout = new QHBoxLayout(auxbtnpanel);
    auxbtnlayout->setMargin(0);
    buttonsl->addWidget(auxbtnpanel);
    buttonsl->addStretch(1);
    chkTree = new QCheckBox(buttons);
    chkTree->setFocusPolicy(Qt::ClickFocus);
    chkTree->setText(tr("tree"));
    buttonsl->addWidget(chkTree);
    btnSearch = new sh::ui::LinkButton(buttons);
    btnSearch->setSizeByFactor(130);
    btnSearch->setIcon(sh::tools::IconManager::instance()->getIcon("edit-find"));
    btnSearch->setFocusPolicy(Qt::ClickFocus);
    connect(btnSearch, &ui::LinkButton::clicked, this, &SearchCustomUi::startSearch);
    buttonsl->addWidget(btnSearch);
    sh::ui::LinkButton* btnClose = new sh::ui::LinkButton(buttons);
    btnClose->setText(tr("Close"));
    btnClose->setSizeByFactor(110);
    btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    btnClose->setFocusPolicy(Qt::ClickFocus);
    connect(btnClose, &ui::LinkButton::clicked, this, &SearchCustomUi::closeSearch);
    buttonsl->addWidget(btnClose);
    cfg = sh::search::SearchConfiguration::deserialize(_eurl->hostname());
    populateUI();
    connect(chkTree, &QCheckBox::toggled, this, [=](){ cfg->showAsTree = chkTree->isChecked(); });

    std::shared_ptr<sh::search::Search> _search = sh::search::SearchManager::instance()->findSearch(_eurl);
    if (_search)
    {
        sh::search::Search* search = _search.get();
        connect(search, &sh::search::Search::isSearchingChanged, this, [=](){
            setLoadingInfoVisible(search->isSearching());
        });
        setLoadingInfoVisible(search->isSearching());
    }
    else
        setLoadingInfoVisible(false);
}

sh::ui::LinkButton* sh::search::SearchCustomUi::makeAuxButton()
{
    sh::ui::LinkButton* btn = new sh::ui::LinkButton(auxbtnpanel);
    btn->setFocusPolicy(Qt::ClickFocus);
    auxbtnpanel->layout()->addWidget(btn);
    return btn;
}

void sh::search::SearchCustomUi::populateUI()
{
    chkTree->setChecked(cfg->showAsTree);
    while(!auxbtnpanel->layout()->isEmpty())
        delete auxbtnpanel->layout()->itemAt(0)->widget();
    if (cfg->criteria.isEmpty())
        cfg->criteria.append( sh::search::SearchManager::instance()->factories().first()->construct());
    if (cfg->criteria.length() == 1)
    {
        std::shared_ptr<sh::search::SearchCriterion> criterion = cfg->criteria.first();
        delete layout()->itemAt(0)->widget();
        QWidget* newwidget = criterion->factory()->editor(criterion, [=](){return makeAuxButton();});
        newwidget->installEventFilter(this);
        dynamic_cast<QVBoxLayout*>(layout())->insertWidget(0, newwidget);
        btnMode->setText(tr("Search by: %1 (&change)").arg(criterion->factory()->description()));
    }
    else
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().detailsAreInteresting().details(tr("More than one search criterion is currently not supported.")));
}

bool sh::search::SearchCustomUi::eventFilter(QObject*, QEvent* e)
{
    if (e->type() == QEvent::Show)
    {
        if (_eurl->hostname().isEmpty())
        {
            QWidget* w = layout()->itemAt(0)->widget();
            while (w && w->layout() && !w->layout()->isEmpty())
                w = w->layout()->itemAt(0)->widget();
            if (w)
                w->setFocus();
        }
    }
    return false;
}

void sh::search::SearchCustomUi::modeChangeMenu()
{
    QMenu menu;
    int i=0;
    sh::base::Operation op;
    foreach(std::shared_ptr<sh::search::SearchCriterionFactory> fact, sh::search::SearchManager::instance()->factories())
    {
        if (fact->isVisibleFor(&op, _node->parentnode()))
        {
            QAction* a = menu.addAction(fact->description());
            connect(a, &QAction::triggered, this, ([&](int _i){ return [=](){
                std::shared_ptr<sh::search::SearchCriterionFactory> fact = sh::search::SearchManager::instance()->factories().at(_i);
                cfg->criteria.clear();
                cfg->criteria.append(fact->construct());
                populateUI();
            };})(i) );
        }
        i++;
    }
    menu.exec(btnMode->mapToGlobal(QPoint(0, btnMode->height())));
}

void sh::search::SearchCustomUi::startSearch()
{
    foreach(std::shared_ptr<sh::search::SearchCriterion> c, cfg->criteria)
        c->factory()->editorupdateconfig(layout()->itemAt(0)->widget(), c);
    std::shared_ptr<sh::search::Search> search = sh::search::SearchManager::instance()->requestSearch(_eurl->outermostInnerEurl(), cfg->serialize(), _node->parentnode());
    sh::tools::MainThread::dispatcher()->invokeAsync([=](){
        sh::ui::MainWindow::mainWindow()->jumpToNode(search->searchnode());
    });
    std::shared_ptr<sh::search::Search> osearch = sh::search::SearchManager::instance()->findSearch(_eurl);
    if (osearch && osearch!=search)
        sh::search::SearchManager::instance()->removeSearch(osearch);
}

void sh::search::SearchCustomUi::setLoadingInfoVisible(bool v)
{
    btnSearch->setText(v ? tr("Searching ...") : tr("Search"));
}

void sh::search::SearchCustomUi::closeSearch()
{
    std::shared_ptr<sh::search::Search> osearch = sh::search::SearchManager::instance()->findSearch(_eurl);
    if (osearch)
    {
        std::shared_ptr<sh::base::FilesystemNode> pnode = osearch->searchnode()->parentnode();
        sh::search::SearchManager::instance()->removeSearch(osearch);
        sh::tools::MainThread::dispatcher()->invokeAsync([=](){
            sh::ui::MainWindow::mainWindow()->jumpToNode(pnode);
        });
    }
}

void sh::search::SearchCustomUi::keyPressEvent(QKeyEvent* e)
{
    if (e->key() == Qt::Key_Enter || e->key() == Qt::Key_Return)
        startSearch();
    else if (e->key() == Qt::Key_Escape)
        closeSearch();
}

sh::search::SearchCustomUi::~SearchCustomUi()
{
    delete cfg;
}
