/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/filesystemnode.h"
#include "base/filesystemmodel.h"
#include "base/filesystemhandlerregister.h"
#include "tools/singletoninitializer.h"
#include "tools/iconmanager.h"
#include "ui/linkbutton.h"
#include "searchfilesystemhandler.h"
#include "searchconfiguration.h"
#include "searchcustomui.h"


namespace sh{namespace search{INITIALIZE_SINGLETON_IN_MAINTHREAD(90000, SearchFilesystemHandler)}}

sh::search::SearchFilesystemHandler::SearchFilesystemHandler(sh::base::FilesystemModel *model)
        : FilesystemHandler(model)
{
}

void sh::search::SearchFilesystemHandler::getreal(std::shared_ptr<const sh::base::Eurl> seurl, std::shared_ptr<sh::base::FilesystemHandler>* handler, std::shared_ptr<const sh::base::Eurl>* eurl)
{
    *eurl = seurl->outermostInnerEurl()->withAppendedSegments(seurl->path());
    *handler = sh::base::FilesystemHandlerRegister::instance()->findHandler((*eurl)->scheme());
}

void sh::search::SearchFilesystemHandler::itemlist(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, sh::base::FilesystemNodeType, sh::base::FilesystemNodeListEditor*)
{
}

sh::base::FilesystemNodeType sh::search::SearchFilesystemHandler::getType(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QList<std::shared_ptr<sh::base::FilesystemNode>> nl = sh::base::FilesystemModel::instance()->tryGetNodesForEurl(eurl);
    if (!nl.isEmpty())
        return nl.first()->nodetype();
    return sh::base::FilesystemNodeType::NONE;
}

qint64 sh::search::SearchFilesystemHandler::getSize(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return 0;
}

QDateTime sh::search::SearchFilesystemHandler::getMtime(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    std::shared_ptr<sh::base::FilesystemHandler> ehandler;
    std::shared_ptr<const sh::base::Eurl> eeurl;
    getreal(eurl, &ehandler, &eeurl);
    return ehandler->getMtime(op, eeurl);
}

void sh::search::SearchFilesystemHandler::setMtime(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QDateTime)
{
}

QString sh::search::SearchFilesystemHandler::getMimetype(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return "application/x-shallot-search-result";
}

QString sh::search::SearchFilesystemHandler::getLinkTarget(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return QString();
}

bool sh::search::SearchFilesystemHandler::canGetFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

std::shared_ptr<QIODevice> sh::search::SearchFilesystemHandler::getFileContent(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return 0;
}

bool sh::search::SearchFilesystemHandler::canCreateDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createDirectory(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
}

bool sh::search::SearchFilesystemHandler::canCreateLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createLink(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
}

bool sh::search::SearchFilesystemHandler::canCreateFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::createFile(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QIODevice*, HandlerTransfer*)
{
}

bool sh::search::SearchFilesystemHandler::canDeleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::deleteItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
}

bool sh::search::SearchFilesystemHandler::canRenameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>)
{
    return false;
}

void sh::search::SearchFilesystemHandler::renameItem(sh::base::Operation*, std::shared_ptr<const sh::base::Eurl>, QString)
{
}

QList<std::shared_ptr<sh::actions::AbstractActionItem> > sh::search::SearchFilesystemHandler::getActions(const QList<std::shared_ptr<const sh::base::Eurl>>)
{
    return QList<std::shared_ptr<sh::actions::AbstractActionItem> >();
}

void sh::search::SearchFilesystemHandler::getCustomUi(std::shared_ptr<sh::base::FilesystemNode> node, QString* header, std::shared_ptr<QWidget>* customui)
{
    if (node->eurl()->outerUrlIsRootDirectory())
    {
        *header = tr("Search");
        QWidget* w = new sh::search::SearchCustomUi(node);
        *customui = std::shared_ptr<QWidget>(w);
    }
}

void sh::search::SearchFilesystemHandler::configureItems(sh::base::Operation*, QList<std::shared_ptr<sh::base::FilesystemNode> > nodes)
{
    foreach(std::shared_ptr<sh::base::FilesystemNode> node, nodes)
    {
        if (node->eurl()->outerUrlIsRootDirectory())
        {
            sh::search::SearchConfiguration* cfg = sh::search::SearchConfiguration::deserialize(node->eurl()->hostname());
            if (cfg->criteria.isEmpty())
                node->setDisplayName(tr("Search"));
            else
            {
                QString dsc = cfg->criteria.first()->valuedescription();
                if (dsc.length()>9)
                    dsc = dsc.left(7) + "...";
                QString andmore = (cfg->criteria.length()>1) ? tr(" &more") : "";
                node->setDisplayName(tr("Search '%1'%2").arg(dsc).arg(andmore));
            }
            node->setIcon(sh::tools::IconManager::instance()->getIcon("edit-find"));
            delete cfg;
        }
    }
}

std::shared_ptr<sh::search::SearchFilesystemHandler> sh::search::SearchFilesystemHandler::instance()
{
    static std::shared_ptr<sh::search::SearchFilesystemHandler> _instance = std::make_shared<sh::search::SearchFilesystemHandler>(sh::base::FilesystemModel::instance());
    return _instance;
}

void sh::search::SearchFilesystemHandler::_initialize()
{
    sh::base::FilesystemHandlerRegister::instance()->addHandler("search", SearchFilesystemHandler::instance());
}

void sh::search::SearchFilesystemHandler::_shutdown()
{
}
