/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SEARCHMANAGER_H
#define SEARCHMANAGER_H

#include <memory>
#include <QList>
#include <QMap>
#include <QMutex>
#include <QObject>


namespace sh{
    namespace base{ class Eurl; class FilesystemNode; }

    namespace search{

        class Search;
        class SearchCriterionFactory;

        /** \brief
         * Manager for searches in the filesystem.
         */
        class SearchManager : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static SearchManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** @brief Base value for display indexes of core (i.e. very important) search criteria.
             *
             * Used in SearchManager::addFactory.
             */
            static const int REGISTER_SEARCHCRITERION_INDEX_CORE;
            /** @brief Base value for display indexes of search criteria with normal importance.
             *
             * Used in SearchManager::addFactory.
             */
            static const int REGISTER_SEARCHCRITERION_INDEX_NORMAL;
            /** @brief Base value for display indexes of exotic search criteria.
             *
             * Used in SearchManager::addFactory.
             */
            static const int REGISTER_SEARCHCRITERION_INDEX_EXOTIC;


            std::shared_ptr<sh::search::Search> requestSearch(std::shared_ptr<const sh::base::Eurl> eurl, QString searchconfig, std::shared_ptr<sh::base::FilesystemNode> searchrootnode);
            void removeSearch(std::shared_ptr<sh::search::Search> search);

            std::shared_ptr<sh::search::Search> findSearch(std::shared_ptr<const sh::base::Eurl> eurl);

            virtual ~SearchManager();

            void addFactory(int index, std::shared_ptr<SearchCriterionFactory> f);
            QList<std::shared_ptr<SearchCriterionFactory>> factories();

        private:
            QList<std::shared_ptr<sh::search::Search>> _searches;
            QList<std::weak_ptr<sh::search::Search>> _weak_searches;
            SearchManager();
            QMap<int, std::shared_ptr<SearchCriterionFactory>> _factories;

        };

    }
}

#endif // SEARCHMANAGER_H
