/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SETTING_H
#define SETTING_H

#include <memory>
#include <QObject>
#include <QString>
#include "settings/settingsregistration.h" //convenience include


namespace sh{
    namespace ui{
        namespace fileview{class FileViewControl;}
    }

    namespace settings{

        /** \brief
         * Groups of settings.
         *
         * A choice here does not make a logical impact. It is merely a matter of graphical presentation.
         */
        enum SettingGroup
        {
            GLOBAL =1,
            GUI =2,
            FILEHANDLING =3,
            DATAVIEW =4,
            BEHAVIOR =5,
            SPECIAL =6
        };

        /** \brief
         * Abstract base class for a Shallot setting.
         *
         * Those have the greatest flexibility. They have to be stored manually by the user with many options.
         * See Shallot documentation for details.
         *
         * Use sh::settings::SettingsRegistration for registering an implementation.
         */
        class Setting
        {
        public:
            /** \brief Is (for subclasses) intended to be directly constructed from everywhere or by
             * registering a factory somewhere.
             */
            Setting();
            virtual ~Setting();
            /** \brief
             * Gets the internal name.
             */
            virtual QString name() = 0;
            /** \brief
             * Gets the description text.
             */
            virtual QString description() = 0;
            /** \brief
             * Gets the group;
             */
            virtual SettingGroup group() = 0;
            /** \brief
             * Is this an advanced setting?
             */
            virtual bool isAdvancedSetting() = 0;
            /** \brief
             * Called from Shallot core when the value was set (for a per-fileview setting).
             */
            virtual void setValue( sh::ui::fileview::FileViewControl*, QString ) {}
            /** \brief
             * Called from Shallot core when the value was set (for a not-per-fileview setting).
             */
            virtual void setValue( QString ) {}
            /** \brief
             * Does this setting apply globally or just for a certain subtree of nodes?
             */
            virtual bool isGlobal() = 0;
            /** \brief
             * Does this setting apply for each fileview individually or for the complete main window?
             */
            virtual bool isPerFileview() = 0;
            /** \brief
             * Get the currently set value.
             */
            virtual QString getValue( sh::ui::fileview::FileViewControl* filelist ) = 0;
            /** \brief
             * Gets a human readable description text for a value.
             */
            virtual QString valueDescription(QString value);
            /** \brief
             * Low-level function which gets the description text of a group.
             */
            static QString getGroupDescription(int g);
        };

    }
}

#define HAS_SETTING_REGISTRATION private: static const sh::settings::SettingsRegistration* registration;

#define REGISTER_SETTING(c) const sh::settings::SettingsRegistration* c::registration = new sh::settings::SettingsRegistration(std::make_shared<c>());

#endif // SETTING_H
