/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SETTINGSMANAGER_H
#define SETTINGSMANAGER_H

#include <memory>
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QVariant>
#include <QXmlDefaultHandler>
#include <QXmlSimpleReader>
#include <QXmlStreamWriter>


namespace sh{
    namespace base{ class Eurl; }
    namespace ui {
        namespace fileview {class FileViewControl;}
    }

    namespace settings{

        class Setting;
        class SettingsRegistration;
        class ProfileNode;

        /** \brief
         * Manager for Shallot settings.
         *
         * This is the more flexible way of shallot customization.
         * It includes everything you see in the 'Manage saved settings' or 'Save settings'.
         *
         * See sh::settings::Setting for more.
         */
        class SettingsManager : public QObject
        {
            Q_OBJECT
            friend class sh::settings::SettingsRegistration;
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static SettingsManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            virtual ~SettingsManager();

            void applyPerEurlSettingsToFileView(sh::ui::fileview::FileViewControl *list);
            void applyGlobalSettingsToFileView(sh::ui::fileview::FileViewControl *list);
            void applyPerEurlSettingsToMainWindow();
            void applyGlobalSettingsToMainWindow();

            QList<sh::settings::ProfileNode *> getNodesForProfile(QString profile);
            sh::settings::ProfileNode* getNodeByNodeId(QString nodeId);
            std::shared_ptr<sh::settings::Setting> getSettingByName(QString name);
            QList<std::shared_ptr<sh::settings::Setting> > getAllSettings();
            QStringList getProfileList();
            void removeProfile(QString profile);
            void removeNode(QString nodeId);
            void storeProfile(std::shared_ptr<const sh::base::Eurl> eurl, QString profilename, QMap<QString,QString> values, QMap<QString,int> onlyinfileviewindex,
                              bool withSubfolders, QStringList inheritFrom);

        signals:
            void profilesChanged();

        private:
            class _SettingsFinderStructure
            {
            public:
                _SettingsFinderStructure( QMap<QString, QString> settings, QMap<QString, int> onlyinfileviewindexes, QStringList inheritFrom, bool withSubfolders );
                _SettingsFinderStructure( );
                QMap<QString, QString> _settings;
                QMap<QString, int> _onlyinfileviewindexes;
                QStringList _inheritFrom;
                bool _withSubfolders;
                int _onlyFileviewIndex;
            };

            class _MyHandler : public QXmlDefaultHandler
            {
            public:
                _MyHandler(sh::settings::ProfileNode* node, QString filename);
                bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
                bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);
            private:
                void throwError();
                bool _stateStarted = true;
                bool _stateInProfile = false;
                sh::settings::ProfileNode* _node;
                QString _filename;
            };

            void applyToFileView(sh::ui::fileview::FileViewControl *list, QMap<QString, QString> settings);
            void applyToMainWindow(QMap<QString, QString> settings);

            QMap<QString, QString> _getSettings(std::shared_ptr<const sh::base::Eurl> eurl, QString profilename, bool directNode, int fileviewindex);
            QList<_SettingsFinderStructure> _getSettings_flat(std::shared_ptr<const sh::base::Eurl> eurl, QString profilename);
            QMap<QString, QMap<QString, QString>> _getSettings_cache;
            explicit SettingsManager();
            void readStoredProfiles();
            QMap<QString, std::shared_ptr<sh::settings::Setting>> _settings;
            QMap<std::shared_ptr<const sh::base::Eurl>, QMap<QString, QList<sh::settings::ProfileNode*> > > _profileNodes;
            QMap<QString, sh::settings::ProfileNode*> _profileNodesById;
            QList<QString> _profileNames;
            void invalidateCache();

        };

    }
}

#endif // SETTINGSMANAGER_H
