/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifdef LIBSECRET

#include <tuple>
#include <libsecret-1/libsecret/secret.h>
#include <QSettings>
#include "configuration/configurationmanager.h"
#include "tools/logger.h"
#include "tools/singletoninitializer.h"
#include "accountsmanager.h"
#include "libsecretaccountsprovider.h"


namespace sh{ namespace tools{ namespace accounts{INITIALIZE_STATIC_IN_MAINTHREAD(100000, LibsecretAccountsProvider)}}}

sh::tools::accounts::LibsecretAccountsProvider::LibsecretAccountsProvider()
{
}

void sh::tools::accounts::LibsecretAccountsProvider::_initialize()
{
    AccountsManager::instance()->registerProvider(std::make_shared<LibsecretAccountsProvider>());
}

void sh::tools::accounts::LibsecretAccountsProvider::_shutdown()
{
}

void sh::tools::accounts::LibsecretAccountsProvider::findAccounts(std::shared_ptr<Account> pattern, QList<std::shared_ptr<Account>>& result)
{
    GError *error = NULL;
    GHashTable *attributes = g_hash_table_new(g_str_hash, g_str_equal);
    QByteArray cauthtype = pattern->authtype.toUtf8();
    QByteArray cdomain = pattern->domain.toUtf8();
    QByteArray cpath = pattern->path.toUtf8();
    QByteArray cprotocol = pattern->protocol.toUtf8();
    QByteArray cserver = pattern->server.toUtf8();
    QByteArray cserverport = QString::number(pattern->serverport).toUtf8();
    QByteArray cusername = pattern->username.toUtf8();
    if (!pattern->authtype.isEmpty())
        g_hash_table_insert(attributes, sauthtype.data(), cauthtype.data());
    if (!pattern->domain.isEmpty())
        g_hash_table_insert(attributes, sdomain.data(), cdomain.data());
    if (!pattern->path.isNull())
        g_hash_table_insert(attributes, spath.data(), cpath.data());
    if (!pattern->protocol.isEmpty())
        g_hash_table_insert(attributes, sprotocol.data(), cprotocol.data());
    if (!pattern->server.isEmpty())
        g_hash_table_insert(attributes, sserver.data(), cserver.data());
    if (pattern->serverport > 0)
        g_hash_table_insert(attributes, sserverport.data(), cserverport.data());
    if (!pattern->username.isNull())
        g_hash_table_insert(attributes, susername.data(), cusername.data());
    gchar* password = secret_password_lookupv_sync(SECRET_SCHEMA_COMPAT_NETWORK, attributes, NULL, &error);
    g_hash_table_destroy(attributes);
    if (error != NULL) {
        LOG_DEBUG(QString(error->message))
        g_error_free (error);
    } else if (password != NULL) {
        QString pwd = QString::fromUtf8(password);
        secret_password_free(password);
        auto newacct = std::make_shared<Account>(*pattern);
        newacct->authinfo = pwd.toUtf8();
        result.append(newacct);
    }
}

bool sh::tools::accounts::LibsecretAccountsProvider::storeAccount(std::shared_ptr<Account> account)
{
    if (account->authtype == "password") {  // maybe we support other later when we know to deal with them
        GError *error = NULL;
        GHashTable *attributes = g_hash_table_new(g_str_hash, g_str_equal);
        QByteArray cauthinfo = account->authinfo;
        QByteArray cauthtype = account->authtype.toUtf8();
        QByteArray cdomain = account->domain.toUtf8();
        QByteArray cpath = account->path.toUtf8();
        QByteArray cprotocol = account->protocol.toUtf8();
        QByteArray cserver = account->server.toUtf8();
        QByteArray cserverport = QString::number(account->serverport).toUtf8();
        QByteArray cusername = account->username.toUtf8();
        QString label = account->username + "@" + account->server;
        if (account->serverport > 0)
            label += ":" + QString::number(account->serverport);
        QByteArray clabel = label.toUtf8();
        g_hash_table_insert(attributes, sauthtype.data(), cauthtype.data());
        if (!account->domain.isEmpty())
            g_hash_table_insert(attributes, sdomain.data(), cdomain.data());
        if (!account->path.isNull())
            g_hash_table_insert(attributes, spath.data(), cpath.data());
        g_hash_table_insert(attributes, sprotocol.data(), cprotocol.data());
        g_hash_table_insert(attributes, sserver.data(), cserver.data());
        if (account->serverport > 0)
            g_hash_table_insert(attributes, sserverport.data(), cserverport.data());
        g_hash_table_insert(attributes, susername.data(), cusername.data());
        secret_password_storev_sync(SECRET_SCHEMA_COMPAT_NETWORK, attributes, SECRET_COLLECTION_DEFAULT, clabel.data(), cauthinfo.data(), NULL, &error);
        g_hash_table_destroy(attributes);
        if (error != NULL) {
            LOG_DEBUG(QString(error->message))
            g_error_free (error);
        }
        else
            return true;
    }
    return false;
}

#endif
