/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef BENCHMARKING_H
#define BENCHMARKING_H

#include <functional>
#include <QElapsedTimer>
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QStack>
#include <QTimer>

#ifdef BENCHMARKING
#define BENCHMARK_BEGIN(x,var) var = sh::tools::Benchmarking::instance()->beginBenchmark(x);
#define BENCHMARK_END(var) sh::tools::Benchmarking::instance()->endBenchmark(var);
#define BENCHMARK_SCOPE(x) auto _bench = sh::tools::ScopeBenchmarker(x);
#else
#define BENCHMARK_BEGIN(x,var)
#define BENCHMARK_END(var)
#define BENCHMARK_SCOPE(x)
#endif


namespace sh{
    namespace tools{

        /** \brief
         * Tools for performance measurements.
         */
        class Benchmarking : public QObject
        {
            Q_OBJECT
        private:
            explicit Benchmarking(QObject *parent = 0);
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            inline static Benchmarking* instance()
            {
                static Benchmarking* _instance = new Benchmarking();
                return _instance;
            }
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** \brief A pair of a benchmark keyname and a duration value.
             */
            typedef QPair<QString,qint64> BenchmarkFrame;

#ifdef BENCHMARKING

            inline int beginBenchmark(QString name)
            {
                QMutexLocker l(&mutex);
                benchTimestamps[_index_counter] = Benchmarking::BenchmarkFrame(name, elapsedTimer.nsecsElapsed());
                return _index_counter++;
            }

            inline void endBenchmark(int id)
            {
                QMutexLocker _l(&mutex);
                qint64 now = elapsedTimer.nsecsElapsed();
                auto frameiter = benchTimestamps.find(id);
                QString name = frameiter->first;
                qint64 acc = now - frameiter->second;
                benchTimestamps.erase(frameiter);
                benchResults[name] += acc; // autom. initialized to 0
            }

            void addReporter(std::function<QMap<QString,QString>()> reporter);

        private:
            QString report();
            QList<std::function<QMap<QString,QString>()>> reporters;
            int _index_counter=0;
            QTimer* benchmarkReporter;
            QMap<QString, qint64> benchResults;
            QMap<int, BenchmarkFrame> benchTimestamps;
            QElapsedTimer elapsedTimer;
            QMutex mutex;
#endif
        };

#ifdef BENCHMARKING
        class ScopeBenchmarker
        {
        public:
            int x;
            ScopeBenchmarker(QString name) {
                BENCHMARK_BEGIN(name,int x)
                this->x = x;
            }
            ~ScopeBenchmarker() {
                BENCHMARK_END(x)
            }
        };
#endif

    }
}

#endif // BENCHMARKING_H
