/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/eurl.h"
#include "configuration/configurationmanager.h"
#include "configuration/configurationvalue.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "bookmarkmanager.h"


namespace sh{namespace tools{ INITIALIZE_SINGLETON_IN_MAINTHREAD(50300, BookmarkManager)}}

sh::tools::BookmarkManager::BookmarkManager() : QObject(), _bookmarksmutex(QMutex::Recursive)
{
    cfgvalBookmarks = sh::configuration::ConfigurationManager::instance()->registerConfigValue(
                "sh::tools::BookmarkManager.Bookmarks", "",
                sh::configuration::ConfigurationValue::ValueTypeString());
    cfgvalBookmarks->consumeValue(this, [=](){
        readBookmarks();
    });
}

void sh::tools::BookmarkManager::readBookmarks()
{
    QMutexLocker _l1(&_bookmarksmutex);
    _bookmarks.clear();
    QString scfgval = cfgvalBookmarks->getConfigValueString();
    QSet<QString> names;
    if (!scfgval.isEmpty())
    {
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            QStringList lst = scfgval.split("&");
            while (lst.length() >= 5) // <<< change this as well when new properties are introduced!
            {
                QString name = QUrl::fromPercentEncoding(lst.takeFirst().toUtf8());
                QString folder = QUrl::fromPercentEncoding(lst.takeFirst().toUtf8());
                QString label = QUrl::fromPercentEncoding(lst.takeFirst().toUtf8());
                QString eurl = QUrl::fromPercentEncoding(lst.takeFirst().toUtf8());
                QString tags = QUrl::fromPercentEncoding(lst.takeFirst().toUtf8());
                QStringList lfolder;
                if (!folder.isEmpty())
                    lfolder = folder.split('/');
                if (!names.contains(name))
                {
                    names.insert(name);
                    _bookmarks.append(std::make_shared<sh::tools::Bookmark>(name, lfolder, label, sh::base::Eurl::fromString(eurl), tags));
                }
            }
        });
    }
    emit changed();
}

void sh::tools::BookmarkManager::writeBookmarks()
{
    static QByteArray includePercentEnc("&");
    static QByteArray excludePercentEnc("");
    QStringList result;
    {
        QMutexLocker _l1(&_bookmarksmutex);
        foreach(std::shared_ptr<Bookmark> bookmark, _bookmarks)
        {
            result << QUrl::toPercentEncoding(bookmark->id(), excludePercentEnc, includePercentEnc);
            result << QUrl::toPercentEncoding(bookmark->folder().join('/'), excludePercentEnc, includePercentEnc);
            result << QUrl::toPercentEncoding(bookmark->label(), excludePercentEnc, includePercentEnc);
            result << QUrl::toPercentEncoding(bookmark->eurl()->asString(), excludePercentEnc, includePercentEnc);
            result << QUrl::toPercentEncoding(bookmark->tags(), excludePercentEnc, includePercentEnc);
        }
    }
    cfgvalBookmarks->setConfigValue(result.join('&'));
}

QList<std::shared_ptr<sh::tools::Bookmark> > sh::tools::BookmarkManager::getBookmarks()
{
    QMutexLocker _l1(&_bookmarksmutex);
    return _bookmarks;
}

bool sh::tools::BookmarkManager::hasBookmarks()
{
    QMutexLocker _l1(&_bookmarksmutex);
    return !_bookmarks.isEmpty();
}

QString sh::tools::BookmarkManager::addBookmark(QList<QString> folder, QString label, std::shared_ptr<const sh::base::Eurl> eurl, QString tags)
{
    QMutexLocker _l1(&_bookmarksmutex);
    sh::configuration::ConfigurationManager::ConfigFilesLock sfl;
    readBookmarks();
    QString id;
    while (id.isEmpty())
    {
        id = QString("%1").arg(qrand());
        for(auto itr = _bookmarks.begin(); itr!=_bookmarks.end(); itr++)
            if ((*itr)->id() == id)
            {
                id = QString();
                break;
            }
    }
    QStringList _folder;
    foreach(QString s, folder)
        _folder << s.replace("/", "_");
    _bookmarks.append(std::make_shared<sh::tools::Bookmark>(id, _folder, label, eurl, tags));
    writeBookmarks();
    emit changed();
    return id;
}

void sh::tools::BookmarkManager::removeBookmark(QString id)
{
    QMutexLocker _l1(&_bookmarksmutex);
    sh::configuration::ConfigurationManager::ConfigFilesLock sfl;
    readBookmarks();
    for(auto itr = _bookmarks.begin(); itr!=_bookmarks.end(); itr++)
        if ((*itr)->id() == id)
        {
            _bookmarks.erase(itr);
            break;
        }
    writeBookmarks();
    emit changed();
}

void sh::tools::BookmarkManager::_changeBookmark(QString id,
                                                 std::function<QString(std::shared_ptr<Bookmark>)> label,
                                                 std::function<std::shared_ptr<const sh::base::Eurl>(std::shared_ptr<Bookmark>)> eurl,
                                                 std::function<QList<QString>(std::shared_ptr<Bookmark>)> folder,
                                                 std::function<QString(std::shared_ptr<Bookmark>)> tags)
{
    QMutexLocker _l1(&_bookmarksmutex);
    sh::configuration::ConfigurationManager::ConfigFilesLock sfl;
    readBookmarks();
    for(auto itr = _bookmarks.begin(); itr!=_bookmarks.end(); itr++)
        if ((*itr)->id() == id)
        {
            _bookmarks.insert(_bookmarks.erase(itr), std::make_shared<sh::tools::Bookmark>(id, folder(*itr), label(*itr), eurl(*itr), tags(*itr)));
            break;
        }
    writeBookmarks();
    emit changed();
}

void sh::tools::BookmarkManager::changeBookmark(QString id, QString label, std::shared_ptr<const sh::base::Eurl> eurl)
{
    _changeBookmark(id, [=](std::shared_ptr<Bookmark>){return label;},
                        [=](std::shared_ptr<Bookmark>){return eurl;},
                        [=](std::shared_ptr<Bookmark> b){return b->folder();},
                        [=](std::shared_ptr<Bookmark> b){return b->tags();});
}

void sh::tools::BookmarkManager::changeBookmarkTags(QString id, QString tags)
{
    _changeBookmark(id, [=](std::shared_ptr<Bookmark> b){return b->label();},
                        [=](std::shared_ptr<Bookmark> b){return b->eurl();},
                        [=](std::shared_ptr<Bookmark> b){return b->folder();},
                        [=](std::shared_ptr<Bookmark>){return tags;});
}

void sh::tools::BookmarkManager::moveBookmarkUp(QString id)
{
    moveBookmark(id, -1);
}

void sh::tools::BookmarkManager::moveBookmarkDown(QString id)
{
    moveBookmark(id, 1);
}

void sh::tools::BookmarkManager::moveBookmark(QString id, int direction)
{
    QMutexLocker _l1(&_bookmarksmutex);
    sh::configuration::ConfigurationManager::ConfigFilesLock sfl;
    readBookmarks();
    int pos = -1;
    for (int i=0;i<_bookmarks.length();i++)
    {
        if (_bookmarks[i]->id()==id)
        {
            pos = i;
            break;
        }
    }
    if (pos > -1)
    {
        int dpos = pos;
        dpos += direction;
        while (dpos>0 && dpos<_bookmarks.length() && _bookmarks[dpos]->folder().join('/') != _bookmarks[pos]->folder().join('/') )
        {
            dpos += direction;
        }
        if (direction>0)
            dpos += direction;
        if (dpos < 0)
            dpos = 0;
        if (dpos > _bookmarks.length())
            dpos = _bookmarks.length();
        _bookmarks.insert(dpos, _bookmarks[pos]);
        _bookmarks.removeAt( (direction<0) ? (pos+1) : pos);
    }
    writeBookmarks();
    emit changed();
}

void sh::tools::BookmarkManager::moveBookmarkToFolder(QString id, QList<QString> folder)
{
    _changeBookmark(id, [=](std::shared_ptr<Bookmark> b){return b->label();},
                        [=](std::shared_ptr<Bookmark> b){return b->eurl();},
                        [=](std::shared_ptr<Bookmark>){return folder;},
                        [=](std::shared_ptr<Bookmark> b){return b->tags();});
}

void sh::tools::BookmarkManager::_initialize()
{
}

void sh::tools::BookmarkManager::_shutdown()
{
}

sh::tools::Bookmark::Bookmark(QString id, QStringList folder, QString label, std::shared_ptr<const sh::base::Eurl> eurl, QString tags) : _id(id), _folder(folder), _label(label), _eurl(eurl), _tags(tags)
{
}

QString sh::tools::Bookmark::id()
{
    return _id;
}

QList<QString> sh::tools::Bookmark::folder()
{
    return _folder;
}

QString sh::tools::Bookmark::label()
{
    return _label;
}

std::shared_ptr<const sh::base::Eurl> sh::tools::Bookmark::eurl()
{
    return _eurl;
}

QString sh::tools::Bookmark::tags()
{
    return _tags;
}
