/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actioncopytree.h"
#include "actions/common/actionmovetree.h"
#include "base/eurl.h"
#include "exceptions/argumentexception.h"
#include "tools/singletoninitializer.h"
#include "tools/threaddispatcher.h"
#include "dataexchange.h"


namespace sh{namespace tools{INITIALIZE_SINGLETON_IN_MAINTHREAD(80010, DataExchange)}}

QString sh::tools::DataExchange::FilelistTypeShallot = "x-special/shallot-copied-files";
QString sh::tools::DataExchange::FilelistTypeGnome = "x-special/gnome-copied-files";
QString sh::tools::DataExchange::FilelistTypeUrilist = "text/uri-list";
QString sh::tools::DataExchange::FilelistTypePlaintext = "text/plain";

#ifdef Q_OS_WIN
QString sh::tools::DataExchange::linebreak = "\r\n";
#else
QString sh::tools::DataExchange::linebreak = "\n";
#endif

sh::tools::DataExchange::DataExchange()
{
}

sh::tools::DataExchange* sh::tools::DataExchange::instance()
{
    static sh::tools::DataExchange* _instance = new sh::tools::DataExchange();
    return _instance;
}

void sh::tools::DataExchange::_initialize()
{
}

void sh::tools::DataExchange::_shutdown()
{
}

bool sh::tools::DataExchange::containsFileEntries(const QMimeData* data)
{
    CHECK_CORRECT_CALLER_THREAD
    return data->hasFormat(FilelistTypeGnome)
            || data->hasFormat(FilelistTypeShallot) || data->hasFormat(FilelistTypeUrilist);
}

QMimeData* sh::tools::DataExchange::getMimeDataFromFilesystemNodes(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes, sh::tools::DataExchange::DataExchangeType type)
{
    CHECK_CORRECT_CALLER_THREAD
    QMimeData* mimeData = new QMimeData;
    QStringList urls;
    QStringList urlsNative;
    bool hasinners = false;
    foreach (std::shared_ptr<sh::base::FilesystemNode> node, nodes) {
        std::shared_ptr<const sh::base::Eurl> eurl = node->eurl();
        if (!eurl->hasInnerUrls() && node->handler()->isWellKnownScheme())
            urls.append(eurl->asString());
        else
            hasinners = true;
        urlsNative.append(eurl->asString());
    }
    QString typestring;
    switch (type) {
    case sh::tools::DataExchange::DataExchangeType::COPY:
        typestring="copy";
        break;
    case sh::tools::DataExchange::DataExchangeType::MOVE:
        typestring="cut";
        break;
    default:
        throw sh::exceptions::ArgumentException(sh::exceptions::ExceptionData().details("unallowed value for 'type' parameter"));
        break;
    }
    QString nativeurllist = urlsNative.join(linebreak);
    QByteArray forShallot = (typestring+linebreak+nativeurllist).toUtf8();
    QByteArray forPlaintext = (nativeurllist).toUtf8();
    mimeData->setData(FilelistTypeShallot, forShallot);
    mimeData->setData(FilelistTypePlaintext, forPlaintext);
    if (!hasinners) {
        QString urllist = urls.join(linebreak);
        QByteArray forGnome = (typestring+linebreak+urllist).toUtf8();
        QByteArray forFilelist = (urllist).toUtf8();
        mimeData->setData(FilelistTypeGnome, forGnome);
        mimeData->setData(FilelistTypeUrilist, forFilelist);
    }
    return mimeData;
}

QList<std::shared_ptr<const sh::base::Eurl>> sh::tools::DataExchange::getSources(const QMimeData *src, bool* isCut)
{
    CHECK_CORRECT_CALLER_THREAD
    QList<std::shared_ptr<const sh::base::Eurl>> srcs;
    QStringList filelist;
    if (isCut)
        *isCut = false;
    if (src->hasFormat(FilelistTypeShallot))
    {
        filelist = QString( src->data(FilelistTypeShallot) ).replace("\r\n","\n").split("\n");
        if (isCut)
            *isCut = filelist.first().trimmed()=="cut";
        filelist.removeFirst();
    }
    else if (src->hasFormat(FilelistTypeGnome))
    {
        filelist = QString( src->data(FilelistTypeGnome) ).replace("\r\n","\n").split("\n");
        if (isCut)
            *isCut = filelist.first().trimmed()=="cut";
        filelist.removeFirst();
    }
    else if (src->hasFormat(FilelistTypeUrilist))
    {
        filelist = QString( src->data(FilelistTypeUrilist) ).replace("\r\n","\n").split("\n");
    }
    foreach(QString file, filelist)
    {
        if (!file.trimmed().isEmpty())
            srcs.append( sh::base::Eurl::fromString(file) );
    }
    return srcs;
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::tools::DataExchange::createPasteAction(const QMimeData* src, std::shared_ptr<const sh::base::Eurl> dest)
{
    CHECK_CORRECT_CALLER_THREAD
    bool isCut;
    QList<std::shared_ptr<const sh::base::Eurl>> sources = getSources(src, &isCut);
    if (isCut)
        return std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::common::ActionMoveTree(sources, dest));
    else
        return std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::common::ActionCopyTree(sources, dest));
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::tools::DataExchange::createCopyAction(const QMimeData* src, std::shared_ptr<const sh::base::Eurl> dest)
{
    CHECK_CORRECT_CALLER_THREAD
    return std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::common::ActionCopyTree(getSources(src), dest));
}

std::shared_ptr<sh::actions::AbstractActionItem> sh::tools::DataExchange::createMoveAction(const QMimeData* src, std::shared_ptr<const sh::base::Eurl> dest)
{
    CHECK_CORRECT_CALLER_THREAD
    return std::shared_ptr<sh::actions::AbstractActionItem>(new sh::actions::common::ActionMoveTree(getSources(src), dest));
}

