/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/eurl.h"
#include "base/operation.h"
#include "tools/filetypes/freedesktoporgmimetypeinformationretrievalstrategy.h"
#include "tools/filetypes/freedesktoporgtoolsmimetypedeterminationmethod.h"
#include "tools/filetypes/freedesktoporgtoolsopenmethoddeterminationmethod.h"
#include "tools/filetypes/suffixlistmimetypedeterminationmethod.h"
#include "tools/filetypes/unixfiletoolmimetypedeterminationmethod.h"
#include "tools/filetypes/userdefinedopenmethoddeterminationstrategy.h"
#include "tools/singletoninitializer.h"
#include "tools/iconmanager.h"
#include "filetypemanager.h"


namespace sh{namespace tools{namespace filetypes{INITIALIZE_SINGLETON_IN_MAINTHREAD(30010, FileTypeManager)}}}

sh::tools::filetypes::FileTypeManager::FileTypeManager(QObject *parent) :
    QObject(parent)
{
#ifdef Q_OS_UNIX
    _mimetypeDeterminationMethods.append(
                std::make_shared<sh::tools::filetypes::UnixFileToolMimetypeDeterminationStrategy>(this));
    _mimetypeDeterminationMethods.append(
                std::make_shared<sh::tools::filetypes::FreedesktopOrgToolsMimetypeDeterminationStrategy>(this));
    _openMethodDeterminationMethods.append(
                std::make_shared<sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy>());
    _mimetypeInformationRetrievalMethods.append(
                std::make_shared<sh::tools::filetypes::FreeDesktopOrgMimetypeInformationRetrievalStrategy>());
#endif
    _openMethodDeterminationMethods.append(sh::tools::filetypes::UserDefinedOpenMethodDeterminationStrategy::instance());
    _mimetypeDeterminationMethods.append(
                std::make_shared<sh::tools::filetypes::SuffixListMimetypeDeterminationStrategy>());
}

sh::tools::filetypes::OpenMethod::OpenMethod(QString name, QString command, QStringList arguments, QIcon icon, int precedence)
    :name(name), command(command), arguments(arguments), icon(icon.isNull() ? sh::tools::IconManager::instance()->getIcon("empty") : icon), precedence(precedence)
{
}

sh::tools::filetypes::FileTypeManager* sh::tools::filetypes::FileTypeManager::instance()
{
    static sh::tools::filetypes::FileTypeManager* _instance = new sh::tools::filetypes::FileTypeManager();
    return _instance;
}

void sh::tools::filetypes::FileTypeManager::_initialize()
{
}

void sh::tools::filetypes::FileTypeManager::_shutdown()
{
}

QString sh::tools::filetypes::FileTypeManager::determineMimetype(sh::base::Operation* op, std::shared_ptr<const sh::base::Eurl> eurl)
{
    QList<std::shared_ptr<const sh::base::Eurl>> l;
    l.append(eurl);
    QMap<std::shared_ptr<const sh::base::Eurl>, QString> m = determineMimetype(op, l);
    return m[eurl];
}

QMap<std::shared_ptr<const sh::base::Eurl>, QString> sh::tools::filetypes::FileTypeManager::determineMimetype(sh::base::Operation* op, QList<std::shared_ptr<const sh::base::Eurl>> items)
{
    QMap<std::shared_ptr<const sh::base::Eurl>, QString> result;
    foreach(auto method, _mimetypeDeterminationMethods)
    {
        QMap<std::shared_ptr<const sh::base::Eurl>, QString> _items = method->determineMimetype(op, items);
        foreach(std::shared_ptr<const sh::base::Eurl> _determined, _items.keys())
        {
            items.removeOne(_determined);
            result[_determined] = _items[_determined];
        }
        if (items.isEmpty())
            break;
    }
    foreach(std::shared_ptr<const sh::base::Eurl> _notDetermined, items)
        result[_notDetermined] = "application/octet-stream";
    return result;
}

QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > sh::tools::filetypes::FileTypeManager::getOpenMethods(QString mimetype, QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
{
    QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > result;
    QStringList allmimetypes;
    allmimetypes << mimetype;
    QStringList _mimetypes = allmimetypes;
    while (!_mimetypes.isEmpty()) {
        QStringList _omimetypes = _mimetypes;
        _mimetypes.clear();
        foreach(QString omimetype, _omimetypes)
            foreach(auto method, _mimetypeInformationRetrievalMethods)
                _mimetypes.append(method->getMimetypeSubclasses(omimetype));
        allmimetypes << _mimetypes;
    }
    foreach(QString _mimetype, allmimetypes)
        foreach(auto method, _openMethodDeterminationMethods) {
            auto oms = method->getOpenMethods(_mimetype, nodes);
            foreach (auto om, oms) {
                bool alreadyin = false;
                foreach (auto x, result) {
                    alreadyin = (x->name == om->name) && (x->command == om->command) && (x->arguments == om->arguments);
                    if (alreadyin)
                        break;
                }
                if (!alreadyin)
                    result.append(om);
            }
        }
    qStableSort(result.begin(), result.end(), [](const std::shared_ptr<sh::tools::filetypes::OpenMethod>& a, const std::shared_ptr<sh::tools::filetypes::OpenMethod>& b){
        if (a->precedence < b->precedence)
            return true;
        if (a->precedence == b->precedence) {
            if (a->command < b->command)
                return true;
            if (a->command == b->command && a->name < b->name)
                return true;
        }
        return false;
    });
    return result;
}

QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > sh::tools::filetypes::FileTypeManager::getAllOpenMethods()
{
    QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > result;
    foreach(auto method, _openMethodDeterminationMethods)
        result.append(method->getOpenMethods("", {}));
    qSort(result.begin(), result.end(), [](const std::shared_ptr<sh::tools::filetypes::OpenMethod>& a, const std::shared_ptr<sh::tools::filetypes::OpenMethod>& b){
        return a->name.toLower() < b->name.toLower();
    });
    return result;
}

sh::tools::filetypes::FileTypeManager::MimetypeDeterminationStrategy::~MimetypeDeterminationStrategy()
{

}

sh::tools::filetypes::FileTypeManager::OpenMethodDeterminationStrategy::~OpenMethodDeterminationStrategy()
{

}

QStringList sh::tools::filetypes::FileTypeManager::MimetypeInformationRetrievalStrategy::getMimetypeSubclasses(QString)
{
    return QStringList();
}

sh::tools::filetypes::FileTypeManager::MimetypeInformationRetrievalStrategy::~MimetypeInformationRetrievalStrategy()
{

}
