/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDir>
#include <QQueue>
#include <QSettings>
#include <QRegularExpression>
#include "exceptions/ioexception.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "tools/iconmanager.h"
#include "freedesktoporgtoolsopenmethoddeterminationmethod.h"


QString sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::_parseValue(QStringList content, QString key)
{
    foreach(QString line, content)
    {
        QString st = line.trimmed();
        if (st.startsWith(key))
        {
            QString vp = st.mid(key.length()).trimmed();
            if (vp.startsWith("="))
            {
                return vp.mid(1).trimmed();
            }
        }
    }
    return QString();
}

sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::FreedesktopOrgToolsOpenMethodDeterminationStrategy()
{
    sh::tools::ThreadPool::enqueueForce([=](){
        QString langFull = QLocale::system().name();
        int sepidx = langFull.indexOf("_");
        QString langPart = (sepidx>-1 ? langFull.mid(0,sepidx) : langFull);
        QString _nameLangGeneric = "Name";
        QString _nameLangFull = QString("Name[%1]").arg(langFull);
        QString _nameLangPart = QString("Name[%1]").arg(langPart);
        // read mimeapps.list
        QMap<QString, QStringList> addedAssociations;
        QString fsentry = QDir::homePath() + "/.config/mimeapps.list";
        QFile f(fsentry);
        if (f.open(QIODevice::ReadOnly))
        {
            sh::exceptions::Exception::executeGuarded([&](){
                QByteArray ba = f.readAll();
                if (f.error() != QFile::NoError)
                    throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Problems while reading '%1': %2").arg(fsentry).arg(f.errorString())).retryable(false));
                QStringList _cont = QString::fromUtf8(ba).split("\n");
                _cont << "[]";
                QStringList cont;
                bool ignorenext = true;
                foreach(QString _cline, _cont)
                {
                    if (_cline.startsWith("["))
                    {
                        if (!ignorenext)
                        {
                            foreach(QString cline, cont) {
                                int ie = cline.indexOf("=");
                                if (ie > 0) {
                                    QString mimetype = cline.mid(0, ie);
                                    QStringList desktopfiles = cline.mid(ie+1).split(";", QString::SkipEmptyParts);
                                    foreach(QString desktopfile, desktopfiles)
                                        addedAssociations[desktopfile].append(mimetype);
                                }
                            }
                        }
                        ignorenext = (!_cline.toLower().startsWith("[added associations]"));
                        cont.clear();
                    }
                    else
                    {
                        cont << _cline;
                    }
                }
            }, sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
        }
        // read .desktop files
        QQueue<QString> desktopFileLocations;
        desktopFileLocations << "/usr/share/applications";
        desktopFileLocations << QDir::homePath() + "/.local/share/applications";  // the user directory comes at last, since QMultiMap::find() leads to iteration from most recently inserted one to oldest one (at we want to iterate user stuff first)
        while(!desktopFileLocations.isEmpty())
        {
            QString desktopFileLocation = desktopFileLocations.dequeue();
            foreach(QString sentry, QDir(desktopFileLocation).entryList())
            {
                if (sentry != "." && sentry != "..")
                {
                    sh::exceptions::Exception::executeGuarded([&](){
                        QString fsentry = desktopFileLocation + QDir::separator() + sentry;
                        if (QFileInfo(fsentry).isDir())
                            desktopFileLocations << fsentry;
                        else
                        {
                            sh::exceptions::Exception::executeGuarded([&](){
                                QFile f(fsentry);
                                if (f.open(QIODevice::ReadOnly))
                                {
                                    QByteArray ba = f.readAll();
                                    if (f.error() != QFile::NoError)
                                        throw sh::exceptions::IOException(sh::exceptions::ExceptionData().detailsAreInteresting().details(QObject::tr("Problems while reading '%1': %2").arg(fsentry).arg(f.errorString())).retryable(false));
                                    QStringList _cont = QString::fromUtf8(ba).split("\n");
                                    _cont << "[]";
                                    QStringList cont;
                                    bool ignorenext = true;
                                    foreach(QString _cline, _cont)
                                    {
                                        if (_cline.startsWith("["))
                                        {
                                            if (!ignorenext)
                                            {
                                                QString name;
                                                name = _parseValue(cont, _nameLangFull);
                                                if (name.isEmpty())
                                                    name = _parseValue(cont, _nameLangPart);
                                                if (name.isEmpty())
                                                    name = _parseValue(cont, _nameLangGeneric);
                                                QString rmimetypes = _parseValue(cont, "MimeType");
                                                QString command = _parseValue(cont, "Exec");
                                                QString icon = _parseValue(cont, "Icon");
                                                QString hidden = _parseValue(cont, "Hidden");
                                                QStringList mimetypes = rmimetypes.split(";");
                                                for(int i=mimetypes.length()-1;i>=0;i--)
                                                {
                                                    if  (mimetypes[i].trimmed().isEmpty())
                                                        mimetypes.removeAt(i);
                                                }
                                                foreach(QString associatedMimetype, addedAssociations[sentry])
                                                    mimetypes << associatedMimetype;
                                                QString ecommand;
                                                QStringList ecommandargs;
                                                _parseExecLine(command, &ecommand, &ecommandargs);
                                                if (ecommandargs.contains("%f") || ecommandargs.contains("%F") || ecommandargs.contains("%u") || ecommandargs.contains("%U"))
                                                {
                                                    ApplicationEntry* entry = new ApplicationEntry();
                                                    entry->name = name;
                                                    entry->mimetypes = mimetypes;
                                                    entry->command = ecommand;
                                                    entry->commandargs = ecommandargs;
                                                    entry->hidden = hidden == "true";
                                                    if (!icon.isEmpty())
                                                        entry->icon = sh::tools::IconManager::instance()->getIcon(icon);
                                                    sh::tools::MainThread::dispatcher()->invokeSync([&](){
                                                        if (mimetypes.isEmpty())
                                                            mimetype2applicationEntry.insert(QString(), entry);
                                                        else
                                                            foreach(QString mimetype, mimetypes)
                                                                mimetype2applicationEntry.insert(mimetype, entry);
                                                    });
                                                }
                                            }
                                            ignorenext = (!_cline.toLower().startsWith("[desktop entry]"));
                                            cont.clear();
                                        }
                                        else
                                        {
                                            cont << _cline;
                                        }
                                    }
                                }
                            }, sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
                        }
                    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
                }
            }
        }
    });
}

sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::~FreedesktopOrgToolsOpenMethodDeterminationStrategy()
{
    QSet<ApplicationEntry*> entries;
    foreach(ApplicationEntry* e, mimetype2applicationEntry.values())
        entries.insert(e);
    foreach(ApplicationEntry* e, entries)
        delete e;
}

QString sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::_escapeExecLineToken(QString s)
{
    QString res = "";
    bool forcepick = false;
    foreach(QChar c, s)
    {
        if (forcepick)
        {
            res += c;
            forcepick = false;
        }
        else if (c == '\\')
            forcepick = true;
        else
            res += c;
    }
    return res;
}

void sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::_parseExecLine(QString execline, QString* command, QStringList* arguments)
{
    // not perfect but more or less according to the FreeDesktop specs
    execline = _escapeExecLineToken(execline);
    QStringList result;
    QRegularExpression re("([^\\\\])\"(.*[^\\\\])?\"");
    re.setPatternOptions(re.patternOptions() | QRegularExpression::InvertedGreedinessOption);
    QRegularExpressionMatchIterator iter = re.globalMatch(execline);
    int i = 0;
    while (iter.hasNext()) {
        QRegularExpressionMatch match = iter.next();
        int ostart = match.capturedStart(0);
        int oend = match.capturedEnd(0);
        QString prepadd = match.captured(1);
        int istart = match.capturedStart(2);
        int iend = match.capturedEnd(2);
        result.append( (execline.mid(i, ostart-i)+prepadd).trimmed().split(" ", QString::SkipEmptyParts) );
        result.append(execline.mid(istart, iend-istart));
        i=oend;
    }
    result.append(execline.mid(i).trimmed().split(" ", QString::SkipEmptyParts));
    QStringList result2;
    foreach(QString s, result)
        result2 << _escapeExecLineToken(s);

    if (result2.isEmpty())
        *command = "";
    else
    {
        *command = result2.first();
        result2.removeFirst();
    }
    *arguments = result2;
}

QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > sh::tools::filetypes::FreedesktopOrgToolsOpenMethodDeterminationStrategy::getOpenMethods(QString mimetype, QList<std::shared_ptr<sh::base::FilesystemNode>>)
{
    QList<std::shared_ptr<sh::tools::filetypes::OpenMethod> > result;
    QList<QStringList> seenkeys;
    QStringList seennames;
    bool returnall = mimetype.isEmpty();
    auto i = returnall ? mimetype2applicationEntry.begin() : mimetype2applicationEntry.find(mimetype);
    while (i != mimetype2applicationEntry.end() && (returnall || i.key() == mimetype)) {
        ApplicationEntry* e = *i;
        QStringList mykey;
        mykey << e->command;
        mykey << e->commandargs;
        bool seen = seennames.contains(e->name);
        if (!seen)
            foreach(QStringList k, seenkeys) {
                seen = k == mykey;
                if (seen)
                    break;
            }
        if (!seen)
        {
            seenkeys.append(mykey);
            if (!e->hidden) {
                seennames.append(e->name);
                result.append(std::shared_ptr<sh::tools::filetypes::OpenMethod> (new sh::tools::filetypes::OpenMethod( e->name ,e->command, e->commandargs, e->icon)));
            }
        }
        ++i;
    }
    return result;
}
