/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <errno.h>
#include "tools/logger.h"
#include "tools/mainthread.h"
#include "tools/singletoninitializer.h"
#include "localfilesystemwatcher.h"


namespace sh{namespace tools{INITIALIZE_SINGLETON_IN_MAINTHREAD(80070, LocalFilesystemWatcher)}}

#ifdef INOTIFY
#include <sys/inotify.h>
#include <unistd.h>
#define INOTIFY_BUF_LEN (10 * (sizeof(struct inotify_event) + NAME_MAX + 1))
#endif
#ifdef Q_OS_WIN
#define _WIN32_WINNT 0x0601
#include <SdkDdkVer.h>
#include <windows.h>
#include <FileAPI.h>
#endif

sh::tools::LocalFilesystemWatcher::LocalFilesystemWatcher()
{
#ifdef INOTIFY
    inotifyThread = new sh::tools::LocalFilesystemWatcher::InotifyThread(this);
    inotifyThread->start();
#endif
}

sh::tools::LocalFilesystemWatcher::~LocalFilesystemWatcher()
{
#ifdef INOTIFY
    inotifyThread->terminate();
    while(inotifyThread->isRunning()) {}
    delete inotifyThread;
#endif
}

void sh::tools::LocalFilesystemWatcher::_initialize()
{
}

void sh::tools::LocalFilesystemWatcher::_shutdown()
{
}

#if defined (INOTIFY)

sh::tools::LocalFilesystemWatcher::InotifyThread::InotifyThread(sh::tools::LocalFilesystemWatcher *watcher)
    : QThread(), _watcher(watcher)
{
}

void sh::tools::LocalFilesystemWatcher::InotifyThread::run()
{
    setTerminationEnabled(false);
    char buf[INOTIFY_BUF_LEN] __attribute__ ((aligned(8)));
    ssize_t numRead;
    char *p;
    struct inotify_event *event;
    {
        QMutexLocker l(&_watcher->inotifythreadmutex);
        inotifyFd = inotify_init();
        inotifyInitialized=true;
        _watcher->inotifythreadmutexwait.wakeAll();
        if (inotifyFd == -1)
        {
            LOG_WARNING("inotify initialization failed")
            return;
        }
    }
    while (true) {
        setTerminationEnabled(true);
        numRead = read(inotifyFd, buf, INOTIFY_BUF_LEN);
        setTerminationEnabled(false);
        if (numRead == 0)
        {
            LOG_WARNING("inotify read operation failed with 0")
            return;
        }
        if (numRead == -1 && errno != EINTR)
        {
            LOG_WARNING(QString("inotify read operation failed with -1, errno %1").arg(errno))
            return;
        }
        QMap<int, QList<qint64>> wd2ids;
        {
            QMutexLocker l(&_watcher->inotifymutex);
            wd2ids = _watcher->inotifywd2ids;
        }
        for (p = buf; p < buf + numRead; ) {
            event = (struct inotify_event *) p;
            QList<qint64> ids = wd2ids[event->wd];
            QString name = (event->len > 0) ? event->name : QString();
            if ((event->mask & IN_ATTRIB) || (event->mask & IN_CLOSE_WRITE) || (event->mask & IN_MODIFY) || (event->mask & IN_MOVE_SELF))
                _watcher->emitElementModified(name, ids);
            else if ((event->mask & IN_CREATE) || (event->mask & IN_MOVED_TO))
                _watcher->emitElementCreated(name, ids);
            else if ((event->mask & IN_DELETE) || (event->mask & IN_DELETE_SELF) || (event->mask & IN_MOVED_FROM))
                _watcher->emitElementDeleted(name, ids);
            p += sizeof(struct inotify_event) + event->len;
        }
    }
}

#elif defined (Q_OS_WIN)

sh::tools::LocalFilesystemWatcher::Win32Thread::Win32Thread(sh::tools::LocalFilesystemWatcher *watcher, QString path, qint64 id)
    : QThread(), _watcher(watcher), _path(path), _id(id)
{
}

void sh::tools::LocalFilesystemWatcher::Win32Thread::run()
{
    std::wstring _wspath = _path.toStdWString();
    HANDLE hDir = CreateFile(_wspath.c_str(), FILE_LIST_DIRECTORY, FILE_SHARE_WRITE | FILE_SHARE_READ | FILE_SHARE_DELETE,
                             NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
    if (hDir == INVALID_HANDLE_VALUE) {
        LOG_WARNING(QString("CreateFile on %1 failed with %2").arg(_path).arg(GetLastError()))
        return;
    }
    {
        QMutexLocker _l(&_watcher->id2hdirmutex);
        _watcher->id2hdir[_id] = hDir;
    }
    std::vector<BYTE> buffer(1024*64);
    DWORD dwBytesReturned = 0;
    while (true) {
        if (ReadDirectoryChangesW(hDir, &buffer[0], buffer.size(), FALSE,
                                  FILE_NOTIFY_CHANGE_FILE_NAME|FILE_NOTIFY_CHANGE_DIR_NAME|FILE_NOTIFY_CHANGE_LAST_WRITE,
                                  &dwBytesReturned, NULL, NULL) == 0)
        {
            auto err = GetLastError();
            if (err == ERROR_NOTIFY_ENUM_DIR)
                _watcher->emitElementModified(QString(), {_id});
            else {
                if (err != ERROR_OPERATION_ABORTED)
                    LOG_WARNING(QString("ReadDirectoryChanges on %1 failed with %2").arg(_path).arg(err))
                break;
            }
        }
        else
        {
            if (dwBytesReturned == 0) {
                LOG_WARNING("ReadDirectoryChangesW dwBytesReturned == 0")
                break;
            }
            QList<qint64> ids = {_id};
            FILE_NOTIFY_INFORMATION *fni = reinterpret_cast<FILE_NOTIFY_INFORMATION*>(&buffer[0]);
            while (true) {
                QString fn = QString::fromWCharArray(fni->FileName, fni->FileNameLength / sizeof(wchar_t));
                if (fni->Action == FILE_ACTION_ADDED || fni->Action == FILE_ACTION_RENAMED_NEW_NAME)
                    _watcher->emitElementCreated(fn, ids);
                else if (fni->Action == FILE_ACTION_REMOVED || fni->Action == FILE_ACTION_RENAMED_OLD_NAME)
                    _watcher->emitElementDeleted(fn, ids);
                else if (fni->Action == FILE_ACTION_MODIFIED)
                    _watcher->emitElementModified(fn, ids);
                if (!fni->NextEntryOffset)
                    break;
                fni = reinterpret_cast<FILE_NOTIFY_INFORMATION*>(reinterpret_cast<BYTE*>(fni) + fni->NextEntryOffset);
            }
        }
    }
    {
        QMutexLocker _l(&_watcher->id2hdirmutex);
        _watcher->id2hdir.remove(_id);
    }
    CloseHandle(hDir);
}

#endif

sh::tools::LocalFilesystemWatcher* sh::tools::LocalFilesystemWatcher::instance()
{
    static sh::tools::LocalFilesystemWatcher* _instance = new sh::tools::LocalFilesystemWatcher();
    return _instance;
}

qint64 sh::tools::LocalFilesystemWatcher::addFile(QString path)
{
#if defined (INOTIFY)
    QMutexLocker l1(&inotifythreadmutex);
    QMutexLocker l2(&inotifymutex);
    while (!inotifyThread->inotifyInitialized)
        inotifythreadmutexwait.wait(&inotifythreadmutex);
    if (inotifyThread->inotifyFd == -1)
        return -1;
    QByteArray bpath = path.toUtf8();
    int wd = inotify_add_watch(inotifyThread->inotifyFd, bpath.data(), IN_DONT_FOLLOW|IN_ATTRIB|IN_CLOSE_WRITE|IN_MODIFY|IN_MOVE_SELF|IN_CREATE|IN_MOVED_TO|IN_DELETE|IN_DELETE_SELF|IN_MOVED_FROM);
    if (wd == -1) {
        LOG_WARNING("inotify adding watch failed")
        return -1;
    }
    qint64 eid = nextEid++;
    inotifywd2ids[wd].append(eid);
    id2inotifywd[eid] = wd;
    id2path[eid] = path;
    return eid;
#elif defined (Q_OS_WIN)
    // not supported
    return 0;
#else
    return 0;
#endif
}

void sh::tools::LocalFilesystemWatcher::removeFile(qint64 id)
{
#if defined (INOTIFY)
    QMutexLocker l2(&inotifymutex);
    if (id == -1)
        return;
    int wd = id2inotifywd[id];
    inotifywd2ids[wd].removeOne(id);
    int cnt = inotifywd2ids[wd].length();
    if (cnt == 0)
    {
        inotifywd2ids.remove(wd);
        inotify_rm_watch(inotifyThread->inotifyFd, wd);
    }
    id2inotifywd.remove(id);
    id2path.remove(id);
#elif defined (Q_OS_WIN)
    // not supported
#endif
}

qint64 sh::tools::LocalFilesystemWatcher::addDirectory(QString path)
{
#if defined (INOTIFY)
    return addFile(path);
#elif defined (Q_OS_WIN)
    qint64 id = nextEid++;
    QThread* t = new sh::tools::LocalFilesystemWatcher::Win32Thread(this, path, id);
    t->start();
    return id;
#else
    return 0;
#endif
}

void sh::tools::LocalFilesystemWatcher::removeDirectory(qint64 id)
{
#if defined (INOTIFY)
    removeFile(id);
#elif defined (Q_OS_WIN)
    for(int _tries=0; _tries<100; _tries++) {
        {
            QMutexLocker _l(&id2hdirmutex);
            auto ihdir = id2hdir.find(id);
            if (ihdir != id2hdir.end()) {
                if (CancelIoEx(*ihdir, NULL))
                    break;
                else {
                    auto err = GetLastError();
                    if (err != ERROR_NOT_FOUND) {
                        LOG_WARNING(QString("CancelIoEx failed with %1").arg(err))
                        break;
                    }
                }
            }
            else
                break;
        }
        QThread::msleep(40);
    }
#endif
}

void sh::tools::LocalFilesystemWatcher::emitElementModified(QString name, QList<qint64> ids)
{
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        emit elementModified(name, ids);
    });
}

void sh::tools::LocalFilesystemWatcher::emitElementDeleted(QString name, QList<qint64> ids)
{
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        emit elementDeleted(name, ids);
    });
}

void sh::tools::LocalFilesystemWatcher::emitElementCreated(QString name, QList<qint64> ids)
{
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        emit elementCreated(name, ids);
    });
}
