/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LOCALFILESYSTEMWATCHER_H
#define LOCALFILESYSTEMWATCHER_H

#include <QList>
#include <QMap>
#include <QMutex>
#include <QObject>
#include <QThread>
#include <QWaitCondition>


namespace sh{
    namespace tools{

        /** \brief
         * Used for watching parts of the local filesystem.
         *
         * Depending on the
         * compile flags and your system, this functionality might not be available.
         * If so, the methods are no-ops.
         */
        class LocalFilesystemWatcher : public QObject
        {
            friend class InotifyThread;
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static LocalFilesystemWatcher* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            ~LocalFilesystemWatcher();
            qint64 addFile(QString path);
            void removeFile(qint64 id);
            qint64 addDirectory(QString path);
            void removeDirectory(qint64 id);
        signals:
            void elementModified(QString name, QList<qint64> ids);
            void elementDeleted(QString name, QList<qint64> ids);
            void elementCreated(QString name, QList<qint64> ids);
        private:
            LocalFilesystemWatcher();

#if defined (INOTIFY)
            class InotifyThread : public QThread
            {
                friend class LocalFilesystemWatcher;
            public:
                InotifyThread(LocalFilesystemWatcher* watcher);
            protected:
                void run();
                LocalFilesystemWatcher* _watcher;
                bool inotifyInitialized = false;
                int inotifyFd = -1;
            };

            InotifyThread* inotifyThread;
            QMutex inotifymutex;
            QMap<int, QList<qint64>> inotifywd2ids;
            QMap<qint64, int> id2inotifywd;
            QMap<qint64, QString> id2path;
            qint64 nextEid = 0;
            QMutex inotifythreadmutex;
            QWaitCondition inotifythreadmutexwait;
#elif defined (Q_OS_WIN)
            class Win32Thread : public QThread
            {
                friend class LocalFilesystemWatcher;
            public:
                Win32Thread(LocalFilesystemWatcher* watcher, QString path, qint64 id);
            protected:
                void run();
                LocalFilesystemWatcher* _watcher;
                QString _path;
                qint64 _id;
            };

            QMap<qint64, void*> id2hdir;
            QMutex id2hdirmutex;
            qint64 nextEid = 0;

#endif
            void emitElementModified(QString name, QList<qint64> ids);
            void emitElementDeleted(QString name, QList<qint64> ids);
            void emitElementCreated(QString name, QList<qint64> ids);
        };

    }
}

#endif // LOCALFILESYSTEMWATCHER_H
