/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/filesystemnode.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "localfilesystemwatcher.h"
#include "localfilesystemwatcherconnector.h"
#include "mainthread.h"
#include "singletoninitializer.h"
#include "threadpool.h"
#include "visibleviews.h"


namespace sh{namespace tools{INITIALIZE_SINGLETON_IN_MAINTHREAD(80095, LocalFilesystemWatcherConnector)}}

sh::tools::LocalFilesystemWatcherConnector::LocalFilesystemWatcherConnector()
{
    sh::tools::VisibleViews::instance()->registerOnEnteredHandlers([=](std::shared_ptr<sh::base::FilesystemNode> dir){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        auto deurl = dir->eurl();
        foreach(auto n, getAllLocalParts(dir)) {
            auto neurl = n.first->eurl();
            qint64 id = sh::tools::LocalFilesystemWatcher::instance()->addDirectory(n.second->eurlToLocal(neurl));
            dir2watcherids[deurl].append(id);
            watcherid2dir[id] = neurl;
        }
    });
    sh::tools::VisibleViews::instance()->registerOnLeftHandlers([=](std::shared_ptr<sh::base::FilesystemNode> dir){
        ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
        QList<qint64> ids = dir2watcherids.take(dir->eurl());
        foreach(qint64 id, ids) {
            watcherid2dir.remove(id);
            sh::tools::LocalFilesystemWatcher::instance()->removeDirectory(id);
        }
    });
    connect(sh::tools::LocalFilesystemWatcher::instance(), SIGNAL(elementCreated(QString,QList<qint64>)), this, SLOT(slot_watcher_elementchanged(QString,QList<qint64>)));
    connect(sh::tools::LocalFilesystemWatcher::instance(), SIGNAL(elementModified(QString,QList<qint64>)), this, SLOT(slot_watcher_elementchanged(QString,QList<qint64>)));
    connect(sh::tools::LocalFilesystemWatcher::instance(), SIGNAL(elementDeleted(QString,QList<qint64>)), this, SLOT(slot_watcher_elementchanged(QString,QList<qint64>)));}

void sh::tools::LocalFilesystemWatcherConnector::slot_watcher_elementchanged(QString name, QList<qint64> ids)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    foreach(qint64 iid, ids) {
        auto itr = watcherid2dir.find(iid);
        if (itr != watcherid2dir.end()) {
            if (name.isNull())
                sh::tools::ThreadPool::enqueueForce([=](){
                    auto nodes = sh::base::FilesystemModel::instance()->tryGetNodesForEurl(*itr);
                    foreach(auto node, nodes) {
                        node->requestContainedItems();
                        sh::tools::MainThread::dispatcher()->invokeSync([&](){
                            for(int i=0; i<node->childnodeCount(); i++)
                                node->childnode(i)->requestDetails();
                        });
                    }
                });
            else
                sh::base::FilesystemModel::instance()->refreshData(sh::filesystemhandlers::LocalFilesystemHandler::instance()->localToEurl(
                                                                       sh::filesystemhandlers::LocalFilesystemHandler::instance()->eurlToLocal(*itr) + "/" + name));
            break;
        }
    }
}

void sh::tools::LocalFilesystemWatcherConnector::_initialize()
{
}

void sh::tools::LocalFilesystemWatcherConnector::_shutdown()
{
}

std::shared_ptr<sh::tools::LocalFilesystemWatcherConnector> sh::tools::LocalFilesystemWatcherConnector::instance()
{
    static auto _instance = std::make_shared<sh::tools::LocalFilesystemWatcherConnector>();
    return _instance;
}

QList<QPair<std::shared_ptr<sh::base::FilesystemNode>, sh::filesystemhandlers::LocalFilesystemHandler*>> sh::tools::LocalFilesystemWatcherConnector::getAllLocalParts(std::shared_ptr<sh::base::FilesystemNode> node)
{
    QList<QPair<std::shared_ptr<sh::base::FilesystemNode>, sh::filesystemhandlers::LocalFilesystemHandler*>> result;
    while (node) {
        sh::filesystemhandlers::LocalFilesystemHandler* lhandler = qobject_cast<sh::filesystemhandlers::LocalFilesystemHandler*>(node->handler());
        if (lhandler)
            result.append(QPair<std::shared_ptr<sh::base::FilesystemNode>, sh::filesystemhandlers::LocalFilesystemHandler*>(node, lhandler));
        node = node->parentnode();
    }
    return result;
}
