/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include "exceptions/runtimeexception.h"
#include "exceptions/programexception.h"
#include "logger.h"
#include "singletoninitializer.h"


namespace sh{namespace tools{INITIALIZE_SINGLETON_IN_MAINTHREAD(50010, Logger)}}

sh::tools::Logger::Logger() : QObject(0)
{
#ifdef QT_DEBUG
    log("This is a debug build.");
#else
    log("This is a release build.");
#endif
#ifdef Q_OS_UNIX
    log("Target platform: Unix/Linux.");
#elif defined(Q_OS_WIN)
    log("Target platform: Windows.");
#else
    log("Target platform: Unknown.");
#endif
#ifdef GIO
    log("Gnome-IO support: yes");
#else
    log("Gnome-IO support: no");
#endif
#ifdef PYTHONSCRIPT
    log("Python scripting support: yes");
#else
    log("Python scripting support: no");
#endif
#ifdef INOTIFY
    log("inotify support: yes");
#else
    log("inotify support: no");
#endif
}

sh::tools::Logger* sh::tools::Logger::instance()
{
    static sh::tools::Logger* _instance = new sh::tools::Logger();
    return _instance;
}

void sh::tools::Logger::_initialize()
{
}

void sh::tools::Logger::_shutdown()
{
}

sh::tools::Logger::~Logger()
{
    foreach(sh::tools::Logger::LogMessage* lm, _logMessages)
        delete lm;
}

void sh::tools::Logger::logException(sh::exceptions::Exception& ex, sh::tools::LogSeverity severity, QString source)
{
    log(ex.message() +
        "\nDetails: " + ex.details()
        , severity, source);
    log(QString("Exception info.") +
        "\nName: " + ex.name() +
        ", Resumeable: " + QString(ex.isResumeable()?"Yes":"No") +
        ", Retryable: " + QString(ex.isRetryable()?"Yes":"No") +
        ", ProgramException: " + QString(ex.isProgramException()?"Yes":"No") +
        "\nClasses: " + ex.classes() +
        "\nCallstack:\n" + ex.callstack() +
        ( ex.auxiliary().isEmpty() ? "" : ("\nAuxiliary information:\n" + ex.auxiliary()) )
        , LogSeverity::DEBUG, source);
}

void sh::tools::Logger::log(QString message, sh::tools::LogSeverity severity, QString source)
{
    QDateTime time = QDateTime::currentDateTime();
    int lastslash = source.lastIndexOf("/");
    int lastslash2 = source.lastIndexOf("\\", lastslash+1);
    if (lastslash2>-1)
        lastslash = lastslash2;
    QString _source = source;
    if (lastslash>-1)
        _source=_source.mid(lastslash+1);
    if (_source.length()>25)
        _source = ":"+_source.right(24);
    else
        _source = (_source+QString(" ").repeated(25)).left(25);
    QString fullmsg =
            "LOG: "+time.time().toString(Qt::ISODate)
            + QString("    " + QString("+").repeated((int)severity)).right(4)
            + " " + _source
            + "  " + message + "\n";
    QByteArray foo = fullmsg.toUtf8();
    std::cout << foo.data();
    std::cout.flush();
    sh::tools::Logger::LogMessage* logMessage= new sh::tools::Logger::LogMessage();
    logMessage->message = message;
    logMessage->source = source;
    logMessage->severity = severity;
    logMessage->time=time;
    mutex.lock();
    _logMessages.append(logMessage);
    mutex.unlock();
    emit newLogMessageArrived();
}

QList<sh::tools::Logger::LogMessage*> sh::tools::Logger::logMessages()
{
    QMutexLocker _l(&mutex);
    return _logMessages;
}

QString sh::tools::Logger::getLogAsText(LogSeverity minseverity)
{
    QStringList result;
    foreach(sh::tools::Logger::LogMessage* msg, logMessages())
    {
        if (minseverity <= msg->severity)
        {
            QString ssev;
            switch(msg->severity)
            {
            case sh::tools::LogSeverity::DEBUG:
                ssev="debug";
                break;
            case sh::tools::LogSeverity::INFO:
                ssev="info";
                break;
            case sh::tools::LogSeverity::WARNING:
                ssev="warning";
                break;
            case sh::tools::LogSeverity::ERROR:
                ssev="error";
                break;
            default:
                ssev="unknown";
                break;
            }
            result << QString("%1 (%2) %3:\n%4\n")
                    .arg(msg->time.time().toString(Qt::ISODate))
                    .arg(msg->source)
                    .arg(ssev)
                    .arg(msg->message);
        }
    }
    return result.join("\n");
}
