/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LOGGER_H
#define LOGGER_H

#include <QDateTime>
#include <QMutex>
#include <QObject>
#include "exceptions/exception.h"


namespace sh{
    namespace tools{

        /** \brief The severity of a log message.
         */
        enum class LogSeverity{
            DEBUG,
            INFO,
            WARNING,
            ERROR
        };

        /** \brief
         * The logging manager.
         *
         * Use it for writing messages to the Shallot log and for reading from it.
         */
        class Logger : public QObject
        {
            Q_OBJECT
        public:

            /** \brief A single log message.
             */
            struct LogMessage{
                QString message;
                LogSeverity severity;
                QString source;
                QDateTime time;
            };

            /** \brief
             * Returns the instance of this singleton.
             */
            static Logger* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();
            virtual ~Logger();
            void logException(sh::exceptions::Exception& ex, LogSeverity severity = LogSeverity::ERROR, QString source = QString());
            void log(QString message, LogSeverity severity = LogSeverity::INFO, QString source = QString());
            QString getLogAsText(LogSeverity minseverity = LogSeverity::DEBUG);
            QList<LogMessage*> logMessages();

        signals:
            void newLogMessageArrived();

        private:
            Logger();
            QMutex mutex;
            QList<LogMessage*> _logMessages;
        };

    }
}

#define __LOG_HELPER(msg,sev,file,line) sh::tools::Logger::instance()->log(msg, sev, QString(file) + ":" + QString("%1").arg(line) );

#ifdef QT_DEBUG
#define LOG_DEBUG(msg) __LOG_HELPER(msg, sh::tools::LogSeverity::DEBUG, __FILE__, __LINE__)
#else
#define LOG_DEBUG(msg)
#endif

#define LOG_INFO(msg) __LOG_HELPER(msg, sh::tools::LogSeverity::INFO, __FILE__, __LINE__)
#define LOG_WARNING(msg) __LOG_HELPER(msg, sh::tools::LogSeverity::WARNING, __FILE__, __LINE__)
#define LOG_ERROR(msg) __LOG_HELPER(msg, sh::tools::LogSeverity::ERROR, __FILE__, __LINE__)


#endif // LOGGER_H
