/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mainthread.h"
#include "singletoninitializer.h"
#include "threadpool.h"


sh::tools::SingletonInitializer::SingletonInitializer()
{
}

QList<sh::tools::SingletonInitializer::InitTuple>* sh::tools::SingletonInitializer::inits;

QThread* sh::tools::SingletonInitializer::_initializerThread = 0;

bool sh::tools::SingletonInitializer::_isshutdown = false;
bool sh::tools::SingletonInitializer::_isshuttingdown = false;
QMutex sh::tools::SingletonInitializer::shutdownmutex;

char sh::tools::SingletonInitializer::registerInitializer(int index, std::function<void()> create, std::function<void()> init, std::function<void()> shutdown, std::function<void()> remove)
{
    if (inits==0) // this guarantees that we actually have a list even at very early times
        inits = new QList<InitTuple>();
    inits->append(std::make_tuple(index, create, init, shutdown, remove));
    return 0;
}

void sh::tools::SingletonInitializer::callInitializers()
{
    qSort(inits->begin(), inits->end(), [](const InitTuple& x, const InitTuple& y){
        return std::get<0>(x) < std::get<0>(y);
    });
    foreach(InitTuple tup, *inits)
        std::get<1>(tup)();
    QtConcurrent::run([&](){
        _initializerThread = QThread::currentThread();
        try
        {
            foreach(InitTuple tup, *inits)
                std::get<2>(tup)();
        }
        catch (sh::exceptions::ThreadAbortException& ae)
        {
        }
    });
}

void sh::tools::SingletonInitializer::shutdown()
{
    QMutexLocker _l1(&shutdownmutex);
    if (!_isshutdown)
    {
        _isshuttingdown = true;
        std::reverse(inits->begin(), inits->end());
        foreach(InitTuple tup, *inits)
            std::get<3>(tup)();
        foreach(InitTuple tup, *inits)
            std::get<4>(tup)();
        delete inits;
        _isshutdown = true;
    }
}

void sh::tools::SingletonInitializer::waitShutdown(std::function<void()> onshutdown)
{
    QObject o;
    o.moveToThread(sh::tools::MainThread::dispatcher()->thread());
    QObject::connect(&o, &QObject::destroyed, sh::tools::MainThread::dispatcher()->thread(), [=](){
        QTimer* timer = new QTimer();
        QObject::connect(timer, &QTimer::timeout, [=](){
            if (isShutdown())
            {
                timer->stop();
                timer->deleteLater();
                onshutdown();
            }
        });
        timer->start(100);
    });
}

bool sh::tools::SingletonInitializer::isShutdown()
{
    QMutexLocker _l1(&shutdownmutex);
    return _isshutdown;
}

bool sh::tools::SingletonInitializer::isShuttingDown()
{
    return _isshuttingdown;
}

sh::tools::SingletonInitializer::~SingletonInitializer()
{
    shutdown();
}

void sh::tools::SingletonInitializer::deleteLater()
{
    delete this;
}

QThread* sh::tools::SingletonInitializer::initializerThread()
{
    return _initializerThread;
}
