/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SINGLETONINITIALIZER_H
#define SINGLETONINITIALIZER_H

#include <functional>
#include <memory>
#include <tuple>
#include <QList>
#include <QMutex>
#include <QTimer>
#include <QtConcurrent>


namespace sh{
    namespace tools{

        /** \brief
         * Takes care of initialization and shutdown of infrastructure singletons.
         *
         * It is a very early part of shallot core infrastructure.
         * Singletongs typically participate by using one of those macros:
         *
         * `INITIALIZE_SINGLETON_IN_MAINTHREAD(I,X)`:
         * I is an integer controlling the order (smaller I means earlier initialization).
         * X is the singleton's class name. It must fulfill the following requirements:
         * - has a parameterless constructor (used in create step).
         * - has a `_initialize` method (used in init step).
         * - has a `_shutdown` method (used in shutdown step).
         * - (destructor used in delete step).
         *
         * `INITIALIZE_STATIC_IN_MAINTHREAD(I,X)`:
         * I is an integer controlling the order (smaller I means earlier initialization).
         * X is the static's class name. It must fulfill the following requirements:
         * - has a static `_initialize` method (used in init step).
         * - has a static `_shutdown` method (used in shutdown step).
         *
         * Both macros use SingletonInitializer::registerInitializer. The mentioned methods are used for the callbacks.
         * Note: Not all macros provide callbacks for each step.
         */
        class SingletonInitializer
        {
        public:
            /** \brief
             * Registers (un)initialization methods for each step.
             * They are actually called in the core infrastructure by callInitializers and shutdown.
             * \param index Controls the order.
             * \param create This callback is called for executing the create-step (1st one).
             * \param init This callback is called for executing the init-step (2nd one).
             * \param shutdown This callback is called for executing the shutdown-step (3rd one).
             * \param remove This callback is called for executing the remove-step (4th one).
             */
            static char registerInitializer(int index, std::function<void()> create, std::function<void()> init, std::function<void()> shutdown, std::function<void()> remove);
            /** \brief
             * Executes all create-callbacks at first, then all init-callbacks. In both runs, the callbacks with lower `index` come first.
             * Afterwards, the singletons are considered as up and running.
             */
            void callInitializers();

            static QThread* initializerThread();

            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            SingletonInitializer();
            ~SingletonInitializer();
            /** \brief
             * Executes all shutdown-callbacks at first, then all remove-callbacks. The callback order is the reversed one of callInitializers.
             * Afterwards, the singletons are considered as shut down and removed.
             */
            void shutdown();
            /** \brief
             * Asynchronously waits for complete shutdown and calls the callback afterwards.
             */
            void waitShutdown(std::function<void()> onshutdown);
            /** \brief
             * If the singletons are completely shut down and removed.
             */
            static bool isShutdown();
            /** \brief
             * If the the shutdown process at least began (or is even finished).
             */
            static bool isShuttingDown();
            /** \brief
             * Deletes the instance. Used internally.
             */
            void deleteLater();

            /** \brief
             * Helper functions for singleton deletion. Used internally.
             */
            template<class T>
            static void _delete(std::shared_ptr<T>)
            {
            }

            /** \brief
             * Helper functions for singleton deletion. Used internally.
             */
            template<class T>
            static void _delete(T* v)
            {
                delete v;
            }

        private:
            typedef std::tuple<int, std::function<void()>, std::function<void()>, std::function<void()>, std::function<void()> > InitTuple;
            static QList<InitTuple>* inits;
            static bool _isshutdown;
            static bool _isshuttingdown;
            static QMutex shutdownmutex;
            static QThread* _initializerThread;
        };

    }
}

#define INITIALIZE_SINGLETON_IN_MAINTHREAD(I,X) auto __initialize_ ## X = sh::tools::SingletonInitializer::registerInitializer(I, [](){X::instance();}, [](){X::instance()->_initialize();}, [](){X::instance()->_shutdown();}, [](){sh::tools::SingletonInitializer::_delete(X::instance());});
#define INITIALIZE_STATIC_IN_MAINTHREAD(I,X) auto __initialize__ ## X = sh::tools::SingletonInitializer::registerInitializer(I, [](){}, X::_initialize, X::_shutdown, [](){});

#endif // SINGLETONINITIALIZER_H
