/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef THREADPOOL_H
#define THREADPOOL_H

#include <memory>
#include <QCoreApplication>
#include <QList>
#include <QMutex>
#include <QThread>
#include <QVariant>
#include <QWaitCondition>
#include "exceptions/exception.h"
#include "exceptions/threadingexception.h"
#include "tools/singletoninitializer.h"


namespace sh{
    namespace tools{

#ifdef QT_DEBUG

        class AsyncForbid
        {
        public:
            static QList<QThread*> restrictedThreads;
            static QMutex mutex;
            AsyncForbid() { QMutexLocker _l(&mutex); restrictedThreads.append(QThread::currentThread()); }
            ~AsyncForbid() { QMutexLocker _l(&mutex); restrictedThreads.removeOne(QThread::currentThread()); }
        };

        class AsyncForbidIgnore
        {
        public:
            static QList<QThread*> restrictionIgnoredThreads;
            static QMutex mutex;
            AsyncForbidIgnore() { QMutexLocker _l(&mutex); restrictionIgnoredThreads.append(QThread::currentThread()); }
            ~AsyncForbidIgnore() { QMutexLocker _l(&mutex); restrictionIgnoredThreads.removeOne(QThread::currentThread()); }
        };

#define FORBID_ASYNC_CALLS sh::tools::AsyncForbid __forbid_async_calls_canary;
#define IGNORE_FORBIDDEN_ASYNC_CALLS sh::tools::AsyncForbidIgnore __ignore_forbid_async_calls_canary;
#define _FORBID_ASYNC_CALLS_BARRIER if ( sh::tools::AsyncForbid::restrictedThreads.contains(QThread::currentThread()) && !sh::tools::AsyncForbidIgnore::restrictionIgnoredThreads.contains(QThread::currentThread())) sh::exceptions::Exception::executeGuarded([&](){ throw sh::exceptions::ThreadingException(sh::exceptions::ExceptionData().details("Asynchronous calls are forbidden from this place."));});
#else
#define FORBID_ASYNC_CALLS
#define IGNORE_FORBIDDEN_ASYNC_CALLS
#define _FORBID_ASYNC_CALLS_BARRIER
#endif

        class ThreadPoolThread;

        /** \brief
         * A pool of worker threads doing some short jobs from a queue.
         */
        class ThreadPool : public QObject
        {
            friend class ThreadPoolThread;
            Q_OBJECT
        public:

            /** \brief
             * Enqueues code to the threadpool.
             */
            inline static void enqueueForce(std::function<void ()> fct)
            {
                _FORBID_ASYNC_CALLS_BARRIER
                _enqueueForce(fct);
            }

            /** \brief
             * Enqueues code to the threadpool.
             *
             * It can also conserve some shared pointers until after execution.
             */
            inline static void enqueueForce(std::function<void ()> fct,
                           std::shared_ptr<void> obj1, std::shared_ptr<void> obj2=0, std::shared_ptr<void> obj3=0)
            {
                _enqueueForce([=]() {
                    std::shared_ptr<void> _obj1 = obj1;
                    std::shared_ptr<void> _obj2 = obj2;
                    std::shared_ptr<void> _obj3 = obj3;
                    fct();
                });
            }

            /** \brief
             * Enqueues code to the threadpool and forbids async calls within it.
             * This is useful for ensuring that no code might access a certain object after its deletion.
             *
             * It can also conserve some shared pointers until after execution.
             */
            inline static void enqueueForceBeyondAsyncCallBarrier (std::function<void ()> fct,
                           std::shared_ptr<void> obj1=0, std::shared_ptr<void> obj2=0, std::shared_ptr<void> obj3=0)
            {
                enqueueForce([=]() {
                    FORBID_ASYNC_CALLS
                    fct();
                }, obj1, obj2, obj3);
            }

            /** \brief
             * Enqueues code to the threadpool if currently in `thread`.
             * Otherwise executes directly.
             */
            inline static void enqueueIfIn(std::function<void ()> fct, QThread* thread)
            {
                _FORBID_ASYNC_CALLS_BARRIER
                if (QThread::currentThread() != thread)
                    sh::exceptions::Exception::executeGuarded([&](){
                        fct();
                    }, sh::exceptions::ExecuteGuardFlag::ALL_ERRORS_KILL_SHALLOT);
                else
                    enqueueForce(fct);
            }

            /** \brief
             * Enqueues code to the threadpool if currently in main thread.
             * Otherwise executes directly.
             */
            inline static void enqueueIfInMain(std::function<void ()> fct)
            {
                enqueueIfIn(fct, QCoreApplication::instance()->thread());
            }

            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            static void _shutdown();

            inline static bool isShuttingDown()
            {
                QMutexLocker l1(&poolmutex);
                return _shutsdown;
            }

            inline static bool isShutdown()
            {
                QMutexLocker l1(&poolmutex);
                return _shutsdown && (_threadsalive==0);
            }

            inline static int getThreadCount()
            {
                return THREAD_COUNT;
            }

            inline static void respectThreadAbort()
            {
                if (_shutsdown && _threads.contains((ThreadPoolThread*) QThread::currentThread()))
                    throw sh::exceptions::ThreadAbortException();
            }

            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            static void _initialize();

        private:

            inline static void _enqueueForce(std::function<void ()> fct)
            {
                QMutexLocker l1(&poolmutex);
                tasks.append( fct );
                pooladdedcondition.wakeOne();
            }

            static const int THREAD_COUNT = 5;

            static QMutex poolmutex;
            static QWaitCondition pooladdedcondition;
            static QList<std::function<void ()>> tasks;
            static bool _shutsdown;
            static QList<ThreadPoolThread*> _threads;
            static int _threadsalive;
        };

        /** \brief A thread in the thread pool.
         */
        class ThreadPoolThread : public QThread
        {
        friend class ThreadPool;
        protected:
            /** \brief Constructed only by the infrastructure and made available otherwise.
             */
            ThreadPoolThread();
            void run();
        };

    }
}

#endif // THREADPOOL_H
