/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/filesystemnode.h"
#include "base/operation.h"
#include "detailcolumns/detailcolumnmimetype.h"
#include "mainthread.h"
#include "operationscache.h"
#include "singletoninitializer.h"
#include "thumbnailmanager.h"
#include "threadpool.h"


namespace sh{namespace tools{INITIALIZE_SINGLETON_IN_MAINTHREAD(130080, ThumbnailManager)}}

sh::tools::ThumbnailManager* sh::tools::ThumbnailManager::instance()
{
    static sh::tools::ThumbnailManager* _instance = new sh::tools::ThumbnailManager();
    return _instance;
}

sh::tools::ThumbnailManager::ThumbnailManager() :
    QObject()
{
    maxworkers = ThreadPool::getThreadCount() / 3;
    runningworkers = 0;
    if (maxworkers == 0)
        maxworkers = 1;
    capacity = (QSysInfo::WordSize <= 32) ? 500 : 2000;
}

void sh::tools::ThumbnailManager::_initialize()
{
}

void sh::tools::ThumbnailManager::_shutdown()
{
    {
        QMutexLocker _l1(&workermutex);
        _shuttingdown = true;
    }
    bool _finished = false;
    while (!_finished)
    {
        QThread::msleep(100);
        {
            QMutexLocker _l1(&workermutex);
            _finished = runningworkers == 0;
        }
    }
}

void sh::tools::ThumbnailManager::worker()
{
    bool __shuttingdown = false;
    while(true)
    {
        std::shared_ptr<sh::base::FilesystemNode> node;
        int width = 0;
        int height = 0;
        QList<std::function<void(QIcon)>> callbacks;
        {
            QMutexLocker _l1(&workermutex);
            if (requestQueue.isEmpty() || __shuttingdown)
            {
                runningworkers--;
                break;
            }
            else
            {
                ThumbnailRequest r = requestQueue.takeLast();
                node = r.node.lock();
                width = r.width;
                height = r.height;
                callbacks = r.callbacks;
            }
        }
        if (node)
        {
            sh::exceptions::Exception::executeGuarded([&](){
                std::shared_ptr<sh::base::Operation> op = sh::tools::OperationsCache::instance()->getOperationForContainer(node->eurl()->outermostInnerEurl());
                QString contentType = sh::detailcolumns::DetailColumnMimetype::instance()->requestValue(node, op.get(), true, true).toString();
                QIcon result;
                QList<std::shared_ptr<ThumbnailProvider>> providers;
                {
                    QMutexLocker _l1(&_thumbnailProvidersMutex);
                    providers = _thumbnailProviders;
                }
                foreach(std::shared_ptr<ThumbnailProvider> provider, providers)
                {
                    provider->getThumbnail(op.get(), node, contentType, width, height, &result);
                    if (!result.isNull())
                        break;
                }
                {
                    QMutexLocker _1(&thumbnailslock);
                    auto oldthumb = thumbnails.find(node.get());
                    while(oldthumb!=thumbnails.end() && oldthumb.key()==node.get())
                    {
                        if ((*oldthumb)->width == width && (*oldthumb)->height == height)
                        {
                            delete (*oldthumb);
                            thumbnails.erase(oldthumb);
                            break;
                        }
                        oldthumb++;
                    }
                    thumbnails.insertMulti(node.get(), new Thumbnail(result, QDateTime::currentMSecsSinceEpoch() + (1000 * 90), _curr_accessTime++, node, width, height));
                    _enforce_capacity();
                }
                foreach (auto callback, callbacks)
                    callback(result);
                sh::tools::MainThread::dispatcher()->invokeSync([=](){
                    {
                        QMutexLocker _1(&workermutex);
                        if (_shuttingdown)
                            return;
                    }
                    emit thumbnailAvailable(node);
                });
            }, sh::exceptions::ALL_ERRORS_KILL_SHALLOT|sh::exceptions::AUTOMATIC_RETRY_ENABLED);
            {
                QMutexLocker _l1(&workermutex);
                __shuttingdown = _shuttingdown;
            }
        }
    }
}

void sh::tools::ThumbnailManager::_enforce_capacity()
{
    if (thumbnails.size() > capacity)
    {
        QList<ThumbnailSortStruct> lst;
        QHashIterator<sh::base::FilesystemNode*, Thumbnail*> i(thumbnails);
        while (i.hasNext()) {
            i.next();
            lst.append(ThumbnailSortStruct(i.key(), i.value()->accessTime));
        }
        qSort(lst.begin(), lst.end(), [](const ThumbnailSortStruct& a, const ThumbnailSortStruct& b){
            return a.accessTime < b.accessTime;
        });
        int c = 100;
        foreach(ThumbnailSortStruct ts, lst)
        {
            Thumbnail* tn = thumbnails.take(ts.node);
            delete tn;
            if (--c == 0)
                break;
        }
    }
}

void sh::tools::ThumbnailManager::invalidateCache()
{
    foreach(Thumbnail* t, thumbnails)
        t->validUntil = 0;
}

void sh::tools::ThumbnailManager::requestThumbnail(std::shared_ptr<base::FilesystemNode> node, int width, int height, std::function<void(QIcon)> callback)
{
    if (node->eurl() == 0)
        return;
    {
        QMutexLocker _l1(&workermutex);
        if (!_shuttingdown)
        {
            foreach(ThumbnailRequest n, requestQueue)
            {
                if (n.width == width && n.height == height)
                {
                    std::shared_ptr<sh::base::FilesystemNode> sn = n.node.lock();
                    if (sn == node) {
                        if (callback)
                            n.callbacks.append(callback);
                        return;
                    }
                }
            }
            {
                QMutexLocker _1(&thumbnailslock);
                auto itr = thumbnails.find(node.get());
                if (itr != thumbnails.end())
                    (*itr)->reordered = true;
            }
            QList<std::function<void(QIcon)>> callbacks;
            if (callback)
                callbacks.append(callback);
            requestQueue.append(ThumbnailRequest(node, width, height, callbacks));
            if (runningworkers < maxworkers)
            {
                runningworkers++;
                sh::tools::ThreadPool::enqueueForce([=](){
                    worker();
                });
            }
        }
    }
}

bool sh::tools::ThumbnailManager::getThumbnail(std::shared_ptr<base::FilesystemNode> node, int width, int height, QIcon* result, bool* outdated, bool* reordered)
{
    QMutexLocker _1(&thumbnailslock);
    auto itr = thumbnails.find(node.get());
    while (itr != thumbnails.end() && itr.key() == node.get())
    {
        Thumbnail* t = *itr;
        if (t->width==width && t->height==height && !t->node.expired()) {
            *result = t->icon;
            if (outdated)
                *outdated = (t->validUntil - QDateTime::currentMSecsSinceEpoch()) < 0;
            if (reordered)
                *reordered = t->reordered;
            t->accessTime = _curr_accessTime++;
            return true;
        }
        itr++;
    }
    return false;
}

sh::tools::ThumbnailManager::Thumbnail::Thumbnail(QIcon icon, qint64 validUntil, qint64 accessTime, std::weak_ptr<base::FilesystemNode> node, int width, int height)
    : icon(icon), validUntil(validUntil), accessTime(accessTime), reordered(false), node(node), width(width), height(height)
{
}

sh::tools::ThumbnailManager::ThumbnailSortStruct::ThumbnailSortStruct(base::FilesystemNode *node, qint64 accessTime)
    : node(node), accessTime(accessTime)
{
}

sh::tools::ThumbnailManager::ThumbnailRequest::ThumbnailRequest(std::weak_ptr<sh::base::FilesystemNode> node, int width, int height, QList<std::function<void(QIcon)>> callbacks)
                : node(node), width(width), height(height), callbacks(callbacks)
{
}

bool sh::tools::ThumbnailManager::ThumbnailRequest::operator==(sh::tools::ThumbnailManager::ThumbnailRequest const& b)
{
    return (node.lock() == b.node.lock()) && (width==b.width) && (height==b.height);
}

void sh::tools::ThumbnailManager::addThumbnailProvider(int index, std::shared_ptr<sh::tools::ThumbnailProvider> provider)
{
    QMutexLocker _l1(&_thumbnailProvidersMutex);
    _thumbnailProviderMap.insertMulti(index, provider);
    _thumbnailProviders = _thumbnailProviderMap.values();
}

const int sh::tools::ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_CORE = 1000000;

const int sh::tools::ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_NORMAL = 2000000;

const int sh::tools::ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_EXOTIC = 3000000;

const int sh::tools::ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_FALLBACK = 4000000;
