/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef THUMBNAILMANAGER_H
#define THUMBNAILMANAGER_H

#include <functional>
#include <memory>
#include <QDateTime>
#include <QIcon>
#include <QMap>
#include <QMutex>
#include <QObject>


namespace sh{
    namespace base{class FilesystemNode;class Operation;}

    namespace tools {

        class ThumbnailProvider
        {
        public:
            virtual void getThumbnail(sh::base::Operation* operation, std::shared_ptr<sh::base::FilesystemNode> node, QString contentType, int width, int height, QIcon* icon) =0;
        };

        /** \brief
         * Creates thumbnail images for filesystem nodes.
         *
         * This is a cached source.
         */
        class ThumbnailManager : public QObject
        {
            Q_OBJECT
        public:
            /** \brief
             * Returns the instance of this singleton.
             */
            static ThumbnailManager* instance();
            /** \brief
             * Only called by the Shallot infrastructure for initialization.
             */
            void _initialize();
            /** \brief
             * Only called by the Shallot infrastructure for shutdown.
             */
            void _shutdown();

            /** @brief Base value for display indexes of core (i.e. very important) thumbnail providers.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_CORE;
            /** @brief Base value for display indexes of thumbnail providers with normal importance.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_NORMAL;
            /** @brief Base value for display indexes of exotic thumbnail providers.
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_EXOTIC;
            /** @brief Base value for display indexes of fallback thumbnail providers (those who try to get something when everything else failed).
             *
             * Used in ThumbnailProvider::addThumbnailProvider.
             */
            static const int REGISTER_THUMBNAILPROVIDER_INDEX_FALLBACK;

            void requestThumbnail(std::shared_ptr<sh::base::FilesystemNode> node, int width, int height, std::function<void(QIcon)> callback = 0);
            bool getThumbnail(std::shared_ptr<sh::base::FilesystemNode> node, int width, int height, QIcon *result, bool *outdated=0, bool *reordered=0);
            void invalidateCache();

            class ThumbnailRequest
            {
            public:
                ThumbnailRequest(std::weak_ptr<sh::base::FilesystemNode> node, int width, int height, QList<std::function<void(QIcon)>> callbacks);
                bool operator==(ThumbnailRequest const& b);
                std::weak_ptr<sh::base::FilesystemNode> node;
                int width;
                int height;
                QList<std::function<void(QIcon)>> callbacks;
            };

            void addThumbnailProvider(int index, std::shared_ptr<ThumbnailProvider> provider);

        signals:
            void thumbnailAvailable(std::shared_ptr<sh::base::FilesystemNode> node);

        private:
            class Thumbnail
            {
            public:
                Thumbnail(QIcon icon, qint64 validUntil, qint64 accessTime, std::weak_ptr<sh::base::FilesystemNode> node, int width, int height);
                QIcon icon;
                qint64 validUntil;
                qint64 accessTime;
                bool reordered;
                std::weak_ptr<sh::base::FilesystemNode> node;
                int width;
                int height;
            };

            class ThumbnailSortStruct
            {
            public:
                ThumbnailSortStruct(sh::base::FilesystemNode* node, qint64 accessTime);
                sh::base::FilesystemNode* node;
                qint64 accessTime;
            };

            ThumbnailManager();
            QHash<sh::base::FilesystemNode*, Thumbnail*> thumbnails;
            QMutex thumbnailslock;
            QList<ThumbnailRequest> requestQueue;
            int maxworkers;
            int runningworkers;
            QMutex workermutex;
            bool _shuttingdown = false;
            int capacity;
            qint64 _curr_accessTime = 0;
            void worker();
            void _enforce_capacity();
            QMap<int, std::shared_ptr<ThumbnailProvider>> _thumbnailProviderMap;
            QList<std::shared_ptr<ThumbnailProvider>> _thumbnailProviders;
            QMutex _thumbnailProvidersMutex;
        };

    }
}

#endif // THUMBNAILMANAGER_H
