/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/filesystemnode.h"
#include "base/operation.h"
#include "configuration/configurationmanager.h"
#include "tools/singletoninitializer.h"
#include "ffmpegvideothumbnailprovider.h"


std::shared_ptr<sh::configuration::ConfigurationValue> sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider::cfgvalFfmpegPath =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider.FfmpegPath", "/usr/bin/ffmpeg",
            sh::configuration::ConfigurationValue::ValueTypeLocalFilePath(),
            QObject::tr("Path to the 'ffmpeg' tool"),
            sh::configuration::ConfigurationCategory::CategoryExternalTools,
            QObject::tr("This tool is possibly used for getting infos about video files."));

namespace sh{namespace tools{namespace thumbnailproviders{INITIALIZE_STATIC_IN_MAINTHREAD(130100, FfmpegVideoThumbnailProvider)}}}

sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider::FfmpegVideoThumbnailProvider()
            : ThumbnailProvider(), reDuration( QRegularExpression("^.*duration=([0-9.]+)$", QRegularExpression::MultilineOption) )
{
    cfgvalFfmpegPath->consumeValue(this, [=](){
        pathToFfmpegTool = cfgvalFfmpegPath->getConfigValueString();
        pathToFfprobeTool = QFileInfo(pathToFfmpegTool + "/../ffprobe").absoluteFilePath();
    });
}

void sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider::_initialize()
{
    ThumbnailManager::instance()->addThumbnailProvider(ThumbnailManager::REGISTER_THUMBNAILPROVIDER_INDEX_CORE+1000, std::make_shared<FfmpegVideoThumbnailProvider>());
}

void sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider::_shutdown()
{
}

void sh::tools::thumbnailproviders::FfmpegVideoThumbnailProvider::getThumbnail(sh::base::Operation *operation, std::shared_ptr<sh::base::FilesystemNode> node, QString contentType, int width, int height, QIcon* icon)
{
    if (contentType.startsWith("video/"))
    {
        QString vidfile;
        sh::exceptions::Exception::executeGuarded([&](){
            vidfile = operation->fetchFile(node->eurl());
        }, sh::exceptions::ExecuteGuardFlag::AUTOMATIC_RETRY_ENABLED|sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
        if (!vidfile.isNull())
        {
            QProcess ffprobeprocess;
            ffprobeprocess.setProgram(pathToFfprobeTool);
            QStringList pargs;
            pargs << "-v" << "error" << "-show_entries" << "format=duration" << vidfile;
            ffprobeprocess.setArguments(pargs);
            ffprobeprocess.start();
            double duration = 0;
            if (ffprobeprocess.waitForFinished())
            {
                QByteArray ba = ffprobeprocess.readAll();
                QString t = QString::fromUtf8(ba);
                mutexReDuration.lock();
                QRegularExpressionMatch match = reDuration.match(t);
                if (match.hasMatch())
                    duration = match.captured(1).toDouble();
                mutexReDuration.unlock();
                if (duration > 0) {
                    QProcess ffmpegprocess;
                    ffmpegprocess.setProgram(pathToFfmpegTool);
                    QStringList args;
                    args << "-y" << "-ss" << QString::number((int)(duration/2)) << "-i" << vidfile << "-vf"
                         << QString("scale=w=%1:h=%2:force_original_aspect_ratio=decrease").arg(width).arg(height)
                         << "-an" << "-vframes" << "1" << "-f" << "mjpeg" << "-";
                    ffmpegprocess.setArguments(args);
                    ffmpegprocess.start();
                    if (ffmpegprocess.waitForFinished())
                    {
                        QByteArray bba = ffmpegprocess.readAll();
                        QImage img;
                        img.loadFromData(bba);
                        *icon = QIcon(QPixmap::fromImage(img));
                    }
                }
            }
        }
    }
}
