/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDesktopWidget>
#include "actions/actionexecutioninfo.h"
#include "configuration/configurationmanager.h"
#include "tools/mainthread.h"
#include "tools/iconmanager.h"
#include "ui/feedbackpanels/msgbox.h"
#include "ui/feedbackpanels/gridform.h"
#include "ui/feedbackpanels/credentials.h"
#include "ui/feedbackpanels/helpers/gridforminnersimplechooser.h"
#include "ui/feedbackpanels/feedbackpanel.h"
#include "actionexecutioninfodialog.h"
#include "ui_actionexecutioninfodialog.h"


sh::ui::ActionExecutionInfoDialog::ActionExecutionInfoDialog(sh::actions::ActionExecutionInfo* info, QWidget *parent) :
    QDialog(parent), info(info),
    ui(new Ui::ActionExecutionInfoDialog)
{
    ui->setupUi(this);
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    ui->btnCancel->setSizeByFactor(100);
    ui->btnCancel->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    ui->btnBackground->setSizeByFactor(100);
    ui->btnBackground->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    sm->style()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    sm->style()->createSheet().fontsize_header().applyTo(ui->label);
    sm->style()->createSheet().textcolor_lighter1().applyTo(ui->labelfromverb).applyTo(ui->labeltoverb);
    sm->style()->createSheet().background_highlighted().applyTo(ui->panelFeedback);
    ui->panelFeedback->setVisible(false);
    setWindowFlags( (windowFlags() | Qt::CustomizeWindowHint) & ~Qt::WindowCloseButtonHint);
}

sh::ui::ActionExecutionInfoDialog::~ActionExecutionInfoDialog()
{
    delete ui;
}

void sh::ui::ActionExecutionInfoDialog::keyPressEvent(QKeyEvent *e)
{
    if (e->key()==Qt::Key_Escape && currentFeedbackPanel)
        currentFeedbackPanel->cancelRequested();
    else
        QDialog::keyPressEvent(e);
}

void sh::ui::ActionExecutionInfoDialog::reject()
{
    // ignore reject
}

void sh::ui::ActionExecutionInfoDialog::setDetails(QString fv, QString fob, QString tv, QString tob)
{
    QFontMetrics fmetrics(ui->labelfromobj->font());
    QString elidedfromObj = fmetrics.elidedText(fob, Qt::ElideMiddle, ui->labelfromobj->width());
    QFontMetrics tmetrics(ui->labeltoobj->font());
    QString elidedtoObj = tmetrics.elidedText(tob, Qt::ElideMiddle, ui->labeltoobj->width());
    ui->labelfromobj->setText( elidedfromObj );
    ui->labelfromverb->setText( fv );
    ui->labeltoobj->setText( elidedtoObj );
    ui->labeltoverb->setText( tv );
}

void sh::ui::ActionExecutionInfoDialog::setHead(QString txt)
{
    ui->label->setText(txt);
}

void sh::ui::ActionExecutionInfoDialog::setProgress(bool isDeterminate, quint64 done, quint64 all, QString text)
{
    if (isDeterminate)
    {
        if (ui->progressBar->maximum() != 1000)
            ui->progressBar->setMaximum(1000);
        double dpval = ((double)done) / ((double)all);
        int pval = (int)(dpval * 1000);
        ui->progressBar->setValue( pval );
    }
    else
    {
        if (ui->progressBar->maximum() != 0)
            ui->progressBar->setMaximum(0);
    }
    ui->labelestimation->setText(text);
}

void sh::ui::ActionExecutionInfoDialog::askFeedbackViaWidget(sh::ui::feedbackpanels::FeedbackPanel* widget)
{
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        if (currentFeedbackPanel)
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details("Nested feedback panels are not supported.").resumeable(true));
        currentFeedbackPanel = widget;
        widget->setParent(ui->panelFeedback);
        ui->panelFeedback->layout()->addWidget(widget);
        ui->panelFeedback->setVisible(true);
        ui->progressBar->setVisible(false);
        ui->labelestimation->setVisible(false);
        widget->setVisible(true);
        info->setManualInterventionNeeded(true);
        int prefHeight = widget->preferredHeight();
        if (prefHeight != -1) {
            int outerHeight = height() - widget->height();
            int screenHeight = QApplication::desktop()->screenGeometry().height();
            int avail = screenHeight - outerHeight - 100;
            prefHeight *= 1.3;
            int useheight = (avail<prefHeight)?avail:prefHeight;
            int useouterheight = useheight + outerHeight;
            if (height() < useouterheight)
                resize(width(), useouterheight);
        }
        show();
    });
    widget->waitUntilClosed();
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        info->setManualInterventionNeeded(false);
        widget->setParent(ui->panelFeedback);
        ui->panelFeedback->setVisible(false);
        ui->progressBar->setVisible(true);
        ui->labelestimation->setVisible(true);
        currentFeedbackPanel = 0;
    });
}

int sh::ui::ActionExecutionInfoDialog::messageBox(QString text, QList<QString> answers, QString icon, int defaultanswer, int cancelanswer, QList<QString> answericons)
{
    sh::ui::feedbackpanels::MsgBox *msgbox = 0;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::feedbackpanels::MsgBox(text, answers, icon, 0, false, defaultanswer, cancelanswer, answericons);
    });
    askFeedbackViaWidget(msgbox);
    int result = msgbox->answer;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::ActionExecutionInfoDialog::inputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer)
{
    sh::ui::feedbackpanels::MsgBox *msgbox = 0;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::feedbackpanels::MsgBox(text, answers, icon, *value, false, defaultanswer, cancelanswer);
    });
    askFeedbackViaWidget(msgbox);
    *value = msgbox->text;
    int result = msgbox->answer;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::ActionExecutionInfoDialog::multilineInputBox(QString text, QList<QString> answers, QString *value, QString icon, int defaultanswer, int cancelanswer)
{
    sh::ui::feedbackpanels::MsgBox *msgbox = 0;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        msgbox = new sh::ui::feedbackpanels::MsgBox(text, answers, icon, *value, true, defaultanswer, cancelanswer);
    });
    askFeedbackViaWidget(msgbox);
    *value = msgbox->text;
    int result = msgbox->answer;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete msgbox;
    });
    return result;
}

int sh::ui::ActionExecutionInfoDialog::gridform(QString text, QList<QPair<QString, QWidget *> > form, QStringList answers, int defaultanswer, int cancelanswer)
{
    sh::ui::feedbackpanels::GridForm *gridform = 0;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        gridform = new sh::ui::feedbackpanels::GridForm(text, form, answers, defaultanswer, cancelanswer);
    });
    askFeedbackViaWidget(gridform);
    int result = gridform->answer;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete gridform;
    });
    return result;
}

int sh::ui::ActionExecutionInfoDialog::simpleChooserGridform(QString text, sh::actions::ActionExecutionUserFeedback::GridformEntries *entries, QStringList answers, int defaultanswer, int cancelanswer)
{
    QList<QPair<QString, QWidget*>> form;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        foreach(sh::actions::ActionExecutionUserFeedback::GridformEntry* entry, entries->list)
        {
            sh::ui::feedbackpanels::helpers::GridFormInnerSimpleChooser* w = new sh::ui::feedbackpanels::helpers::GridFormInnerSimpleChooser(&entry->choices, this);
            QPair<QString, QWidget*> row(entry->label, w);
            form.append(row);
        }
    });
    int result = gridform(text, form, answers, defaultanswer, cancelanswer);
    return result;
}

void sh::ui::ActionExecutionInfoDialog::on_btnCancel_clicked()
{
    info->cancel();
}

void sh::ui::ActionExecutionInfoDialog::on_btnBackground_clicked()
{
    setBackground(true);
}

void sh::ui::ActionExecutionInfoDialog::setLogicallyVisible(bool v)
{
    _logicallyVisible = v;
    _computevisibility();
}

void sh::ui::ActionExecutionInfoDialog::_computevisibility()
{
    if ((_logicallyVisible && !_isbackground) || _isforeground_forced)
    {
        show();
    }
    else
    {
        hide();
    }
}

bool sh::ui::ActionExecutionInfoDialog::isLogicallyVisible()
{
    return _logicallyVisible;
}

bool sh::ui::ActionExecutionInfoDialog::isBackground()
{
    return _isbackground;
}

void sh::ui::ActionExecutionInfoDialog::setBackground(bool v)
{
    if (!_isforeground_forced)
    {
        _isbackground = v;
        _computevisibility();
    }
}

void sh::ui::ActionExecutionInfoDialog::setForceForeground(bool v)
{
    _isforeground_forced = v;
    if (v)
    {
        ui->panelButtons->setVisible(false);
    }
    else
    {
        ui->panelButtons->setVisible(true);
    }
    _computevisibility();
}

bool sh::ui::ActionExecutionInfoDialog::credentialsDialog(QString text, bool showDomain, bool showUsername, bool showPassword, bool showAnonymous, bool showRemember, QString* domain, QString* username, QString* password, bool *isAnonymous, bool* isRemember)
{
    sh::ui::feedbackpanels::Credentials *form = 0;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        form = new sh::ui::feedbackpanels::Credentials(*domain, *username, *password, text, showDomain, showUsername, showPassword, showAnonymous, showRemember);
    });
    askFeedbackViaWidget(form);
    bool result = form->accepted;
    *domain = form->domain;
    *username = form->username;
    *password = form->password;
    *isAnonymous = form->anonymous;
    *isRemember = form->remember;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete form;
    });
    return result;
}
