/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QPainter>
#include <QStatusBar>
#include <QMouseEvent>
#include <QColor>
#include <QApplication>
#include <QPalette>
#include <QScreen>
#include <QPaintEngine>
#include "actions/actionexecutioninfo.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "ui/logviewdialog.h"
#include "actionexecutioninfopanel.h"


sh::ui::ActionExecutionInfoPanel::ActionExecutionInfoPanel(sh::actions::ActionExecutionInfo *info, QColor color, QWidget *parent) :
    QWidget(parent), info(info), gradient(100, 100)
{
    QColor color1 = (color.isValid()) ? color : QApplication::palette().color(QPalette::Highlight);
    setWidth(100);
    int h,s,l;
    color1.getHsl(&h, &s, &l);
    colorBase = QColor::fromHsl(h,200,200);
    colorProgress1 = colorBase.lighter(125);
    colorProgress2 = colorProgress1.lighter(145);
    colorFrame = colorBase.darker(130);
    colorText = colorBase.darker(170);
    colorTextDark = colorBase.darker(190);
    QLinearGradient lgrad;
    lgrad.setStart(0,0);
    lgrad.setFinalStop(100, 0);
    lgrad.setColorAt(0, colorProgress1);
    lgrad.setColorAt(0.25, colorProgress2);
    lgrad.setColorAt(0.5, colorProgress1);
    lgrad.setColorAt(0.75, colorProgress2);
    lgrad.setColorAt(1, colorProgress1);
    QPainter* p = new QPainter(&gradient);
    p->fillRect(QRect(0,0,100,100), QBrush(lgrad));
    delete p;
    _indeterminateAnimationTimer = new QTimer(this);
    _indeterminateAnimationTimer->setInterval(100);
    _indeterminateAnimationTimer->setSingleShot(false);
    connect(_indeterminateAnimationTimer, SIGNAL(timeout()), this, SLOT(slot_indeterminateanimationtimer()));
}

void sh::ui::ActionExecutionInfoPanel::showEvent(QShowEvent* e)
{
    QWidget::showEvent(e);
    emit visibilityChanged();
}

void sh::ui::ActionExecutionInfoPanel::hideEvent(QHideEvent* e)
{
    QWidget::hideEvent(e);
    emit visibilityChanged();
}

void sh::ui::ActionExecutionInfoPanel::_check_indeterminateanimationtimer_enable()
{
    if (isVisible() && !_progressDeterminate && !_isforeground_forced)
    {
        if (!_indeterminateAnimationTimer->isActive())
            _indeterminateAnimationTimer->start();
    }
    else if (_indeterminateAnimationTimer->isActive())
        _indeterminateAnimationTimer->stop();
}

void sh::ui::ActionExecutionInfoPanel::slot_indeterminateanimationtimer()
{
    if (_indeterminateAnimationCounter==0)
        _indeterminateAnimationCounter=100;
    _indeterminateAnimationCounter-=2;
    update();
}

void sh::ui::ActionExecutionInfoPanel::setLabel1(QString s)
{
    lbl1 = s;
    update();
}

void sh::ui::ActionExecutionInfoPanel::setLabel2(QString s)
{
    lbl2 = s;
    update();
}

void sh::ui::ActionExecutionInfoPanel::setForceForeground(bool v)
{
    _isforeground_forced = v;
    update();
}

void sh::ui::ActionExecutionInfoPanel::setProgress(bool isProgressDeterminate, quint64 progressDone, quint64 progressAll)
{
    _progressDeterminate = isProgressDeterminate;
    _progressDone = progressDone;
    _progressAll = progressAll;
    update();
}

void sh::ui::ActionExecutionInfoPanel::paintEvent(QPaintEvent*)
{
    _check_indeterminateanimationtimer_enable();
    QPainter painter( this );
    QString defaultFont = painter.font().defaultFamily();
    QRect viewp = painter.viewport();
    int innerwidth = viewp.width()-2;
    int innerheight = viewp.height()-2;
    if (_progressDeterminate)
    {
        painter.fillRect(viewp, QBrush(colorProgress1));
        painter.fillRect(1, 1, innerwidth*((double)_progressDone/(double)_progressAll), innerheight, QBrush(colorProgress2));
    }
    else
    {
        for (int x=0;x*100<(width()+100);x++) {
            for (int y=0;y*100<height();y++) {
                painter.drawPixmap(x*100-_indeterminateAnimationCounter, y*100, 100, 100, gradient);
            }
        }
    }
    int lh = height()*100/QApplication::screens()[0]->logicalDotsPerInchY();
    painter.setFont(QFont(defaultFont, lh*40/100));
    painter.setPen(QPen(colorText));
    QTextOption textopt;
    textopt.setWrapMode(QTextOption::NoWrap);
    textopt.setAlignment(Qt::AlignVCenter|Qt::AlignLeft);
    painter.drawText(QRect(5,1,innerwidth-2, innerheight), lbl1, textopt);
    if (_isforeground_forced)
    {
        painter.setFont(QFont(defaultFont, lh*60/100, QFont::Bold));
        QTextOption textopt;
        textopt.setAlignment(Qt::AlignVCenter|Qt::AlignRight);
        painter.setPen(QPen(colorProgress2));
        for (int ax=-2;ax<=2;ax++)
            for (int ay=-2;ay<=2;ay++)
            {
                painter.drawText(QRect(2,1,innerwidth-2+ax, innerheight+ay), "?!", textopt);
            }
        painter.setPen(QPen(colorTextDark));
        painter.drawText(QRect(2,1,innerwidth-2, innerheight), "?!", textopt);
    }
    painter.setPen(QPen(colorFrame));
    painter.drawRect(1,1,innerwidth, innerheight);
}

QSize sh::ui::ActionExecutionInfoPanel::sizeHint() const
{
    return QSize(_width, 1);
}

void sh::ui::ActionExecutionInfoPanel::setWidth(int width)
{
    _width = width*200/QApplication::screens()[0]->logicalDotsPerInchX();
    updateGeometry();
}

void sh::ui::ActionExecutionInfoPanel::mousePressEvent(QMouseEvent*)
{
    emit clicked();
}

sh::ui::ActionExecutionInfoPanel* sh::ui::ActionExecutionInfoPanel::showErrorPanel()
{
    sh::ui::ActionExecutionInfoPanel* result;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        result = new sh::ui::ActionExecutionInfoPanel(0, QColor(100,30,30));
        result->setProgress(false,0,0);
        result->setWidth(50);
        result->setLabel1(tr("Problem"));
        sh::ui::MainWindow::mainWindow()->insertIntoStatusbar(0, result);
        connect(result, &sh::ui::ActionExecutionInfoPanel::clicked, [=](){
            sh::ui::LogViewDialog* logdlg = new sh::ui::LogViewDialog(sh::ui::MainWindow::mainWindow(),
                    tr("Some problems occurred.\n\nIf you encounter wrong program behavior, you should restart Shallot."),
                    tr("This can be a software error in Shallot or it is the fault of a plugin. If you are interested in technical details, please read the log."),
                    sh::tools::LogSeverity::ERROR);
            logdlg->show();
            connect(logdlg, &QDialog::finished, [=](int){
                logdlg->deleteLater();
            });
            result->deleteLater();
        });
    });
    return result;
}
