/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QTextStream>
#include "configuration/configurationmanager.h"
#include "tools/logger.h"
#include "tools/iconmanager.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "ui/logviewdialog.h"
#include "exceptiondialog.h"
#include "ui_exceptiondialog.h"


sh::ui::ExceptionDialog::ExceptionDialog(bool mayClose, QWidget *parent) :
    QDialog(parent), mayClose(mayClose), cancelled(false),
    ui(new Ui::ExceptionDialog)
{
    ui->setupUi(this);
    sh::exceptions::Exception::executeGuarded([&](){
        ui->btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("arrow-right-double"));
        ui->btnRetry->setIcon(sh::tools::IconManager::instance()->getIcon("view-refresh"));
        ui->btnCancel->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    }, sh::exceptions::ExecuteGuardFlag::IGNORE_ALL_RESUMEABLE_ERRORS_SILENTLY);
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    sm->style()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    sm->style()->createSheet().fontsize_header().applyTo(ui->lblHead);
    sm->style()->createSheet()
            .fontsize_smaller()
            .textcolor_lighter1().applyTo(ui->lblLogfile);
    on_btnShowDetails_toggled(false);
}

sh::ui::ExceptionDialog::~ExceptionDialog()
{
    delete ui;
}

int sh::ui::ExceptionDialog::exceptionDialog(QString error1, QString error2, QString details, QString icon, bool mayRetry, bool mayClose, bool mayCancel, bool showLoglabelAndDetails)
{
    QMutexLocker _l1(&_inShutdownMutex);
    if (_inShutdown)
        return QDialog::Rejected;
    int result;
    bool cancelled = false;
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        sh::ui::ExceptionDialog dlg(mayClose);
        dlg.ui->lblHead->setText(error1);
        dlg.ui->label->setText(error2);
        dlg.ui->txtDetails->setText(details);
        dlg.ui->lblIcon->setPixmap(sh::tools::IconManager::instance()->getPixmap(icon, 48));
        dlg.ui->btnClose->setVisible(mayClose);
        dlg.ui->btnExit->setVisible(!mayClose);
        dlg.ui->btnRetry->setVisible(mayRetry);
        dlg.ui->btnCancel->setVisible(mayCancel);
        dlg.ui->btnShowDetails->setVisible(showLoglabelAndDetails);
        dlg.resize(600,450);
        result = dlg.exec();
        cancelled = dlg.cancelled;
    });
    if (cancelled)
        throw sh::exceptions::CancelException();
    return result;
}

int sh::ui::ExceptionDialog::exceptionDialog(sh::exceptions::Exception& ex)
{
    bool isretryable=ex.isRetryable() && ex.isResumeable();
    bool isresumeable=ex.isResumeable();
    bool showloghint=true;
    QString details=ex.details();
    QString error1 = isresumeable
            ? QObject::tr("Some operation failed to succeed.")
            : sh::exceptions::Exception::SHALLOT_MUST_CLOSE_TEXT;
    QString error2 = ex.message();
    if (ex.isDetailsAreInteresting())
        error2 += "\n\n" + ex.details();
    QString icon=(isresumeable
                 ?( isretryable?"error-retryable":"error-normal" )
                 :"error-critical");
    bool mayCancel = !sh::exceptions::ExceptionHandlerSettingsManager::instance()->handlerSettings()->exceptionHandler_cancel.isEmpty()
            && isresumeable;
    return exceptionDialog(error1, error2, details, icon, isretryable, isresumeable, mayCancel, showloghint);
}

void sh::ui::ExceptionDialog::on_btnExit_clicked()
{
    _inShutdown = true;
    this->done(QDialog::Rejected);
}

void sh::ui::ExceptionDialog::on_btnClose_clicked()
{
    this->done(QDialog::Rejected);
}

void sh::ui::ExceptionDialog::on_btnCancel_clicked()
{
    cancelled = true;
    this->done(QDialog::Rejected);
}

void sh::ui::ExceptionDialog::on_btnRetry_clicked()
{
    this->done(QDialog::Accepted);
}

void sh::ui::ExceptionDialog::on_btnShowLog_clicked()
{
    on_btnClose_clicked();
    sh::ui::LogViewDialog* dlg = new sh::ui::LogViewDialog(sh::ui::MainWindow::mainWindow(), QString(), QString(), sh::tools::LogSeverity::ERROR);
    dlg->show();
}

void sh::ui::ExceptionDialog::on_btnExitAndSaveLog_clicked()
{
    QString log = sh::tools::Logger::instance()->getLogAsText();
    QString filename = QDir::homePath()+QString("/shallot.crashlog.%1.txt").arg(QDateTime::currentDateTime().toString(Qt::ISODate));
    QFile file(filename);
    if (file.open(QIODevice::ReadWrite)) {
        QTextStream stream(&file);
        stream << log << endl;
        file.close();
    }
    on_btnExit_clicked();
}

void sh::ui::ExceptionDialog::on_btnShowDetails_toggled(bool checked)
{
    ui->btnExitAndSaveLog->setVisible(checked &&!mayClose);
    ui->btnShowLog->setVisible(checked && mayClose);
    ui->lblLogfile->setVisible(checked);
    ui->txtDetails->setVisible(checked);
    if (checked)
        ui->txtDetails->setFocus();
}

bool sh::ui::ExceptionDialog::_inShutdown = false;

QMutex sh::ui::ExceptionDialog::_inShutdownMutex;
