/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QScrollBar>
#include "base/filesystemnode.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "configuration/configurationmanager.h"
#include "configuration/uistyle.h"
#include "tools/iconmanager.h"
#include "tools/singletoninitializer.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "filepropertydialog.h"
#include "ui_filepropertydialog.h"


namespace sh{namespace ui{namespace filepropertydialog{INITIALIZE_STATIC_IN_MAINTHREAD(50000, FilePropertyDialog)}}}

sh::ui::filepropertydialog::FilePropertyDialog::FilePropertyDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::FilePropertyDialog)
{
    ui->setupUi(this);

    scrollAreaLayout = qobject_cast<QVBoxLayout*>(ui->scrollAreaWidgetContents->layout());

    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("window-close"));
    sm->style()->createSheet()
                .background_liketitlebar()
            .applyTo(ui->panelHead);
    sm->style()->createSheet()
                .fontsize_header()
            .applyTo(ui->label);
    sm->style()->createSheet()
                .background_highlighted()
            .applyTo(ui->scrollAreaWidgetContents);
    connect(ui->scrollArea->verticalScrollBar(), &QScrollBar::valueChanged, this, &FilePropertyDialog::selectTabByScrollPosition);
    ui->scrollAreaWidgetContents->setFocus();
}

QList<sh::ui::filepropertydialog::FilePropertyDialogTabFactory*> sh::ui::filepropertydialog::FilePropertyDialog::getpropertytabs()
{
    QList<FilePropertyDialogTabFactory*> result;
    QList<int> keys = propertytabs.keys();
    qSort(keys);
    foreach(int i, keys)
        foreach(std::shared_ptr<FilePropertyDialogTabFactory> f, propertytabs.values(i))
            result.append(f.get());
    return result;
}

void sh::ui::filepropertydialog::FilePropertyDialog::showItems(QList<std::shared_ptr<sh::base::FilesystemNode> > nodes)
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    this->nodes = nodes;
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    foreach(FilePropertyDialogTabFactory* tabfactory, getpropertytabs())
    {
        std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> tab = tabfactory->construct();
        if(tab)
        {
            tab->_dialog = shared_from_this();
            QList<QString> properties = tab->properties();
            QMap<int, QWidget*> widgetmap;
            if (!properties.isEmpty())
            {
                QWidget* widget = new QWidget(this);
                QVBoxLayout* layout = new QVBoxLayout(widget);
                widget->setLayout(layout);
                QLabel* headlabel = new QLabel(this);
                QString titleWithoutMnemonic;
                bool lastWasAmp = false;
                foreach(QChar c, tab->title())
                {
                    if (c != '&' || lastWasAmp)
                        titleWithoutMnemonic += c;
                    lastWasAmp = (c == '&');
                }
                headlabel->setText(titleWithoutMnemonic);
                headlabel->setWordWrap(true);
                headlabel->setAlignment(Qt::AlignHCenter);
                sm->style()->createSheet()
                            .fontsize_byFactor(150)
                        .applyTo(headlabel);
                layout->addWidget(headlabel);
                for (int i=0; i<properties.length(); i++)
                {
                    QLabel* label = new QLabel(this);
                    label->setText(properties[i]);
                    label->setWordWrap(true);
                    sm->style()->createSheet()
                                .fontsize_byFactor(130)
                                .customcss("padding:0px 0px 0px 30px")
                            .applyTo(label);
                    layout->addWidget(label);
                    QWidget* pwidget = tab->createWidget(i);
                    pwidget->setParent(this);
                    pwidget->installEventFilter(this);
                    layout->addWidget(pwidget);
                    widgetmap[i] = pwidget;
                    layout->addSpacing(1); //will be set later
                }
                scrollAreaLayout->addWidget(widget);
                QWidget* internaltabwidget = new QWidget(this);
                _internaltabwidgets.append(internaltabwidget);
                _tabheads.append(tab->title());
            }
            tab->_nodes = nodes;
            tabs[tab] = widgetmap;
        }
    }
    scrollAreaLayout->addStretch(1);
    show();
    foreach(std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> tab, tabs.keys())
        foreach(QWidget* w, tabs[tab].values())
            w->hide();
    on_btnRefresh_clicked();
}

void sh::ui::filepropertydialog::FilePropertyDialog::_updateWidget(FilePropertyDialogTab* tab, int i, QWidget* widget, std::shared_ptr<sh::base::Operation> op)
{
    sh::tools::ThreadPool::enqueueForce([=](){
        sh::exceptions::Exception::executeGuarded_errorpanel([&](){
            tab->updateWidget(i, widget, op.get());
        });
    }, shared_from_this());
}

bool sh::ui::filepropertydialog::FilePropertyDialog::eventFilter(QObject *obj, QEvent *event)
{
    if (isVisible()/*otherwise dialog is closing*/
            && (event->type() == QEvent::ShowToParent || event->type() == QEvent::HideToParent))
    {
        bool newVisibility = (event->type() == QEvent::ShowToParent);
        QWidget* wobj = qobject_cast<QWidget*>(obj);
        QWidget* wpar = wobj->parentWidget();
        QVBoxLayout* wparlayout = qobject_cast<QVBoxLayout*>(wpar->layout());
        int i = wparlayout->indexOf(wobj);
        // handle the label visibility
        QWidget* lbl = qobject_cast<QWidget*>(wparlayout->itemAt(i-1)->widget());
        QSpacerItem* spacer = (QSpacerItem*)(wparlayout->itemAt(i+1));
        lbl->setVisible(newVisibility);
        spacer->changeSize(1, newVisibility?20:0);// this fails sometimes :-/ but not a major thing...

        // check if the parent box should be visible
        bool boxvisible = false;
        wpar->setVisible(true); // otherwise our tests fail
        for(int i=1; i<wparlayout->count(); i++)
        {
            QWidget* w = qobject_cast<QWidget*>(wparlayout->itemAt(i)->widget());
            if (w && w->isVisible())
            {
                boxvisible = true;
                break;
            }
        }
        wpar->setVisible(boxvisible);
        // handle the tab visibility
        int tabidx = scrollAreaLayout->indexOf(wpar);
        if (boxvisible)
        {
            int wtabidx = 0;
            bool insert = true;
            while (wtabidx < ui->tabWidget->count())
            {
                QWidget* w = ui->tabWidget->widget(wtabidx);
                int _tabidx = _internaltabwidgets.indexOf(w);
                if (_tabidx == tabidx)
                {
                    insert = false;
                    break;
                }
                else if (_tabidx > tabidx)
                    break;
                wtabidx++;
            }
            if (insert)
            {
                int _i = ui->tabWidget->currentIndex();
                ui->tabWidget->insertTab(wtabidx, _internaltabwidgets[tabidx], _tabheads[tabidx]);
                if (_i == 0)
                    ui->tabWidget->setCurrentIndex(_i);
            }
        }
        else
        {
            QWidget* w =  _internaltabwidgets[tabidx];
            int i = ui->tabWidget->indexOf(w);
            if (i!=-1)
                ui->tabWidget->removeTab(i);
        }
    }
    return QObject::eventFilter(obj, event);
}

sh::ui::filepropertydialog::FilePropertyDialog::~FilePropertyDialog()
{
    delete ui;
}

QHash<int, std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTabFactory>> sh::ui::filepropertydialog::FilePropertyDialog::propertytabs;

void sh::ui::filepropertydialog::FilePropertyDialog::showForNodes(QList<std::shared_ptr<sh::base::FilesystemNode>> nodes)
{
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> dlg = std::make_shared<sh::ui::filepropertydialog::FilePropertyDialog>();
    dlg->setWindowTitle( nodes.length()==1 ? tr("Properties: %1").arg(nodes.first()->eurl()->basename()) : tr("Properties: %1 items").arg(nodes.length()) );
    dlg->ui->label->setText( nodes.length()==1 ? tr("Properties of '%1'").arg(nodes.first()->eurl()->basename()) : tr("Properties of %1 items").arg(nodes.length()) );
    QList<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog>>* _dlg = new QList<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog>>({dlg});
    connect(dlg.get(), &QDialog::rejected, [=](){
        delete _dlg;
    });
    dlg->showItems(nodes);
}

void sh::ui::filepropertydialog::FilePropertyDialog::addTabFactory(int i, std::shared_ptr<FilePropertyDialogTabFactory> factory)
{
    propertytabs.insertMulti(i, factory);
}

void sh::ui::filepropertydialog::FilePropertyDialog::on_btnClose_clicked()
{
    reject();
}

sh::ui::filepropertydialog::FilePropertyDialogTabFactoryByFunction::FilePropertyDialogTabFactoryByFunction(std::function<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab>()> fct)
                : fct(fct)
{
}

std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> sh::ui::filepropertydialog::FilePropertyDialogTabFactoryByFunction::construct()
{
    return fct();
}

void sh::ui::filepropertydialog::FilePropertyDialog::_initialize()
{
}

void sh::ui::filepropertydialog::FilePropertyDialog::_shutdown()
{
}

const int sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE = 10000;
const int sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_VERYIMPORTANT = 20000;
const int sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_IMPORTANT = 30000;
const int sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_NORMAL = 40000;
const int sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_EXOTIC = 50000;

sh::ui::filepropertydialog::FilePropertyDialogTabFactory::~FilePropertyDialogTabFactory()
{
}

void sh::ui::filepropertydialog::FilePropertyDialogTab::refresh()
{
    sh::tools::MainThread::dispatcher()->invokeSync([=](){
        std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> dlg = _dialog.lock();
        if (dlg)
            dlg->on_btnRefresh_clicked();
    });
}

QList<std::shared_ptr<sh::base::FilesystemNode> > sh::ui::filepropertydialog::FilePropertyDialogTab::nodes()
{
    return _nodes;
}

QWidget* sh::ui::filepropertydialog::FilePropertyDialogTab::widgetAt(int i)
{
    return _dialog.lock()->tabs[shared_from_this()][i];
}

std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> sh::ui::filepropertydialog::FilePropertyDialogTab::dialog()
{
    return _dialog.lock();
}

sh::ui::filepropertydialog::FilePropertyDialogTab::~FilePropertyDialogTab()
{
}

void sh::ui::filepropertydialog::FilePropertyDialog::on_btnRefresh_clicked()
{
    if (!isVisible())
        return;
    std::shared_ptr<sh::base::Operation> op = std::make_shared<sh::base::Operation>();
    foreach(std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> tab, tabs.keys())
        foreach(int i, tabs[tab].keys())
            _updateWidget(tab.get(), i, tabs[tab][i], op);
}

void sh::ui::filepropertydialog::FilePropertyDialog::on_tabWidget_currentChanged(int index)
{
    if (_skip_on_tabWidget_currentChanged)
        return;
    _skip_selectTabByScrollPosition++;
    int tindex = _internaltabwidgets.indexOf(ui->tabWidget->widget(index));
    QLayoutItem* li = scrollAreaLayout->itemAt(tindex);
    if (li)
    {
        int ypos = li->geometry().y();
        ui->scrollArea->verticalScrollBar()->setValue(ypos * (ui->scrollArea->verticalScrollBar()->maximum()) / std::max(1, ui->scrollAreaWidgetContents->height()-ui->scrollArea->height()));
    }
    _skip_selectTabByScrollPosition--;
}

void sh::ui::filepropertydialog::FilePropertyDialog::selectTabByScrollPosition()
{
    if (_skip_selectTabByScrollPosition)
        return;
    QWidget* w = ui->scrollArea->childAt(10, 10);
    int iw = scrollAreaLayout->indexOf(w);
    if (iw!=-1)
    {
        int i = ui->tabWidget->indexOf(_internaltabwidgets[iw]);
        _skip_on_tabWidget_currentChanged++;
        ui->tabWidget->setCurrentIndex(i);
        _skip_on_tabWidget_currentChanged--;
    }
}
