/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILEPROPERTYDIALOG_H
#define FILEPROPERTYDIALOG_H

#include <functional>
#include <memory>
#include <QDialog>
#include <QHeaderView>
#include <QMap>
#include <QVBoxLayout>


namespace Ui {
class FilePropertyDialog;
}

namespace sh {
    namespace base {
        class FilesystemNode;
        class Operation;
    }
    namespace ui {
        namespace filepropertydialog {
            class FilePropertyDialogTabFactory;
            class FilePropertyDialogTab;

            /**
             * @brief The 'Properties' dialog
             */
            class FilePropertyDialog : public QDialog, public std::enable_shared_from_this<FilePropertyDialog>
            {
                friend class FilePropertyDialogTab;
                Q_OBJECT
            public:
                /**
                 * @brief Shows the Properties dialog for some nodes (typically one one).
                 */
                static void showForNodes(QList<std::shared_ptr<sh::base::FilesystemNode> > nodes);
                /**
                 * @brief Adds a FilePropertyDialogTabFactory so the Properties dialogs show its
                 * content.
                 *
                 * See also the REGISTER_FILEPROPERTYDIALOGTAB macro.
                 *
                 * @param i An index which controls the display order. Use one of the
                 * REGISTER_FILEPROPERTYDIALOGTAB_INDEX_* values in FilePropertyDialogTabFactory as base values.
                 */
                static void addTabFactory(int i, std::shared_ptr<FilePropertyDialogTabFactory> factory);
                ~FilePropertyDialog();

                /** \brief
                 * Only called by the Shallot infrastructure for initialization.
                 */
                static void _initialize();
                /** \brief
                 * Only called by the Shallot infrastructure for shutdown.
                 */
                static void _shutdown();

                /** \brief Constructed only by the infrastructure and made available otherwise.
                 */
                explicit FilePropertyDialog(QWidget *parent = 0);

            public slots:
                void on_btnClose_clicked();
                void on_btnRefresh_clicked();
                void on_tabWidget_currentChanged(int index);
                void selectTabByScrollPosition();

            protected:
                bool eventFilter(QObject *obj, QEvent *event);
                void showItems(QList<std::shared_ptr<sh::base::FilesystemNode> > nodes);
                QList<sh::ui::filepropertydialog::FilePropertyDialogTabFactory*> getpropertytabs();

            private:
                Ui::FilePropertyDialog *ui;
                void _updateWidget(FilePropertyDialogTab *tab, int i, QWidget* widget, std::shared_ptr<sh::base::Operation> op);
                QMap<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab>, QMap<int, QWidget*>> tabs;
                QList<QWidget*> _internaltabwidgets;
                QList<QString> _tabheads;
                int _skip_on_tabWidget_currentChanged = 0;
                int _skip_selectTabByScrollPosition = 0;
                QList<std::shared_ptr<sh::base::FilesystemNode> > nodes;
                QVBoxLayout* scrollAreaLayout;
                static QHash<int, std::shared_ptr<FilePropertyDialogTabFactory>> propertytabs;
            };

            /**
             * @brief Abstract base class for one tab in the Properties dialog.
             *
             * Subclass and register this class for providing additional content in the
             * Properties dialog.
             *
             * Register it by adding a FilePropertyDialogTabFactory via FilePropertyDialog::addTabFactory.
             */
            class FilePropertyDialogTab : public std::enable_shared_from_this<FilePropertyDialogTab>
            {
                friend class FilePropertyDialog;
            public:
                /** @brief The tab title.
                 */
                virtual QString title() =0;
                /** @brief The list of properties.
                 *
                 * Each property is displayed with a header and a widget (created in createWidget)
                 * with some content (populated in updateWidget).
                 */
                virtual QList<QString> properties() =0;
                /** @brief Creates a widget for the i-th property.
                 */
                virtual QWidget* createWidget(int i) =0;
                /** @brief Populates the widget for the i-th property with actual content.
                 */
                virtual void updateWidget(int i, QWidget* widget, sh::base::Operation *op) =0;
                /**
                 * @brief Refreshes the complete content.
                 *
                 * Typically called after some user actions in a property widget require that all the other content must be refreshed.
                 */
                void refresh();
                /** @brief The nodes to show.
                 */
                QList<std::shared_ptr<sh::base::FilesystemNode> > nodes();

                QWidget* widgetAt(int i);
                std::shared_ptr<FilePropertyDialog> dialog();

                virtual ~FilePropertyDialogTab();

            private:
                std::weak_ptr<FilePropertyDialog> _dialog;
                QList<std::shared_ptr<sh::base::FilesystemNode> > _nodes;
            };

            /** \brief Abstract factory for creating property dialog tabs for the selected nodes.
             *
             * Register an instance of this class with FilePropertyDialog::addTabFactory in order to make a property tab implementation available.
             * See also FilePropertyDialogTabFactoryByFunction.
             */
            class FilePropertyDialogTabFactory
            {
            public:
                /** @brief Base value for display indexes of core (i.e. maximum important) tabs.
                 *
                 * Used in FilePropertyDialog::addTabFactory.
                 */
                static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE;
                /** @brief Base value for display indexes of very important tabs.
                 *
                 * Used in FilePropertyDialog::addTabFactory.
                 */
                static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_VERYIMPORTANT;
                /** @brief Base value for display indexes of important tabs.
                 *
                 * Used in FilePropertyDialog::addTabFactory.
                 */
                static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_IMPORTANT;
                /** @brief Base value for display indexes of tabs with medium importance.
                 *
                 * Used in FilePropertyDialog::addTabFactory.
                 */
                static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_NORMAL;
                /** @brief Base value for display indexes of exotic (i.e. less important) tabs.
                 *
                 * Used in FilePropertyDialog::addTabFactory.
                 */
                static const int REGISTER_FILEPROPERTYDIALOGTAB_INDEX_EXOTIC;
                /** @brief Constructs a fresh instance of a FilePropertyDialogTab implementation.
                 */
                virtual std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> construct() =0;
                virtual ~FilePropertyDialogTabFactory();
            };

            /**
             * @brief Implementation of FilePropertyDialogTabFactory for making custom FilePropertyDialogTab implementations available.
             *
             * This implementation relies on an external factory function which must be provided to the constructor.
             */
            class FilePropertyDialogTabFactoryByFunction : public FilePropertyDialogTabFactory
            {
            public:
                FilePropertyDialogTabFactoryByFunction(std::function<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab>()> fct);
                virtual std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab> construct();
            private:
                std::function<std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab>()> fct;
            };

        }
    }
}

#define REGISTER_FILEPROPERTYDIALOGTAB(I,T) sh::ui::filepropertydialog::FilePropertyDialog::addTabFactory((I), std::make_shared<sh::ui::filepropertydialog::FilePropertyDialogTabFactoryByFunction>([](){ return (std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTab>) new T(); }))

#endif // FILEPROPERTYDIALOG_H
