/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "actions/common/actionabstracttransfertree.h"
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "base/filesystemnode.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "detailcolumns/detailcolumnextendedattributes.h"
#include "tools/singletoninitializer.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "filepropertydialogtabactionsview.h"
#include "filepropertydialogtabextendedattributes.h"
#include "filepropertydialogtabtableview.h"


namespace sh{namespace ui{namespace filepropertydialog{INITIALIZE_STATIC_IN_MAINTHREAD(56001, FilePropertyDialogTabExtendedAttributes)}}}

sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::FilePropertyDialogTabExtendedAttributes()
{
}

QString sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::title()
{
    return QObject::tr("E&xtended attributes");
}

QList<QString> sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::properties()
{
    QStringList result;
    result << "";
    return result;
}

QWidget* sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::createWidget(int)
{
    FilePropertyDialogTabActionsView* aview = new FilePropertyDialogTabActionsView();
    FilePropertyDialogTabTableView* content = new FilePropertyDialogTabTableView(aview);
    aview->setContent(content);
    aview->setButtons(QStringList({QObject::tr("Add attribute"), QObject::tr("Change attribute"), QObject::tr("Remove attribute")}));
    QObject::connect(aview, &FilePropertyDialogTabActionsView::buttonTriggered, [=](int i){slot_buttontriggered(i);});
    return aview;
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::slot_buttontriggered(int i)
{
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes> me =
            std::static_pointer_cast<sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes>(shared_from_this());
    sh::tools::ThreadPool::enqueueForce([=](){
        std::shared_ptr<sh::actions::ActionActionItem> a;
        if (i == 0)
            a = std::make_shared<ActionAddAttribute>(me); // add
        else if (i == 1)
            a = std::make_shared<ActionChangeAttribute>(me); // change
        else if (i == 2)
            a = std::make_shared<ActionRemoveAttribute>(me); // remove
        a->initializeSync();
        a->execute();
    });
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::updateWidget(int, QWidget* widget, sh::base::Operation* op)
{
    QList<QPair<QString, QString>> values;
    if (nodes().length() == 1)
    {
        QVariant vv = sh::detailcolumns::DetailColumnExtendedAttributes::instance()->requestValue(nodes().first(), op);
        QMap<QString, QByteArray> vals = sh::detailcolumns::DetailColumnExtendedAttributes::getMapByValue(vv);
        foreach(QString key, vals.keys())
        {
            QByteArray bval = vals[key];
            QString sval = QObject::tr("binary or large content (%1 byte)").arg(bval.length());
            if (bval.length() < 1000)
            {
                QTextCodec* utf8 = QTextCodec::codecForName("UTF-8");
                QTextCodec::ConverterState converterstate;
                QString _sval = utf8->toUnicode(bval.data(), bval.length(), &converterstate);
                if (!converterstate.invalidChars && !converterstate.remainingChars)
                    sval = _sval;
            }
            values.append(QPair<QString, QString>(key, sval));
        }
    }
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        FilePropertyDialogTabTableView* tw = qobject_cast<FilePropertyDialogTabTableView*>(qobject_cast<FilePropertyDialogTabActionsView*>(widget)->content());
        tw->setContent(values);
        widget->show();
    });
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::_initialize()
{
    REGISTER_FILEPROPERTYDIALOGTAB(sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_EXOTIC-1000, sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes);
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::_shutdown()
{
}

sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionAddAttribute::ActionAddAttribute(std::shared_ptr<FilePropertyDialogTabExtendedAttributes> tab)
            : tab(tab)
{
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionAddAttribute::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead(tr("Adding extended attribute"));
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> _dlg = tab->dialog();
    if (!_dlg)
        return;
    QString name = info->userfeedback()->simpleInputBox(QObject::tr("Please enter the name for the new attribute."), "");
    if (!name.isEmpty())
    {
        QString value = "";
        if (info->userfeedback()->multilineInputBox(QObject::tr("Please enter the content for the new attribute. You should only proceed if you know, what you are doing. Bad changes here may lead to problems."),
                                                                QStringList({QObject::tr("OK"), QObject::tr("Cancel")}), &value, QString(), -1, 0) == 0)
        {
            foreach (std::shared_ptr<sh::base::FilesystemNode> node, tab->nodes()) {
                node->handler()->setExtendedAttribute(info->operation(), node->eurl(), name, value.toUtf8());
            }
            tab->refresh();
        }
    }
}

sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionChangeAttribute::ActionChangeAttribute(std::shared_ptr<FilePropertyDialogTabExtendedAttributes> tab)
            : tab(tab)
{
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionChangeAttribute::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead(tr("Changing extended attribute"));
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> _dlg = tab->dialog();
    if (!_dlg)
        return;
    sh::ui::filepropertydialog::FilePropertyDialogTabTableView* table = qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabTableView*>(
            qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabActionsView*>(tab->widgetAt(0))->content());
    QModelIndexList mil = table->selectionModel()->selectedIndexes();
    if (mil.length() != 1)
    {
        info->userfeedback()->simpleMessageBox(QObject::tr("Please select one attribute in the list before."));
    }
    else
    {
        QString att = table->model()->headerData(mil[0].row(), Qt::Vertical).toString();
        foreach(std::shared_ptr<sh::base::FilesystemNode> node, tab->nodes())
        {
            QByteArray bval = node->handler()->getExtendedAttribute(info->operation(), node->eurl(), att);
            QTextCodec* utf8 = QTextCodec::codecForName("UTF-8");
            QTextCodec::ConverterState converterstate;
            QString _sval = utf8->toUnicode(bval.data(), bval.length(), &converterstate);
            if (converterstate.invalidChars || converterstate.remainingChars)
            {
                info->userfeedback()->simpleMessageBox(QObject::tr("This attribute seems to contain binary data. Please remove and recreate it instead."));
            }
            else
            {
                QString value = _sval;
                if (info->userfeedback()->multilineInputBox(QObject::tr("Please modify the attribute content. You should only proceed if you know, what you are doing. Bad changes here may lead to problems."),
                                                                        QStringList({QObject::tr("OK"), QObject::tr("Cancel")}), &value, QString(), -1, 0) == 0)
                {
                    node->handler()->setExtendedAttribute(info->operation(), node->eurl(), att, value.toUtf8());
                    tab->refresh();
                }
            }
        }
    }
}

sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionRemoveAttribute::ActionRemoveAttribute(std::shared_ptr<FilePropertyDialogTabExtendedAttributes> tab)
            : tab(tab)
{
}

void sh::ui::filepropertydialog::FilePropertyDialogTabExtendedAttributes::ActionRemoveAttribute::action(sh::actions::ActionExecutionInfo* info)
{
    info->setHead(tr("Removing extended attribute"));
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> _dlg = tab->dialog();
    if (!_dlg)
        return;
    sh::ui::filepropertydialog::FilePropertyDialogTabTableView* table = qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabTableView*>(
            qobject_cast<sh::ui::filepropertydialog::FilePropertyDialogTabActionsView*>(tab->widgetAt(0))->content());
    QModelIndexList mil = table->selectionModel()->selectedIndexes();
    if (mil.length() == 0)
    {
        info->userfeedback()->simpleMessageBox(QObject::tr("Please select attributes in the list before."));
    }
    else
    {
        QStringList atts;
        for(int i=0; i<mil.length(); i++)
            atts << table->model()->headerData(mil[i].row(), Qt::Vertical).toString();
        bool doit = true;
        if (atts.length() == 1)
            doit = sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes == info->userfeedback()->simpleMessageBox(QObject::tr("Do you really want to delete the attribute '%1'?").arg(atts[0]), (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes|(int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::No);
        else
            doit = sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes == info->userfeedback()->simpleMessageBox(QObject::tr("Do you really want to delete the %1 selected attributes?").arg(atts.length()), (int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::Yes|(int)sh::actions::ActionExecutionUserFeedback::MessageBoxButton::No);
        if (doit)
        {
            foreach (std::shared_ptr<sh::base::FilesystemNode> node, tab->nodes()) {
                foreach(QString att, atts)
                    node->handler()->removeExtendedAttribute(info->operation(), node->eurl(), att);
            }
            tab->refresh();
        }
    }
}
