/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QLabel>
#include "actions/common/actionabstracttransfertree.h"
#include "actions/actionexecutioninfo.h"
#include "actions/actionexecutionuserfeedback.h"
#include "base/filesystemnode.h"
#include "base/eurl.h"
#include "base/operation.h"
#include "detailcolumns/detailcolumncustomattributes.h"
#include "tools/singletoninitializer.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "ui/actionexecutioninfodialog.h"
#include "ui/feedbackpanels/unixpermissions.h"
#include "filepropertydialogtabactionsview.h"
#include "filepropertydialogtabunixpermissions.h"

#ifdef Q_OS_UNIX
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#endif


namespace sh{namespace ui{namespace filepropertydialog{INITIALIZE_STATIC_IN_MAINTHREAD(56003, FilePropertyDialogTabUnixPermissions)}}}

sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::FilePropertyDialogTabUnixPermissions()
{
}

QString sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::title()
{
    return QObject::tr("&Permissions");
}

QList<QString> sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::properties()
{
    QStringList result;
    result << "";
    return result;
}

QWidget* sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::createWidget(int)
{
    FilePropertyDialogTabActionsView* aview = new FilePropertyDialogTabActionsView();
    QLabel* lbl = new QLabel();
    lbl->setWordWrap(true);
    aview->setContent(lbl);
    aview->setButtons(QStringList({QObject::tr("Change permissions")}));
    QObject::connect(aview, &FilePropertyDialogTabActionsView::buttonTriggered, [=](int i){slot_buttontriggered(i);});
    return aview;
}

void sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::slot_buttontriggered(int)
{
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions> me =
            std::static_pointer_cast<sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions>(shared_from_this());
    sh::tools::ThreadPool::enqueueForce([=](){
        std::shared_ptr<sh::actions::ActionActionItem> a = std::make_shared<ActionChange>(me);
        a->initializeSync();
        a->execute();
    });
}

void sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::updateWidget(int, QWidget* widget, sh::base::Operation* op)
{
#ifdef Q_OS_UNIX
    QString desc = "";
    if (nodes().length() == 1)
    {
        QVariant vv = sh::detailcolumns::DetailColumnCustomAttributes::instance()->requestValue(nodes().first(), op);
        QMap<QString, QString> vals = sh::detailcolumns::DetailColumnCustomAttributes::getMapByValue(vv);
        if (vals.contains("unix_permissions"))
        {
            QStringList vl = vals["unix_permissions"].split(";");
            int flags = vl[0].toInt();
            int uid = vl[1].toInt();
            int gid = vl[2].toInt();
            if (flags & S_ISUID)
                desc += QObject::tr("set user on execute") + "\n";
            if (flags & S_ISGID)
                desc += QObject::tr("set group on execute") + "\n";
            if (flags & S_ISVTX)
                desc += QObject::tr("sticky") + "\n";
            QString usermay = userPermissionsDescription(flags, S_IRUSR, S_IWUSR, S_IXUSR);
            QString groupmay = userPermissionsDescription(flags, S_IRGRP, S_IWGRP, S_IXGRP);
            QString othersmay = userPermissionsDescription(flags, S_IROTH, S_IWOTH, S_IXOTH);
            desc += QObject::tr("user '%1' may: %2").arg(getUserName(uid)).arg(usermay) + "\n";
            desc += QObject::tr("group '%1' may: %2").arg(getGroupName(gid)).arg(groupmay) + "\n";
            desc += QObject::tr("others may: %1").arg(othersmay);
        }
    }
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        qobject_cast<QLabel*>(qobject_cast<FilePropertyDialogTabActionsView*>(widget)->content())->setText(desc);
        widget->setVisible(!desc.isEmpty());
    });
#endif
}

QString sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::userPermissionsDescription(int flags, int rflag, int wflag, int xflag)
{
    QStringList result;
    if (flags & rflag)
        result << QObject::tr("read");
    if (flags & wflag)
        result << QObject::tr("write");
    if (flags & xflag)
        result << QObject::tr("execute");
    if (result.isEmpty())
        result << QObject::tr("nothing");
    return result.join(", ");
}

QString sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::getUserName(int uid)
{
#ifdef Q_OS_UNIX
    struct passwd *pw;
    pw = getpwuid((uid_t)uid);
    if (pw)
        return QString(pw->pw_name);
#endif
    return QString::number(uid);
}

QString sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::getGroupName(int gid)
{
#ifdef Q_OS_UNIX
    struct group *gr;
    gr = getgrgid((gid_t)gid);
    if (gr)
        return QString(gr->gr_name);
#endif
    return QString::number(gid);
}

QMap<int, QString> sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::getAllUsers()
{
    QMap<int, QString> result;
#ifdef Q_OS_UNIX
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        struct passwd *grpp;
        setpwent();
        while (true) {
            grpp = getpwent();
            if (!grpp)
                break;
            result[grpp->pw_uid] = grpp->pw_name;
        }
        endpwent();
    });
#endif
    return result;
}

QMap<int, QString> sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::getAllGroups()
{
    QMap<int, QString> result;
#ifdef Q_OS_UNIX
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        struct group *grpp;
        setgrent();
        while (true) {
            grpp = getgrent();
            if (!grpp)
                break;
            result[grpp->gr_gid] = grpp->gr_name;
        }
        endgrent();
    });
#endif
    return result;
}

void sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::_initialize()
{
    REGISTER_FILEPROPERTYDIALOGTAB(sh::ui::filepropertydialog::FilePropertyDialogTabFactory::REGISTER_FILEPROPERTYDIALOGTAB_INDEX_CORE+1000, sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions);
}

void sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::_shutdown()
{
}

sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::ActionChange::ActionChange(std::shared_ptr<FilePropertyDialogTabUnixPermissions> tab)
            : tab(tab)
{
}

void sh::ui::filepropertydialog::FilePropertyDialogTabUnixPermissions::ActionChange::action(sh::actions::ActionExecutionInfo* info)
{
#ifdef Q_OS_UNIX
    info->setHead(tr("Changing permissions"));
    std::shared_ptr<sh::ui::filepropertydialog::FilePropertyDialog> _dlg = tab->dialog();
    if (!_dlg)
        return;
    sh::ui::ActionExecutionInfoDialog* duserfeedback = dynamic_cast<sh::ui::ActionExecutionInfoDialog*>(info->userfeedback());
    sh::ui::feedbackpanels::UnixPermissions* widget = 0;
    std::shared_ptr<sh::base::FilesystemNode> node = tab->nodes().first();
    QVariant vv = sh::detailcolumns::DetailColumnCustomAttributes::instance()->requestValue(node, info->operation());
    QMap<QString, QString> vals = sh::detailcolumns::DetailColumnCustomAttributes::getMapByValue(vv);
    QStringList vl = vals["unix_permissions"].split(";");
    int flags = vl[0].toInt();
    int uid = vl[1].toInt();
    int gid = vl[2].toInt();
    bool userMayRead = (flags & S_IRUSR);
    bool userMayWrite = (flags & S_IWUSR);
    bool userMayExecute = (flags & S_IXUSR);
    bool groupMayRead = (flags & S_IRGRP);
    bool groupMayWrite = (flags & S_IWGRP);
    bool groupMayExecute = (flags & S_IXGRP);
    bool othersMayRead = (flags & S_IROTH);
    bool othersMayWrite = (flags & S_IWOTH);
    bool othersMayExecute = (flags & S_IXOTH);
    bool sticky = (flags & S_ISVTX);
    bool setuid = (flags & S_ISUID);
    bool setgid = (flags & S_ISGID);
    QMap<int, QString> musers = getAllUsers();
    QMap<int, QString> mgroups = getAllGroups();
    QString currentuser = musers.value(uid);
    QString currentgroup = mgroups.value(gid);
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        widget = new sh::ui::feedbackpanels::UnixPermissions();
        widget->setflags(userMayRead, userMayWrite, userMayExecute, groupMayRead, groupMayWrite, groupMayExecute, othersMayRead, othersMayWrite, othersMayExecute, sticky, setuid, setgid);
        widget->setusers(musers.values(), currentuser);
        widget->setgroups(mgroups.values(), currentgroup);
    });
    duserfeedback->askFeedbackViaWidget(widget);
    if (widget->accepted)
    {
        if (!widget->user.isNull())
            uid = musers.key(widget->user);
        if (!widget->group.isNull())
            gid = mgroups.key(widget->group);
        flags = 0;
        if (widget->userMayRead)
            flags |= S_IRUSR;
        if (widget->userMayWrite)
            flags |= S_IWUSR;
        if (widget->userMayExecute)
            flags |= S_IXUSR;
        if (widget->groupMayRead)
            flags |= S_IRGRP;
        if (widget->groupMayWrite)
            flags |= S_IWGRP;
        if (widget->groupMayExecute)
            flags |= S_IXGRP;
        if (widget->othersMayRead)
            flags |= S_IROTH;
        if (widget->othersMayWrite)
            flags |= S_IWOTH;
        if (widget->othersMayExecute)
            flags |= S_IXOTH;
        if (widget->sticky)
            flags |= S_ISVTX;
        if (widget->setuid)
            flags |= S_ISUID;
        if (widget->setgid)
            flags |= S_ISGID;
        QStringList vperm;
        vperm << "unix_permissions";
        vperm << QString("%1;%2;%3!").arg(flags).arg(uid).arg(gid);
        sh::detailcolumns::DetailColumnCustomAttributes::instance()->applyValueByNode(node, info->operation(), vperm);
        tab->refresh();
    }
    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        delete widget;
    });
#endif
}
