/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <QKeyEvent>
#include "actions/actionsmanager.h"
#include "actions/abstractactionitem.h"
#include "actions/actionactionitem.h"
#include "base/filesystemnode.h"
#include "base/filesystemmodel.h"
#include "base/filesystemmodelfileviewproxy.h"
#include "configuration/configurationmanager.h"
#include "configuration/uistyle.h"
#include "exceptions/argumentexception.h"
#include "settings/settingsregistration.h"
#include "settings/settingsmanager.h"
#include "tools/mainthread.h"
#include "tools/threadpool.h"
#include "ui/fileview/filelist.h"
#include "ui/fileview/fileiconview.h"
#include "ui/fileview/fileviewevents.h"
#include "ui/fileview/fileviewcontrol.h"
#include "ui/mainwindow.h"
#include "filesystempanel.h"
#include "ui_filesystempanel.h"


sh::ui::FilesystemPanel::FilesystemPanel(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::FilesystemPanel)
{
    ui->setupUi(this);
    activeColor = sh::configuration::ConfigurationManager::instance()->style()->backgroundColor().name();
    inactiveColor = sh::configuration::ConfigurationManager::instance()->style()->inactiveBackgroundColor().name();
    addView();
    activeView = _views[0];
    activeView->focus();
}

sh::ui::FilesystemPanel::~FilesystemPanel()
{
    delete ui;
    _views.clear();
    foreach(sh::ui::fileview::FileViewControl* o, _viewcontrols)
        delete o;
}

void sh::ui::FilesystemPanel::addView(bool reinitialize)
{
    int i = _views.length();
    sh::ui::fileview::FileViewControl* viewctrl = new sh::ui::fileview::FileViewControl(this, i);
    _viewcontrols.append(viewctrl);
    if (reinitialize)
        viewctrl->columndimensions()->remove();
    std::shared_ptr<sh::ui::fileview::FileView> list;
    buildView(&list, viewctrl, true);
    connect(viewctrl->events(), &sh::ui::fileview::FileViewEvents::nodesActivated, this, [=](QList<std::shared_ptr<sh::base::FilesystemNode>> nodes){
        sh::base::FilesystemNodeType nodetype = (nodes.length()==1) ? nodes.first()->nodetype() : sh::base::FilesystemNodeType::NONE;
        if (nodetype == sh::base::FilesystemNodeType::Directory || nodetype == sh::base::FilesystemNodeType::SpecialTreeOnlyDirectory)
        {
            emit folderActivated(nodes[0], true);
        }
        else if (nodetype == sh::base::FilesystemNodeType::Link && !nodes[0]->linkDestinationNodes().isEmpty() && nodes[0]->linkDestinationNodes().first()->nodetype() == sh::base::FilesystemNodeType::Directory)
        {
            emit folderActivated(nodes[0]->linkDestinationNodes().first(), true);
        }
        else
        {
            auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(nodes, activeView->node());
            sh::actions::ActionsManager::instance()->getActions(instantiation, [&](QList<std::shared_ptr<sh::actions::ActionInstantiation> > selacts, QList<std::shared_ptr<sh::actions::ActionInstantiation>>){
                sh::tools::MainThread::dispatcher()->invokeAsync([=](){
                    QList<std::shared_ptr<sh::actions::AbstractActionItem>> acts;
                    for(auto actsit : selacts)
                        acts.append(actsit->createdAction);
                    sh::actions::ActionsManager::instance()->getDefaultAction(acts, [=](std::shared_ptr<sh::actions::ActionActionItem> action){
                        if (action)
                            action->execute();
                    });

                });
            }, [](std::shared_ptr<sh::actions::ActionInstantiation>){
            });
        }
    });
    connect(viewctrl->events(), &sh::ui::fileview::FileViewEvents::viewOptionChanged, this, [=](){
        int index = qobject_cast<sh::ui::fileview::FileViewEvents*>(sender())->index();
        emit viewOptionChanged(index);
    });
    connect(viewctrl->events(), &sh::ui::fileview::FileViewEvents::selectionChanged, this, [=](){
        emit selectionChanged();
    });
    emit viewCountChanged();
    sh::settings::SettingsManager::instance()->applyGlobalSettingsToFileView(viewctrl);
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(viewctrl);
    list->focus();
}

void sh::ui::FilesystemPanel::buildView(std::shared_ptr<sh::ui::fileview::FileView>* list, sh::ui::fileview::FileViewControl* viewctrl, bool isNew)
{
    QWidget* main;
    if (isNew)
    {
        main = new QWidget(ui->splitter);
        QVBoxLayout* mainlayout = new QVBoxLayout(main);
        mainlayout->setMargin(0);
        main->setLayout(mainlayout);
        _mainviews.insert(viewctrl->index(), main);
        ui->splitter->insertWidget(viewctrl->index(), main);
    }
    else
        main = _mainviews[viewctrl->index()];
    if (viewctrl->viewmode() == sh::ui::fileview::FileViewMode::ListView)
        *list = std::shared_ptr<sh::ui::fileview::FileView>(new sh::ui::fileview::FileList(main));
    else if (viewctrl->viewmode() == sh::ui::fileview::FileViewMode::IconView)
        *list = std::shared_ptr<sh::ui::fileview::FileView>(new sh::ui::fileview::FileIconview(main));
    else
        sh::exceptions::Exception::executeGuarded([&](){
            throw sh::exceptions::ProgramException(sh::exceptions::ExceptionData().details(QString("%1 is not a valid mode.").arg((int)viewctrl->viewmode())));
        });
    if (!isNew)
        _views.removeAt(viewctrl->index());
    _views.insert(viewctrl->index(), *list);
    (*list)->configure(viewctrl);
    (*list)->setBackgroundColor(inactiveColor);
    QWidget *llst = (*list)->as_qwidget();
    dynamic_cast<QVBoxLayout*>(main->layout())->insertWidget(0, (*list)->as_qwidget());
    llst->installEventFilter(this);
}

void sh::ui::FilesystemPanel::setViewMode(int i, sh::ui::fileview::FileViewMode mode)
{
    if (_viewcontrols[i]->viewmode() != mode)
    {
        bool isActive = i == activeView->control()->index();
        sh::ui::fileview::FileViewControl* viewctrl = _viewcontrols[i];
        viewctrl->_viewmode = mode;
        std::shared_ptr<sh::ui::fileview::FileView> v;
        {
            IGNORE_FORBIDDEN_ASYNC_CALLS
            buildView(&v, viewctrl, false);
        }
        if (isActive)
        {
            sh::tools::MainThread::dispatcher()->invokeAsync([=](){
                v->focus(); // seems to fail sometimes without async
            });
        }
        else
            emit viewmodeChanged(i); // for active ones, this comes nevertheless
    }
}

sh::ui::fileview::FileViewMode sh::ui::FilesystemPanel::viewMode(int i)
{
    return _viewcontrols[i]->viewmode();
}

void sh::ui::FilesystemPanel::removeView(int i)
{
    std::shared_ptr<sh::ui::fileview::FileView> ov = _views.takeAt(i);
    dynamic_cast<QWidget*>(ov.get())->setParent(0);
    delete _viewcontrols.takeAt(i);
    QWidget* mw = _mainviews.takeAt(i);
    _customwidgets.remove(mw);
    delete mw;
    _viewcontrols.last()->columndimensions()->remove(); // this just looks weird ;)
    for(int j = i; j< _viewcontrols.length(); j++)
        _viewcontrols[j]->setIndex(j);
    if (ov == activeView)
        _views[std::min(i, _views.length()-1)]->focus();
    emit viewCountChanged();
}

sh::ui::fileview::FileViewControl* sh::ui::FilesystemPanel::currentView()
{
    return activeView->control();
}

int sh::ui::FilesystemPanel::currentViewIndex()
{
    return activeView->control()->index();
}

void sh::ui::FilesystemPanel::reloadAll(bool requestFetch)
{
    CHECK_CORRECT_CALLER_THREAD
    foreach(std::shared_ptr<sh::ui::fileview::FileView> list, _views) {
        list->reload(requestFetch);
        for(int i=0; i<list->node()->childnodeCount(); i++)
            list->node()->childnode(i)->requestDetails();
    }
}

void sh::ui::FilesystemPanel::selectFolder(std::shared_ptr<base::FilesystemNode> folder, int index)
{
    CHECK_CORRECT_CALLER_THREAD
    sh::ui::fileview::FileViewControl* v = (index == -1) ? activeView->control() : _viewcontrols[index];
    v->gotoDir(folder);
}

void sh::ui::FilesystemPanel::setCustomWidget(int i, std::shared_ptr<QWidget> w)
{
    QWidget* main = _mainviews[i];
    std::shared_ptr<QWidget> custom = _customwidgets.take(main);
    if (custom)
        main->layout()->removeWidget(custom.get());
    if (w)
    {
        main->layout()->addWidget(w.get());
        QVBoxLayout* l = dynamic_cast<QVBoxLayout*>(main->layout());
        l->setStretch(0, 1);
        l->setStretch(1, 0);
        _customwidgets[main] = w;
    }
}

int sh::ui::FilesystemPanel::viewsCount()
{
    CHECK_CORRECT_CALLER_THREAD
    return _views.length();
}

sh::ui::fileview::FileViewControl *sh::ui::FilesystemPanel::view(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    return _viewcontrols[i];
}

bool sh::ui::FilesystemPanel::eventFilter(QObject *object, QEvent *event)
{
    if (event->type() == QEvent::FocusIn && !_skip_eventfilter)
    {
        sh::ui::fileview::FileView* _w = dynamic_cast<sh::ui::fileview::FileView*>(object);
        if (_w) {
            activeView = _w->shared_from_this();
            for(int i=0; i<_views.length();i++)
                _views[i]->setBackgroundColor(inactiveColor);
            _w->setBackgroundColor(activeColor);
            emit currentViewChanged();
            emit viewmodeChanged(_w->control()->index());
            emit viewOptionChanged(_w->control()->index());
            emit folderActivated( _w->node(), false );
            emit selectionChanged();
        }
    }
    return QWidget::eventFilter(object, event);
}

std::shared_ptr<sh::ui::fileview::FileView> sh::ui::FilesystemPanel::viewwidget(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    return _views[i];
}

QList<std::shared_ptr<sh::base::FilesystemNode> > sh::ui::FilesystemPanel::selectedNodes()
{
    CHECK_CORRECT_CALLER_THREAD
    return activeView->selectedNodes();
}
