/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QHeaderView>
#include <QMenu>
#include <QDragEnterEvent>
#include <QApplication>
#include <QDrag>
#include <QScrollBar>
#include "actions/actionactionitem.h"
#include "actions/common/actionnavigation.h"
#include "base/detailcolumn.h"
#include "base/eurl.h"
#include "base/filesystemhandler.h"
#include "base/filesystemhandlerregister.h"
#include "base/filesystemmodel.h"
#include "base/filesystemmodelfileviewproxy.h"
#include "configuration/configurationmanager.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "tools/dataexchange.h"
#include "tools/visibleviews.h"
#include "ui/fileview/fileviewcontrol.h"
#include "ui/columndimensions.h"
#include "ui/mainwindow.h"
#include "filelist.h"


std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::fileview::FileList::cfgvalFileListIconSize =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::fileview::FileList.FileListIconSize", -1,
            sh::configuration::ConfigurationValue::ValueTypeInteger(),
            QObject::tr("Icon size in file list panels"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("For default size, set it to -1."));

sh::ui::fileview::FileList::FileList(QWidget *parent) :
    QTreeView(parent), sh::ui::fileview::FileView()
{
    QSize defaultIconSize = iconSize();
    cfgvalFileListIconSize->consumeValue(this, [=](){
        int s = cfgvalFileListIconSize->getConfigValueInt();
        if (s<0)
            setIconSize(defaultIconSize);
        else
            setIconSize(QSize(s,s));
    });
    setRootIsDecorated(false);
    setAcceptDrops(true);
    setUniformRowHeights(true);
    setSortingEnabled(true);
    setAnimated(false);
    setAllColumnsShowFocus(true);
    setSelectionMode(SelectionMode::ExtendedSelection);
    header()->setStretchLastSection(false);
    setItemsExpandable(false);
    connect(header(), &QHeaderView::sortIndicatorChanged, this, [=](int c,Qt::SortOrder o){
        viewctrl->setSort(c, o);
    });
    connect(header(), SIGNAL(sectionResized(int,int,int)), this, SLOT(slot_sectionResized(int,int,int)));
}

void sh::ui::fileview::FileList::setSort(int column, Qt::SortOrder order)
{
    sortByColumn(column, order);
}

QString sh::ui::fileview::FileList::getColumnNameForIndex(int index)
{
    return (index==0) ? "int_name" : _node->getDetailColumnByIndex(index-1)->name();
}

void sh::ui::fileview::FileList::slot_sectionResized(int index, int, int newsize)
{
    if (_skip_slot_sectionResized)
        return;
    viewctrl->columndimensions()->setWidth(getColumnNameForIndex(index), newsize);
}

void sh::ui::fileview::FileList::_adaptColumnWidth(int index)
{
    _skip_slot_sectionResized = true;
    int defaultwidth = (index==0)? 200 : _node->getDetailColumnByIndex(index-1)->defaultWidth();
    header()->resizeSection(index, viewctrl->columndimensions()->getWidth(getColumnNameForIndex(index), defaultwidth));
    _skip_slot_sectionResized = false;
}

void sh::ui::fileview::FileList::dragEnterEvent(QDragEnterEvent *event)
{
    dnd_dragEnterEvent(event);
}

void sh::ui::fileview::FileList::mousePressEvent(QMouseEvent *event)
{
    if (!dnd_mousePressEvent(event))
        QTreeView::mousePressEvent(event);
}

void sh::ui::fileview::FileList::mouseMoveEvent(QMouseEvent *event)
{
    if (!dnd_mouseMoveEvent(event))
        QTreeView::mouseMoveEvent(event);
}

void sh::ui::fileview::FileList::dragMoveEvent(QDragMoveEvent *event)
{
    dnd_dragMoveEvent(event);
}

void sh::ui::fileview::FileList::dragLeaveEvent(QDragLeaveEvent *event)
{
    dnd_dragLeaveEvent(event);
}

void sh::ui::fileview::FileList::dropEvent(QDropEvent *event)
{
    dnd_dropEvent(event);
}

void sh::ui::fileview::FileList::keyPressEvent(QKeyEvent *event)
{
    evaluateKeyPress(event);
    QTreeView::keyPressEvent(event);
}

void sh::ui::fileview::FileList::_reload_goto_modelloaded()
{
    connect(_model, &QAbstractItemModel::columnsInserted, this, [=](const QModelIndex&, int a, int b){
        for (int i=a; i<=b; i++)
            _adaptColumnWidth(i);
    });
    for(int i=0; i<header()->count(); i++)
        _adaptColumnWidth(i);
}

void sh::ui::fileview::FileList::onBackgroundColorSet()
{
    this_qwidget->setStyleSheet("QTreeView{background:" + bgColor + "}");
}

