/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QMenu>
#include <QApplication>
#include <QScrollBar>
#include "actions/actionactionitem.h"
#include "actions/actionsmanager.h"
#include "base/detailcolumn.h"
#include "base/eurl.h"
#include "base/filesystemhandler.h"
#include "base/filesystemhandlerregister.h"
#include "base/filesystemmodel.h"
#include "base/filesystemmodelfileviewproxy.h"
#include "configuration/configurationmanager.h"
#include "configuration/uistyle.h"
#include "settings/settingsmanager.h"
#include "tools/threadpool.h"
#include "tools/mainthread.h"
#include "tools/dataexchange.h"
#include "tools/visibleviews.h"
#include "ui/fileview/fileviewcontrol.h"
#include "ui/columndimensions.h"
#include "ui/filesystempanel.h"
#include "ui/mainwindow.h"
#include "fileview.h"
#include "filevieweventsimpl.h"


sh::ui::fileview::FileView::FileView()
{
    QColor _actColor = sh::ui::MainWindow::mainWindow()->palette().color(QPalette::Active, QPalette::Base );
    QColor _dndTargetColor = _actColor.darker(130);
    dndTargetColor = _dndTargetColor.name();
}

sh::ui::fileview::FileView::~FileView() {
    if (_model)
        delete _model;
}

void sh::ui::fileview::FileView::configure(sh::ui::fileview::FileViewControl *viewctrl)
{
    this->viewctrl = viewctrl;
    this_qobject = dynamic_cast<QObject*>(this);
    this_qwidget = dynamic_cast<QWidget*>(this);
    this_qabstractitemview = dynamic_cast<QAbstractItemView*>(this);
    this_qwidget->installEventFilter(new sh::ui::fileview::FileView::EventFilter(this_qobject, this, control()->_events));
    this_qabstractitemview->setFrameShape(QFrame::NoFrame);
    if (this_qabstractitemview)
    {
        this_qabstractitemview->setContextMenuPolicy(Qt::CustomContextMenu);
        QObject::connect(this_qabstractitemview, &QWidget::customContextMenuRequested, this_qabstractitemview, [=](QPoint pos){
            getContextMenuForActions()->exec( this_qabstractitemview->viewport()->mapToGlobal(pos));
        });
        QObject::connect(this_qabstractitemview, &QAbstractItemView::activated, this_qabstractitemview, [=](){
            control()->_events->emit_nodesActivated(selectedNodes());
        });
    }
    this_qwidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setSort(viewctrl->sortColumn(), viewctrl->sortOrder());
    viewctrl->gotoDir(viewctrl->node());
}

sh::ui::fileview::FileView::EventFilter::EventFilter(QObject *parent, sh::ui::fileview::FileView *fileview, sh::ui::fileview::FileViewEventsImpl* fileviewevents)
    : QObject(parent), fileview(fileview), fileviewevents(fileviewevents)
{
}

bool sh::ui::fileview::FileView::EventFilter::eventFilter(QObject*, QEvent *event)
{
    if (event->type() == QEvent::KeyPress)
    {
        QKeyEvent* keyevent = (QKeyEvent*) event;
        if (keyevent->key() == Qt::Key_Backspace)
        {
            std::shared_ptr<sh::base::FilesystemNode> node = fileview->node();
            if (node->parentnode())
                fileviewevents->emit_nodesActivated(QList<std::shared_ptr<sh::base::FilesystemNode>>({node->parentnode()}));
        }
        else if (keyevent->key() == Qt::Key_Down && fileview->_model->rowCount() == 1 && !fileview->this_qabstractitemview->selectionModel()->hasSelection())
        {
            fileview->this_qabstractitemview->selectAll();
        }
    }
    return false;
}

std::shared_ptr<QMenu> sh::ui::fileview::FileView::getContextMenuForActions()
{
    auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectedNodes(), node());
    return sh::actions::ActionsManager::instance()->mountActionItemsInMenu(instantiation);
}

QThread* sh::ui::fileview::FileView::thread()
{
    return this_qobject->thread();
}

void sh::ui::fileview::FileView::deleteLater()
{
    this_qobject->deleteLater();
}

void sh::ui::fileview::FileView::gotoDir(std::shared_ptr<sh::base::FilesystemNode> n)
{
    if (_node != n)
    {
        _reload_goto(true, true, n);
        if (n != sh::base::FilesystemModel::instance()->rootNode())
            viewctrl->historyTracker()->visitValue(n->eurl());
    }
}

void sh::ui::fileview::FileView::reload(bool requestFetch)
{
    _reload_goto(requestFetch, false, _node);
}

void sh::ui::fileview::FileView::_reload_goto(bool requestFetch, bool makemodel, std::shared_ptr<base::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    QAbstractItemModel* oldmodel = _model;
    if ( (makemodel||requestFetch) && node)
    {
        if (!makemodel && _model)
            _model->triggerReloadData();
        sh::tools::ThreadPool::enqueueForce([=](){
            node->requestContainedItems();
            //this fetches all files in the view and their details in one for loop (on top of the armada of threads for the visible ones)
            sh::tools::MainThread::dispatcher()->invokeSync([&](){
                int childnodecount = node->childnodeCount();
                for(int i=0; i<childnodecount; i++)
                    node->childnode(i)->requestDetails();
            });
        });
    }
    if (makemodel)
    {
        if (_node != node)
        {
            _visibleviewslifetimecanary = 0;
            if (node)
                _visibleviewslifetimecanary = sh::tools::VisibleViews::instance()->viewEnteredDirectory(node);
        }
        _node = node;
        int myasynccounter = ++_async_gotodir_index;
        if (node) {
            std::weak_ptr<sh::ui::fileview::FileView> _canary = shared_from_this();
            sh::tools::MainThread::dispatcher()->invokeAsync([=](){ // we must do that async, otherwise model machinery kills us in a subtile way (when this is called between beginRemoveRows and endRemoveRows)
                std::shared_ptr<sh::ui::fileview::FileView> _scanary = _canary.lock();
                if (!_scanary)
                    return;
                if (myasynccounter != _async_gotodir_index)
                    return;
                QModelIndex index = sh::base::FilesystemModel::instance()->getIndexForNode( node );
                _model = sh::base::FilesystemModel::instance()->createFileviewProxy(index, _modelwithtooltip, [=](std::shared_ptr<sh::base::FilesystemNode> bestValid) {
                    // this happens when this model dies (i.e. directory was removed)
                    _reload_goto(true, true, 0);
                    sh::tools::MainThread::dispatcher()->invokeAsync([=](){
                        viewctrl->gotoDir(bestValid);
                    });
                });
                if (_thumbnailsize!=-1)
                    _model->setThumbnail(true, _thumbnailsize);
                QObject::connect(sh::base::FilesystemModel::instance(), &QAbstractItemModel::rowsInserted, _model,
                                 [=](const QModelIndex& parent, int a, int b){
                    if (index==parent)
                    {
                        for(int i=a;i<=b;i++)
                            node->childnode(i)->requestDetails();
                    }
                });
                _model->setSizeFormattingMode(control()->getSizeFormattingMode());
                _model->setHiddenFilesVisible(control()->hiddenFilesVisible());
                QItemSelectionModel* oldsm = this_qabstractitemview->selectionModel();
                _model->sort(_sort_column, _sort_order);
                QObject::connect(_model, &QAbstractItemModel::columnsInserted, [=](QModelIndex, int, int){  // we must reset the sorting, since new columns may change it (within qt)
                    sh::tools::MainThread::dispatcher()->invokeAsync([=](){
                        if (_model)
                            _model->sort(_sort_column, _sort_order);
                    });
                });
                this_qabstractitemview->setModel(_model);
                delete oldsm;
                _reload_goto_modelloaded();
                QObject::connect(this_qabstractitemview->selectionModel(), &QItemSelectionModel::selectionChanged, this_qabstractitemview, [=](const QItemSelection&, const QItemSelection&){
                    control()->_events->emit_selectionChanged();
                });
            });
        }

        _reload_goto_flushmodel();
        if (oldmodel)
            delete oldmodel;
    }
}

void sh::ui::fileview::FileView::setBackgroundColor(QString c)
{
    bgColor = c;
    onBackgroundColorSet();
}

QObject* sh::ui::fileview::FileView::as_qobject()
{
    return this_qobject;
}

QWidget* sh::ui::fileview::FileView::as_qwidget()
{
    return this_qwidget;
}

QAbstractItemView* sh::ui::fileview::FileView::as_qabstractitemview()
{
    return this_qabstractitemview;
}

sh::base::FilesystemModelFileviewProxy* sh::ui::fileview::FileView::modelproxy()
{
    return _model;
}

QString sh::ui::fileview::FileView::backgroundColor()
{
    return bgColor;
}

void sh::ui::fileview::FileView::setSizeFormatting(sh::base::SizeFormatting v)
{
    if (_model)
        _model->setSizeFormattingMode(v);
}

QList<std::shared_ptr<sh::base::FilesystemNode>> sh::ui::fileview::FileView::selectedNodes()
{
    QList<std::shared_ptr<sh::base::FilesystemNode>> nodes;
    foreach(const QModelIndex idx, selectedModelIndices())
    {
        if (idx.column()==0){
            QModelIndex didx = _model->mapToSource(idx);
            std::shared_ptr<sh::base::FilesystemNode> node = sh::base::FilesystemModel::instance()->getNodeForIndex(didx);
            nodes.append(node);
        }
    }
    return nodes;
}

void sh::ui::fileview::FileView::setHiddenFilesVisible(bool value)
{
    if (_model)
        _model->setHiddenFilesVisible(value);
}

std::shared_ptr<sh::base::FilesystemNode> sh::ui::fileview::FileView::node()
{
    return _node;
}

void sh::ui::fileview::FileView::evaluateKeyPress(QKeyEvent *event)
{
    int key = event->key();
    if (key != Qt::Key_Down && key != Qt::Key_Up && key != Qt::Key_Right &&
                    key != Qt::Key_Left && key != Qt::Key_PageDown && key != Qt::Key_PageUp &&
                    key != Qt::Key_Home && key != Qt::Key_End && key != Qt::Key_Control &&
                    key != Qt::Key_Meta && key != Qt::Key_Alt && key != Qt::Key_Shift && key != Qt::Key_Space)
    {
        sh::actions::ActionsManager::instance()->getActionForShortcut(selectedNodes(), _node, event->key(), event->modifiers(), [=](std::shared_ptr<sh::actions::AbstractActionItem> a){
            sh::actions::ActionActionItem* aaction = qobject_cast<sh::actions::ActionActionItem*>(a.get());
            if (aaction)
                aaction->execute();  // a noop for disabled actions
        });
    }
}

bool sh::ui::fileview::FileView::dnd_dragEnterEvent(QDragEnterEvent *event)
{
    if (sh::tools::DataExchange::instance()->containsFileEntries(event->mimeData())){
        _tmp_bgColor = backgroundColor();
        setBackgroundColor(dndTargetColor);
        event->acceptProposedAction();
        return true;
    }
    else
    {
        event->dropAction();
        return false;
    }
}

bool sh::ui::fileview::FileView::dnd_mousePressEvent(QMouseEvent *event)
{
    int btn = event->button();
    if (btn == Qt::LeftButton) {
        _dragStartPosition = event->pos();
        _dragIsValid = this_qabstractitemview->indexAt(event->pos()).isValid();
    }
    else if (btn == Qt::BackButton || btn == Qt::ForwardButton) {
        auto lst = (btn == Qt::BackButton) ? viewctrl->historyTracker()->backwardList() : viewctrl->historyTracker()->forwardList();
        if (!lst.isEmpty()) {
            auto e = lst.first();
            viewctrl->historyTracker()->revisitValue(e.index());
            sh::ui::MainWindow::mainWindow()->jumpToEurl(e.value());
        }
        return true;
    }
    return false;
}

bool sh::ui::fileview::FileView::dnd_mouseMoveEvent(QMouseEvent *event)
{
    if ((event->buttons() & Qt::LeftButton))
    {
        if ((event->pos() - _dragStartPosition).manhattanLength() >= QApplication::startDragDistance())
        {
            QList<std::shared_ptr<sh::base::FilesystemNode>> nodes = selectedNodes();
            if (nodes.length()>0 && _dragIsValid) {
                QDrag *drag = new QDrag(this_qobject);
                QMimeData* mimeData = sh::tools::DataExchange::instance()->getMimeDataFromFilesystemNodes(nodes, sh::tools::DataExchange::DataExchangeType::MOVE);
                drag->setMimeData(mimeData);
                drag->exec(Qt::CopyAction | Qt::MoveAction);
            }
        }
        return true;
    }
    else
        return false;
}

void sh::ui::fileview::FileView::dnd_dragMoveEvent(QDragMoveEvent *event)
{
    event->acceptProposedAction();
}

void sh::ui::fileview::FileView::dnd_dragLeaveEvent(QDragLeaveEvent *event)
{
    setBackgroundColor(_tmp_bgColor);
    event->accept();
}

bool sh::ui::fileview::FileView::dnd_dropEvent(QDropEvent *event)
{
    setBackgroundColor(_tmp_bgColor);
    if (event->source() != this_qobject)
    {
        std::shared_ptr<sh::actions::AbstractActionItem> action;
        if (event->dropAction()==Qt::CopyAction)
            action = sh::tools::DataExchange::instance()->createCopyAction( event->mimeData(), node()->eurl());
        else // if (event->dropAction()==Qt::MoveAction)
            action = sh::tools::DataExchange::instance()->createMoveAction( event->mimeData(), node()->eurl());
        sh::tools::ThreadPool::enqueueForce([=](){
            sh::actions::ActionActionItem* a = qobject_cast<sh::actions::ActionActionItem*>(action.get());
            a->initializeSync();
            a->execute();
        });
        return true;
    }
    return false;
}

QModelIndexList sh::ui::fileview::FileView::selectedModelIndices()
{
    return this_qabstractitemview->selectionModel()->selection().indexes();
}

void sh::ui::fileview::FileView::_reload_goto_modelloaded()
{
}

void sh::ui::fileview::FileView::_reload_goto_flushmodel()
{
    QItemSelectionModel* oldsm = this_qabstractitemview->selectionModel();
    this_qabstractitemview->setModel(_model = 0);
    delete oldsm;
}

void sh::ui::fileview::FileView::focus()
{
    this_qwidget->setFocus();
}

sh::ui::fileview::FileViewControl* sh::ui::fileview::FileView::control()
{
    return viewctrl;
}

void sh::ui::fileview::FileView::setSort(int column, Qt::SortOrder order)
{
    _sort_column = column;
    _sort_order = order;
    if (_model)
        _model->sort(column, order);
}
