/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FILEVIEW_H
#define FILEVIEW_H

#include <memory>
#include <QAbstractItemModel>
#include <QDragEnterEvent>
#include <QDrag>
#include <QHeaderView>
#include <QObject>
#include "base/filesystemmodelfileviewproxy.h"


namespace sh{
    namespace base{class FilesystemNode;}
    namespace configuration{class ConfigurationValue;}
    namespace actions{
        class AbstractActionItem;
        namespace common{
            class ActionHistoryNavigateForward;
            class ActionHistoryNavigateBackward;
            class ActionNavigateInHistory;
        }
    }
    namespace ui {
        namespace fileview {

            class FileViewControl;
            class FileViewEventsImpl;

            /** \brief
             * Abstract base class for views for the contents of one directory.
             */
            class FileView : public std::enable_shared_from_this<FileView>
            {
                friend class sh::actions::common::ActionHistoryNavigateForward;
                friend class sh::actions::common::ActionHistoryNavigateBackward;
                friend class sh::actions::common::ActionNavigateInHistory;
                friend class sh::ui::fileview::FileViewControl;
            public:
                /** \brief Constructed only by the infrastructure and made available otherwise.
                 */
                FileView();
                virtual ~FileView();
                void setBackgroundColor(QString c);
                QString backgroundColor();
                QList<std::shared_ptr<sh::base::FilesystemNode> > selectedNodes();
                std::shared_ptr<sh::base::FilesystemNode> node();
                sh::base::FilesystemModelFileviewProxy* modelproxy();
                void reload(bool requestFetch);
                void setSizeFormatting(sh::base::SizeFormatting v);
                void gotoDir(std::shared_ptr<sh::base::FilesystemNode> n);
                void setHiddenFilesVisible(bool value);
                QThread* thread();
                virtual void configure(sh::ui::fileview::FileViewControl *viewctrl);
                void deleteLater();
                virtual void focus();
                virtual void setSort(int column, Qt::SortOrder order);

                QObject* as_qobject();
                QWidget* as_qwidget();
                QAbstractItemView* as_qabstractitemview();

                sh::ui::fileview::FileViewControl* control();

            protected:
                virtual QModelIndexList selectedModelIndices();

                virtual void _reload_goto(bool requestFetch, bool makemodel, std::shared_ptr<sh::base::FilesystemNode> node);
                virtual void _reload_goto_modelloaded();
                virtual void _reload_goto_flushmodel();

                virtual void onBackgroundColorSet() =0;

                void evaluateKeyPress(QKeyEvent *event);

                bool dnd_dragEnterEvent(QDragEnterEvent *event);
                bool dnd_mousePressEvent(QMouseEvent *event);
                bool dnd_mouseMoveEvent(QMouseEvent *event);
                void dnd_dragMoveEvent(QDragMoveEvent *event);
                void dnd_dragLeaveEvent(QDragLeaveEvent *event);
                bool dnd_dropEvent(QDropEvent *event);

                std::shared_ptr<QMenu> getContextMenuForActions();

                QString bgColor;
                QString dndTargetColor;
                sh::base::FilesystemModelFileviewProxy* _model = 0;
                std::shared_ptr<sh::base::FilesystemNode> _node = 0;
                sh::ui::fileview::FileViewControl *viewctrl;

                QObject* this_qobject;
                QWidget* this_qwidget;
                QAbstractItemView* this_qabstractitemview;
                int _thumbnailsize = -1;
                bool _modelwithtooltip = false;

            private:
                class EventFilter : public QObject
                {
                public:
                    EventFilter(QObject *parent, FileView* fileview, sh::ui::fileview::FileViewEventsImpl *fileviewevents);
                    bool eventFilter(QObject *object, QEvent *event);
                private:
                    FileView* fileview;
                    sh::ui::fileview::FileViewEventsImpl* fileviewevents;
                };

                QString _tmp_bgColor;
                QPoint _dragStartPosition;
                int _async_gotodir_index = 0;
                bool _dragIsValid = false;
                std::shared_ptr<QObject> _visibleviewslifetimecanary = 0;
                int _sort_column = 0;
                Qt::SortOrder _sort_order = Qt::AscendingOrder;
            };

        }
    }
}

#endif // FILEVIEW_H
