/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDockWidget>
#include "base/eurl.h"
#include "base/filesystemmodel.h"
#include "base/filesystemhandler.h"
#include "ui/columndimensions.h"
#include "ui/filesystempanel.h"
#include "ui/fileview/fileview.h"
#include "ui/fileview/fileiconview.h"
#include "ui/fileview/filevieweventsimpl.h"
#include "settings/settingsmanager.h"
#include "tools/threaddispatcher.h"
#include "fileviewcontrol.h"


sh::ui::fileview::FileViewControl::FileViewControl(sh::ui::FilesystemPanel* panel, int i) :
    QObject(), _panel(panel), i(i)
{
    CHECK_CORRECT_CALLER_THREAD
    _columndimensions = new sh::ui::ColumnDimensions(i);
    _historyTracker = new sh::tools::HistoryTracker<std::shared_ptr<const sh::base::Eurl>>();
    _events = new sh::ui::fileview::FileViewEventsImpl(this, i);
    _node = sh::base::FilesystemModel::instance()->rootNode();
    _sortColumn = 0;
    _sortOrder = Qt::AscendingOrder;
}

sh::ui::fileview::FileViewControl::~FileViewControl()
{
    CHECK_CORRECT_CALLER_THREAD
    delete _columndimensions;
    delete _historyTracker;
    delete _events;
}

sh::ui::ColumnDimensions* sh::ui::fileview::FileViewControl::columndimensions()
{
    CHECK_CORRECT_CALLER_THREAD
    return _columndimensions;
}

sh::tools::HistoryTracker<std::shared_ptr<const sh::base::Eurl>>* sh::ui::fileview::FileViewControl::historyTracker()
{
    CHECK_CORRECT_CALLER_THREAD
    return _historyTracker;
}

sh::ui::fileview::FileViewMode sh::ui::fileview::FileViewControl::viewmode()
{
    CHECK_CORRECT_CALLER_THREAD
    return _viewmode;
}

void sh::ui::fileview::FileViewControl::setIndex(int i)
{
    CHECK_CORRECT_CALLER_THREAD
    this->i = i;
    _columndimensions->index = i;
    _columndimensions->store();
}

int sh::ui::fileview::FileViewControl::index()
{
    CHECK_CORRECT_CALLER_THREAD
    return i;
}

std::shared_ptr<sh::base::FilesystemNode> sh::ui::fileview::FileViewControl::node()
{
    CHECK_CORRECT_CALLER_THREAD
    return _node;
}

void sh::ui::fileview::FileViewControl::gotoDir(std::shared_ptr<sh::base::FilesystemNode> n)
{
    CHECK_CORRECT_CALLER_THREAD
    _node = n;
    QString header;
    std::shared_ptr<QWidget> customui;
    if (n && n->handler())
        n->handler()->getCustomUi(n, &header, &customui);
    if (customui)
    {
        QDockWidget* dcustomui = new QDockWidget();
        dcustomui->setFeatures(QDockWidget::DockWidgetVerticalTitleBar);
        dcustomui->setWindowTitle(header);
        dcustomui->setWidget(customui.get());
        customui->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
        connect(dcustomui, &QObject::destroyed, [=](){
            customui->setParent(0);
        });
        _panel->setCustomWidget(i, std::shared_ptr<QWidget>(dcustomui));
    }
    else
        _panel->setCustomWidget(i, 0);
    _panel->viewwidget(i)->gotoDir(n);
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(this);
}

sh::base::SizeFormatting sh::ui::fileview::FileViewControl::getSizeFormattingMode()
{
    CHECK_CORRECT_CALLER_THREAD
    return _sizeformattingmode;
}

void sh::ui::fileview::FileViewControl::setSizeFormattingMode(sh::base::SizeFormatting mode)
{
    CHECK_CORRECT_CALLER_THREAD
    _sizeformattingmode = mode;
    _panel->viewwidget(i)->setSizeFormatting(mode);
    _events->emit_viewOptionChanged();
}

bool sh::ui::fileview::FileViewControl::hiddenFilesVisible()
{
    CHECK_CORRECT_CALLER_THREAD
    return _hiddenfilesvisible;
}

void sh::ui::fileview::FileViewControl::setHiddenFilesVisible(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    _hiddenfilesvisible = v;
    _panel->viewwidget(i)->setHiddenFilesVisible(v);
    _events->emit_viewOptionChanged();
}

int sh::ui::fileview::FileViewControl::iconDimension()
{
    CHECK_CORRECT_CALLER_THREAD
    return _icondimension;
}

void sh::ui::fileview::FileViewControl::setIconDimension(int v)
{
    CHECK_CORRECT_CALLER_THREAD
    _icondimension = v;
    sh::ui::fileview::FileIconview* fiv = qobject_cast<sh::ui::fileview::FileIconview*>(_panel->viewwidget(i)->as_qwidget());
    if (fiv)
        fiv->setThumbDimension(1.0 + 0.15 * v);
}

sh::ui::fileview::FileViewEvents* sh::ui::fileview::FileViewControl::events()
{
    CHECK_CORRECT_CALLER_THREAD
    return _events;
}

void sh::ui::fileview::FileViewControl::setSort(int column, Qt::SortOrder order)
{
    _sortColumn = column;
    _sortOrder = order;
    _panel->viewwidget(i)->setSort(column, order);
}

int sh::ui::fileview::FileViewControl::sortColumn()
{
    return _sortColumn;
}

Qt::SortOrder sh::ui::fileview::FileViewControl::sortOrder()
{
    return _sortOrder;
}
