/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/eurl.h"
#include "base/filesystemmodel.h"
#include "tools/iconmanager.h"
#include "mainwindow.h"
#include "jumpbar.h"
#include "ui_jumpbar.h"


sh::ui::JumpBar::JumpBar(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::JumpBar)
{
    ui->setupUi(this);
    ui->btnOk->setIcon( sh::tools::IconManager::instance()->getIcon("dialog-ok") );
    ui->btnAbort->setIcon( sh::tools::IconManager::instance()->getIcon("dialog-close") );
    connect(ui->lineEdit, SIGNAL(enterPressed()), this, SLOT(on_btnOk_clicked()));
    connect(ui->lineEdit, SIGNAL(escapePressed()), this, SLOT(on_btnAbort_clicked()));
    setButtonMode();
}

sh::ui::LineEditWithKeyboardShortcuts::LineEditWithKeyboardShortcuts(QWidget *parent)
    :QLineEdit(parent)
{

}

void sh::ui::LineEditWithKeyboardShortcuts::keyPressEvent(QKeyEvent * event)
{
    if (event->key() == Qt::Key_Enter || event->key() == Qt::Key_Return)
    {
        emit enterPressed();
    }
    else if (event->key() == Qt::Key_Escape)
    {
        emit escapePressed();
    }
    else
        QLineEdit::keyPressEvent(event);
}


sh::ui::JumpBar::~JumpBar()
{
    delete ui;
}

void sh::ui::JumpBar::resizeEvent(QResizeEvent * event)
{
    if (!isTextMode)
        _buildButtons();
    QWidget::resizeEvent(event);
}


void sh::ui::JumpBar::setButtonMode()
{
    ui->editbar->setVisible(false);
    ui->buttonsbar->setVisible(true);
    isTextMode = false;
    _buildButtons();
}

void sh::ui::JumpBar::setTextMode()
{
    ui->buttonsbar->setVisible(false);
    ui->editbar->setVisible(true);
    isTextMode = true;
    QString eurl;
    if (_node)
        eurl = _node->eurl()->asString();
    ui->lineEdit->setText(eurl);
    ui->lineEdit->selectAll();
    ui->lineEdit->setFocus();
}

std::shared_ptr<sh::base::FilesystemNode> sh::ui::JumpBar::node()
{
    return _node;
}

QSize sh::ui::JumpBar::sizeHint() const
{
    return QSize( ui->editbar->sizeHint().width() , ui->lineEdit->sizeHint().height() );
}


void sh::ui::JumpBar::_buildButtons()
{
    int xcur = 0;
    int h = height();
    int fw = width();
    foreach (QObject* owidget, ui->buttonsbar->children()) {
        QWidget* widget = dynamic_cast<QWidget*>(owidget);
        if (widget)
            delete widget;
    }
    QList<QToolButton*> btns;
    if (_node)
        for (std::shared_ptr<sh::base::FilesystemNode> __node=_node;__node!=0;__node=__node->parentnode()) {
            QToolButton* btn = new QToolButton(ui->buttonsbar);
            btn->setText(__node->displayName());
            btn->setToolTip(__node->eurl()->asString());
            btn->setAutoRaise(true);
            btn->setFocusPolicy(Qt::ClickFocus);
            int w = btn->sizeHint().width();
            xcur += w;
            if (xcur < fw)
            {
                btns.insert(0, btn);
                connect(btn, &QToolButton::clicked, [=](){
                    this->_node = __node;
                    _buildButtons();
                    emit nodeChanged();
                });
            }
            else
            {
                delete btn;
                break;
            }
        }
    xcur = 0;
    foreach (QToolButton* btn, btns) {
        int w = btn->sizeHint().width();
        btn->setGeometry(xcur, 0, w, h);
        btn->setVisible(true);
        xcur += w;
    }
    QPushButton* btnfree = new QPushButton(ui->buttonsbar);
    btnfree->setFlat(true);
    btnfree->setText("");
    btnfree->setIcon(QIcon());
    btnfree->setCursor(Qt::IBeamCursor);
    btnfree->setGeometry(xcur, 0, fw-xcur, h);
    btnfree->setVisible(true);
    btnfree->setFocusPolicy(Qt::ClickFocus);
    connect(btnfree, &QPushButton::clicked, [=](){
        setTextMode();
    });
}

void sh::ui::JumpBar::setNode(std::shared_ptr<base::FilesystemNode> node)
{
    if (node != _node)
    {
        _node = node;
        if (isTextMode)
        {
            ui->lineEdit->setText( _node->eurl()->asString() );
        }
        else
        {
            _buildButtons();
        }
    }
}

void sh::ui::JumpBar::on_btnOk_clicked()
{
    sh::exceptions::Exception::executeGuarded([&](){
        emit eurlRequested(sh::base::Eurl::fromString(ui->lineEdit->text()));
    },sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::JumpBar::on_btnAbort_clicked()
{
    setButtonMode();
}
