/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QTextStream>
#include "configuration/configurationmanager.h"
#include "tools/iconmanager.h"
#include "logviewdialog.h"
#include "ui_logviewdialog.h"


sh::ui::LogViewDialog::LogViewDialog(QWidget *parent, QString headtext, QString subheadtxt, tools::LogSeverity minseverity) :
    QDialog(parent), ui(new Ui::LogViewDialog)
{
    ui->setupUi(this);
    if (!headtext.isNull())
        ui->lblHead->setText(headtext);
    if (!subheadtxt.isNull())
        ui->lblSubhead->setText(subheadtxt);
    else
        ui->lblSubhead->hide();
    ui->btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    sm->style()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    sm->style()->createSheet().fontsize_header().applyTo(ui->lblHead);
    sm->style()->createSheet().textcolor_lighter1().applyTo(ui->lblSubhead);
    connect(sh::tools::Logger::instance(), &sh::tools::Logger::newLogMessageArrived, this, [=](){
        _reloadLog();
    });
    setMinimumSeverity(minseverity);
    ui->scrollArea->setFocus();
}

void sh::ui::LogViewDialog::_reloadLog()
{
    QWidget* ow = new QWidget(this);
    QVBoxLayout* olayout = new QVBoxLayout(ow);
    QList<sh::tools::Logger::LogMessage*> messages = sh::tools::Logger::instance()->logMessages();
    std::reverse(messages.begin(), messages.end());
    foreach(sh::tools::Logger::LogMessage* logmsg, messages)
    {
        if (logmsg->severity >= _minseverity)
        {
            QString ssev;
            QString backcolor;
            if (logmsg->severity == sh::tools::LogSeverity::DEBUG)
            {
                ssev = "";
                backcolor = "#fefefe";
            }
            else if (logmsg->severity == sh::tools::LogSeverity::INFO)
            {
                ssev = "+";
                backcolor = "#eee";
            }
            else if (logmsg->severity == sh::tools::LogSeverity::WARNING)
            {
                ssev = "++";
                backcolor = "#feff84";
            }
            else if (logmsg->severity == sh::tools::LogSeverity::ERROR)
            {
                ssev = "+++";
                backcolor = "#ffbcbc";
            }
            else
            {
                ssev = "?";
                backcolor = "#febcff";
            }
            QFrame* w = new QFrame(this);
            w->setStyleSheet("background:"+backcolor);
            QHBoxLayout* layout = new QHBoxLayout(w);
            layout->setSpacing(8);
            layout->setMargin(5);
            QLabel* ldate = new QLabel(this);
            ldate->setAlignment(Qt::AlignLeft|Qt::AlignTop);
            ldate->setText(logmsg->time.time().toString(Qt::ISODate));
            layout->addWidget(ldate);
            QWidget* wb = new QWidget();
            QLayout* blayout = new QVBoxLayout(wb);
            blayout->setSpacing(2);
            blayout->setMargin(0);
            layout->addWidget(wb);
            layout->setStretchFactor(wb, 1);
            if (!logmsg->source.isEmpty()) {
                QLabel* lsource = new QLabel(this);
                sh::configuration::ConfigurationManager::instance()->style()->createSheet().fontsize_smaller().applyTo(lsource);
                lsource->setAlignment(Qt::AlignLeft);
                lsource->setText(logmsg->source);
                lsource->setWordWrap(true);
                lsource->setTextInteractionFlags(Qt::TextSelectableByMouse);
                blayout->addWidget(lsource);
            }
            QLabel* lmessage = new QLabel(this);
            lmessage->setAlignment(Qt::AlignLeft);
            lmessage->setText(logmsg->message);
            lmessage->setWordWrap(true);
            lmessage->setTextInteractionFlags(Qt::TextSelectableByMouse);
            blayout->addWidget(lmessage);
            QLabel* lseverity = new QLabel(this);
            lseverity->setAlignment(Qt::AlignLeft|Qt::AlignTop);
            lseverity->setText(ssev);
            layout->addWidget(lseverity);
            olayout->addWidget(w);
        }
    }
    olayout->addStretch(1);
    delete ui->scrollArea->widget();
    ui->scrollArea->setWidget(ow);
}

sh::ui::LogViewDialog::~LogViewDialog()
{
    delete ui;
}

void sh::ui::LogViewDialog::on_btnClose_clicked()
{
    this->close();
}

void sh::ui::LogViewDialog::on_btnSaveToFile_clicked()
{
    QFileDialog savedlg;
    savedlg.selectFile(QDir::homePath()+"/"+QString("shallot.log.%1.txt").arg(
                           QDateTime::currentDateTime().toString("yyyy-MM-ddThh:mm:ss")));
    savedlg.setAcceptMode(QFileDialog::AcceptSave);
    if (savedlg.exec() == QDialog::Accepted)
    {
        QString filename = savedlg.selectedFiles()[0];
        QFile file(filename);
        if (file.open(QIODevice::ReadWrite)) {
            QTextStream stream(&file);
            stream << sh::tools::Logger::instance()->getLogAsText() << endl;
            file.close();
        }
    }
}

void sh::ui::LogViewDialog::setMinimumSeverity(sh::tools::LogSeverity minseverity)
{
    _minseverity = minseverity;
    QString tpl = tr("%1 (&change)");
    switch(minseverity)
    {
    case sh::tools::LogSeverity::DEBUG:
        ui->btnSeverity->setText(tpl.arg(tr("very verbose")));
        break;
    case sh::tools::LogSeverity::INFO:
        ui->btnSeverity->setText(tpl.arg(tr("shows infos")));
        break;
    case sh::tools::LogSeverity::WARNING:
        ui->btnSeverity->setText(tpl.arg(tr("shows warnings")));
        break;
    case sh::tools::LogSeverity::ERROR:
        ui->btnSeverity->setText(tpl.arg(tr("shows errors")));
        break;
    }
    _reloadLog();
}

void sh::ui::LogViewDialog::on_btnSeverity_clicked()
{
    QMenu menu;
    QAction* actd = menu.addAction(tr("very &verbose"));
    connect(actd, &QAction::triggered, this, [=](){ setMinimumSeverity(sh::tools::LogSeverity::DEBUG); });
    QAction* acti = menu.addAction(tr("show &infos"));
    connect(acti, &QAction::triggered, this, [=](){ setMinimumSeverity(sh::tools::LogSeverity::INFO); });
    QAction* actw = menu.addAction(tr("show &warnings"));
    connect(actw, &QAction::triggered, this, [=](){ setMinimumSeverity(sh::tools::LogSeverity::WARNING); });
    QAction* acte = menu.addAction(tr("show &errors"));
    connect(acte, &QAction::triggered, this, [=](){ setMinimumSeverity(sh::tools::LogSeverity::ERROR); });
    menu.exec(ui->btnSeverity->mapToGlobal(QPoint(0, ui->btnSeverity->height())));
}
