/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QSortFilterProxyModel>
#include <QModelIndex>
#include <QDir>
#include <QMessageBox>
#include <QWidget>
#include "actions/actionsmanager.h"
#include "actions/actionactionitem.h"
#include "base/filesystemmodeldirectorytreeproxy.h"
#include "base/filesystemmodeldirectorytreeproxyvisibilityenforcements.h"
#include "base/filesystemmodel.h"
#include "base/filesystemnode.h"
#include "configuration/configurationmanager.h"
#include "filesystemhandlers/localfilesystemhandler.h"
#include "settings/settingsmanager.h"
#include "settings/setting.h"
#include "tools/mainthread.h"
#include "tools/threadpool.h"
#include "tools/thumbnailmanager.h"
#include "ui/filesystempanel.h"
#include "ui/filedetailspanel.h"
#include "ui/actionexecutioninfopanel.h"
#include "ui/fileview/fileview.h"
#include "ui/fileview/fileviewcontrol.h"
#include "jumpbar.h"
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "toolbarbutton.h"


/** \namespace Ui
 * \brief Qt-internal for ui.
 */

/** \class _ProjectInfo
 * \brief Some internal program information.
 *
 * This is auto-generated content only used by the core.
 */

/** \namespace sh
 * \brief Shallot root.
 */

/** \namespace sh::actions
 * \brief Infrastructure for actions (submenus, executable ones, and more), as shown in the toolbar and menus and used at some other places.
 *
 * See the abstract base class sh::actions::AbstractActionItem for more.
 */

/** \namespace sh::actions::common
 * \brief Implementations of actions.
 *
 * Subclasses of sh::actions::AbstractActionItem (and possibly some auxiliary stuff).
 * Many action implementations used in Shallot are here.
 * You can find other actions in the sibling namespaces and a few very special ones at different places.
 */

/** \namespace sh::actions::viewmenu
 * \brief Implementation of the view menu.
 *
 * (The view menu is next to the main menu.)
 * Subclasses of sh::actions::AbstractActionItem (and possibly some auxiliary stuff).
 * You can find other actions in the sibling namespaces and a few very special ones at different places.
 */

/** \namespace sh::actions::mainmenu
 * \brief Implementation of the main menu.
 *
 * Subclasses of sh::actions::AbstractActionItem (and possibly some auxiliary stuff).
 * You can find other actions in the sibling namespaces and a few very special ones at different places.
 */

/** \namespace sh::base
 * \brief Core logic.
 *
 * Includes common data structures and some basic infrastructure.
 * Some other basic things are in sh::tools.
 * Some other namespaces in sh contain parts of the infrastructure as well.
 */

/** \namespace sh::configuration
 * \brief Configuration, so everything you see in 'Finetuning', and some other data (e.g. get shallot program directory).
 */

/** \namespace sh::detailcolumns
 * \brief Implementations of detail columns.
 *
 * Subclasses of sh::base::DetailColumn (and possibly some auxiliary stuff).
 * They are shown in the fileviews and can be queried in code.
 */

/** \namespace sh::exceptions
 * \brief Exceptions.
 *
 * The class hierarchy of Shallot exceptions as well as some utilities for
 * exception handling.
 */

/** \namespace sh::filesystemhandlers
 * \brief Implementations of filesystem handlers.
 *
 * Subclasses of sh::base::FilesystemHandler (and possibly some auxiliary stuff).
 * They implement how to access various types of filesystems.
 *
 * Some special purpose handlers reside outside of this namespace.
 */

/** \namespace sh::paneldetails
 * \brief Infrastructure for panel details.
 *
 * Can be shown in the bottom part of the main window.
 * Beyond the infrastructure, it also includes some very basic detail providers.
 *
 * See sh::paneldetails::PanelDetailManager for more.
 */

/** \namespace sh::scripting
 * \brief The Shallot scripting interface.
 */

/** \namespace sh::scripting::api
 * \brief Adapter functions and classes for interaction with scripting.
 *
 * Some programming interfaces in the core are not directly usable
 * for the scripting api. Scripting-friendly surrogates for them are here.
 */

/** \namespace sh::search
 * \brief Infrastructure for searches in the filesystem.
 *
 * See sh::search::SearchManager for more.
 */

/** \namespace sh::search::criteria
 * \brief Implementations of search criteria.
 *
 * Subclasses of sh::search::SearchCriterion.
 */

/** \namespace sh::settings
 * \brief Infrastructure for settings, so everything you see in 'Manage saved settings'.
 *
 * See the abstract base class sh::settings::Setting and sh::settings::SettingsManager for more.
 */

/** \namespace sh::settings::common
 * \brief Implementations of settings.
 *
 * Subclasses of sh::settings::Setting (and possibly some auxiliary stuff).
 * Everything here could be listed in the 'Save settings' dialog.
 */

/** \namespace sh::tools
 * \brief Auxiliary stuff.
 *
 * Some of them are freqently used and highly important.
 */

/** \namespace sh::tools::accounts.
 * \brief Account and password storage.
 *
 * See sh::tools::accounts::AccountsManager for more.
 */

/** \namespace sh::tools::filetypes
 * \brief Tools for determining a file's type (pdf, image, mp3, et al) and ways how to deal with that.
 *
 * See sh::tools::filetypes::FileTypeManager for more.
 */

/** \namespace sh::ui
 * \brief User interface.
 *
 * This is the presentation layer.
 */

/** \namespace sh::ui::feedbackpanels
 * \brief User interface support for user communication in action executions.
 *
 * Those small ui controls are mainly used in sh::ui::ActionExecutionInfoDialog.
 */

/** \namespace sh::ui::fileview
 * \brief File views (as filelists, icon grids).
 *
 * They are used in the main window for listing the content of a directory.
 */

/** \namespace sh::ui::filepropertydialog
 * \brief All stuff used for/in the Properties dialog.
 *
 * This also contains infrastructure elements.
 */

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalFolderTreeIconSize =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.FolderTreeIconSize", -1,
            sh::configuration::ConfigurationValue::ValueTypeInteger(),
            QObject::tr("Icon size in folder tree"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("For default size, set it to -1."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalToolbarLocation =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.ToolbarLocation", "N",
            sh::configuration::ConfigurationValue::ValueTypeString(),
            QObject::tr("Main toolbar location"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("N, E, S or W."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelLocation =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelLocation", "S",
            sh::configuration::ConfigurationValue::ValueTypeString(),
            QObject::tr("Detail panel location"),
            sh::configuration::ConfigurationCategory::CategoryGUI, QString(),
            QObject::tr("N, E, S or W."));

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelDimensions =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelDimensions", "-1",
            sh::configuration::ConfigurationValue::ValueTypeString());

std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::MainWindow::cfgvalDetailPanelThumbnailDimensions =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue(
            "sh::ui::MainWindow.DetailPanelThumbnailDimensions", "-1",
            sh::configuration::ConfigurationValue::ValueTypeString());

sh::ui::MainWindow* sh::ui::MainWindow::_mainWindow = 0;

sh::ui::MainWindow* sh::ui::MainWindow::mainWindow()
{
    return _mainWindow;
}

sh::ui::MainWindow::MainWindow(sh::tools::SingletonInitializer *singletonInitializer, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow), _titlePattern("{dirname}"), singletonInitializer(singletonInitializer)
{
    _mainWindow = this;
    ui->setupUi(this);
    _thumbnailwidth = logicalDpiX();
    QSize treeViewDefaultIconSize = ui->treeView->iconSize();
    cfgvalFolderTreeIconSize->consumeValue(this, [=](){
        int s = cfgvalFolderTreeIconSize->getConfigValueInt();
        if (s<0)
            ui->treeView->setIconSize(treeViewDefaultIconSize);
        else
            ui->treeView->setIconSize(QSize(s,s));
    });
    cfgvalToolbarLocation->consumeValue(this, [=](){
        QString s = cfgvalToolbarLocation->getConfigValueString();
        if (s != toolbarPosition())
        {
            setToolbarPosition(s);
            slot_toolbar_moved();
        }
    });
    cfgvalDetailPanelLocation->consumeValue(this, [=](){
        QString s = cfgvalDetailPanelLocation->getConfigValueString();
        if (s != detailPanelPosition())
        {
            setDetailPanelPosition(s);
            slot_detailbar_moved();
        }
    });
    cfgvalDetailPanelDimensions->consumeValue(this, [=](){
        _resizedetailpanel();
    });
    cfgvalDetailPanelThumbnailDimensions->consumeValue(this, [=](){
        _resizethumbnail();
    });
    fspanel = new sh::ui::FilesystemPanel(ui->filesystempanel);
    ui->filesystempanel->layout()->addWidget(fspanel);
    jumpbar = new sh::ui::JumpBar(this);
    static_cast<QVBoxLayout*>( ui->centralWidget->layout() )->insertWidget(0,jumpbar);
    connect(ui->mainToolBar, SIGNAL(topLevelChanged(bool)), this, SLOT(slot_toolbar_moved()));
    connect(ui->filedetailsdockwidget, SIGNAL(topLevelChanged(bool)), this, SLOT(slot_detailbar_moved()));
    connect(ui->treeView, SIGNAL(collapsed(QModelIndex)), this, SLOT(slot_treeview_collapsedexpanded(QModelIndex)));
    connect(ui->treeView, SIGNAL(expanded(QModelIndex)), this, SLOT(slot_treeview_collapsedexpanded(QModelIndex)));

    ui->splitter->setStretchFactor(0,0);
    ui->splitter->setStretchFactor(1,1);    

    _statusbar_fullheight = ui->statusBar->height();
    ui->statusBar->setMinimumHeight(0);

    connect(&_statusbar_timer, &QTimer::timeout, this, &sh::ui::MainWindow::slot_statusbartimer);

    ui->filedetailsdockwidget->setAutoFillBackground(true);
    QPalette pal;
    pal.setColor(backgroundRole(), sh::configuration::ConfigurationManager::instance()->style()->windowColorHighlighted());
    ui->filedetailsdockwidget->setPalette(pal);
}

void sh::ui::MainWindow::_initialize()
{

    connect(fspanel, &sh::ui::FilesystemPanel::viewmodeChanged, this, [=](int){
        _refresh_detailthumbnailvisibility();
    });

    connect(fspanel, &sh::ui::FilesystemPanel::currentViewChanged, this, [=](){
        _refresh_detailthumbnailvisibility();
    });

    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        treemodel = sh::base::FilesystemModelDirectoryTreeProxy::instance().get();
        ui->treeView->setModel( treemodel );
        sh::actions::ActionsManager::instance()->setToolbar(ui->mainToolBar);
    });

    connect(fspanel, &sh::ui::FilesystemPanel::folderActivated , [=](std::shared_ptr<sh::base::FilesystemNode> node, bool realSwitch){
        QModelIndex mi = sh::base::FilesystemModel::instance()->getIndexForNode( node );
        _jumpToIndex(mi, (realSwitch?-1:1) );
    });

    connect(fspanel, &sh::ui::FilesystemPanel::selectionChanged , [=](){
        auto selectednodes = fspanel->selectedNodes();
        ui->filedetailspanel->setNodes(selectednodes);
        _request_detailthumbnail();
        auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(selectednodes, getCurrentDirectory());
        sh::actions::ActionsManager::instance()->refreshSituation(instantiation);
    });

    connect(ui->treeView->selectionModel(), &QItemSelectionModel::currentChanged , [=](QModelIndex a, QModelIndex){
        if (sh::base::FilesystemModel::instance()->getNodeForIndex(a) != getCurrentDirectory())
        {
            QModelIndex t = treemodel->mapToSource(a);
            _jumpToIndex(t, 2);
        }
    });

    connect(jumpbar, &sh::ui::JumpBar::nodeChanged , [=](){
        QModelIndex mi = sh::base::FilesystemModel::instance()->getIndexForNode( jumpbar->node() );
        _jumpToIndex(mi, 3);
    });

    connect(jumpbar, &sh::ui::JumpBar::eurlRequested , [=](std::shared_ptr<const sh::base::Eurl> eurl){
        jumpToEurl(eurl);
        fspanel->_skip_eventfilter = true;
        jumpbar->setButtonMode();
        fspanel->_skip_eventfilter = false;
        fspanel->viewwidget(fspanel->currentViewIndex())->focus();
    });

    connect(ui->detailssplitter, &QSplitter::splitterMoved, [=](int, int){
        _thumbnailwidth = ui->detailssplitter->sizes().first();
        _thumbnail_resized();
    });

    ui->detailsThumbnail->installEventFilter(this);
    ui->filedetailsdockwidget->installEventFilter(this);

    sh::tools::MainThread::dispatcher()->invokeSync([&](){
        setFileDetailsPanelVisible(true);
        QString startpath = QApplication::arguments().length() < 2 ? QDir::currentPath() : QApplication::arguments()[1];
        jumpToEurl(sh::base::Eurl::fromString(startpath));
    });

}

bool sh::ui::MainWindow::eventFilter(QObject *obj, QEvent *event)
{
    if (obj == ui->detailsThumbnail) {
        if (event->type() == QEvent::Resize && !_request_detailthumbnail_skip)
            _request_detailthumbnail(false);
    }
    else if (obj == ui->filedetailsdockwidget) {
        if (event->type() == QEvent::Resize)
            _detailpanel_resized();
    }
    return QObject::eventFilter(obj, event);
}

void sh::ui::MainWindow::_refresh_detailthumbnailvisibility()
{
    if (fspanel->currentView()->viewmode() == sh::ui::fileview::FileViewMode::ListView) {
        _thumbnailwidth = std::abs(_thumbnailwidth);
        ui->detailssplitter->setSizes({_thumbnailwidth, 1});
    }
    else {
        _thumbnailwidth = -std::abs(_thumbnailwidth);
        ui->detailssplitter->setSizes({0, 1});
    }
}

void sh::ui::MainWindow::_request_detailthumbnail(bool force) {
    if (_thumbnailrequestongoing && !force /* this is laziness; we could implement the same for forced requests as well */) {
        _thumbnailrequestfollowup = true;
        return;
    }
    quint64 mythumbnailrequestid = ++_lastthumbnailrequestid;
    if (_thumbnailwidth > 0) {
        auto selectednodes = fspanel->selectedNodes();
        if (selectednodes.length() == 1) {
            QSize dtsize = ui->detailsThumbnail->size();
            int dtwidth = dtsize.width();
            int dtheight = dtsize.height();
            if (ui->detailssplitter->orientation() == Qt::Horizontal)
                dtwidth = _thumbnailwidth;
            else
                dtheight = _thumbnailwidth;
            QSize thsize = QSize(dtwidth, dtheight);
            if (force || _thumbnaillastreqwidth != thsize.width() || _thumbnaillastreqheight != thsize.height()) {
                _thumbnaillastreqwidth = thsize.width();
                _thumbnaillastreqheight = thsize.height();
                _thumbnailrequestongoing = true;
                ui->detailsThumbnail->setPixmap(QPixmap());
                sh::tools::ThumbnailManager::instance()->requestThumbnail(selectednodes.first(), _thumbnaillastreqwidth/*thsize.width()*/, _thumbnaillastreqheight/*thsize.height()*/, [=](QIcon result){
                    if (sh::tools::SingletonInitializer::isShuttingDown())
                        return;
                    sh::tools::MainThread::dispatcher()->invokeSync([&](){
                        if (mythumbnailrequestid != _lastthumbnailrequestid)
                            return;
                        _request_detailthumbnail_skip = true;
                        ui->detailssplitter->setSizes({0, 1});
                        _request_detailthumbnail_skip = false;
                        if (!result.isNull()) {
                            _request_detailthumbnail_skip = true;
                            ui->detailssplitter->setSizes({_thumbnailwidth, 1});
                            ui->detailsThumbnail->setPixmap(result.pixmap(thsize));
                            _request_detailthumbnail_skip = false;
                        }
                        _thumbnailrequestongoing = false;
                        if (_thumbnailrequestfollowup) {
                            _thumbnailrequestfollowup = false;
                            _request_detailthumbnail(false);
                        }
                    });
                });
            }
        }
        else {
            ui->detailssplitter->setSizes({0, 1});
            ui->detailsThumbnail->setPixmap(QPixmap());
        }
    }
    else
        ui->detailsThumbnail->setPixmap(QPixmap());
}

void sh::ui::MainWindow::slot_statusbartimer()
{
    int h = ui->statusBar->minimumHeight();
    int step = _statusbar_targetheight - h;
    if (step==0)
        _statusbar_timer.stop();
    else
    {
        int ostep = step;
        step /= 2;
        if (step==0)
            step= (ostep>0)?1:-1;
        ui->statusBar->setMinimumHeight(h+step);
    }
}

void sh::ui::MainWindow::_newToStatusbar_helper(sh::ui::ActionExecutionInfoPanel* w)
{
    CHECK_CORRECT_CALLER_THREAD
    setStatusBarVisibility(true);
    _statusbar_childs.insert(w);
    connect(w, &QObject::destroyed, this, [=](){
        _statusbar_childs.remove(w);
        setStatusBarVisibility(!_statusbar_childs.isEmpty());
    });
    connect(w, &sh::ui::ActionExecutionInfoPanel::visibilityChanged, this, [=](){
        if (w->isVisible())
            _statusbar_childs.insert(w);
        else
            _statusbar_childs.remove(w);
        setStatusBarVisibility(!_statusbar_childs.isEmpty());
    });
}

void sh::ui::MainWindow::appendToStatusbar(sh::ui::ActionExecutionInfoPanel* w)
{
    statusBar()->addWidget(w);
    _newToStatusbar_helper(w);
}

void sh::ui::MainWindow::insertIntoStatusbar(int i, sh::ui::ActionExecutionInfoPanel* w)
{
    statusBar()->insertWidget(i, w);
    _newToStatusbar_helper(w);
}

void sh::ui::MainWindow::setStatusBarVisibility(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    if ( (v && (_statusbar_targetheight == 0)) || (!v && (_statusbar_targetheight != 0)) )
    {
        _statusbar_targetheight = v?_statusbar_fullheight:0;
        if (!_statusbar_timer.isActive())
            _statusbar_timer.start(80);
    }
}

void sh::ui::MainWindow::quit()
{
    if (uninitthread)
        return;
    delete fspanel;
    fspanel = 0;
    uninitthread = new UninitializeHelperThread(singletonInitializer, [=](){
        close();
    });
    uninitthread->start();
}

void sh::ui::MainWindow::closeEvent ( QCloseEvent * event )
{
    if (sh::tools::SingletonInitializer::isShutdown())
    {
        delete uninitthread;
        QApplication::instance()->exit();
    }
    else
    {
        if (sh::actions::ActionActionItem::runningActions()>0)
        {
            event->ignore();
            QMessageBox::information(this, tr("Shallot"), tr("Shallot cannot be closed while actions are running."));
            return;
        }
        event->accept();
        quit();
    }
}

sh::ui::MainWindow::UninitializeHelperThread::UninitializeHelperThread(sh::tools::SingletonInitializer* singletonInitializer, std::function<void()> onsuccess)
    :singletonInitializer(singletonInitializer), onsuccess(onsuccess)
{

}

void sh::ui::MainWindow::UninitializeHelperThread::run()
{
    singletonInitializer->shutdown();
    singletonInitializer->waitShutdown(onsuccess);
}

void sh::ui::MainWindow::_jumpToIndex(QModelIndex index, int skip)
{
    CHECK_CORRECT_CALLER_THREAD
    std::shared_ptr<sh::base::FilesystemNode> node = _currentDirectory = sh::base::FilesystemModel::instance()->getNodeForIndex(index);
    if (skip!=1) {
        // update file panel
        fspanel->selectFolder(node);
    }
    if (skip!=2) {
        // update tree
        if (_treeSticky) {
            QModelIndex tindex = treemodel->mapFromSource(index);
            if (tindex.isValid())
                ui->treeView->selectionModel()->setCurrentIndex(tindex, QItemSelectionModel::ClearAndSelect);
            else
                ui->treeView->clearSelection();
        }
    }
    if (skip!=3)
    {
        // update jumpbar
        jumpbar->setNode(node);
    }
    // update actions
    sh::tools::MainThread::dispatcher()->invokeAsync([=](){ //async is important here (error handling, ...)
        auto instantiation = std::make_shared<sh::actions::ActionInstantiation>(fspanel->activeView->selectedNodes(), node);
        sh::actions::ActionsManager::instance()->refreshSituation(instantiation);
        slot_toolbar_moved();
    });
    // global stuff
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToMainWindow();
    ui->filedetailspanel->setNodes(QList<std::shared_ptr<sh::base::FilesystemNode>>());
    _writeTitle();
    emit currentDirectoryChanged();
}

void sh::ui::MainWindow::_jumpToNode(std::shared_ptr<sh::base::FilesystemNode> node, int skip)
{
    CHECK_CORRECT_CALLER_THREAD
    if (node->isAlive())
    {
        QModelIndex idx = sh::base::FilesystemModel::instance()->getIndexForNode(node);
        _jumpToIndex(idx, skip);
        fspanel->viewwidget(fspanel->currentViewIndex())->focus();
    }
}

void sh::ui::MainWindow::jumpToNode(std::shared_ptr<sh::base::FilesystemNode> node)
{
    CHECK_CORRECT_CALLER_THREAD
    _jumpToNode(node, -1);
}

void sh::ui::MainWindow::jumpToEurl(std::shared_ptr<const sh::base::Eurl> eurl)
{
    sh::tools::ThreadPool::enqueueIfInMain([=](){
        QList<std::shared_ptr<sh::base::FilesystemNode>> nodes = sh::base::FilesystemModel::instance()->findNodesForEurl(eurl);
        if (!nodes.isEmpty())
        {
            sh::tools::MainThread::dispatcher()->invokeSync([&](){
                _jumpToNode(nodes.first(), -1);
            });
        }
    });
}

std::shared_ptr<sh::base::FilesystemNode> sh::ui::MainWindow::getCurrentDirectory()
{
    CHECK_CORRECT_CALLER_THREAD
    return _currentDirectory;
}

QString sh::ui::MainWindow::toolbarPosition()
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::ToolBarArea area = toolBarArea(ui->mainToolBar);
    QString location = "N";
    switch(area){
    case Qt::LeftToolBarArea:
        location="W"; break;
    case Qt::RightToolBarArea:
        location="E"; break;
    case Qt::TopToolBarArea:
        location="N"; break;
    case Qt::BottomToolBarArea:
        location="S"; break;
    default: break;
    }
    return location;
}

void sh::ui::MainWindow::setToolbarPosition(QString pos)
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::ToolBarArea area;
    if (pos == "S")
        area = Qt::BottomToolBarArea;
    else if (pos == "W")
        area = Qt::LeftToolBarArea;
    else if (pos == "E")
        area = Qt::RightToolBarArea;
    else /* if (pos == sh::ui::ToolbarButton::Position::BOTTOM) */
        area = Qt::TopToolBarArea;
    addToolBar(area, ui->mainToolBar);
}

QString sh::ui::MainWindow::detailPanelPosition()
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::DockWidgetArea area = dockWidgetArea(ui->filedetailsdockwidget);
    QString location = "N";
    switch(area){
    case Qt::LeftDockWidgetArea:
        location="W"; break;
    case Qt::RightDockWidgetArea:
        location="E"; break;
    case Qt::TopDockWidgetArea:
        location="N"; break;
    case Qt::BottomDockWidgetArea:
        location="S"; break;
    default: break;
    }
    return location;
}

void sh::ui::MainWindow::setDetailPanelPosition(QString pos)
{
    CHECK_CORRECT_CALLER_THREAD
    Qt::DockWidgetArea area;
    if (pos == "S")
        area = Qt::BottomDockWidgetArea;
    else if (pos == "W")
        area = Qt::LeftDockWidgetArea;
    else if (pos == "E")
        area = Qt::RightDockWidgetArea;
    else  // if (pos == sh::ui::ToolbarButton::Position::BOTTOM)
        area = Qt::TopDockWidgetArea;
    addDockWidget(area, ui->filedetailsdockwidget);
}

void sh::ui::MainWindow::slot_detailbar_moved()
{
    Qt::DockWidgetArea area = dockWidgetArea(ui->filedetailsdockwidget);
    Qt::Orientation orientation = (area == Qt::LeftDockWidgetArea || area == Qt::RightDockWidgetArea)
            ? Qt::Vertical : Qt::Horizontal;
    QString location = detailPanelPosition();
    cfgvalDetailPanelLocation->setConfigValue(location);
    ui->detailssplitter->setOrientation(orientation);
    ui->filedetailspanel->setOrientation(orientation);
    _resizethumbnail();
    _resizedetailpanel();
    _myorientation = orientation;
}

void sh::ui::MainWindow::slot_toolbar_moved()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString location = toolbarPosition();
    cfgvalToolbarLocation->setConfigValue(location);
    sh::ui::ToolbarButton::Position plocation = sh::ui::ToolbarButton::Position::TOP;
    if (location == "N")
        plocation = sh::ui::ToolbarButton::Position::TOP;
    else if (location == "E")
        plocation = sh::ui::ToolbarButton::Position::RIGHT;
    else if (location == "S")
        plocation = sh::ui::ToolbarButton::Position::BOTTOM;
    else if (location == "W")
        plocation = sh::ui::ToolbarButton::Position::LEFT;
    foreach(QObject* w, ui->mainToolBar->children())
    {
        sh::ui::ToolbarButton* btn = qobject_cast<sh::ui::ToolbarButton*>(w);
        if (btn != 0)
            btn->setLocation(plocation);
    }
}

void sh::ui::MainWindow::slot_treeview_collapsedexpanded(const QModelIndex &index)
{
    QModelIndex fmindex = qobject_cast<const QAbstractProxyModel*>(index.model())->mapToSource(index);
    std::shared_ptr<sh::base::FilesystemNode> node = sh::base::FilesystemModel::instance()->getNodeForIndex(fmindex);
    sh::base::FilesystemModelDirectoryTreeProxyVisibilityEnforcements::instance()->nodeCollapsedInTree(node);
}

void sh::ui::MainWindow::_thumbnail_resized()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    QString x = cfgvalDetailPanelThumbnailDimensions->getConfigValueString();
    int w = -1;
    int h = -1;
    QStringList lx = x.split(";");
    if (lx.length() == 2) {
        bool ok = true;
        w = lx[0].toInt(&ok);
        if (!ok)
            w = -1;
        h = lx[1].toInt(&ok);
        if (!ok)
            h = -1;
    }
    if (ui->detailssplitter->orientation() == Qt::Horizontal)
        w = _thumbnailwidth;
    else
        h = _thumbnailwidth;
    cfgvalDetailPanelThumbnailDimensions->setConfigValue(QString("%1;%2").arg(w).arg(h));
}

void sh::ui::MainWindow::_resizethumbnail()
{
    QString s = cfgvalDetailPanelThumbnailDimensions->getConfigValueString();
    QStringList sl = s.split(";");
    if (sl.length() == 2) {
        bool ok = true;
        int x = (ui->detailssplitter->orientation() == Qt::Horizontal ? sl[0] : sl[1]).toInt(&ok);
        if (ok && x != -1) {
            _thumbnailwidth = x;
            ui->detailssplitter->setSizes({x, 1});
        }
    }
}

void sh::ui::MainWindow::_detailpanel_resized()
{
    ONLY_ALLOWED_TO_CALL_FROM_MAIN_THREAD
    Qt::DockWidgetArea dockarea = dockWidgetArea(ui->filedetailsdockwidget);
    if ((dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea) != (_myorientation == Qt::Horizontal))
        return;
    QString x = cfgvalDetailPanelDimensions->getConfigValueString();
    int w = -1;
    int h = -1;
    QStringList lx = x.split(";");
    if (lx.length() == 2) {
        bool ok = true;
        w = lx[0].toInt(&ok);
        if (!ok)
            w = -1;
        h = lx[1].toInt(&ok);
        if (!ok)
            h = -1;
    }
    QSize dss = ui->filedetailsdockwidget->size();
    if (dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea)
        h = dss.height();
    else
        w = dss.width();
    cfgvalDetailPanelDimensions->setConfigValue(QString("%1;%2").arg(w).arg(h));
}

void sh::ui::MainWindow::_resizedetailpanel()
{
    QString s = cfgvalDetailPanelDimensions->getConfigValueString();
    QStringList sl = s.split(";");
    if (sl.length() == 2) {
        bool ok = true;
        int x;
        Qt::Orientation ortho;
        Qt::DockWidgetArea dockarea = dockWidgetArea(ui->filedetailsdockwidget);
        if (dockarea == Qt::TopDockWidgetArea || dockarea == Qt::BottomDockWidgetArea) {
            x = sl[1].toInt(&ok);
            ortho = Qt::Vertical;
        }
        else {
            x = sl[0].toInt(&ok);
            ortho = Qt::Horizontal;
        }
        if (ok && x != -1)
            resizeDocks({ui->filedetailsdockwidget}, {x}, ortho);
    }
}

sh::ui::MainWindow::~MainWindow()
{
    CHECK_CORRECT_CALLER_THREAD
    delete ui;
    _mainWindow = 0;
}

void sh::ui::MainWindow::setTreeVisible(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    ui->treeView->setVisible(v);
    emit viewOptionChanged();
}

bool sh::ui::MainWindow::treeVisible()
{
    CHECK_CORRECT_CALLER_THREAD
    return ui->treeView->isVisible();
}

void sh::ui::MainWindow::setTreeSticky(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    _treeSticky = v;
    emit viewOptionChanged();
}

bool sh::ui::MainWindow::treeSticky()
{
    CHECK_CORRECT_CALLER_THREAD
    return _treeSticky;
}

void sh::ui::MainWindow::setFileDetailsPanelVisible(bool v)
{
    CHECK_CORRECT_CALLER_THREAD
    if (v != _detailsPanelVisible)
    {
        _detailsPanelVisible = v;
        ui->filedetailsdockwidget->setVisible(v);
        emit viewOptionChanged();
    }
}

bool sh::ui::MainWindow::fileDetailsPanelVisible()
{
    CHECK_CORRECT_CALLER_THREAD
    return _detailsPanelVisible;
}

QString sh::ui::MainWindow::getCurrentProfile()
{
    CHECK_CORRECT_CALLER_THREAD
    return _currentProfile;
}

void sh::ui::MainWindow::setCurrentProfile(QString profile)
{
    CHECK_CORRECT_CALLER_THREAD
    _currentProfile = profile;
    sh::settings::SettingsManager::instance()->applyGlobalSettingsToMainWindow();
    sh::settings::SettingsManager::instance()->applyPerEurlSettingsToMainWindow();
    for(int i=0; i<fspanel->viewsCount(); i++)
    {
        sh::ui::fileview::FileViewControl* viewctrl = fspanel->view(i);
        sh::settings::SettingsManager::instance()->applyGlobalSettingsToFileView(viewctrl);
        sh::settings::SettingsManager::instance()->applyPerEurlSettingsToFileView(viewctrl);
    }
    emit currentProfileChanged();
}

void sh::ui::MainWindow::reloadProfile()
{
    CHECK_CORRECT_CALLER_THREAD
    setCurrentProfile(getCurrentProfile());
}

sh::ui::FilesystemPanel* sh::ui::MainWindow::filesystemPanel()
{
    CHECK_CORRECT_CALLER_THREAD
    return fspanel;
}

void sh::ui::MainWindow::setTitlePattern(QString value)
{
    CHECK_CORRECT_CALLER_THREAD
    if (_titlePattern != value) {
        _titlePattern = value;
        _writeTitle();
    }
}

QString sh::ui::MainWindow::getTitlePattern()
{
    CHECK_CORRECT_CALLER_THREAD
    return _titlePattern;
}

void sh::ui::MainWindow::_writeTitle()
{
    CHECK_CORRECT_CALLER_THREAD
    QString tt = _titlePattern;
    std::shared_ptr<sh::base::FilesystemNode> _node =getCurrentDirectory();
    QString dir;
    QString dirname;
    if (_node==0) {
        dirname = dir = tr("(root)");
    }
    else {
        dir = _node->eurl()->asString();
        dirname = _node->displayName();
    }
    QString t = "";
    while (!tt.isEmpty()) {
        int a = tt.indexOf("{");
        if (a > -1) {
            int b = tt.indexOf("}", a);
            if (b > -1) {
                QString token = tt.mid(a+1, b-a-1);
                t += tt.left(a);
                if (token == "dir")
                    t += dir;
                else if (token == "dirname")
                    t += dirname;
                else
                    t += tt.mid(a, b-a+1);
                tt = tt.mid(b+1);
                continue;
            }
        }
        t += tt;
        break;
    }
    setWindowTitle(t);
}

void sh::ui::MainWindow::jumpbarSetTextMode()
{
    jumpbar->setTextMode();
}

void sh::ui::MainWindow::jumpbarSetButtonMode()
{
    jumpbar->setButtonMode();
}

sh::ui::MainWindow::MyFlexibleLabel::MyFlexibleLabel(QWidget* parent)
                : QLabel(parent)
{
}

QSize sh::ui::MainWindow::MyFlexibleLabel::sizeHint() const
{
    return QSize(1, 1);
}
