/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <functional>
#include <QLabel>
#include <QMainWindow>
#include <QModelIndex>
#include <QSet>
#include "exceptions/exception.h"
#include "tools/threaddispatcher.h"
#include "ui/toolbarbutton.h"


namespace Ui {
class MainWindow;
}

namespace sh{
    namespace base{ class Eurl; class FilesystemNode;class FilesystemModelDirectoryTreeProxy; }
    namespace configuration{ class ConfigurationValue; }
    namespace tools{ class SingletonInitializer; }

    namespace ui{
        class ActionExecutionInfoPanel;
        class FilesystemPanel;
        class JumpBar;

        /** \brief
         * The main window.
         */
        class MainWindow : public QMainWindow
        {
            Q_OBJECT

        public:
            explicit MainWindow(sh::tools::SingletonInitializer* singletonInitializer, QWidget *parent = 0);
            ~MainWindow();
            /** \brief
             * Returns the instance of this singleton.
             */
            static MainWindow* mainWindow();
            void _initialize();

            sh::ui::FilesystemPanel* filesystemPanel();

            /** \brief
             * Let the current view jump to another location.
             */
            void jumpToEurl(std::shared_ptr<const sh::base::Eurl>eurl);
            /** \brief
             * Let the current view jump to another location.
             */
            void jumpToNode(std::shared_ptr<sh::base::FilesystemNode> node);

            void setTitlePattern(QString value);
            QString getTitlePattern();

            void setCurrentProfile(QString profile);
            QString getCurrentProfile();

            void setTreeVisible(bool v);
            bool treeVisible();

            void setTreeSticky(bool v);
            bool treeSticky();

            void setFileDetailsPanelVisible(bool v);
            bool fileDetailsPanelVisible();

            /** \brief
             * Gets the sh::base::FilesystemNode selected in the current view.
             */
            std::shared_ptr<sh::base::FilesystemNode> getCurrentDirectory();

            void appendToStatusbar(sh::ui::ActionExecutionInfoPanel* w);
            void insertIntoStatusbar(int i, sh::ui::ActionExecutionInfoPanel* w);

            void reloadProfile();

            void jumpbarSetTextMode();
            void jumpbarSetButtonMode();

            QString toolbarPosition();
            void setToolbarPosition(QString pos);

            QString detailPanelPosition();
            void setDetailPanelPosition(QString pos);

            void quit();

            /** \brief
             * Needed internally in main window in order to have a label which really does not request any sizes.
             */
            class MyFlexibleLabel : public QLabel {
            public:
                MyFlexibleLabel(QWidget* parent = 0);
                QSize sizeHint() const override;
            };

        signals:
            void currentDirectoryChanged();
            void viewOptionChanged();
            void currentProfileChanged();

        protected:
            virtual void closeEvent ( QCloseEvent * event );
            bool eventFilter(QObject *obj, QEvent *event);

        private slots:
            void slot_detailbar_moved();
            void slot_toolbar_moved();
            void slot_statusbartimer();
            void slot_treeview_collapsedexpanded(const QModelIndex &index);

        private:
            void _jumpToNode(std::shared_ptr<sh::base::FilesystemNode> node, int skip);
            void _jumpToIndex(QModelIndex idx, int skip);
            void _writeTitle();
            void _newToStatusbar_helper(sh::ui::ActionExecutionInfoPanel *w);
            void _refresh_detailthumbnailvisibility();
            void _request_detailthumbnail(bool force=true);
            void _thumbnail_resized();
            void _resizethumbnail();
            void _detailpanel_resized();
            void _resizedetailpanel();
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalFolderTreeIconSize;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalToolbarLocation;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelLocation;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelDimensions;
            static std::shared_ptr<sh::configuration::ConfigurationValue> cfgvalDetailPanelThumbnailDimensions;
            bool _treeSticky = true;
            bool _detailsPanelVisible = true;
            Ui::MainWindow *ui;
            QString _titlePattern;
            QString _currentProfile;
            sh::ui::FilesystemPanel* fspanel;
            sh::base::FilesystemModelDirectoryTreeProxy* treemodel = 0;
            sh::ui::JumpBar* jumpbar;
            static MainWindow* _mainWindow;
            QTimer _statusbar_timer;
            int _statusbar_fullheight;
            int _statusbar_targetheight=0;
            QSet<sh::ui::ActionExecutionInfoPanel*> _statusbar_childs;
            void setStatusBarVisibility(bool v);
            sh::tools::SingletonInitializer* singletonInitializer;
            std::shared_ptr<sh::base::FilesystemNode> _currentDirectory = 0;
            quint64 _lastthumbnailrequestid = 0;
            bool _thumbnailrequestongoing = false;
            bool _thumbnailrequestfollowup = false;
            int _thumbnailwidth = 0;
            int _thumbnaillastreqwidth = 0;
            int _thumbnaillastreqheight = 0;
            bool _request_detailthumbnail_skip = false;
            Qt::Orientation _myorientation = Qt::Horizontal;

            class UninitializeHelperThread : public QThread
            {
            public:
                UninitializeHelperThread(sh::tools::SingletonInitializer* singletonInitializer, std::function<void()> onsuccess);
                void run();
            private:
                sh::tools::SingletonInitializer* singletonInitializer;
                std::function<void()> onsuccess;
            };

            UninitializeHelperThread* uninitthread = 0;
        };

    }
}

#endif // MAINWINDOW_H
