/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QInputDialog>
#include "base/eurl.h"
#include "base/filesystemnode.h"
#include "configuration/configurationmanager.h"
#include "tools/iconmanager.h"
#include "tools/bookmarkmanager.h"
#include "mainwindow.h"
#include "managebookmarksdialog.h"
#include "ui_managebookmarksdialog.h"


sh::ui::ManageBookmarksDialog::ManageBookmarksDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::ManageBookmarksDialog)
{
    ui->setupUi(this);
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    ui->treeWidget->setFocus();
    ui->btnDelete->setIcon(sh::tools::IconManager::instance()->getIcon("list-remove"));
    ui->btnNew->setIcon(sh::tools::IconManager::instance()->getIcon("list-add"));
    ui->btnUp->setIcon(sh::tools::IconManager::instance()->getIcon("arrow-up"));
    ui->btnDown->setIcon(sh::tools::IconManager::instance()->getIcon("arrow-down"));
    ui->btnMove->setIcon(sh::tools::IconManager::instance()->getIcon("transform-move"));
    sm->style()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    connect(sh::tools::BookmarkManager::instance(), &sh::tools::BookmarkManager::changed, this, [=](){
        readBookmarks();
    });
    readBookmarks();
    _refresh_values();
}

sh::ui::ManageBookmarksDialog::~ManageBookmarksDialog()
{
    delete ui;
}

void sh::ui::ManageBookmarksDialog::on_btnClose_clicked()
{
    accept();
}

void sh::ui::ManageBookmarksDialog::readBookmarks()
{
    QHash<QString, QTreeWidgetItem*> items;
    ui->treeWidget->clear();
    QTreeWidgetItem* root = new QTreeWidgetItem(ui->treeWidget, QStringList({tr("Bookmarks"), "", "", ""}));
    root->setIcon(0, sh::tools::IconManager::instance()->getIcon("folder-bookmark"));
    items[""] = root;
    QStringList folders;
    foreach(std::shared_ptr<sh::tools::Bookmark> bookmark, sh::tools::BookmarkManager::instance()->getBookmarks())
        folders.append(bookmark->folder().join('/'));
    qSort(folders.begin(), folders.end());
    foreach(QString folder, folders)
    {
        if (!folder.isEmpty())
        {
            QTreeWidgetItem* parent = root;
            QString fsgm;
            foreach(QString sgm, folder.split('/'))
            {
                if (!fsgm.isEmpty())
                    fsgm += "/";
                fsgm += sgm;
                auto itr = items.find(fsgm);
                if (itr != items.end())
                {
                    parent = *itr;
                }
                else
                {
                    QTreeWidgetItem* item = new QTreeWidgetItem(parent, QStringList({sgm, "", "", sgm}));
                    item->setIcon(0, sh::tools::IconManager::instance()->getIcon("folder-bookmark"));
                    items[fsgm] = item;
                    parent = item;
                }
            }
        }
    }
    foreach(std::shared_ptr<sh::tools::Bookmark> bookmark, sh::tools::BookmarkManager::instance()->getBookmarks())
    {
        QTreeWidgetItem* parent = items[bookmark->folder().join('/')];
        new QTreeWidgetItem(parent, QStringList({bookmark->label().isEmpty()?bookmark->eurl()->asString():bookmark->label(),
                                                 bookmark->id(), bookmark->eurl()->asString(), bookmark->label()}));
    }
    ui->treeWidget->expandAll();
}

void sh::ui::ManageBookmarksDialog::_refresh_values()
{
    bool selected = !ui->treeWidget->selectedItems().isEmpty();
    bool leafselected = selected && !ui->treeWidget->selectedItems().first()->data(2, Qt::DisplayRole).toString().isEmpty();
    bool rootselected = selected && ui->treeWidget->selectedItems().first() == ui->treeWidget->topLevelItem(0);
    ui->panelValues->setEnabled(leafselected);
    ui->btnDelete->setEnabled(selected && !rootselected);
    ui->btnUp->setEnabled(leafselected);
    ui->btnDown->setEnabled(leafselected);
    ui->btnMove->setEnabled(leafselected);
    if (!leafselected)
    {
        ui->edtLabel->clear();
        ui->edtLocation->clear();
    }
}

void sh::ui::ManageBookmarksDialog::on_treeWidget_itemSelectionChanged()
{
    _refresh_values();
    if (ui->edtLabel->isEnabled())
    {
        ui->edtLabel->setText(ui->treeWidget->selectedItems().first()->data(3, Qt::DisplayRole).toString());
        ui->edtLocation->setText(ui->treeWidget->selectedItems().first()->data(2, Qt::DisplayRole).toString());
        ui->btnStore->setEnabled(false);
    }
}

void sh::ui::ManageBookmarksDialog::on_btnDelete_clicked()
{
    QTreeWidgetItem* item = ui->treeWidget->selectedItems().first();
    if (item->childCount()>0)
    {
        QMessageBox::warning(this, tr("Remove entry"), tr("Collections can only be removed if empty."));
    }
    else
    {
        int btn = QMessageBox::question(this, tr("Remove entry"), tr("Do you really want to delete '%1'?").arg(item->data(0, Qt::DisplayRole).toString()));
        if (btn == QMessageBox::Yes)
        {
            if (!item->data(2, Qt::DisplayRole).toString().isEmpty())
                sh::tools::BookmarkManager::instance()->removeBookmark(item->data(1, Qt::DisplayRole).toString());
            else
                delete item;
        }
    }
}

void sh::ui::ManageBookmarksDialog::on_edtLabel_textChanged(const QString&)
{
    ui->btnStore->setEnabled(true);
}

void sh::ui::ManageBookmarksDialog::on_edtLocation_textChanged(const QString&)
{
    ui->btnStore->setEnabled(true);
}

void sh::ui::ManageBookmarksDialog::on_btnStore_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::exceptions::Exception::executeGuarded([&](){
        sh::tools::BookmarkManager::instance()->changeBookmark(id, ui->edtLabel->text(), sh::base::Eurl::fromString(ui->edtLocation->text()));
        _selectbookmark(id);
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::ManageBookmarksDialog::on_btnNew_clicked()
{
    QMenu menu;
    QFont font = menu.font();
    font.setBold(true);
    QAction* actNewBookmark = menu.addAction(tr("New bookmark"));
    actNewBookmark->setFont(font);
    QAction* actNewCollection = menu.addAction(tr("New collection of bookmarks"));
    QAction* a = menu.exec(ui->btnNew->mapToGlobal(QPoint(0, ui->btnNew->height())));
    QStringList folder;
    QTreeWidgetItem* selecteditem;
    if (!ui->treeWidget->selectedItems().isEmpty())
    {
        selecteditem = ui->treeWidget->selectedItems().first();
        if (!selecteditem->data(2, Qt::DisplayRole).toString().isEmpty())
            selecteditem = selecteditem->parent();
        QTreeWidgetItem* _item = selecteditem;
        while (_item != ui->treeWidget->topLevelItem(0))
        {
            folder.insert(0, _item->data(3, Qt::DisplayRole).toString());
            _item = _item->parent();
        }
    }
    else
        selecteditem = ui->treeWidget->topLevelItem(0);
    if (a == actNewBookmark)
    {
        QString id = sh::tools::BookmarkManager::instance()->addBookmark(folder, "", sh::ui::MainWindow::mainWindow()->getCurrentDirectory()->eurl());
        _selectbookmark(id);
        ui->edtLabel->setFocus();
    }
    else if (a == actNewCollection)
    {
        QString name = QInputDialog::getText(this, tr("New collection"), tr("Please enter the name for the new collection."), QLineEdit::Normal, "");
        name = name.replace("/", "_");
        if (!name.isEmpty())
        {
            bool exists = false;
            for(int i=0;i<selecteditem->childCount();i++)
            {
                if (selecteditem->child(i)->data(3, Qt::DisplayRole).toString() == name)
                {
                    exists = true;
                    break;
                }
            }
            if (!exists)
            {
                QTreeWidgetItem* firstNonfolder = 0;
                for(int i=0;i<selecteditem->childCount();i++)
                {
                    if (!selecteditem->child(i)->data(2, Qt::DisplayRole).toString().isEmpty())
                    {
                        firstNonfolder = selecteditem->child(i);
                        break;
                    }
                }
                QTreeWidgetItem* newitem;
                if (firstNonfolder)
                {
                    newitem = new QTreeWidgetItem(QStringList({name, "", "", name}));
                    selecteditem->insertChild(selecteditem->indexOfChild(firstNonfolder), newitem);
                }
                else
                    newitem = new QTreeWidgetItem(selecteditem, QStringList({name, "", "", name}));
                newitem->setIcon(0, sh::tools::IconManager::instance()->getIcon("folder-bookmark"));
            }
            QStringList folder;
            folder << name;
            QTreeWidgetItem* x = selecteditem;
            while(x != ui->treeWidget->topLevelItem(0))
            {
                folder.insert(0, x->data(3, Qt::DisplayRole).toString());
                x = x->parent();
            }
            ui->treeWidget->expandItem(selecteditem);
            _selectfolder(folder);
        }
    }
}

void sh::ui::ManageBookmarksDialog::_selectfolder(QStringList name)
{
    QTreeWidgetItem* node = ui->treeWidget->topLevelItem(0);
    while(!name.isEmpty())
    {
        QString segm = name.takeFirst();
        bool found = false;
        for(int i=0; i<node->childCount(); i++)
        {
            QTreeWidgetItem* itm = node->child(i);
            if (itm->data(3, Qt::DisplayRole).toString() == segm)
            {
                found = true;
                node = itm;
                break;
            }
        }
        if (!found)
            return;
    }
    if (!ui->treeWidget->selectedItems().isEmpty())
        ui->treeWidget->selectedItems().first()->setSelected(false);
    node->setSelected(true);
}

void sh::ui::ManageBookmarksDialog::_selectbookmark(QString id)
{
    QList<QTreeWidgetItem*> stack;
    stack.append(ui->treeWidget->topLevelItem(0));
    while(!stack.isEmpty())
    {
        QTreeWidgetItem* itm = stack.takeFirst();
        if (itm->data(1, Qt::DisplayRole).toString() == id)
        {
            if (!ui->treeWidget->selectedItems().isEmpty())
                ui->treeWidget->selectedItems().first()->setSelected(false);
            itm->setSelected(true);
            break;
        }
        for(int i=0; i<itm->childCount(); i++)
            stack.append(itm->child(i));
    }
}

void sh::ui::ManageBookmarksDialog::on_btnUp_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::tools::BookmarkManager::instance()->moveBookmarkUp(id);
    _selectbookmark(id);
}

void sh::ui::ManageBookmarksDialog::on_btnDown_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    sh::tools::BookmarkManager::instance()->moveBookmarkDown(id);
    _selectbookmark(id);
}

void sh::ui::ManageBookmarksDialog::on_btnMove_clicked()
{
    QString id = ui->treeWidget->selectedItems().first()->data(1, Qt::DisplayRole).toString();
    QMenu menu;
    QSet<QString> folders;
    folders.insert("/");
    typedef QPair<QTreeWidgetItem*, QString> StackItem;
    QList<StackItem> stack;
    stack.append(StackItem(ui->treeWidget->topLevelItem(0), ""));
    while(!stack.isEmpty())
    {
        StackItem stackitem = stack.takeLast();
        QTreeWidgetItem* witem = stackitem.first;
        for(int i=0; i<witem->childCount(); i++)
        {
            QTreeWidgetItem* cwitem = witem->child(i);
            if (cwitem->data(2, Qt::DisplayRole).toString().isEmpty())
            {
                QString folder = stackitem.second+"/"+cwitem->data(3, Qt::DisplayRole).toString();
                stack.append(StackItem(cwitem, folder));
                folders.insert(folder);
            }
        }
    }
    QStringList lfolders = folders.toList();
    qSort(lfolders.begin(), lfolders.end());
    foreach(QString f, lfolders)
        menu.addAction(f);
    QAction* a = menu.exec(ui->btnMove->mapToGlobal(QPoint(0, ui->btnMove->height())));
    if (a)
    {
        QStringList newfolder = a->text().split('/');
        newfolder.takeFirst();
        if (newfolder.first().isEmpty())
            newfolder.takeFirst();
        sh::tools::BookmarkManager::instance()->moveBookmarkToFolder(id, newfolder);
        _selectbookmark(id);
    }
}
