/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "base/eurl.h"
#include "configuration/configurationmanager.h"
#include "settings/settingsmanager.h"
#include "settings/profilenode.h"
#include "settings/setting.h"
#include "tools/iconmanager.h"
#include "ui/filesystempanel.h"
#include "ui/settinguiframe.h"
#include "manageprofilesdialog.h"
#include "ui_manageprofilesdialog.h"
#include "mainwindow.h"


sh::ui::ManageProfilesDialog::ManageProfilesDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::ManageProfilesDialog)
{
    ui->setupUi(this);
    ui->listView->setEditTriggers(QAbstractItemView::NoEditTriggers);
    ui->listView->setSelectionBehavior(QAbstractItemView::SelectRows);
    ui->listView->setSelectionMode(QAbstractItemView::SingleSelection);
    ui->btnClose->setSizeByFactor(120);
    ui->btnClose->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    connect (ui->listView, SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)), this, SLOT(slot_listViewactivated(QListWidgetItem*,QListWidgetItem*)));
    _createSettingsList();
    _createProfileList(true);
    ui->listView->setFocus();
    ui->btnDelNode->setIcon(sh::tools::IconManager::instance()->getIcon("format-remove-node"));
    ui->btnDelProfile->setIcon(sh::tools::IconManager::instance()->getIcon("edit-clear-list"));
    QString wc = sh::configuration::ConfigurationManager::instance()->style()->backgroundColor().name();
    ui->listFrame->setStyleSheet(QString("QFrame{background:%1;}").arg(wc));
    ui->scrollAreaWidgetContents->setStyleSheet(QString("background:%1;").arg(wc));
    ui->label_2->setStyleSheet(QString("font-size:%1pt; margin:8px; color: %2;")
                        .arg(sh::configuration::ConfigurationManager::instance()->style()->fontSizeInPt(115))
                        .arg(sh::configuration::ConfigurationManager::instance()->style()->foregroundColorLighter1().name())
                             );
    ui->panelSelectProfile->setStyleSheet(QString(".QWidget{background:%1;}")
                        .arg(sh::configuration::ConfigurationManager::instance()->style()->windowColorLikeTitlebar().name()));
    ui->lblHead->setStyleSheet(QString("font-size:%1pt;")
                        .arg(sh::configuration::ConfigurationManager::instance()->style()->fontSizeInPt(130)));
}

void sh::ui::ManageProfilesDialog::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Delete)
    {
        on_btnDelNode_clicked();
    } else
        QDialog::keyPressEvent(event);
}

void sh::ui::ManageProfilesDialog::resizeEvent(QResizeEvent* event)
{
    if (event)
        QDialog::resizeEvent(event);
    int w1=1, w2=1;
    if (ui->listView->count()>0) {
        w1=ui->listView->sizeHintForColumn(0);
        w2=width() - w1 - 130;
    }
    ui->splitter->setSizes(QList<int>({w1,w2}));
}

void sh::ui::ManageProfilesDialog::_createProfileList(bool selectCurrent)
{
    QString _oldprofile = ui->comboBox->currentText();
    QStringList profilelist;
    profilelist = sh::settings::SettingsManager::instance()->getProfileList();
    ui->comboBox->clear();
    ui->comboBox->addItems(profilelist);
    QString currprofile;
    if (selectCurrent)
    {
        currprofile = sh::ui::MainWindow::mainWindow()->getCurrentProfile();
    } else {
        currprofile = _oldprofile;
    }
    if (profilelist.contains(currprofile))
        ui->comboBox->setCurrentText(currprofile);
    else if (profilelist.contains("main"))
        ui->comboBox->setCurrentText("main");
    else if (profilelist.length()>0)
        ui->comboBox->setCurrentText(profilelist[0]);
}

sh::ui::ManageProfilesDialog::~ManageProfilesDialog()
{
    delete ui;
}

void sh::ui::ManageProfilesDialog::_createNodesList()
{
    QString profilename = ui->comboBox->currentText();
    _nodes.clear();
    ui->listView->clear();
    QIcon icoN = sh::tools::IconManager::instance()->getIcon("arrow-right");
    QIcon icoG = sh::tools::IconManager::instance()->getIcon("arrow-right-double");
    foreach(sh::settings::ProfileNode* n,  sh::settings::SettingsManager::instance()->getNodesForProfile(profilename))
    {
        QString text="";
        if (n->eurl())
            text = n->eurl()->asString();
        QListWidgetItem *item = new QListWidgetItem(text, ui->listView);
        ui->listView->addItem(item);
        _nodes[item]=n->nodeId();
        item->setIcon(icoN);
    }
    ui->listView->model()->sort(0);
    for (int i=0;i<ui->listView->count();i++)
    {
        if (ui->listView->item(i)->text().isEmpty())
        {
            ui->listView->item(i)->setIcon( icoG );
            ui->listView->item(i)->setText(tr("global"));
        }
        else
            break;
    }
    if (_nodes.keys().length()>0) {
        ui->listView->selectionModel()->setCurrentIndex( ui->listView->model()->index(0,0), QItemSelectionModel::SelectCurrent);
        resizeEvent(0);
    }
}

void sh::ui::ManageProfilesDialog::on_comboBox_currentTextChanged(const QString &)
{
    _createNodesList();
}

void sh::ui::ManageProfilesDialog::_createSettingsList()
{
    QWidget* l = ui->scrollAreaWidgetContents;
    QVBoxLayout* lo = (QVBoxLayout*)l->layout();

    QLayoutItem* it;
    while ((it = lo->takeAt(0)) != 0) {
        if (it->widget())
            delete it->widget();
        delete it;
    }
    int _fontSize1 = sh::configuration::ConfigurationManager::instance()->style()->fontSizeInPt(90);
    QString _fgColor1 = sh::configuration::ConfigurationManager::instance()->style()->foregroundColor().name();
    QString _fgColor2 = sh::configuration::ConfigurationManager::instance()->style()->foregroundColorLighter1().name();
    if (_currentNode)
    {
        QLabel* lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("Those settings are stored in the selected node:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1) );
        lo->addWidget(lbl);
        bool empty=true;
        foreach(QString settingname, _currentNode->getSettingNames())
        {
            std::shared_ptr<sh::settings::Setting> setting = sh::settings::SettingsManager::instance()->getSettingByName(settingname);
            if (setting != 0)
            {
                empty=false;
                QString value = _currentNode->getSettingValue(settingname);
                int onlyinfileview = _currentNode->getSettingOnlyInFileviewIndex(settingname);
                sh::ui::SettingUIFrame* frame = new sh::ui::SettingUIFrame(setting.get(), true, value,
                                                           onlyinfileview!=-1 ? tr("only for file list #%1").arg(onlyinfileview+1) : "",
                                                           "", l);
                l->layout()->addWidget(frame);
            }
        }
        if (empty)
        {
            lbl = new QLabel(l);
            lbl->setWordWrap(true);
            lbl->setText(tr("(none)"));
            lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
            lo->addWidget(lbl);
        }
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("also applies on subfolders:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText((_currentNode->withSubfolders() ? tr("yes"): tr("no")));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("Inherits from other profiles:"));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor1));
        lo->addWidget(lbl);
        lbl = new QLabel(l);
        lbl->setWordWrap(true);
        QStringList inhe = _currentNode->inheritsFrom();
        if (inhe.length()==0)
            lbl->setText(tr("(none)"));
        else if (inhe.length()==1)
            lbl->setText(inhe.first());
        else {
            QString txt = QStringList(inhe.mid(0, inhe.length()-1))
                    .join(", ")+" "+tr("and")+" "+inhe.last();
            lbl->setText(txt);
        }
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2; margin:4px;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
        lo->addStretch(1);
    }
    else
    {
        QLabel* lbl = new QLabel(l);
        lbl->setWordWrap(true);
        lbl->setText(tr("Nothing selected."));
        lbl->setStyleSheet(QString("font-size:%1pt; color:%2;").arg(_fontSize1).arg(_fgColor2));
        lo->addWidget(lbl);
    }
}

void sh::ui::ManageProfilesDialog::slot_listViewactivated(QListWidgetItem* current,QListWidgetItem*)
{
    if (current)
        _currentNode = sh::settings::SettingsManager::instance()->getNodeByNodeId( _nodes[current] );
    else
        _currentNode = 0;
    ui->btnDelNode->setEnabled(current!=0);
    _createSettingsList();
}

void sh::ui::ManageProfilesDialog::on_btnClose_clicked()
{
    this->close();
}

void sh::ui::ManageProfilesDialog::on_btnDelNode_clicked()
{
    sh::exceptions::Exception::executeGuarded([&](){
        if (_currentNode){
            QString nodeId = _currentNode->nodeId();
            int btn = QMessageBox::question(this, tr("Remove entry"), tr("Do you really want to delete the selected entry from the profile?"));
            if (btn == QMessageBox::Yes)
            {
                sh::settings::SettingsManager::instance()->removeNode(nodeId);
                _createProfileList(false);
            }
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}

void sh::ui::ManageProfilesDialog::on_btnDelProfile_clicked()
{
    sh::exceptions::Exception::executeGuarded([&](){
        QString profile = ui->comboBox->currentText();
        int nodecount = _nodes.keys().count();
        QString _q;
        if (nodecount!=1)
            _q = tr("Do you really want to delete the complete profile '%2' with its %1 entries?").arg(nodecount);
        else
            _q = tr("Do you really want to delete the complete profile '%1' with its one entry?");
        int btn = QMessageBox::question(this, tr("Remove profile"), _q.arg(profile));
        if (btn == QMessageBox::Yes)
        {
            sh::settings::SettingsManager::instance()->removeProfile(profile);
            _createProfileList(false);
        }
    }, sh::exceptions::ExecuteGuardFlag::RESUMEABLE_PROGRAM_ERROR_END_WITH_USER_FEEDBACK_HERE);
}
