/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QInputDialog>
#include "base/eurl.h"
#include "configuration/configurationmanager.h"
#include "tools/filetypes/filetypemanager.h"
#include "tools/iconmanager.h"
#include "openwithdialog.h"
#include "ui_openwithdialog.h"


sh::ui::OpenWithDialog::OpenWithDialog(std::shared_ptr<const base::Eurl> forDirectory, QWidget *parent) :
    QDialog(parent),
    _rememberfordirectory(forDirectory),
    ui(new Ui::OpenWithDialog)
{
    ui->setupUi(this);
    on_chkRememberDir_toggled(0);

    ui->btnOk->setSizeByFactor(120);
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();

    ui->btnCancel->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    ui->btnOk->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-ok"));

    sm->style()->createSheet()
                .background_highlighted()
            .applyTo(ui->listView);
    sm->style()->createSheet()
                .background_liketitlebar()
            .applyTo(ui->panelHead);
    sm->style()->createSheet()
                .fontsize_header()
            .applyTo(ui->lblHead);
    sm->style()->createSheet()
                .fontsize_smaller()
            .applyTo(ui->lblName);
    ui->btnAnotherAncestor->setSizeByFactor(80);
    ui->listView->setFocus();
}

void sh::ui::OpenWithDialog::setProgramList(QList<sh::tools::filetypes::OpenMethod*> openMethods)
{
    _listmodel = new OpenWithDialogModel(openMethods, this);
    this->_openMethods = openMethods;
    ui->listView->setModel(_listmodel);
    connect(ui->listView->selectionModel(), &QItemSelectionModel::selectionChanged, this, [=](const QItemSelection& sel){
        ui->btnOk->setEnabled(true);
        _chosenMethod = _openMethods[sel.indexes().first().row()];
        ui->lblName->setText(_chosenMethod->name);
    });
}

sh::tools::filetypes::OpenMethod* sh::ui::OpenWithDialog::chosenMethod()
{
    return _chosenMethod;
}

bool sh::ui::OpenWithDialog::rememberForMimetype()
{
    return ui->chkRememberType->isChecked();
}

std::shared_ptr<const sh::base::Eurl> sh::ui::OpenWithDialog::rememberForDirectory()
{
    return (ui->chkRememberDir->isChecked()) ? _rememberfordirectory : 0;
}

bool sh::ui::OpenWithDialog::rememberForFile()
{
    return ui->chkRememberFile->isChecked();
}

sh::ui::OpenWithDialog::~OpenWithDialog()
{
    delete ui;
    if (_manualMethod)
        delete _manualMethod;
}

void sh::ui::OpenWithDialog::on_btnCancel_clicked()
{
    reject();
}

void sh::ui::OpenWithDialog::on_btnOk_clicked()
{
    accept();
}

void sh::ui::OpenWithDialog::on_btnSelectOther_clicked()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("Open with"));
    if (!filename.isEmpty())
    {
        if (_manualMethod)
            delete _manualMethod;
        QString name = QFileInfo(filename).fileName();
        QString command = filename;
        QStringList arguments({"%f"});
        _manualMethod = _chosenMethod = new sh::tools::filetypes::OpenMethod(name, command, arguments);
        ui->btnOk->setEnabled(true);
        ui->lblName->setText(_chosenMethod->name);
    }
}

void sh::ui::OpenWithDialog::on_listView_activated(const QModelIndex&)
{
    if (ui->listView->selectionModel()->hasSelection())
        accept();
}

sh::ui::OpenWithDialogModel::OpenWithDialogModel(QList<sh::tools::filetypes::OpenMethod*> methods, QObject *parent)
        : QStringListModel(parent), _methods(methods)
{
    QStringList names;
    foreach(sh::tools::filetypes::OpenMethod* m, _methods)
        names.append(m->name);
    setStringList(names);
}

QVariant sh::ui::OpenWithDialogModel::data(const QModelIndex &index, int role) const
{
    if (index.isValid() && role == Qt::DecorationRole)
    {
        QIcon i;
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Normal, QIcon::On), QIcon::Normal, QIcon::On);
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Active, QIcon::On), QIcon::Active, QIcon::On);
        i.addPixmap(_methods[index.row()]->icon.pixmap(32, 32, QIcon::Selected, QIcon::On), QIcon::Selected, QIcon::On);
        return i;
    }
    else
        return QStringListModel::data(index, role);
}

void sh::ui::OpenWithDialog::on_btnAnotherAncestor_clicked()
{
    QString ndir = QInputDialog::getText(this, tr("Another ancestor"), tr("Please specify which directory to store this application choice for."), QLineEdit::Normal, _rememberfordirectory?_rememberfordirectory->asString():"");
    try {
        _rememberfordirectory = sh::base::Eurl::fromString(ndir);
        QString btntxt = _rememberfordirectory->basename();
        if (btntxt.length() > 20)
            btntxt = "..." + btntxt.right(17);
        ui->btnAnotherAncestor->setText(btntxt);
    }
    catch (sh::base::EurlMisformattedException&) {
    }
}

void sh::ui::OpenWithDialog::on_chkRememberDir_toggled(bool)
{
    ui->btnAnotherAncestor->setVisible(ui->chkRememberDir->isChecked());
}
