/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration/configurationmanager.h"
#include "settings/setting.h"
#include "settings/settingsmanager.h"
#include "tools/iconmanager.h"
#include "ui/mainwindow.h"
#include "ui/filesystempanel.h"
#include "ui/settinguiframe.h"
#include "storeprofiledialog.h"
#include "ui_storeprofiledialog.h"


std::shared_ptr<sh::configuration::ConfigurationValue> sh::ui::StoreProfileDialog::cfgvalLevel =
        sh::configuration::ConfigurationManager::instance()->registerConfigValue("sh::ui::StoreProfileDialog.Level", 0,
                              sh::configuration::ConfigurationValue::ValueTypeInteger());

sh::ui::StoreProfileDialog::StoreProfileDialog(QString eurl, QWidget *parent) :
    QDialog(parent), ui(new Ui::StoreProfileDialog), _eurl(eurl)
{
    ui->setupUi(this);
    sh::configuration::ConfigurationManager* sm = sh::configuration::ConfigurationManager::instance();
    QWidget* l = ui->scrollAreaWidgetContents;

    sm->style()->createSheet().background_highlighted().applyTo(l);
    sm->style()->createSheet().background_liketitlebar().applyTo(ui->panelHead);
    sm->style()->createSheet().fontsize_header().applyTo(ui->label);

    ui->toolButton_2->setSizeByFactor(120);
    ui->toolButton->setSizeByFactor(120);
    ui->toolButton_2->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-close"));
    ui->toolButton->setIcon(sh::tools::IconManager::instance()->getIcon("dialog-ok"));
    int lastgroup = -1;
    foreach(std::shared_ptr<sh::settings::Setting> setting, sh::settings::SettingsManager::instance()->getAllSettings())
    {
        int g = setting->group();
        if (lastgroup == -1 || lastgroup != g)
        {
            QLabel* lbl = new QLabel(l);
            QString txt = sh::settings::Setting::getGroupDescription(g);
            lbl->setText(txt);
            lbl->setAlignment(Qt::AlignHCenter);
            l->layout()->addWidget(lbl);
            lastgroup = g;
            _headerlabels[g] = lbl;
        }
        QString currval = setting->getValue( sh::ui::MainWindow::mainWindow()->filesystemPanel()->currentView() );
        if (!currval.isNull())
        {
            sh::ui::SettingUIFrame* frame = new sh::ui::SettingUIFrame(setting.get(), false, currval,
                                                       setting->isPerFileview() ? tr("only for file list #%1").arg(sh::ui::MainWindow::mainWindow()->filesystemPanel()->currentViewIndex()+1) : "",
                                                       sh::ui::MainWindow::mainWindow()->filesystemPanel()->currentViewIndex(), l);
            l->layout()->addWidget(frame);
            _widgets[frame] = setting;
        }
    }
    foreach(QString _prof, sh::settings::SettingsManager::instance()->getProfileList())
    {
        _inheritProfileNameList.append(_prof);
        QListWidgetItem* item = new QListWidgetItem(_prof, ui->listInherit);
        item->setCheckState(Qt::Unchecked);
        ui->listInherit->addItem(item);
    }
    ui->lineEditProfilename->addItems( sh::settings::SettingsManager::instance()->getProfileList() );
    ui->lineEditProfilename->setCurrentText( sh::ui::MainWindow::mainWindow()->getCurrentProfile() );
    QSpacerItem* spacer = new QSpacerItem(0,0, QSizePolicy::Fixed, QSizePolicy::Expanding);
    l->layout()->addItem(spacer);
    int lvl = cfgvalLevel->getConfigValueInt();
    _settingsLevel(1, false); // needed for initialization
    _settingsLevel(0, false); // needed for initialization
    _settingsLevel(lvl, false);
    _setglobal(true);
    ui->scrollAreaWidgetContents->setFocus();
}

sh::ui::StoreProfileDialog::~StoreProfileDialog()
{
    delete ui;
}

QList<sh::ui::StoreProfileDialog::CheckedSetting> sh::ui::StoreProfileDialog::getCheckedSettings()
{
    QList<sh::ui::StoreProfileDialog::CheckedSetting> result;
    foreach(sh::ui::SettingUIFrame* frame, _widgets.keys())
    {
        std::shared_ptr<sh::settings::Setting> setting = _widgets[frame];
        if (frame->isChecked() && frame->isEnabled())
        {
            QVariant a = frame->additionalCheckValue();
            int ofv = -1;
            if (!a.isNull())
                ofv = a.toInt();
            result.append(CheckedSetting(setting.get(), ofv));
        }
    }
    return result;
}

QString sh::ui::StoreProfileDialog::profilename()
{
    return ui->lineEditProfilename->currentText();
}

bool sh::ui::StoreProfileDialog::withSubdirectories()
{
    return ui->checkBoxWithSubdirectories->isChecked();
}

QStringList sh::ui::StoreProfileDialog::inheritsFrom()
{
    QStringList result;
    for(int i=0; i<ui->listInherit->count();i++)
    {
        QListWidgetItem* itm = ui->listInherit->item(i);
        if (itm->checkState() == Qt::Checked){
            result << itm->text();
        }
    }
    return result;
}

void sh::ui::StoreProfileDialog::_setglobal(bool v)
{
    _global = v;
    if (_global)
    {
        ui->lblEurl->setText(tr("all directories"));
        ui->btnGlobal->setText(tr("restrict &to current directory"));
        ui->checkBoxWithSubdirectories->setChecked(true);
        ui->checkBoxWithSubdirectories->setEnabled(false);
    }
    else
    {
        ui->lblEurl->setText(_eurl);
        ui->btnGlobal->setText(tr("store &globally"));
        ui->checkBoxWithSubdirectories->setEnabled(true);
    }
    _settingsLevel(_level, false);
}

bool sh::ui::StoreProfileDialog::hasGlobal()
{
    return _global;
}

void sh::ui::StoreProfileDialog::_settingsLevel(int level, bool persist)
{
    _level = level;
    if (persist)
        cfgvalLevel->setConfigValue(level);
    ui->lineEditProfilename->setVisible(level>=1);
    ui->lblProfName->setVisible(level>=1);
    if (level==0)
        ui->lineEditProfilename->setCurrentText( sh::ui::MainWindow::mainWindow()->getCurrentProfile() );
    foreach(QLabel* lbl, _headerlabels.values())
        lbl->setVisible(false);
    foreach(sh::ui::SettingUIFrame* widget, _widgets.keys())
    {
        std::shared_ptr<sh::settings::Setting> setting = _widgets[widget];
        bool show = true;
        if (setting->isAdvancedSetting() && level<1)
            show = false;
        if (setting->isGlobal() && !_global)
            show = false;
        widget->setVisible(show);
        widget->setEnabled(show);
        if (show) {
            widget->setAdditionalCheckVisible(level >= 1);
            _headerlabels[setting->group()]->setVisible(true);
        }
        widget->setChecked(show && (level == 0));
    }
    if (level<1)
    {
        for(int i=0; i<ui->listInherit->count();i++)
            ui->listInherit->item(i)->setCheckState(Qt::Unchecked);
    }
    ui->label_5->setVisible(level>=1);
    ui->listInherit->setVisible(level>=1);
    ui->btnAddProf->setVisible(level>=1);
    QString text;
    if (level==0) {
        text=tr("basic stuff");
    } else {
        text=tr("advanced stuff");
    }
    ui->btn22->setText(text+" "+tr("(&change)"));
    ui->checkBoxWithSubdirectories->setVisible(level>=1);
    if (level==0)
        ui->checkBoxWithSubdirectories->setChecked(true);
    ui->btnGlobal->setVisible(level>=1);
    if (level==0 && !_global)
        _setglobal(true);
}

void sh::ui::StoreProfileDialog::on_btnAddProf_clicked()
{
    QString nname = QInputDialog::getText(this, tr("Profile"), tr("Please enter the name of the profile you want to inherit from."));
    if (!nname.isEmpty() && !nname.trimmed().isEmpty()){
        nname=nname.trimmed();
        if (!_inheritProfileNameList.contains(nname))
        {
            QListWidgetItem* lwi = new QListWidgetItem(nname, ui->listInherit);
            lwi->setCheckState(Qt::Checked);
            ui->listInherit->addItem(lwi);
            _inheritProfileNameList << nname;
        }
    }
}

void sh::ui::StoreProfileDialog::on_btnGlobal_clicked()
{
    _setglobal(!_global);
}

void sh::ui::StoreProfileDialog::on_btn22_clicked()
{
    QMenu menu;
    QAction* actd = menu.addAction(tr("simple mode"));
    connect(actd, &QAction::triggered, this, [=](){ _settingsLevel(0, true); });
    QAction* acti = menu.addAction(tr("advanced mode"));
    connect(acti, &QAction::triggered, this, [=](){ _settingsLevel(1, true); });
    menu.exec(ui->btn22->mapToGlobal(QPoint(0, ui->btn22->height())));
}

void sh::ui::StoreProfileDialog::on_toolButton_2_clicked()
{
    reject();
}

void sh::ui::StoreProfileDialog::on_toolButton_clicked()
{
    accept();
}

sh::ui::StoreProfileDialog::CheckedSetting::CheckedSetting(sh::settings::Setting* setting, int onlyInFileview)
    : setting(setting), onlyInFileview(onlyInFileview)
{
}
