/*
Shallot is a Qt based file manager.
Copyright (C) 2014-2018 Josef Hahn

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QStylePainter>
#include <QStyleOptionToolButton>
#include <QMouseEvent>
#include "configuration/configurationmanager.h"
#include "configuration/uistyle.h"
#include "tools/mainthread.h"
#include "ui/mainwindow.h"
#include "toolbarbutton.h"


sh::ui::ToolbarButton::ToolbarButton(QWidget *parent) :
    QToolButton(parent)
{
    location = Position::TOP;
    unsetClickAction();
    QFont f;
    f.setPointSizeF(f.pointSizeF() * 1.1);
    setFont(f);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
    connect(this, SIGNAL(clicked()), this, SLOT(slot_clicked()));
    setMouseTracking(true);
    setFocusPolicy(Qt::ClickFocus);
    sh::ui::MainWindow::mainWindow()->installEventFilter(this);
}

bool sh::ui::ToolbarButton::eventFilter(QObject*, QEvent *event)
{
    if (!_drawmnemonics && event->type() == QEvent::KeyPress)
    {
        QKeyEvent* keyevent = (QKeyEvent*) event;
        if (keyevent->key() == Qt::Key_Alt)
        {
            _drawmnemonics = true;
            repaint();
        }
    }
    else if (_drawmnemonics && event->type() == QEvent::KeyRelease)
    {
        _drawmnemonics = false;
        repaint();
    }
    else if (event->type() == QEvent::WindowActivate || event->type() == QEvent::WindowDeactivate)
    {
        _drawmnemonics = false;
        repaint();
    }
    return false;
}

void sh::ui::ToolbarButton::_openMenu()
{
    int x=0, y=0;
    QSize _s = _menu->sizeHint();
    int mw = _s.width();
    int mh = _s.height();
    QPoint _gp = this->mapToGlobal(QPoint(0,0));
    int ancx = _gp.x();
    int ancy = _gp.y();
    int h = height();
    int w = width();
    switch(location)
    {
    case Position::TOP:
        x=ancx;
        y=ancy+h;
        break;
    case Position::BOTTOM:
        x=ancx;
        y=ancy-mh;
        break;
    case Position::LEFT:
        x=ancx+w;
        y=ancy;
        break;
    case Position::RIGHT:
        x=ancx-mw;
        y=ancy;
        break;
    }
    _menu->exec(QPoint(x,y));
}

void sh::ui::ToolbarButton::slot_clicked()
{
    QPoint p = mapFromGlobal(QCursor::pos());
    int x = p.x();
    int w = width();
    int oph = opad / 2;
    if (_menu && (
                (location==Position::RIGHT && x<=ew+oph) ||
                (location!=Position::RIGHT && x >= (w-ew-oph) )
        ))
    {
        _openMenu();
    }
    else
    {
        _clickAction();
    }
}

void sh::ui::ToolbarButton::setClickAction( std::function<void ()> action )
{
    _hasClickAction = true;
    _clickAction = action;
}

void sh::ui::ToolbarButton::trigger()
{
    if (isEnabled())
    {
        if (_menu)
            _openMenu();
        else
            _clickAction();
    }
}

void sh::ui::ToolbarButton::unsetClickAction()
{
    _hasClickAction = false;
    _clickAction = [&](){ this->_openMenu(); };
}

void sh::ui::ToolbarButton::computeArrowAndDividerLineCoordinates()
{
    int w = width();
    int h = height();
    int oph = opad / 2;
    arrow = QPainterPath();
    if (location == Position::LEFT || location == Position::RIGHT)
    {
        iedim = h-2*opad;
        xt = opad + textpad / 2;
        if (_text1.isEmpty())
            yt2 = h - (h - thorig) / 2 - oph;
        else {
            yt2 = h - std::max(0.0, h - 1.9 * thorig) / 2 - oph;
            yt1 = yt2 - 0.8 * thorig;
        }
        xi = oph;
        yi = (h - iedim) / 2 + oph;
        if (location == Position::RIGHT)
        {
            xt += ew;
            xi += ew;
        }
        int arrowx;
        int arrowy = h/2;
        int arrd;
        expy1 = oph * 4;
        expy2 = h - opad * 2;
        if (location == Position::LEFT)
        {
            arrd = -1;
            arrowx = w - oph - ew/2;
            expx2 = expx1 = w - oph - ew;
        }
        else
        {
            arrd = 1;
            arrowx = ew/2 + oph;
            expx2 = expx1 = oph + ew;
        }
        arrow.moveTo(arrowx+arrd*ew/4, arrowy-eh/4);
        arrow.lineTo(arrowx-arrd*ew/4, arrowy);
        arrow.lineTo(arrowx+arrd*ew/4, arrowy+eh/4);
        arrow.closeSubpath();
    }
    else
    {
        iedim = std::min(w-2*opad, h-2*opad);
        xt = (w - ew - twmax + textpad) / 2;
        if (_text1.isEmpty())
            yt2 = h - ( h - thorig) / 3 - oph;
        else {
            yt2 = h - std::max(0.0, h - 1.9 * thorig) / 2 - oph;
            yt1 = yt2 - 0.8 * thorig;
        }
        yi = opad;
        xi = (w - ew - iedim) / 2 + opad;
        expx1 = w - oph - ew;
        expy1 = oph * 4;
        expx2 = w - oph - ew;
        expy2 = h - opad * 2;
        int arrowx = w - oph - ew / 2;
        int arrowy = h / 2;
        int arrd;
        if (location == Position::TOP)
            arrd = -1;
        else
            arrd = 1;
        arrow.moveTo(arrowx-ew/4, arrowy+arrd*eh/4);
        arrow.lineTo(arrowx+ew/4, arrowy+arrd*eh/4);
        arrow.lineTo(arrowx, arrowy-arrd*eh/4);
        arrow.closeSubpath();
    }
}

void sh::ui::ToolbarButton::paintEvent(QPaintEvent*)
{
    QStylePainter p(this);
    QStyleOptionToolButton opt;
    initStyleOption(&opt);
    p.drawComplexControl(QStyle::CC_ToolButton, opt);
    bool _drawenabled = isEnabled();
    if (_menu && _drawenabled && _hovered) {
        p.setPen( QPen( palette().color(QPalette::Normal, QPalette::Window).darker(140) ));
        if (_hasClickAction)
            p.drawLine( expx1, expy1, expx2, expy2 );

        QColor brushcolor = (!_hasClickAction||_arrowhovered)
                  ? palette().color(QPalette::Normal, QPalette::Highlight)
                  : palette().color(QPalette::Normal, QPalette::Window).darker(200);
        p.fillPath(arrow, QBrush(brushcolor));
    }
    QIcon::Mode imode = QIcon::Normal;
    if (_drawenabled)
    {
        if (_hovered)
        {
            if ( !(_menu && _hasClickAction) || !_arrowhovered)
                imode = QIcon::Selected;
        }
    }
    else
        imode = QIcon::Disabled;
    if (!_drawmnemonics) {
        if (!_hovered)
            p.setOpacity(0.6);
        p.drawPixmap(xi, yi, iedim, iedim, _icon.pixmap(iedim, iedim, imode));
        p.setOpacity(1.0);
    }
    p.setFont(font());
    if (_drawenabled)
        p.setPen( QPen( palette().color(QPalette::Normal, QPalette::Text) ));
    else
        p.setPen( QPen( palette().color(QPalette::Disabled, QPalette::Text) ));
    QRectF txtrect1(xt, yt1-thorig, tworig1, thorig);
    QRectF txtrect2(xt, yt2-thorig, tworig2, thorig);
    if (_drawmnemonics || !_drawIconOnly) {
        if (_hovered || _drawmnemonics) {
            p.setOpacity(0.4);
            QPen orgpen = p.pen();
            QPen backpen = orgpen;
            QColor c = orgpen.color();
            backpen.setColor(QColor(255-c.red(), 255-c.green(), 255-c.blue()));
            p.setPen(backpen);
            for (int dx=-2; dx<=2; dx++)
                for (int dy=-2; dy<=2; dy++)
                    p.drawText(txtrect1.adjusted(dx, dy, dx, dy), Qt::TextShowMnemonic, _text1);
            for (int dx=-2; dx<=2; dx++)
                for (int dy=-2; dy<=2; dy++)
                    p.drawText(txtrect2.adjusted(dx, dy, dx, dy), Qt::TextShowMnemonic, _text2);
            p.setPen(orgpen);
            p.setOpacity(1.0);
        }
        else
            p.setOpacity(0.6);
        p.drawText(txtrect1, Qt::TextShowMnemonic, _text1);
        p.drawText(txtrect2, Qt::TextShowMnemonic, _text2);
        p.setOpacity(1.0);
    }
    if (!_hovered) {
        p.setOpacity(_drawmnemonics ? 0.2 : 0.4);
        p.drawPixmap(xi, yi, iedim, iedim, _icon.pixmap(iedim, iedim, imode));
        p.setOpacity(1.0);
    }
}

bool sh::ui::ToolbarButton::event(QEvent* e)
{
    if (e->type() == QEvent::Enter)
    {
        _hovered = true;
        update();
    }
    if (e->type()==QEvent::Leave)
    {
        _hovered = false;
        update();
    }
    return QToolButton::event(e);
}

void sh::ui::ToolbarButton::mouseMoveEvent(QMouseEvent* event)
{
    bool __arrowhovered = event->x() > expx1;
    if (location == Position::RIGHT)
        __arrowhovered = !__arrowhovered;
    if (__arrowhovered != _arrowhovered)
    {
        _arrowhovered = __arrowhovered;
        update();
    }
}

void sh::ui::ToolbarButton::resizeEvent(QResizeEvent*)
{
    computeArrowAndDividerLineCoordinates();
}

void sh::ui::ToolbarButton::setLocation(Position location)
{
    this->location = location;
    _calcDims();
}

void sh::ui::ToolbarButton::_calcDims()
{
    QFontMetrics fm(font());
    tworig1 = fm.width(_textWithoutMnemonic1);
    tw1 = tworig1 + textpad;
    tworig2 = fm.width(_textWithoutMnemonic2);
    tw2 = tworig2 + textpad;
    twmax = std::max(tw1, tw2);
    tworigmax = std::max(tworig1, tworig2);
    th = thorig = fm.height();
    iw = logicalDpiX() / 4;
    ih = logicalDpiY() / 4;
    ew = ((hasExpander() || location == Position::RIGHT) ? logicalDpiX() : 0) / 8;
    eh = ((hasExpander() || location == Position::RIGHT) ? logicalDpiY() : 0) / 8;
    computeArrowAndDividerLineCoordinates();
    updateGeometry();
}

QSize sh::ui::ToolbarButton::sizeHint() const
{
    int w,h;
    if (location == Position::LEFT || location == Position::RIGHT) {
        w = twmax + iw + ew;
        h = std::max(ih, (int)((_text1.isEmpty()?1.0:1.9) * thorig));
    }
    else {
        w = std::max(twmax, iw) + ew;
        h = std::max(ih, (int)(1.9 * thorig));
    }
    return QSize(w+opad, h+opad);
}

void sh::ui::ToolbarButton::setText(const QString &text)
{
    _text = text;
    if (text.startsWith("&") && text.length() == 2)
    {
        _text1 = "";
        _text2 = text;
        _textWithoutMnemonic1 = "";
        _textWithoutMnemonic2 = text.mid(1);
        _drawIconOnly = true;
    }
    else
    {
        _text1 = "";
        _text2 = text;
        if (text.length() > 8) {
            int i = text.length() / 2;
            for(int j=0; j<i; j++)
                for (int d=-1; d<=1; d+=2) {
                    int ip = i+d*j;
                    if (text[ip] == ' ') {
                        _text1 = text.mid(0, ip);
                        _text2 = text.mid(ip+1);
                        goto breakout;  // never use goto!
                    }
                }
            breakout: {}
        }
        _textWithoutMnemonic1 = QString();
        _textWithoutMnemonic2 = QString();
        bool _lastwasamp = false;
        foreach(QChar c, _text1)
        {
            if (c != '&' || _lastwasamp)
                _textWithoutMnemonic1 += c;
            _lastwasamp = (c == '&');
        }
        _lastwasamp = false;
        foreach(QChar c, _text2)
        {
            if (c != '&' || _lastwasamp)
                _textWithoutMnemonic2 += c;
            _lastwasamp = (c == '&');
        }
        _drawIconOnly = text.isNull();
    }
    _calcDims();
    repaint();
}

QString sh::ui::ToolbarButton::text() const
{
    return _text;
}

void sh::ui::ToolbarButton::setIcon(const QIcon &icon)
{
    _icon = icon;
    _calcDims();
    repaint();
}

QIcon sh::ui::ToolbarButton::icon()
{
    return _icon;
}

bool sh::ui::ToolbarButton::hasExpander()
{
    return _menu != 0;
}

void sh::ui::ToolbarButton::setSubmenu(QMenu* menu)
{
    _menu = menu;
    _calcDims();
    repaint();
}

QMenu* sh::ui::ToolbarButton::submenu()
{
    return _menu;
}
